!     Last change:  LP   16 Mar 2005   10:12 am
!****
!****   SUBROUTINES used in the program EDMA and identical with BayMEM
!****
!****
!****   Made on 29.10.98 by Martin Schneider
!****   Modified on 19.02.99 by Martin Schneider
!****   and then many times by Lukas Palatinus

!        Determine length of first non space string in DString and data
!******************************************************************
      SUBROUTINE DataLength (ioDString,oDData)
!******************************************************************
      IMPLICIT NONE
      CHARACTER ioDString*(*)
      CHARACTER oDData*132
      INTEGER DLength, OverAll, i
      CHARACTER*1 Space, Tab
      PARAMETER (Space=' ')
      OverAll=LEN(ioDString)
      Tab=CHAR(9)
      DO i=1,OverAll
        IF ((ioDString(i:i).EQ.Space).OR.(ioDString(i:i).EQ.Tab)) THEN
          DLength=i-1
          oDData=ioDString(1:DLength)
          ioDString=ioDString(DLength+1:OverAll)
          GOTO 2010
        END IF
      END DO
2010  END SUBROUTINE DataLength

!       Finding the fist position of occurance of 'char'
!******************************************************************
      SUBROUTINE NextPos (iString, iChar, oPos)
!******************************************************************
      IMPLICIT NONE
      CHARACTER*(*) :: iString
      CHARACTER*1 :: iChar
      INTEGER StringLength, i, oPos
      StringLength=LEN(iString)
      DO i=1,StringLength
        IF (iString(i:i)==iChar) THEN
          oPos=i
          GOTO 1010
        END IF
      END DO 
      oPos=-1
1010  END SUBROUTINE NextPos

!       Cutting the first spaces or tabs if any off the string
!******************************************************************
      SUBROUTINE SpaceErase (ioString)
!******************************************************************
      IMPLICIT NONE
      CHARACTER*(*) ioString
      INTEGER StringLength, i
      CHARACTER*1 Space, Tab
      PARAMETER (Space=' ')
      Tab=CHAR(9)
      StringLength=LEN(ioString)
      DO i=1,StringLength
        IF ((ioString(i:i).NE.Space).AND.(ioString(i:i).NE.Tab)) THEN
          ioString=ioString(i:StringLength)
          GOTO 1010
        END IF
      END DO 
1010  END SUBROUTINE SpaceErase

! Function counting the true length of the string
!*******************************
      FUNCTION Idel(l)
!*******************************
      character*(*) l
      integer idel
      
      Idel=len(l)
1000  IF(Idel/=0) THEN
        if(l(Idel:Idel)==' ') THEN
          Idel=Idel-1
          GO TO 1000
        END IF
      END IF
      END FUNCTION Idel

!       Applying a symmetrie operation to a vector, integer version
!******************************************************************
      SUBROUTINE SymOp (iSymN, ixyz, oSxyz)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      
      INTEGER, INTENT (IN) :: iSymN              ! Number of symmetry operation
      INTEGER, INTENT (IN) :: ixyz(MaxDim)       ! Vector
      INTEGER, INTENT (OUT) :: oSxyz(MaxDim)     ! Symmetrie related vector to xyz
      
      INTEGER :: i, ii                          ! Counters      

      oSxyz=0
      DO i=1, SDim
        DO ii=1, SDim
          oSxyz(i)=oSxyz(i)+ixyz(ii)*RMatI(i,ii,iSymN)
        END DO
        oSxyz(i)=oSxyz(i)+TVecI(i,iSymN)
      END DO
      END SUBROUTINE SymOp

!       Applying a symmetrie operation to a vector, integer version
!******************************************************************
      SUBROUTINE SymOpR (iSymN, ixyz, oSxyz)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      
      INTEGER, INTENT (IN) :: iSymN              ! Number of symmetry operation
      REAL, INTENT (IN) :: ixyz(MaxDim)       ! Vector
      REAL, INTENT (OUT) :: oSxyz(MaxDim)     ! Symmetrie related vector to xyz
      
      INTEGER :: i, ii                          ! Counters      

      oSxyz=0
      DO i=1, SDim
        DO ii=1, SDim
          oSxyz(i)=oSxyz(i)+ixyz(ii)*RMat(i,ii,iSymN)
        END DO
        oSxyz(i)=oSxyz(i)+TVec(i,iSymN)
      END DO
      END SUBROUTINE SymOpR

!******************************************************************
      SUBROUTINE Arraylocation (iDhkl, oArrLoc, iDelta)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      
      INTEGER, INTENT (IN) :: iDhkl (MaxDim)   ! hkl to compute from
      INTEGER, INTENT (IN) :: iDelta           ! Difference between elements, 2 in case of complex array, 1 else  
      INTEGER :: htmp                   ! Temporary h,k,l... buffer
      INTEGER, INTENT (OUT) :: oArrLoc   ! Computed array location of structure factor
      LOGICAL, ALLOCATABLE :: Warning(:)! Warning about too small divisions
      LOGICAL :: Warn                   ! dto.
      INTEGER :: i, ii                  !  Internal loop counters
      
      ALLOCATE (Warning(SDim))
      DO i=1, SDim
        Warning(i)=.FALSE.
      END DO
      Warn=.FALSE.
      oArrLoc=0

      DO ii=1, SDim
!       Mapping into positive
        htmp=iDhkl(ii)
        DO WHILE (htmp>=NVox(ii))
          htmp=htmp-NVox(ii)
        END DO
        DO WHILE (htmp<0)
          htmp=htmp+NVox(ii)
        END DO
        oArrLoc=oArrLoc+htmp*N(ii-1)*iDelta
      END DO
      END SUBROUTINE Arraylocation

!       Calculate Coordinates from array location
!******************************************************************
      SUBROUTINE Coordinates (iArrLoc, iDelta, oDxyz)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      
      INTEGER, INTENT (Out) :: oDxyz (MaxDim)    ! xyz to compute
      INTEGER, INTENT (IN) :: iDelta             ! Difference between elements, 2 in case of complex array, 1 else  
      INTEGER, INTENT (IN) :: iArrLoc    ! Array location of location
      INTEGER :: LocRest                ! Rest of ArrLoc for calculation of next coordinate
      INTEGER :: i                      !  Internal loop counters

      oDxyz=0
      LocRest=iArrLoc/iDelta
!       Reversed step direction
      DO i=SDim, 1, -1
        oDxyz(i)=INT(LocRest/N(i-1))
        LocRest=LocRest-oDxyz(i)*N(i-1)
      END DO
      END SUBROUTINE Coordinates


      
!       Conversion of integers to double precision
!       without artefacts from binary conversion
!*******************************************************************************
      FUNCTION INT2DBL (Int)
!*******************************************************************************

      DOUBLE PRECISION :: INT2DBL
      DOUBLE PRECISION :: DBuffer
      INTEGER, INTENT (IN) :: Int
      CHARACTER :: Buffer*20
      
      WRITE (Buffer(1:16), '(I15)') Int
      Buffer(17:20)='.0D0'
      READ (Buffer, '(ES20.6)') DBuffer
      INT2DBL=DBuffer
      END FUNCTION INT2DBL


!       Remapping a vector into the unit cell
!******************************************************************
      SUBROUTINE Remap (xyz)
!******************************************************************
      
      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      
      INTEGER :: xyz (MaxDim)
      INTEGER :: i
      INTEGER :: sgn
      
      DO i=1, SDim
        sgn=SIGN(1, xyz(i))
        SELECT CASE (sgn)
          CASE (-1)
            xyz(i)=xyz(i)+(1+INT(xyz(i)/NVox(i)))*NVox(i)
          CASE DEFAULT
            xyz(i)=MOD(xyz(i),NVox(i))
        END SELECT
      END DO

      END SUBROUTINE Remap

! Function that determines the value of the Recl factor / differs between DEC and other systems
!******************************************************
       FUNCTION GetRecLFac()
!******************************************************
       IMPLICIT NONE
       character*13 :: filename
       integer GetRecLFac
       integer :: size,status, nameln,i,orecl
       orecl=4
       filename='getreclen.tmp'
       nameln=13
       open (99,access='DIRECT', file=filename, form='UNFORMATTED',STATUS='UNKNOWN',recl=orecl)
       i=132
       write(99,rec=1) i
       close(99)
       call fstatc(filename, nameln, size, status)
       GetRecLFac=orecl**2/size
       open(99,file='getreclen.tmp')
       close(99,status='delete')
       end FUNCTION GetRecLFac
