!     Last change:  LP   16 Mar 2005   10:09 am
!****************************************************************
!* PROGRAM EDMA - for topological analysis of discrete          *
!* electron density maps                                        *
!* Lukas Palatinus                                              *
!* Sander van Smaalen                                           *
!* version 05/01/2005                                           *
!****************************************************************
      program EDMA
      
      USE GLOBALDEFINITIONS
      USE GLOBALVARIABLES
      
      IMPLICIT NONE
      
      INTEGER :: i,ii,iii,pcnt ! counters
      INTEGER :: ipc ! Cljf
      INTEGER :: GetRecLFac !Function returning RecLFac for direct-acces files
      INTEGER :: ArrLoc,ArrLocR !Positions in rho arrays. Any name ending with R applies to the real dimension density
      INTEGER :: Dxyz(MaxDim),DxyzCur(MaxDim),Dxyz2(MaxDim),Dxyz3(MaxDim)  !dummy coordinates
      INTEGER :: StartC(MaxDim),StartCR(MaxDim),EndC(MaxDim),EndCR(MaxDim) !starting and ending coordinates of the "hypercubes" of electron density
      INTEGER :: Orig(MaxDim)  !Origin of the density section for spline interpolation
      INTEGER :: Idel !function returning length of the string
      DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: y,y2 !dummy variables, need in the Splines subroutine, but not used here
      DOUBLE PRECISION,DIMENSION(MaxDim) :: t,x4 !vectors of t a x in additional dimensions
      DOUBLE PRECISION,DIMENSION(MaxDim) :: pt !precise coordinates returned by peak serching routine
      DOUBLE PRECISION :: grad, rout,disc,dpos,dsc(MaxDim),Gr !"rho out", discrepancy, dummy position, scaling of the output coordinates
      INTEGER, DIMENSION(MaxDim) :: DerOrd,AxOrd !Order of the derivatives(not used here), ored of the axes in the input map
      INTEGER, DIMENSION(MaxDim) :: CumDifVec !Cummulative difference vector - calculation of the center of charge 
      CHARACTER*(24) :: filefmt='( (''_'',f4.2),''.m81'')' !format template for the output t-maps
      TYPE EVstr_type
        character*(linelength) :: line
      END TYPE EVstr_type
      Type(EVstr_type), DIMENSION(3) :: EVstr
      DOUBLE PRECISION :: rhorsigma
      LOGICAL :: cpyes, succ, succ2, NextCoo, ongridstep, defaultongridstep, hadloop !flag, funciton returning next coordinate in the density array
 
!      open(2,status='scratch') !Cljf
      RecLFac=GetRecLFac()
      CALL Input !read in all the parameters
      WRITE(filefmt(2:2),'(i1)') SDim-RDim !Serves to create output filenames
      ALLOCATE(NVoxR(1:RDim),STAT=AllocError) !pixel division of the real space density
      IF (AllocError.GT.0) STOP 'Error allocating NVoxR! Too little memory?'
      ALLOCATE(NVoxRim(1:RDim),STAT=AllocError) !pixel division with additional areas of the neighboring cells
      IF (AllocError.GT.0) STOP 'Error allocating NVoxR! Too little memory?'
      Aper=0
      DO i=1,SDim-RDim
        DO ii=1,RDim
          IF (.not.projection) THEN !Otherwise we project the superspace density along the internal coordinates and that is periodic
            if (abs(Qvec(ii,i))>eps) Aper(ii)=1 !! indicates particular direction being an aperiodic one
	  END IF
	END DO
      END DO
      Rim=0
      DO i=1,RDim
          Rim(i)=(int(tolerance*NVox(CO(i))/Cell(i))+1+range/2+1)       ! add range and tolerance to the unit cell 
            !Muptiplication by Aper assures periodic directions have Rim=0
          Rim(i)=Aper(i)*max(Rim(i),int(addborder*NVox(CO(i)))+1) ! if com is computed, take maximum of what is needed
      END DO    
      DO i=1,RDim ! Permute the superspace pixel division acording to the order of the axes, fill all different NVox* arrays
	NVoxr(i)=NVox(CO(i))	
	NVoxRim(i)=NVoxR(i)+2*Rim(i)
	NVoxTotR=NVoxTotR*NVoxR(i)
	NVoxTotRim=NVoxTotRim*NVoxRim(i)
        NRim(i)= NVoxRim(i)*NRim(i-1)
      END DO
      cApvx=Cell(7)/dble(NVoxTotR) !cubic Angstrom per voxel
      ALLOCATE(rho(0:NVoxTot-1),STAT=AllocError) !Superspace density
      IF (AllocError.GT.0) STOP 'Error allocating rho! Too little memory?'
      ALLOCATE(rhor(0:NVoxTotRim-1),STAT=AllocError) !real dimension density
      IF (AllocError.GT.0) STOP 'Error allocating rhor! Too little memory?'
      IF (cpformat/='none') THEN ! gradient of the density to quickly locate the CP's
        ALLOCATE(rhogr(0:NVoxTotRim-1),STAT=AllocError) !real dimension density
        IF (AllocError.GT.0) STOP 'Error allocating rhor! Too little memory?'
      END IF
      ALLOCATE(srho(0:range**(max(SDim-RDim,RDim))-1),STAT=AllocError) !Part of the density for the spline interpolation
      IF (AllocError.GT.0) STOP 'Error allocating srho! Too little memory?'
!     Arrays, which are not in use, but must be supplied for the Splines subroutine
      ALLOCATE(y(range),STAT=AllocError)
      IF (AllocError.GT.0) STOP 'Error allocating dummy spline variables! Too little memory?'
      ALLOCATE(y2(range),STAT=AllocError)
      IF (AllocError.GT.0) STOP 'Error allocating dummy spline variables! Too little memory?'
      DO i=1,SDim
        DerOrd(i)=0
	AxOrd(i)=i
      END DO
      write(*,*) 'Reading the density map ',Inrhofile(1:Idel(inrhofile))
      SELECT CASE(InRhoType)
        CASE ('jana')
	  CALL ReadJana(InRhoFile,AllocError) !read in the density
        CASE ('xplor') 
	  CALL ReadXplor(InRhoFile,AllocError) !read in the density
      END SELECT
!     If fullcell no is given, check if the map really has the symmetry given in the unit cell
      IF (.not.fullcellpeaks) CALL CheckSymmetry
      IF (AllocError/=0) STOP 'Error reading input electron density'
! If only projection of the density onto the realspace is to be analyzed, then
      IF ((projection).and.(SDim>RDim)) THEN
! 1/ project the density onto realspace
        Dxyz=0
        DO i=1,RDim
          StartCR(i)=0 !Fill the "lower front left" and "upper back right" corner of the electron density section considered in the calculation of the maxima and charge
          EndCR(i)=NVoxR(i)-1
          Dxyz(i)=StartCR(i)
	END DO
        write(*,'(a)') ' Projecting the electron density along the internal coordinates:'
	DO
	  Call ArrayLocationR(Dxyz,ArrLocR,1) !get array location for current coordinates
!         Permute accvording to the order of axes in the input file
          DO i=1,SDim
	    Dxyz2(i)=Dxyz(CO(i))
	  END DO
!         Now cycle over all internal dimensions and sum the density
          rhor(ArrlocR)=0.0D0
121       Call ArrayLocation(Dxyz2,ArrLoc,1)
          rhor(ArrlocR)=rhor(ArrlocR)+rho(ArrLoc)
!           move one point further
          Cnt2=RDim+1
141       Dxyz2(CO(cnt2))=Dxyz2(CO(cnt2))+1
          IF (Dxyz2(CO(cnt2))>NVox(CO(Cnt2))-1) THEN
            Dxyz2(CO(cnt2))=0
            cnt2=cnt2+1
            if (cnt2==SDim+1) goto 161
	    goto 141
          END IF      
          goto 121
161       continue        
          DO i=RDim+1,SDim
            rhor(ArrLocR)=rhor(ArrLocR)/NVox(CO(i))
          END DO
!           move one point further
          IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz,AxOrd)) EXIT
        END DO
! 2/ change the dimension accordingly
        SDim=RDim
! And that's all what has to be be done!
      END IF

!       Write the header of the output file, if needed
      IF ((maxformat/='none').or.(cpformat/='none')) THEN
        OPEN(UNIT=90,file=outrhofile(1:idel(outrhofile))//'.coo')
        WRITE(90,'(2a)')   '# Analysis of the density map ',inrhofile(1:idel(inrhofile))
        WRITE(90,'(a,i3)') '# Dimension of the map  : ',SDim
        Buffer='(a, i5)'
	WRITE(Buffer(4:4),'(i1)') SDim
	WRITE(90,Buffer)   '# Division of the map   : ',(NVox(i),i=1,SDim)
        Buffer='(a, i1)'
	WRITE(Buffer(4:4),'(i1)') SDim
        WRITE(90,Buffer)   '# Order of the axes     : ',(CO(i),i=1,SDim)
        WRITE(90,'(a,i3)') '# Range for spline interpolation: ',range
        IF (SDim/=RDim) THEN!applies only to the more-dimensional case
          Buffer='(a, f9.4)'
          WRITE(Buffer(4:4),'(i1)') 2*RDim
          WRITE(90,Buffer) '# Cell: ',(Cell(i),i=1,2*RDim)
          WRITE(90,'(a)') '# q-vectors:'
          Buffer='(a2, f9.5)'
	  WRITE(Buffer(5:5),'(i1)') RDim
          DO i=1,SDim-RDim
	    write(90,Buffer) '# ',(qvec(ii,i),ii=1,RDim)
	  END DO
          IF (.not.projection) THEN
	    WRITE(90,'(a)') '# limits and steps of t:'
	    WRITE(90,'(a)') '#   start     end     step'
            DO i=1,SDim-RDim
	      write(90,'(''# '',3f9.5)') (TList(ii,i),ii=1,3)
	    END DO
          ELSE
	    WRITE(90,'(a)') '# The density will be projected onto 3D before analysis,\.'
	  END IF
	END IF
	IF (com) THEN
	  WRITE(90,'(a)', ADVANCE='NO') '# The center of charge and total charge of atomic basins will be computed using '
	  SELECT CASE (basinmode)
	    CASE(ongridmode)
              WRITE(90,'(a)') 'an on-grid algorithm (fast, but possibly inaccurate)'
	    CASE(neargridmode)
              WRITE(90,'(a,i2,a)') 'a near-grid algorithm with ',EdgeRemovalCycles,' cycles of edge removal.'
	    CASE(pedanticmode)
              WRITE(90,'(a,i2,a)') 'a pedantic algorithm (very slow, but the most accurate).'
	  END SELECT
	  WRITE(90,'(2(a,f7.3))') '# The density in the incommensurate directions is calculated from',-addborder,' to',1.+addborder
	  WRITE(90,'(a,f7.3,a)') '# The density limit for the pixel to be accepted in the com calculation is ',chlim,'*Rho(max)'
	ELSE
	  WRITE(90,'(a)') '# The center of charge and total charge of atomic basins will not be computed.'
	END IF
        WRITE(90,'(2a)') '# The scale of the output coordinates will be ',scale
        IF (maxformat(1:idel(maxformat))=='atoms') THEN
	  write(90,'(a,f6.3,a,f9.3,a)') '# Positions of following atoms +- ',tolerance, &
	                       ' and with density over', plim, ' will be searched for maxima at each t'
          IF (position=='relative') THEN
            write(90,'(a)') '# Coordinates are related to the input coordinates of the atoms:'
	  ELSE IF (position=='absolute') THEN
            write(90,'(a)') '# Coordinates are related to the origin of the unit cell.'
	  END IF
          Buffer='(a2,a8, f9.5)'
          WRITE(Buffer(8:8),'(i1)') RDim
          DO i=1,AtomNr
            write(90,Buffer) '#   ',Atoms(i)%Name,(Atoms(i)%Coo(ii),ii=1,RDim)
	  END DO
          write(90,'()')
	ELSE IF (maxformat/='none') THEN
	  write(90,'(a,f9.3,a)') '# All maxima with density above ',plim,' will be printed for each t.'
          write(90,'()')
        END IF
        IF (cpformat(1:idel(cpformat))=='list') THEN
	  write(90,'(a,f6.3,a,f9.3,a)') '# Following positions +- ',tolerance, &
	                       ' and with density over', plim, ' will be searched for crtitical points at each t'
          IF (position=='relative') THEN
            write(90,'(a)') '# Coordinates are related to the input coordinates of the critical points:'
	  ELSE IF (position=='absolute') THEN
            write(90,'(a)') '# Coordinates are related to the origin of the unit cell.'
	  END IF
          Buffer='(a2,a8, f9.5)'
          WRITE(Buffer(8:8),'(i1)') RDim
          DO i=1,CPlistNr
            write(90,Buffer) '#   ',CPlist(i)%Name,(Cplist(i)%Coo(ii),ii=1,RDim)
	  END DO
	ELSE IF (cpformat/='none') THEN
	  write(90,'(a,f9.3,a)') '# All critical points with density above ',plim,' will be printed for each t.'
        END IF
        IF (cpformat/='none') then
	  IF (expandedcp) THEN
	    write(90,'(a)') '# Eigenvectors of the critical points will be listed.'
            write(90,'()')
	  ELSE
	    write(90,'(a)') '# Eigenvectors of the critical points will not be listed.'
            write(90,'()')
	  END IF
	END IF
	
        CLOSE (90)
      END IF
      IF ((maxformat=='all').or.(cpformat=='all')) THEN
        OPEN(UNIT=91,file=outrhofile(1:idel(outrhofile))//'.coo',POSITION='APPEND') !we can write data directly to the output file
      ELSE IF (maxformat=='atoms') THEN
        OPEN(UNIT=91,file=outrhofile(1:idel(outrhofile))//'.tmp') !we need to write to the temporary file and then rewrite the data into convenient format later
        fullcellpeaks=.true. !We don't know where are the atoms located -> search the whole unit cell
      END IF
      IF (cpformat=='list')THEN
        OPEN(UNIT=92,file=outrhofile(1:idel(outrhofile))//'.tmc') !we need to write to the temporary file and then rewrite the data into convenient format later
      END IF
!     cycle over the additional dimensions according to the t-settings
      DO i=1,SDim-RDim
        t(i)=Tlist(1,i) !Fill actual t with the start point of the intervals
      END DO
50    continue            
        IF (SDim==RDim) THEN !periodic "3D" case, we just need to assign rho to rhor
          !just if we haven't made projection, otherwise we have filled rhor already
	  if (.not.projection) THEN
            rhor=rho
            deallocate(rho)
          ENDIF
          if (maxformat=='atoms') write(91,*) !maintain the format of the temporary file
          if (cpformat=='list') write(92,*) !maintain the format of the temporary file
	  goto 170 ! and we can skip to the determination of the t-section
        ELSE
	  fullcellpeaks=.true. !The symmetry does not work in the t-sections of the modulated density 
	END IF
        write(*,'(a)',ADVANCE='NO') 't= '
        IF ((maxformat=='all').or.(cpformat=='all')) write(91,'(a)',ADVANCE='NO') '# t= '
        DO i=1,SDim-RDim
          write(*,'(f7.3)',ADVANCE='NO') t(i) !list the actual t's on the screen...
          IF ((maxformat/='none').or.(cpformat=='all'))  write(91,'(f7.3)',ADVANCE='NO') t(i) !...and in the output file
          IF (cpformat=='list')  write(92,'(f7.3)',ADVANCE='NO') t(i) !...and in the output file
	END DO
        write(*,*)
        IF ((maxformat/='none').or.(cpformat=='all')) write(91,*)
        IF (cpformat=='list') write(92,*)
!       cycle over the R-Dimensional real space:              
	DO i=1,RDim
          StartCR(i)=-Rim(i) !Fill the "lower front left" and "upper back right" corner of the electron density section considered in the calculation of the maxima and charge
          EndCR(i)=NVoxR(i)+Rim(i)-1
          Dxyz(i)=StartCR(i)
	END DO
        pcnt=0 !counts the pixel which were already searched - just for writing out the progress
        write(*,'(a)') ' Extracting the electron density in real dimensions:'
        call send(-1) !Cljf
	DO
	  pcnt=pcnt+1
!Cljf
          IF (mod(pcnt,int(float(NVoxTotRim)/100.))==0) THEN
	    write(*,'(i3,2a)',ADVANCE='NO') nint(100.*float(pcnt)/float(NVoxTotRim)),'%',char(13) !write out progress
            ipc =  nint(100.*float(pcnt)/float(NVoxTotRim))
            write(*,'(i3,2a)',ADVANCE='NO') ipc,'%',char(13) !write out progress
            call fplivc
            call send(ipc)
	  END IF
	  Call ArrayLocationR(Dxyz,ArrLocR,1) !get array location for current coordinates
!         calculate x4 from t's
          x4=0.0D0
          DO i=1,SDim-RDim
            DO ii=1,RDim
              x4(i)=x4(i)+qvec(ii,i)*DBLE(Dxyz(ii))/DBLE(NVoxR(ii))
            END DO
            x4(i)=(x4(i)+t(i))*NVox(CO(i+RDim))
          END DO
!           fill the srho array
          DO ii=1,SDim-RDim
            Dxyz(ii+RDim)=int(x4(ii)+float(range)/2.)+1-range !This construct assures proper setting of the origin in additional dimensions
          END DO
!         Dxyz now contains correct real space coordinates and origin of the "hypercube" in additional dimensions.
!         The indices must be permuted according to CO to reach the correct points in the superspace density
          DO i=1,SDim
	    Dxyz2(i)=Dxyz(CO(i))
	  END DO
          Cnt1=0
!         And now fill the srho ("spline rho") with corresponding section of the censity in additional dimension(s)
120       Call ArrayLocation(Dxyz2,ArrLoc,1)
          srho(Cnt1)=rho(ArrLoc)
!           move one point further
          Cnt2=RDim+1
140       Dxyz2(CO(cnt2))=Dxyz2(CO(cnt2))+1
          IF (Dxyz2(CO(cnt2))>Dxyz(Cnt2)+range-1) THEN
            Dxyz2(CO(cnt2))=Dxyz(Cnt2)
            cnt2=cnt2+1
            if (cnt2==SDim+1) goto 160
	    goto 140
          END IF      
          Cnt1=Cnt1+1
          goto 120    
160       continue        
            !RhoAtPoint returns the electron density at x4 interpolated using bicubic splines
          CALL RhoAtPoint(SDim-RDim,range,DerOrd,AxOrd,x4,Dxyz(RDim+1),srho,rout,y,y2)
          rhor(ArrLocR)=rout
!           move one point further
          IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz,AxOrd)) EXIT
        END DO
        write(*,*)
!       well, we have the density, now extract the maxima
170     CONTINUE
!       if plimkey='relative', we must find the peak density in the map
        IF (plimkey=='relative') THEN
          dpos=0.0D0
          DO i=1,NVoxTotR
	    dpos=max(dpos,rhor(i))
	  END DO
          plim=plim*dpos !Plim was a relative number up to now, let's make it absolute
	ELSEIF (plimkey=='sigma') THEN
          rhorsigma=0.0
          DO i=0,NVoxTotR-1
            rhorsigma=rhorsigma+(rhor(i))**2
          ENDDO
	  rhorsigma=sqrt(rhorsigma/NVoxTotR)
	  plim=plim*rhorsigma	
	END IF
        IF (maxformat=='none') goto 180 !skip writing maxima
	CALL SearchMaxima
        IF (scale=='angstrom') THEN 
          DO i=1,RDim
            dsc(i)=Cell(i) !set scaling of the output coordinates to angstroms
          END DO
         ELSE
           dsc=1.0D0 !no scaling
        END IF
        IF (maxformat=='atoms') THEN
          IF (position=='relative') THEN 
	    dpos=1.0D0 !allow subtraction of the input atomic coordinates
	  ELSE
	    dpos=0.0D0 !absolute coords in the unit cell
	  END IF	    
	  DO i=1,AtomNr ! repeat over all atoms in the list
! Map the periodic directions onto the unit cell. Aperiodic can't be simply mapped
            DO Cnt1=1,RDim !find out the largest difference between current maximum and current atom
              DO WHILE (Atoms(i)%Coo(Cnt1)>=1.0D0)
                Atoms(i)%Coo(Cnt1)=Atoms(i)%Coo(Cnt1)-1.0D0
              END DO
              DO WHILE (Atoms(i)%Coo(Cnt1)<0.0D0)
                Atoms(i)%Coo(Cnt1)=Atoms(i)%Coo(Cnt1)+1.0D0
              END DO
	    END DO
	    DO ii=1,MaxNr !repeat over all maxima           
              disc=0.0D0
	      cpyes=.false.
	      DO Cnt1=1,RDim !find out the largest difference between current maximum and current atom
	        disc=max(disc,Cell(Cnt1)*abs(Maxima(ii)%Coo(Cnt1)-Atoms(i)%Coo(Cnt1)))
	      END DO
              IF (disc<=tolerance) THEN
                IF (com.and.(Maxima(ii)%cc<chlimlist)) CYCLE
                  IF (maxformat=='atoms')  THEN
                  DO cnt2=1,SDim-RDim
                    x4(cnt2)=0.0D0
		    DO iii=1,RDim
                      x4(cnt2)=x4(cnt2)+qvec(iii,cnt2)*Atoms(i)%Coo(iii)
                    END DO
                    x4(cnt2)=x4(cnt2)+t(cnt2)
                    write(91,'(f7.3)',ADVANCE='NO') x4(cnt2)
	          END DO
		END IF
                IF (com) THEN !write coordinates of peak, center of charge, total charge and peak density
                  write(91,'(4X,3f12.6,''   '',3f12.6,''   '',f11.5,2f12.4)',ADVANCE='NO') &
                   (dsc(Cnt1)*(Maxima(ii)%Coo(Cnt1)-Atoms(i)%Coo(Cnt1)*dpos),Cnt1=1,RDim), & 
                   (dsc(Cnt1)*(Maxima(ii)%Com(Cnt1)-Atoms(i)%Coo(Cnt1)*dpos),Cnt1=1,RDim), & 
		   Maxima(ii)%cc, Maxima(ii)%rho, Maxima(ii)%vol
                ELSE !write only coordinates and peak density
                  write(91,'(4X,3f12.6,''   '',f12.4)',ADVANCE='NO') &
                   (dsc(Cnt1)*(Maxima(ii)%Coo(Cnt1)-Atoms(i)%Coo(Cnt1)*dpos),Cnt1=1,RDim), &
		   Maxima(ii)%rho
 	        END IF
                IF (Maxima(ii)%err) THEN
	          write(91,'(a)') '  !'
	        ELSE
	          write(91,'(a)')
                END IF
		cpyes=.true. !flag that the maximum for given atom has been found
		EXIT
	      END IF
	    END DO
            IF (.not.cpyes) write(91,'(a47)') '                    no maximum found           '
	  END DO
	ELSE !we simply list all maxima
          write(91,'(a)') '#  maxima:'
          IF (com) THEN
	    write(91,'(2a)') '#     x           y           z              x_coc       y_coc       z_coc        ',&
	                    'charge     density        vol.'
          ELSE
            write(91,'(a)') '#     x           y           z             density'
	  END IF
	  DO i=1,MaxNr
            IF (com.and.(Maxima(i)%cc<chlimlist)) CYCLE
            IF (com) THEN !write coordinates of peak, center of charge, total charge and peak density
              write(91,'(3f12.6,''   '',3f12.6,''   '',f11.5,2f12.4)',ADVANCE='NO') &
               (dsc(ii)*Maxima(i)%Coo(ii),ii=1,RDim), &
               (dsc(ii)*Maxima(i)%Com(ii),ii=1,RDim),Maxima(i)%cc,Maxima(i)%rho,Maxima(i)%vol 
	    ELSE !write only coordinates and peak density
              write(91,'(3f12.6,''   '',f12.4)',ADVANCE='NO') &
               (dsc(ii)*Maxima(i)%Coo(ii),ii=1,RDim), Maxima(i)%rho
            END IF
            IF (Maxima(i)%err) THEN
	      write(91,'(a)') '  !'
	    ELSE
	      write(91,'(a)')
            END IF
	  END DO
          write(91,*)
          write(91,*)
	END IF
180     IF (cpformat=='none') goto 190 !skip the search for critical points
!       Search for critical points
        CALL SearchCP
        IF (scale=='angstrom') THEN 
          DO i=1,RDim
            dsc(i)=Cell(i) !set scaling of the output coordinates to angstroms
          END DO
         ELSE
           dsc=1.0D0 !no scaling
        END IF
        IF (cpformat=='list') THEN
          IF (position=='relative') THEN 
	    dpos=1.0D0 !allow subtraction of the input atomic coordinates
	  ELSE
	    dpos=0.0D0 !absolute coords in the unit cell
	  END IF	    
	  DO i=1,CPlistNr ! repeat over all cpoints in the list
	    DO ii=1,CPNr !repeat over all found CP
              disc=0.0D0
	      cpyes=.false.
	      DO Cnt1=1,RDim !find out the largest difference between current cpoint and current atom
	        disc=max(disc,Cell(Cnt1)*abs(Cpoint(ii)%Coo(Cnt1)-CPlist(i)%Coo(Cnt1)))
	      END DO
              IF (disc<=tolerance) THEN
                write(92,'(4X,3f12.6,2X,a9,f12.4,3f14.4)') & 
                 (dsc(Cnt1)*(Cpoint(ii)%Coo(Cnt1)-CPlist(i)%Coo(Cnt1)*dpos),Cnt1=1,RDim), &
                  Cpoint(ii)%type, Cpoint(ii)%rho, (Cpoint(ii)%Evl(Cnt1),Cnt1=1,RDim)
	          cpyes=.true. !flag that the maximum for given atom has been found
                IF (expandedcp) THEN
	          DO Cnt1=1,RDim
	            write(92,'(20X,3f12.6)') (CPoint(ii)%Evc(Cnt1,Cnt2),Cnt2=1,RDim)
	          END DO
	        END IF
                EXIT
	      END IF
	    END DO
            IF (.not.cpyes) write(92,'(a47)') '                no critical point found       '
	  END DO
	ELSE !we simply list all critical points
          write(91,'(a)') '#  critical points:'
          write(91,'(a)') '#     x           y           z         type        density       lambda1       lambda2       lambda3'
	  DO i=1,CPNr                        
            write(91,'(3f12.6,2X,a9,f12.4,3f14.4)') &
             (dsc(ii)*Cpoint(i)%Coo(ii),ii=1,RDim), Cpoint(i)%type, Cpoint(i)%rho,(Cpoint(i)%Evl(ii),ii=1,RDim)
            IF (expandedcp) THEN
	      DO Cnt1=1,RDim
	        write(91,'(20X,3f12.6)') (CPoint(i)%Evc(Cnt1,Cnt2),Cnt2=1,RDim)
	      END DO
	    END IF
	  END DO
          write(91,*)
          write(91,*)
	END IF
190     IF ((writetmap).and.((SDim/=RDim).or.(projection))) THEN !write out current tmap, only for more-dimensional densities
          IF (projection) THEN
            Buffer='_proj.m81'
	  ELSE
            WRITE(Buffer,FMT=filefmt) (t(i),i=1,SDim-RDim) !add values of t to the output filename
          END IF
	  Buffer=outrhofile(1:idel(outrhofile))//Buffer(1:Idel(Buffer))
          write(*,'(2a)') 'Saving the current density to ',Buffer(1:Idel(Buffer)) !show the output filename
          CALL WriteJanaR(Buffer(1:idel(Buffer)),.false.,AllocError) !write the map
        END IF
        IF ((writebasin).and.(com)) THEN
          IF (SDim/=RDim) THEN !nD case has different format then 3D case
            WRITE(Buffer,FMT=filefmt) (t(i),i=1,SDim-RDim)
            Buffer=outrhofile(1:idel(outrhofile))//'_basins'//Buffer(1:Idel(Buffer))
          ELSE ! 3D case
            Buffer=outrhofile(1:idel(outrhofile))//'_basins.m81'
          END IF
          write(*,*) Buffer(1:Idel(Buffer))
          DO i=0,NVoxTotRim-1
            rhor(i)=float(basin(i)) !write values from the array basin (containing numbers of maxima) to rhor
          END DO
          CALL WriteJanaR(Buffer(1:idel(Buffer)),.false.,AllocError) !write the map
        END IF
        IF (wantstructure) THEN
          IF (RDim==SDim) THEN
	    Buffer='.'//OutstrType(1:idel(outstrtype))
	  ELSE
	    WRITE(Buffer,'(''_'',f4.2,a)') t(1),'.'//OutstrType(1:idel(outstrtype))
          END IF
	  Buffer=outstrfile(1:idel(outstrfile))//Buffer(1:Idel(Buffer))
          SELECT CASE (OutStrType)
	  CASE ('m40')
	    write(*,'(2a)') ' Saving the maxima in jana m40 format to file ',Buffer(1:Idel(Buffer)) !show the output filename
	  CASE ('cif')
	    write(*,'(2a)') ' Saving the maxima in cif format to file ',Buffer(1:Idel(Buffer)) !show the output filename
	  CASE ('ins')
	    write(*,'(2a)') ' Saving the maxima in SHELX ins format to file ',Buffer(1:Idel(Buffer)) !show the output filename
          END SELECT
	  CALL WriteStructure(buffer)
          call send(-4) !Cljf	  
        END IF
        IF (Compm40.and.(SDim==RDim)) CALL Comparem40
        IF (SDim==RDim) goto 500
!       move one t further
        Cnt2=1
400     t(cnt2)=t(cnt2)+TList(3,Cnt2)
        IF (t(cnt2)>Tlist(2,Cnt2)+eps) THEN
          t(Cnt2)=TList(1,Cnt2)
          cnt2=cnt2+1
          if (cnt2==SDim-RDim+1) goto 500
          goto 400
        END IF      
      goto 50      
500   CONTINUE
      IF (maxformat=='atoms') THEN ! we have the data in the temporary file, we must rearange them!
        OPEN(UNIT=90,file=outrhofile(1:idel(outrhofile))//'.coo',POSITION='APPEND') !we will append the data in convenient format to the main data file
	DO cnt1=1,AtomNr
          write(90,*)
          write(90,*)
	  write(90,'(2a)') '# ',atoms(cnt1)%name(1:idel(atoms(cnt1)%name))
	  rewind(91) !return back to the beginning of the file
          read(91,'(a)',END=700) buffer ! read t
          write(90,'(a)',ADVANCE='NO') '#'
	  DO i=1,2*idel(Buffer)
	    write(90,'(a1)',ADVANCE='NO') ' '
	  END DO
          IF (com) THEN
	    write(90,'(2a)') '         x           y           z              x_coc       y_coc       z_coc        charge     ',&
	                     'density        vol.'
          ELSE
            write(90,'(a)') '         x           y           z             density'
	  END IF
	  rewind(91) !return back to the beginning of the file
600       read(91,'(a)',END=700) buffer ! read t
          write(90,'(a)',ADVANCE='NO') buffer(1:idel(buffer)) !write t
	  DO cnt2=1,AtomNr !read sequentially the records for atoms...
	    read(91,'(a)') Buffer
	    if (cnt2==cnt1) write(90,'(a)') buffer(1:idel(buffer)) !...and write that one that matches
	  END DO
          GOTO 600
700       continue
        END DO
      END IF
      IF (maxformat/='none') THEN 
        CLOSE (90)
        CLOSE (91)
      END IF
!     And do the same with the critical points
      IF (cpformat=='list') THEN ! we have the data in the temporary file, we must rearange them!
        OPEN(UNIT=90,file=outrhofile(1:idel(outrhofile))//'.coo',POSITION='APPEND') !we will append the data in convenient format to the main data file
	DO cnt1=1,CPlistNr
          write(90,*)
          write(90,*)
	  write(90,'(2a)') '# ',CPlist(cnt1)%name(1:idel(CPlist(cnt1)%name))
          rewind(92)
	  read(92,'(a)',END=701) buffer ! read t
          write(90,'(a)',ADVANCE='NO') '#'
          DO i=1,idel(Buffer)
	    write(90,'(a1)',ADVANCE='NO') ' '
	  END DO
	  write(90,'(a)') '         x           y           z         type        density       lambda1       lambda2       lambda3'
	  rewind(92) !return back to the beginning of the file
651       read(92,'(a)',END=701) buffer ! read t
          write(90,'(a)',ADVANCE='NO') buffer(1:idel(buffer)) !write t
	  DO cnt2=1,CPlistNr !read sequentially the records for atoms...
	    read(92,'(a)') Buffer
            IF (expandedcp) THEN
              DO i=1,RDim
                read(92,'(a)') Evstr(i)%line
              END DO
	    END IF
	    if (cnt2==cnt1) THEN
	      write(90,'(a)') buffer(1:idel(buffer)) !...and write that one that matches
              IF (expandedcp) THEN
	        DO i=1,RDim
	          write(90,'(a)') Evstr(i)%line(1:idel(Evstr(i)%line))
	        END DO
	      END IF
	    end if
	  END DO
          GOTO 651
701       continue
        END DO
      END IF
      IF (cpformat=='list') THEN 
        CLOSE (90)
	CLOSE (92)
      END IF
  
      CALL FinalCleanUp      
      call send(-5) !Cljf
                     
      CONTAINS
        SUBROUTINE SearchMaxima
          IMPLICIT NONE
	  !Sobel operator in 3D: http://www.aravind.ca/cs788h_Final_Project/gradient_estimators.htm
          INTEGER, PARAMETER :: wghtpower=2
	  INTEGER :: ListNr,ActMax,MaxAl, PixelCnt, NbCount,HalfNbCount !number of item in the PixList, number of actual maximum, array location for actual maximum
          INTEGER :: ERLoop, EdgeRefFlag, ErcLoop, curbasin, currefbasin, MaxPixCnt
          DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:) :: voxdist, rhonbhood, DeltaRho
          DOUBLE PRECISION, DIMENSION(1:MaxDim) :: StepxG, Gradvec,GradCor, CurGradCor
	  DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:,:) :: MMatrix, Minverse, RMatrix, MRW
          DOUBLE PRECISION :: MaxRho, MaxGrad, dummy
	  LOGICAL :: pedantic

!	  write(*,*) 'Entering Laplacian'
!	  CALL Laplacian
!zero everything what must be zero
          MaxNr=0
          IF (Com) THEN
            IF (.not.allocated(Diffvec)) THEN !We come here for the first time)
              cnt1=0
              MaxPixCnt=0
              DO i=1,RDim
	        cnt1=max(cnt1,NVoxRim(i))
	        MaxPixCnt=MaxPixCnt+NVoxRim(i) !maximum possible length of the path to maximum
	      END DO
              ALLOCATE (DiffVec(0:NVoxTotRim-1,RDim),STAT=AllocError) !allows to determine borders of particular atomic basin
              IF (AllocError.GT.0) STOP 'Error allocating Diffvec! Too little memory?'
              ALLOCATE (Pixlist(1:MaxPixCnt),STAT=AllocError) ! Stores all pixels found on the way to the maximum
              IF (AllocError.GT.0) STOP 'Error allocating PixList! Too little memory?'
              ALLOCATE (Diff(1:MaxPixCnt,1:RDim),STAT=AllocError) ! Stores differences along the gradient
              IF (AllocError.GT.0) STOP 'Error allocating Diff! Too little memory?'
              ALLOCATE (Basin(0:NVoxTotRim-1),STAT=AllocError) !array containing assignment of the pixel to the atomic basin
              IF (AllocError.GT.0) STOP 'Error allocating Basin! Too little memory?'
	    ENDIF
            Basin=0
            DO i=1,MaxMaxNr
              Maxima(i)%cc=0.0
              Maxima(i)%ulc=0.0
              Maxima(i)%err=.false.
              Maxima(i)%vol=0.0D0
              DO ii=1,MaxDim
                Maxima(i)%com(ii)=0.0
              END DO
	    END DO
	  END IF  
          pixelCnt=1
	  DO i=1,RDim !set the starting point of the "cube" (unit cell +- tolerance)
            StartC(i)=-(int(tolerance*NVoxR(i)/Cell(i))+1)
	    EndC(i)=NVoxR(i)+int(tolerance*NVoxR(i)/Cell(i))+1
            Dxyz(i)=StartC(i)
	    AxOrd(i)=i
            PixelCnt=PixelCnt*(NVoxR(i)+int(tolerance*NVoxR(i)/Cell(i))+int(tolerance*NVoxR(i)/Cell(i))+3)
          END DO
          pcnt=0
	  write(*,*) 'Searching for maxima:'
!Cljf
          call send(-2)
	  DO !repeat over all pixels in the "cube"
            pcnt=pcnt+1
	    IF (mod(pcnt,int(float(PixelCnt)/100.))==0) THEN
!Cljf start change
              ipc = nint(100.*float(pcnt)/float(PixelCnt))
              write(*,'(i3,2a)',ADVANCE='NO') ipc,'%',char(13) !write progress
              call fplivc
              call send(ipc)
!Cljf end change
	    END IF
	    Call ArrayLocationR(Dxyz,ArrLoc,1)
            IF (rhor(ArrLoc)<plim) GOTO 1000 !ignore too low (potential) maxima
            cpyes=.true.
!             Cycle over the cube 3x3x3... to find a maximum
            DO i=1,RDim
              StartCR(i)=Dxyz(i)-1
	      EndCR(i)=Dxyz(i)+1
              Dxyz2(i)=StartCR(i)
            END DO
            DO
              Call ArrayLocationR(Dxyz2,Cnt2,1)
	      IF (rhor(ArrLoc)<rhor(Cnt2)) THEN
                cpyes=.false.
                EXIT
              ELSEIF ((rhor(ArrLoc)==rhor(Cnt2)).and.(Arrloc>Cnt2)) THEN !If equal values of different pixels, take only the one with lower ArrLoc
                cpyes=.false.
                EXIT
	      END IF
!              move one point further            
              IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz2,AxOrd)) EXIT
            END DO
	    IF (cpyes) THEN !we have found a maximum, let's save it
              IF (.not.fullcellpeaks) THEN !We want to exclude symmetry equivalent peaks
		succ=.true.
                DO i=1,RDim
		  IF ((Dxyz(i)>=NVox(CO(i))).or.(Dxyz(i)<0)) succ=.false.
                END DO
                IF (.not.succ) goto 1000 !Outside the unit cell, ignore such maximum
                succ2=.true.
                DO i=1,MaxNr
		  DO ii=2,NSym
		    CALL SymOp(ii,Maxima(i)%ICoo,Dxyz2)
                    succ=.false.
                    DO iii=1,RDim
                      DO WHILE (Dxyz2(iii)>=NVox(CO(iii)))
                        Dxyz2(iii)=Dxyz2(iii)-NVox(CO(iii))
                      END DO
                      DO WHILE (Dxyz2(iii)<0)
                        Dxyz2(iii)=Dxyz2(iii)+NVox(CO(iii))
                      END DO
                      IF (abs(Dxyz(iii)-Dxyz2(iii))>1) succ=.true.
                    END DO
		    IF (.not.succ) EXIT
		  END DO
		  IF (.not.succ) THEN !Another equivalent maximum already exists
		    Maxima(i)%mult=Maxima(i)%mult+1
		    succ2=.false.
		  END IF
		END DO
                IF (.not.succ2) GOTO 1000
	      ENDIF
	      !We come here only if the maximum is really new and not symmetricaly equivalent to another maximum
	      CALL GetPeak(Dxyz,pt,rout,.true.) !get the exact position and density of the maximum
              IF ((pt(1)>-1.0E29).and.(rout>plim)) THEN
                MaxNr=MaxNr+1
	        IF (MaxNr>MaxMaxNr-1) THEN !
	          write(*,'(A,I4,A)') 'Warning: The maximum number of ',MaxMaxNr,' maxima in one t-map exceeded.'
	          write(*,'(A)')      'Some maxima have been ommited. Try increasing the density treshold (keyword plim)'
	          write(*,'(A)')      'or increasing the parameter MaxMaxNr in module GlobalDefinition and recompiling the program.'
                  IF (com) THEN 
		    write(*,'(A)')      'Charges and centers of charge will not be calculated.'
                    com=.false.
                  END IF
                  MaxNr=MaxNr-1
                  RETURN
	        END IF
	        DO i=1,RDim
	          Maxima(MaxNr)%ICoo(i)=Dxyz(i)
	        END DO
                  DO i=1,RDim
	            Maxima(MaxNr)%Coo(i)=pt(i)/DBLE(NVoxR(i))
	          END DO
                Maxima(MaxNr)%rho=rout
                Maxima(MaxNr)%err=.false.
	        Maxima(MaxNr)%al=ArrLoc
                Maxima(MaxNr)%mult=1
                IF (com) THEN !assign the number of maximum to the pixel of maximum in the Basin array
	        ! only if not assigned yet, otherwise we have two equivalent maxima (e.g. 0,0,0 and 1,0,0)
		  IF (Basin(Arrloc)==0) THEN
		    DO Cnt1=1,RDim
		      Diffvec(ArrLoc,Cnt1)=0 !Difference of the maximum to itself is zero
		    END DO
		    Basin(Arrloc)=MaxNr
                    !We have to check, if a neighboring pixel does not have the same value. In that case assign the same basin number
                    StartCR=Maxima(MaxNr)%Icoo-1
                    EndCR=Maxima(MaxNr)%Icoo+1
                    Dxyz2=StartCR
                    DO
                      Call ArrayLocationR(Dxyz2,Cnt2,1)
                      IF (rhor(ArrLoc)==rhor(Cnt2)) Basin(Cnt2)=MaxNr
                      IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz2,AxOrd)) EXIT
                    END DO
		  END IF
	        END IF
	      END IF
	    END IF
1000        IF (.not.NextCoo(RDim,StartC,EndC,Dxyz,AxOrd)) EXIT !if .true., returns next Dxyz 
          END DO
          write(*,*)
1010      IF (com) THEN !we want to calculate the center of charge and total charge
	    !Fill the voxdist array and the MRW matrix to prepare the gradient calculation
            NBCount=3**RDim
	    HalfNBCount=3**RDim/2
            allocate(VoxDist(1:NBCount))
	    allocate(RhoNBHood(1:NBCount))
	    allocate(DeltaRho(1:HalfNbCount))
	    allocate(MRW(1:RDim,1:HalfNbCount))
	    allocate(Mmatrix(1:RDim,1:RDim))
	    allocate(Minverse(1:RDim,1:RDim))
            IF (((basinmode==neargridmode).and.(edgeremovalcycles>0)).or.(basinmode==pedanticmode)) THEN
              ALLOCATE (RefBasin(0:NVoxTotRim-1),STAT=AllocError)
              IF (AllocError.GT.0) STOP 'Error allocating RefBasin! Too little memory?'
	    ENDIF

            StartCR=-1; EndCR=1; Dxyz=StartCR
	    i=0
	    DO
              i=i+1
	      if (all(Dxyz(1:3)==0)) THEN
                voxdist(i)=1.0
	      else
	        StepxG(1:RDim)=matmul(dble(Dxyz(1:RDim))/NVox,G)
                voxdist(i)=sqrt(sum((dble(Dxyz(1:RDim))/NVox)*StepxG(1:RDim)))
	      endif
              IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz,AxOrd)) EXIT
            END DO
            
	    !Calculate the matrix G_grid^-1M^-1
	    MMatrix=0.0
	    StartCR=-1; EndCR=1; Dxyz=StartCR
	    iii=0
	    DO
	      iii=iii+1
	      DO i=1,RDim
	        DO ii=1,RDim
		  Mmatrix(i,ii)=MMatrix(i,ii)+dble(Dxyz(i)*Dxyz(ii))/(NVox(i)*NVox(ii)*voxdist(iii)**wghtpower)
	        ENDDO
	      ENDDO
              IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz,AxOrd)) EXIT
            ENDDO	    
	    call invert(MMatrix, Minverse,RDim)	    
	    Minverse=matmul(Gi,Minverse) !Minverse now contains G_grid^-1 M^-1
            !Calculate The matrix W'
	    StartCR=-1; EndCR=1; Dxyz=0; Dxyz(1)=1 !We start from the center of the cube, not from the corner, to obtain only the "positive" vectors
	    iii=0; 
	    DO
	      iii=iii+1
              MRW(1:3,iii)=Dxyz(1:3)/(voxdist(iii+HalfNbCount+1)**wghtpower) !MRW now contains W'
              IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz,AxOrd)) EXIT
            ENDDO	    
	    MRW=matmul(Minverse,MRW)	    

            write(*,'(a)') ' Integrating charge of the atoms: '
!Cljf
            call send(-3)
	    IF (basinmode==ongridmode) THEN
	      ongridstep=.true.; defaultongridstep=.true.; EdgeRefFlag=1; pedantic=.false.
	    ELSEIF (basinmode==neargridmode) THEN
	      ongridstep=.false.; defaultongridstep=.false.; pedantic=.false.
	      IF (EdgeRemovalCycles==0) THEN
	        EdgeRefFlag=1
	      ELSE
	        EdgeRefFlag=2
	      ENDIF
	    ELSEif (basinmode==pedanticmode) THEN
	      ongridstep=.false.; defaultongridstep=.false.; EdgeRefFlag=1; pedantic=.true.
              RefBasin=Basin
	    ENDIF
            DO erloop=1,EdgeRefFlag              
              StartC(1:RDim)=-Rim(1:RDim)+Aper(1:RDim)
	      EndC(1:RDim)=NVoxR(1:RDim)-1+Rim(1:RDim)-Aper(1:RDim)
              DxyzCur(1:RDim)=StartC(1:RDim)
              HadLoop=.false.
              pcnt=0
	      DO !repeat over all pixels
                pcnt=pcnt+1
	        IF (mod(pcnt,int(float(NVoxTotRim)/100.))==0) THEN
                  ipc = nint(100.*float(pcnt)/float(NVoxTotRim))
                  write(*,'(i3,2a)',ADVANCE='NO') ipc,'%',char(13) !write progress
                  call fplivc
                  call send(ipc)
                END IF
                Call ArrayLocationR(DxyzCur,ArrLoc,1) !get location in rhor
                if ((Basin(ArrLoc)>0).or.(Basin(ArrLoc)==-1)) GOTO 1300 !this pixel has been already checked -> go to next pixel
	        PixList=0; ListNr=0
                Dxyz=DxyzCur
	        GradCor=0.0
1100            Call ArrayLocationR(Dxyz,ArrLoc,1) !Now follow the gradient from the Dxyzcur to maximum
		DO i=1,RDim
	           IF ((Aper(i)==1).and.((Dxyz(i)<=-Rim(i)).or.(Dxyz(i)>=NVoxR(i)+Rim(i)-1))) THEN
	             ! we have come to the edge of the known density -> discard these pixels from the analysis
                     actmax=-1		     
		     goto 1200
	           END IF
	        END DO
		IF (.not.pedantic) THEN
		  curbasin=Basin(ArrLoc)
		  currefbasin=Basin(Arrloc)
		ELSE
		  curbasin=Basin(ArrLoc)
		  currefbasin=RefBasin(Arrloc)
		ENDIF
                IF (((currefbasin>0).and.(currefbasin<99000)).or.(currefbasin==-1)) THEN !this pixel has been already assigned
	          actmax=currefbasin !make its maximum the active maximum
		  goto 1200
	        ELSEIF (curbasin>99000) THEN !We have been here already before -> we are in a loop!
		  DO i=1,ListNr-1
		    IF (ArrLoc==PixList(i)) THEN
		      ListNr=i-1
		      ongridstep=.true.
                      hadloop=.true.
		      EXIT
		    ENDIF 
                  ENDDO
	        END IF
	        ListNr=ListNr+1 !add the pixel to the PixList
	        IF (Basin(Arrloc)<99000) Basin(Arrloc)=Basin(Arrloc)+100000 !Flag the pixel as part of the currently analyzed trace
                IF (ListNR>MaxPixCnt) THEN
	          write(*,*) 'Error, internal inconsistency! Please contact the author.'
	          STOP
	        ENDIF
                PixList(ListNr)=ArrLoc
!               Cycle over the hypercube 3x3x3... to fill the list of neighboring density values
                StartCR(1:RDim)=Dxyz(1:RDim)-1; EndCR(1:RDim)=Dxyz(1:RDim)+1; Dxyz2(1:RDim)=StartCR(1:RDim)
                ii=0
	        DO
	          ii=ii+1
                  Call ArrayLocationR(Dxyz2,Cnt2,1)
	          RhoNBHood(ii)=rhor(Cnt2)
                  IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz2,AxOrd)) EXIT
                END DO	      
	        IF (maxval(RhoNBhood)==RhoNBhood(14)) THEN
	          GradVec=0.0
		  MaxAl=ArrLoc
	        ELSEIF (ongridstep) THEN
	          MaxGrad=-1000000.
                  StartCR=Dxyz-1; EndCR=Dxyz+1; Dxyz2=StartCR
                  MaxGrad=-100000.0
	          ii=0
	          DO
	            ii=ii+1
                    Call ArrayLocationR(Dxyz2,Cnt2,1)
	            grad=((rhor(Cnt2)-rhor(ArrLoc))/voxdist(ii))
                    IF (grad>maxgrad) THEN
                      MaxGrad=grad !Remember the highest gradient
                      Dxyz3=Dxyz2 !Remember the coordinates of the highest density
                      MaxAl=Cnt2
                    END IF
!                    move one point further            
                    IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz2,AxOrd)) EXIT
                  ENDDO		
		  GradCor=0.0
		  IF (Basin(MaxAl)<99000) ongridstep=defaultongridstep !Return to the default mode if we are not trapped in a loop
	        ELSE
		  !Prepare DeltaRho
		  DO i=1,HalfNbCount
		    DeltaRho(i)=RhoNBHood(HalfNbCount+1+i)-RhoNBHood(HalfNbCount+1-i)
		  ENDDO
                  GradVec(1:RDim)=matmul(MRW,DeltaRho)	!Calculate the gradient
                  dummy=maxval(abs(GradVec(1:RDim)))	!Get the largest component
                  if (dummy<eps) THEN !This is a minimum in  special position. Disgusting! We have to get out of it in an arbitrary direction
                    GradVec=0.0
                    GRadVec(1)=1.
                  ELSE
	            GradVec(1:3)=GradVec(1:3)/dummy		!Normalize so that the latgest component is 1
	          ENDIF
                  CurGradCor(1:3)=GradCor(1:3)+GradVec(1:3) !Add the current gradient to the sum of previous corrections
                  Dxyz3(1:3)=nint(dble(Dxyz(1:3))+CurGradCor(1:3)) !Find the voxel nearest to the tip of the corrected normalized gradient
		  GradCor(1:3)=dble(Dxyz(1:3)-Dxyz3(1:3))+CurGradCor(1:3) !Calculate the new correction
	          Call ArrayLocation(Dxyz3,MaxAl, 1)
	        ENDIF

                IF (MaxAl==ArrLoc) THEN ! The current pixel is a maximum
		  IF (Basin(ArrLoc)-100000<=0) THEN
	            ! we have found an "uninteresting" maximum, one that is not inside the basic unit cell or bellow plim
                    actmax=-1
		    !Add it to refbasin, if necessary
                  ELSE
		     actmax=Basin(ArrLoc)-100000 !actmax will be asigned to all pixels in PixList
                     Diff(ListNr,1:RDim)=0
		  END IF
	        ELSE !move to the pixel with higher density and repeat the search
		  DO Cnt1=1,RDim
                    Diff(ListNr,Cnt1)=Dxyz3(Cnt1)-Dxyz(Cnt1)
                  END DO
                  Dxyz=Dxyz3
                  goto 1100
	        END IF
1200            continue ! We have found pixel with non-zero ActMax
                IF (actmax>0) CumDifVec(1:RDim)=DiffVec(Arrloc,1:RDim)		
		DO i=ListNr,1,-1
                  IF (.not.pedantic) THEN
		    basin(PixList(i))=actmax !fill the Basin array
                  ELSE
		    basin(PixList(i))=basin(PixList(i))-100000 !Erase the "on-the-current-track" flag
		  ENDIF
		  IF (actmax>0) THEN !if we have some "accepted" maximum, fill the Mxmbox array appropriately 
	            CALL CoordinatesR(PixList(i),1, Dxyz3)
                    CumDifVec(1:RDim)=CumDifVec(1:RDim)+Diff(i,1:RDim) 
                    IF (.not.pedantic) DiffVec(Pixlist(i),1:RDim)=CumDifVec(1:RDim)
                  END IF
	        END DO
                IF (pedantic) THEN
		  basin(Pixlist(1))=actmax
		  IF (actmax>0) DiffVec(Pixlist(1),1:RDim)=CumDifVec(1:RDim)
		ENDIF
		IF (HadLoop) THEN !Set to zero all remaining labels of touched pixels
                  where (Basin>99000) Basin=Basin-100000
		  HadLoop=.false.
	        ENDIF
!		IF (any(Basin>99000)) THEN
!		  write(*,*) 'Neco nevynulovaneho:'
!		  DO i=1,ListNr
!		    write(*,*) PixList(i),Basin(PixList(i)), pedantic
!		  ENDDO
!		  read(*,*)
!		ENDIF
1300            IF (.not.NextCoo(RDim,StartC,EndC,DxyzCur,AxOrd)) EXIT  !cycle to the next non-tested pixel
	      END DO !End of the scan over all pixels
         
              IF ((EdgeRefFlag==2).and.(erloop==1)) THEN !Set to zero all pixels on the border between basins
                write(*,*) 'Erasing the voxel assignment at the borders between the basins:'
		RefBasin=Basin
                pcnt=0		
		DO ercloop=1,EdgeRemovalCycles
		  StartC(1:RDim)=-Rim(1:RDim)+Aper(1:RDim)
	          EndC(1:RDim)=NVoxR(1:RDim)-1+Rim(1:RDim)-Aper(1:RDim)
                  Dxyz=StartC		
		  DO
                    pcnt=pcnt+1
	            IF (mod(pcnt,int(float(NVoxTotRim*EdgeRemovalCycles)/100.))==0) THEN
                      ipc = nint(100.*float(pcnt)/float(NVoxTotRim*EdgeRemovalCycles))
                      write(*,'(i3,2a)',ADVANCE='NO') ipc,'%',char(13) !write progress
                      call fplivc
                      call send(ipc)
                    END IF
                    Call ArrayLocationR(Dxyz,ArrLoc,1) !get location in rhor
		    !For every pixel check the neighborhood, and if any surrounding pixel is different, set this to zero
                    StartCR=Dxyz-1; EndCR=Dxyz+1; Dxyz2=StartCR
                    ii=0
                    DO
	              ii=ii+1
                      Call ArrayLocationR(Dxyz2,Cnt2,1)
	              IF (RefBasin(Cnt2)/=RefBasin(Arrloc)) Basin(ArrLoc)=0
                      IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz2,AxOrd)) EXIT
                    ENDDO
                    IF (.not.NextCoo(RDim,StartC,EndC,Dxyz,AxOrd)) EXIT
		  ENDDO
		  RefBasin=Basin
                  pedantic=.true.
                ENDDO
                !Now assing the positions of the maxima again, to make sure that they were not erased
		DO i=1,MaxNr
                  IF (Basin(Maxima(i)%al)==0) THEN
		    Basin(Maxima(i)%al)=i
                    !We have to check, if a neighboring pixel does not have the same value. In that case assign the same basin number
                    StartCR=Maxima(MaxNr)%Icoo-1
                    EndCR=Maxima(MaxNr)%Icoo+1
                    Dxyz2=StartCR
                    DO
                      Call ArrayLocationR(Dxyz2,Cnt2,1)
                      IF (rhor(ArrLoc)==rhor(Cnt2)) Basin(Cnt2)=i
                      IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz2,AxOrd)) EXIT
                    END DO
		  END IF
	        ENDDO
		RefBasin=Basin
		write(*,*) 'Assigning the voxels at the borders:'
	      ENDIF
	    ENDDO !End of the "edge-refine" loop
!	    !Check the symmetry
!		write(*,*) 'Checking the symmetry:'
!		DO i=1,MaxNr-1
!                  DO ii=i+1,MaxNr
!                    DO iii=2,NSym
!                      CALL SymOp(iii,Maxima(i)%ICoo,Dxyz2)
!	              CALL ArrayLocation(Dxyz2,cnt2,1)
!		      IF (cnt2==Maxima(ii)%al) THEN
!		        where (Basin==ii) Basin=i
!		      ENDIF
!                    ENDDO
!                  ENDDO		  
!		ENDDO
!		StartC(1:RDim)=-Rim(1:RDim)+Aper(1:RDim)
!	        EndC(1:RDim)=NVoxR(1:RDim)-1+Rim(1:RDim)-Aper(1:RDim)
!                Dxyz=StartC
!		pcnt=0
!		DO
!                  Call ArrayLocationR(Dxyz,ArrLoc,1) !get location in rhor
!                  DO ii=2,NSym
!                    CALL SymOp(ii,Dxyz,Dxyz2)
!	            CALL ArrayLocation(Dxyz2,cnt2,1)
!		    IF (Basin(ArrLoc)/=Basin(cnt2)) THEN
!		      write(*,*) Dxyz(1:3), Arrloc, Basin(Arrloc), ':', Dxyz2(1:3), Cnt2, Basin(Cnt2)
!		    ENDIF
!                  ENDDO
!                  IF (.not.NextCoo(RDim,StartC,EndC,Dxyz,AxOrd)) EXIT
!		ENDDO


            deallocate(VoxDist)
	    deallocate(RhoNBHood)
	    deallocate(DeltaRho)
	    deallocate(MRW)
	    deallocate(Mmatrix)
	    deallocate(Minverse)
	    if (allocated(RefBasin)) deallocate(RefBasin)

	    ! OK, we have filled the basin array, now we know exactly what pixel belongs to what basin, we can sum up the charge and calculate the center of charge
            IF (maxformat=='none') RETURN !we are interested only in the atomic basin, not in the maxima
!            Sum up the weighted average of coordinates and at the same time sum up the cummulative charge
            DO cnt1=0,NVoxTotRim-1
              IF (basin(cnt1)<=0) cycle !belongs to "uninteresting maximum"
              CALL CoordinatesR(Maxima(basin(cnt1))%al,1,Dxyz2)
              Maxima(basin(cnt1))%cc=Maxima(basin(cnt1))%cc+max(rhor(Cnt1),0.0D0)
              Maxima(basin(cnt1))%vol=Maxima(basin(cnt1))%vol+1.0D0
              IF (rhor(Cnt1)>chlim*rhor(Maxima(basin(cnt1))%al)) THEN !density is high enough to include it in the com calculation
                ! transform the coordinates to neighboiring cell, if necessary
                DO i=1,RDim
                  Dxyz(i)=Dxyz2(i)-DiffVec(cnt1,i)
                  Maxima(basin(cnt1))%com(i)=Maxima(basin(cnt1))%com(i)+rhor(Cnt1)*DBLE(Dxyz(i))/DBLE(NVoxR(i))
                END DO
                Maxima(basin(cnt1))%ulc=Maxima(basin(cnt1))%ulc+rhor(Cnt1) !sum the charge included in com calculation
              END IF
            END DO            
            dpos=0.0D0
            DO cnt1=1,MaxNr
              IF (Maxima(cnt1)%ulc>eps) THEN
                DO i=1,RDim
                  Maxima(Cnt1)%com(i)=Maxima(Cnt1)%com(i)/Maxima(Cnt1)%ulc
                END DO
	          ELSE
                DO i=1,RDim
                  Maxima(Cnt1)%com(i)=Maxima(Cnt1)%coo(i)
                END DO
	          ENDIF
              Maxima(Cnt1)%cc=Maxima(Cnt1)%cc*cApvx !renormalize it to the real charge
              dpos=max(dpos,Maxima(Cnt1)%cc) !dpos stores now the maximal charge
              Maxima(Cnt1)%ulc=Maxima(Cnt1)%ulc*cApvx
              Maxima(Cnt1)%vol=Maxima(Cnt1)%vol*cApvx
	        END DO
            IF (chlimlistkey=='relative') then
              chlimlist=chlimlistorig*dpos
            ELSE
              chlimlist=chlimlistorig
            ENDIF
            DO cnt1=1,MaxNr
              IF (basin(Maxima(cnt1)%al)/=cnt1) THEN !maximum equivalent to another maximum
	            !everything is the same, only the coordinates must be correspondingly transformed
                cnt2=basin(Maxima(cnt1)%al)
 	            Maxima(cnt1)%rho=Maxima(cnt2)%rho
	            Maxima(cnt1)%cc=Maxima(cnt2)%cc
	            Maxima(cnt1)%ulc=Maxima(cnt2)%ulc
	            Maxima(cnt1)%err=Maxima(cnt2)%err
                Maxima(cnt1)%vol=Maxima(cnt2)%vol
                DO i=1,RDim
		          Maxima(cnt1)%com(i)=Maxima(cnt2)%com(i)+Maxima(cnt1)%coo(i)-Maxima(cnt2)%coo(i)
		        END DO
	          END IF
	        END DO
!         assign an atomic type, if composition provided
!	    DO i=1,MaxNr
!	      maxcc=max(maxcc,Maxima(i)%cc)
!            END DO
!	    DO i=1,MaxNr
!              disc=10000.
!              if (ElNr>0) THEN
!                DO ii=1,ElNr
!		  if (compos(ii,1)==1) CYCLE !Ignore hydrogens in the composition
!                  IF (abs(1.-(Maxima(i)%cc/maxcc)/(dble(Compos(ii,1))/dble(MaxZ)))<disc) then
!                    disc=abs(1.-(Maxima(i)%cc/maxcc)/(dble(Compos(ii,1))/dble(MaxZ)))
!	            Maxima(i)%an=ii
!                  END IF
!                END DO	  
!              else
!                Maxima(i)%an=1
!              END IF
!            END DO
          END IF
	END SUBROUTINE SearchMaxima

        SUBROUTINE SearchCP ! Locate critical points of the density
        INTEGER :: PixelCnt,LoopCnt,Loop
!       Create a map that contains the length of the gradient of rhor
! Combined handling of creating the whole map and only sections around the points from the list causes a bit of mess
        IF (CPformat=='list') THEN
	  LoopCnt=CPlistNr
	ELSE
	  LoopCnt=1
	END IF
	write(*,*) 'Calculating the gradient map:'
        DO Loop=1,LoopCnt
          PixelCnt=LoopCnt
	  DO i=1,RDim !set the starting point of the "cube" (unit cell +- tolerance)
            IF (CPformat=='list') THEN
              StartC(i)=int(CPlist(Loop)%Coo(i)*dble(NVoxR(i)))-(int(tolerance*NVoxR(i)/Cell(i))+1)
              EndC(i)=int(CPlist(Loop)%Coo(i)*dble(NVoxR(i)))+int(tolerance*NVoxR(i)/Cell(i))+2	    
              PixelCnt=PixelCnt*(int(tolerance*NVoxR(i)/Cell(i))+int(tolerance*NVoxR(i)/Cell(i))+4)
	    ELSE          
              StartC(i)=-(int(tolerance*NVoxR(i)/Cell(i))+1)
              EndC(i)=NVoxR(i)+int(tolerance*NVoxR(i)/Cell(i))+1
              PixelCnt=PixelCnt*(NVoxR(i)+int(tolerance*NVoxR(i)/Cell(i))+int(tolerance*NVoxR(i)/Cell(i))+3)
            END IF
            Dxyz(i)=StartC(i)
            AxOrd(i)=i
	  END DO
          pcnt=0
	  DO !repeat over all pixels in the "cube"
            pcnt=pcnt+1
!Cljf
	    IF (mod(pcnt,int(float(PixelCnt)/100.))==0) THEN
              ipc = nint(100.*float(pcnt)/float(PixelCnt)) 
              write(*,'(i3,2a)',ADVANCE='NO') ipc,'%',char(13) !write progress
              call fplivc
              call send(ipc)
	    END IF
	    
	    Call ArrayLocationR(Dxyz,ArrLoc,1)
            DO i=1,3
	      pt(i)=DBLE(Dxyz(i))
	    END DO
	    succ=.true.
	    IF (.not.succ) cycle
!       we calculate the gradient at the point ArrLoc
            Gr=0.0D0
!            call FillRhoCut(AxOrd,Orig,RDim) !fill just one row of rhoalong the actual axis to get the first derivative
            DO i=1,3
              DO cnt1=1,RDim
	        AxOrd(cnt1)=mod(i+cnt1-1,RDim)
		IF (AxOrd(cnt1)==0) AxOrd(cnt1)=RDim
                Orig(cnt1)=Dxyz(cnt1)
	      END DO
              Orig(AxOrd(1))=int(pt(AxOrd(1))-dble(range)/2.+range)+1-range
              call FillRhoCut(AxOrd,Orig,1) !fill just one row of rho along the actual axis to get the first derivative
!	      DerOrd=0
!             DerOrd(i)=1
	      DerOrd=0
	      DerOrd(AxOrd(1))=1
!	      write(*,*) srho(0:range-1)
              call RhoAtPoint(1,range,DerOrd,AxOrd,pt,orig,srho,rout,y,y2) !y and y2 contain parameters of the spline interpolation along the last axis
!              write(*,*) 'Derivace:',Orig(1:3),AxOrd(1:3),rout
              Gr=Gr+rout**2 !gr will contain square of the gradient length
	    END DO
!	    read(*,*)
            rhogr(ArrLoc)=sqrt(Gr)
	    DO i=1,RDim
              AxOrd(i)=i
	    END DO
            IF (.not.NextCoo(RDim,StartC,EndC,Dxyz,AxOrd)) EXIT !if .true., returns next Dxyz 
          END DO            
        END DO
!       Now search for minima in the RhoGr map and check presence of the critical point at each minimum
        write(*,*) 'Searching for critical points:'
        CPNr=0
        DO Loop=1,LoopCnt	
          PixelCnt=LoopCnt
          pcnt=0
	  DO i=1,RDim !set the starting point of the "cube" (unit cell +- tolerance)
            IF (CPformat=='list') THEN
              StartC(i)=int(CPlist(Loop)%Coo(i)*dble(NVoxR(i)))-int(tolerance*NVoxR(i)/Cell(i))
              EndC(i)=int(CPlist(Loop)%Coo(i)*dble(NVoxR(i)))+int(tolerance*NVoxR(i)/Cell(i))+1
              PixelCnt=PixelCnt*(int(tolerance*NVoxR(i)/Cell(i))+int(tolerance*NVoxR(i)/Cell(i))+2)
	    ELSE          
              StartC(i)=-int(tolerance*NVoxR(i)/Cell(i))
              EndC(i)=NVoxR(i)+int(tolerance*NVoxR(i)/Cell(i))
              PixelCnt=PixelCnt*(NVoxR(i)+int(tolerance*NVoxR(i)/Cell(i))+int(tolerance*NVoxR(i)/Cell(i))+3)
            END IF
            Dxyz(i)=StartC(i)
            AxOrd(i)=i
	  END DO
	  DO !repeat over all pixels in the "cube"
!Cljf
        Call ArrayLocationR(Dxyz,ArrLoc,1)
        IF (rhor(ArrLoc)<plim) GOTO 9989
        pcnt=pcnt+1
	    IF (mod(pcnt,max(int(float(PixelCnt)/100.),1))==0) THEN
              ipc =  nint(100.*float(pcnt)/float(PixelCnt))
              write(*,'(i3,2a)',ADVANCE='NO') ipc,'%',char(13) !write progress
              call fplivc
              call send(ipc)
	    END IF
	    cpyes=.true.
!           Cycle over the cube 3x3x3... to find a minimum
            DO i=1,RDim
              StartCR(i)=Dxyz(i)-1
	      EndCR(i)=Dxyz(i)+1
              Dxyz2(i)=StartCR(i)
            END DO
            DO
              Call ArrayLocationR(Dxyz2,Cnt2,1)
	      IF (rhogr(ArrLoc)>rhogr(Cnt2)) THEN
                cpyes=.false.
                EXIT
              END IF
!              move one point further            
              IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz2,AxOrd)) EXIT
            END DO
            IF (cpyes) THEN
	      CALL GetPeak(Dxyz,pt,rout,.false.) !get the exact position and density of the critical point, using spline interpolation
              IF (pt(1)>-1.0E29) THEN
	        IF (MaxNr>MaxMaxNr-1) THEN !
	          write(*,'(A,I4,A)') 'Warning: The maximum number of ',MaxMaxNr,' critical points in one t-map exceeded.'
	          write(*,'(A)')      'Some critical points have been ommited. Try increasing the parameter MaxMaxNr'
	          write(*,'(A)')      'in module GlobalDefinition and recompiling the program.'
                  RETURN
	        END IF
                CPNr=CPNr+1
		Cnt1=0
	        DO i=1,RDim
	          Cpoint(CPNr)%ICoo(i)=Dxyz(i)
                  Cpoint(CPNr)%Coo(i)=pt(i)/DBLE(NVoxR(i))
                  CPoint(CPNr)%Evl(i)=Evl(i)
		  DO ii=1,RDim
		    CPoint(CPNr)%Evc(i,ii)=Evc(i,ii)
		  END DO
                  IF (Evl(i)>0) Cnt1=Cnt1+1
                END DO
		SELECT CASE (Cnt1)
		CASE (0)
		  Cpoint(CPNr)%type='(3,-3)'
		CASE (1)
		  Cpoint(CPNr)%type='(3,-1)'
		CASE (2)
		  Cpoint(CPNr)%type='(3,+1)'
		CASE (3)
		  Cpoint(CPNr)%type='(3,+3)'
		END SELECT
                CPoint(CPNr)%rho=rout
	      END IF
            END IF
9989        IF (.not.NextCoo(RDim,StartC,EndC,Dxyz,AxOrd)) EXIT !if .true., returns next Dxyz
          END DO
	END DO
	END SUBROUTINE SearchCP
      END

! Working subroutine that prints neighborhood of the point cpt
      SUBROUTINE PrintRho(cpt,spn)
      USE GLOBALDEFINITIONS
      USE GLOBALVARIABLES

      IMPLICIT NONE
      integer :: cpt(3),tmp(MaxDim),spn,ArrLoc,i,ii,iii
      
      write(*,*) cpt
      DO i=-spn,spn
        DO ii=-spn,spn
	  DO iii=-spn,spn
            tmp(1)=cpt(1)+iii
            tmp(2)=cpt(2)+ii
            tmp(3)=cpt(3)+i
            Call ArrayLocationR(tmp,ArrLoc,1)	    
	    write(*,'(f8.4)',ADVANCE='NO') rhor(Arrloc)
	  END DO
	  write(*,*)
	END DO
	write(*,*)
      END DO
      END SUBROUTINE PrintRho

!Cljf
      SUBROUTINE send(i)
! dummy subroutine
      integer :: i

      END SUBROUTINE send
