!****
!****   SUBROUTINES used in the program EDMA and differing from BayMEM
!****
!****   Made on 27.1.03 by Lukas Palatinus
!        Compute location of structure factors in the REAL-dimension unit-cell

!     ! Calculate a Fourier transform of the density
!     !******************************************************************
!           SUBROUTINE Laplacian
!     !******************************************************************
!
!           USE GlobalDefinitions
!           USE GlobalVariables
!
!           IMPLICIT NONE
!           integer :: i,FDir, Err
!           integer, dimension(1:RDim) :: hkl
!           double precision :: laplacefact
!           character*132 :: tmprho
!
!           ! Allocate SF
!           IF (ALLOCATED(SF)) DEALLOCATE(SF)
!           write(*,*) 'NVoxTotRim:',NVoxTotRim
!           ALLOCATE(SF(0:2*NVoxTotrim-1))
!           write(*,*) 'Initializing FFT'
!           CALL FFTInit
!           write(*,*) capvx
!           ! Copy rho to SF
!           SF=0.0D0
!           DO i=0,NVoxTotrim-1
!             SF(2*i)=rhor(i)*capvx
!           ENDDO
!           ! Call FFT
!           FDir=1
!           CALL FFT(FDir)
!           write(*,*) NVoxRim
!           !Modify structure factors
!           do i=0,NVoxTotRim-1
!             call CoordinatesRhkl(i,1,hkl)
!             laplacefact=4*pi*pi*sum((hkl/Cell(1:3))**2)
!             SF(2*i)=SF(2*i)*laplacefact
!             SF(2*i+1)=SF(2*i+1)*laplacefact
!     !        IF (any(hkl==NVox/2)) THEN
!     !          SF(2*i)=0.0
!     !          SF(2*i+1)=0.0
!     !        ENDIF
!      !       write(*,*) hkl,SF(2*i),SF(i+1)
!           enddo
!           ! CALL inverse FFT
!           FDir=-1
!           CALL FFT(FDir)
!           DO i=0,NVoxTotRim-1
!             rhor(i)=SF(2*i)/capvx
!           ENDDO
!           !Save the new density
!           tmprho='test.m81'
!           CALL WriteJanaR(trim(tmprho),.false.,Err)
!           DEALLOCATE(SF)
!           write(*,*) 'Finished'
!           STOP
!           END SUBROUTINE Laplacian

! Check if the symmetry of the density corresponds to that given in the input file
!******************************************************************
      SUBROUTINE CheckSymmetry
!******************************************************************
      
      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      LOGICAL :: succ
      INTEGER :: i,ii,iii
      INTEGER :: Dxyz(MaxDim),Dxyz2(MaxDim)
      
        DO i=1,126 ! 127 is a prime number picked on my own choice
          cnt1=int(float(NVoxTot)*float(i)/127.)
       CALL Coordinates(cnt1,1,Dxyz)
          succ=.true.
       DO ii=2,NSym
            CALL SymOp(ii,Dxyz,Dxyz2)
         CALL ArrayLocation(Dxyz2,cnt2,1)
         if ((rho(cnt1)>0.05).and.(abs(rho(cnt1)-rho(cnt2))/abs(rho(cnt1)))>0.01) THEN
!           write(*,*) cnt1,cnt2,rho(cnt1),rho(cnt2)
           succ=.false.
         endif
       ENDDO
       if (.not.succ) EXIT
     ENDDO
     IF (.not.succ) THEN
       write(*,'(a)') 'Warning!!!! The density does not match the symmetry given in the input file!'
       write(*,'(a)') 'The use of the "fullcell no" option might fail!'
     ENDIF
      END SUBROUTINE CheckSymmetry


!******************************************************************
      SUBROUTINE ArraylocationR (iDhkl, oArrLoc, iDelta)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      
      INTEGER :: iDhkl (MaxDim)   ! hkl to compute from
      INTEGER, INTENT (IN) :: iDelta           ! Difference between elements, 2 in case of complex array, 1 else  
      INTEGER :: htmp                   ! Temporary h,k,l... buffer
      INTEGER, INTENT (OUT) :: oArrLoc  ! Computed array location of structure factor
      INTEGER :: ii                     !  Internal loop counters
      
      oArrLoc=0
    
      DO ii=1, RDim
!       Check, if inside the unit cell wih rims - we cannot use the periodicity!
        htmp=iDhkl(ii)+Rim(ii)
        IF (Aper(ii)==1) THEN
          IF ((htmp<0).or.(htmp>NVoxRim(ii)-1)) THEN
            write(*,*) 'NVoxRim:',NVoxRim
         write(*,*) 'Error ArrLoc:',ii, rim,iDhkl
         STOP 'Error in ArrayLocationR - pixel outside the unit cell in an aperiodic direction!'
          END IF
     ELSE
!         Mapping into the unit cell
          htmp=iDhkl(ii)
          DO WHILE (htmp>=NVoxRim(ii))
            htmp=htmp-NVoxRim(ii)
          END DO
          DO WHILE (htmp<0)
            htmp=htmp+NVoxRim(ii)
          END DO
        END IF
     oArrLoc=oArrLoc+htmp*NRim(ii-1)*iDelta
      END DO
      END SUBROUTINE ArraylocationR

!       Calculate Coordinates from array location
!******************************************************************
      SUBROUTINE CoordinatesR (iArrLoc, iDelta, oDxyz)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      
      INTEGER, INTENT (Out) :: oDxyz (MaxDim)    ! xyz to compute
      INTEGER, INTENT (IN) :: iDelta             ! Difference between elements, 2 in case of complex array, 1 else  
      INTEGER, INTENT (IN) :: iArrLoc    ! Array location of location
      INTEGER :: LocRest                 ! Rest of ArrLoc for calculation of next coordinate
      INTEGER :: i                       !  Internal loop counters

      oDxyz=0
      LocRest=iArrLoc/iDelta
!       Reversed step direction
      DO i=RDim, 1, -1
        oDxyz(i)=INT(LocRest/NRim(i-1))
        LocRest=LocRest-oDxyz(i)*NRim(i-1)
        oDxyz(i)=oDxyz(i)-Rim(i)
      END DO
      END SUBROUTINE CoordinatesR

!Subroutine returning the coordinates in the grid for given arraylocation in SF array
!**********************************************
      SUBROUTINE CoordinatesRhkl(iArrLoc, iDelta, ohkl)
!**********************************************
      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE

      INTEGER, INTENT (Out) :: ohkl (MaxDim)    ! hkl to compute
      INTEGER, INTENT (IN) :: iDelta             ! Difference between elements, 2 in case of complex array, 1 else
      INTEGER, INTENT (IN) :: iArrLoc    ! Array location of location
      INTEGER :: LocRest                 ! Rest of ArrLoc for calculation of next coordinate
      INTEGER :: i                       !  Internal loop counters

      ohkl=0
      LocRest=iArrLoc/iDelta
!       Reversed step direction
      DO i=RDim, 1, -1
        ohkl(i)=INT(LocRest/NRim(i-1))
        LocRest=LocRest-ohkl(i)*NRim(i-1)
        IF (ohkl(i)>NVoxRim(i)/2) ohkl(i)=ohkl(i)-NVoxRim(i)
      END DO
      END SUBROUTINE CoordinatesRhkl

! Shift coordinates in the hypercube between StartC and EndC
!*****************************************************************************
      FUNCTION NextCoo(Dim,StartC,EndC,Coo,AxOrd)
!*****************************************************************************
      LOGICAL :: NextCoo
      INTEGER :: Dim,i
      INTEGER, DIMENSION(*) :: StartC,EndC,Coo,AxOrd
        NextCoo=.true.
        i=1
100     Coo(AxOrd(i))=Coo(AxOrd(i))+1
        IF (Coo(AxOrd(i))>EndC(AxOrd(i))) THEN
          Coo(AxOrd(i))=StartC(AxOrd(i))
          i=i+1
          IF (i==Dim+1) THEN
         NextCoo=.false.
         goto 200
       END IF
       goto 100
        END IF      
200     RETURN
      END FUNCTION NextCoo

!*****************************************************************************
      subroutine MatMatMul(M1,M2,Mout,n)
!*****************************************************************************
      IMPLICIT NONE
      INTEGER ::i,ii,iii,n
      DOUBLE PRECISION,DIMENSION(n,n) :: M1,M2,Mout  !matrices
      DO i=1,n
        DO ii=1,n
       Mout(i,ii)=0.0D0
       DO iii=1,n
         Mout(i,ii)=Mout(i,ii)+M1(i,iii)*M2(iii,ii)
       END DO
     END DO
      END DO
      END SUBROUTINE MatMatMul

!*****************************************************************************
      subroutine TrMat(M1,Mout,n)
!*****************************************************************************
      IMPLICIT NONE
      INTEGER ::i,ii,n
      DOUBLE PRECISION,DIMENSION(n,n) :: M1,Mout  !matrices
      DO i=1,n
        DO ii=1,n
          Mout(i,ii)=M1(ii,i)
     END DO
      END DO
      END SUBROUTINE TrMat

!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!c http://www.nauticom.net/www/jdtaft/FortranMatrix.htm
!c    routine to invert a matrix by Gaussian elimination
!c     Ainv=inverse(A)    A(n,n)   Ainv(n,n)
!c     Note: sizes maxed at 64 x 64 - re dimension for larger matrices;
!c           D matrix is extended.
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
       subroutine invert( A, Ainv, n)
       integer :: n,i,j
       double precision, DIMENSION(1:n,1:n) :: A,Ainv
       double precision :: D(1:n,1:2*n), ALPHA, BETA

!  initialize the reduction matrix
       n2 = 2*n
       do i = 1,n
         do j = 1,n
           D(i,j) = A(i,j)
           d(i,n+j) = 0.
         end do
         D(i,n+i) = 1.
       end do

!  do the reduction 
       do i = 1,n
         alpha = D(i,i)
         if(alpha==0.) go to 300
         do j = 1,n2
           D(i,j) = D(i,j)/alpha
         end do
         do k = 1,n
           if((k-i)==0) cycle
           beta = D(k,i)
           do j = 1,n2
             D(k,j) = D(k,j) - beta*D(i,j)
           end do
         end do
       end do

!  copy result into output matrix
       do i = 1,n
         do j = 1,n
           Ainv(i,j) = D(i,j+n)
         end do
       end do

       return
 300   Ainv(1,1)=-1.0D30
       return
       end


! QL diagonalization of matrix / taken from Jana, in Jana probably from Numerical Recipes
!*****************************************************************************
      subroutine ql(a,z,d,n)
!*****************************************************************************
      INTEGER :: n
      DOUBLE PRECISION :: e(20),d(n),a(*),z(*)
      DOUBLE PRECISION, PARAMETER :: tol=.0000001, eps=.0000001
      imd=1-2*mod(n,2)
      nb=n*(n-1)+1
      k=0
      do 20 i=1,n
        do 10 j=1,n
          k=k+1
          if(i.eq.j) then
            z(k)=1.
          else
            z(k)=0.
          endif
10      continue
20    continue
      iz=n*(n+1)/2
      l=n
      do 900 ip=3,n
        i=l
        l=l-1
        h=0.
        iz=iz-i
        do 100 k=1,l
          ind=iz+k
          f=a(ind)
          d(k)=f
          h=h+f**2
100     continue
        if(h.le.tol) then
          e(i)=0.
          h=0.
          go to 890
        endif
        g=-sqrt(h)
        if(f.lt.0.) g=-g
        e(i)=g
        h=h-f*g
        rh=1./h
        dl=f-g
        d(l)=dl
        ind=iz+l
        a(ind)=dl
        f=0.
        ind=1
        do 300 j=1,l
          g=0.
          jk=ind
          do 200 k=1,l
            g=g+a(jk)*d(k)
            id=1
            if(j.le.k) id=k
            jk=jk+id
200       continue
          g=g*rh
          e(j)=g
          f=f+g*d(j)
          ind=ind+j
300     continue
        hh=f*.5*rh
        jk=0
        do 500 j=1,l
          f=d(j)
          g=e(j)-hh*f
          e(j)=g
          do 400 k=1,j
            jk=jk+1
            a(jk)=a(jk)-f*e(k)-g*d(k)
400       continue
500     continue
890     jk=iz+i
        d(i)=a(jk)
        a(jk)=h
900   continue
      d(1)=a(1)
      e(2)=a(2)
      d(2)=a(3)
      b=0.
      f=0.
      do 2000 l=1,n
        ln=l*n-n
        if(l.eq.n) go to 1800
        l1=l+1
        h=eps*(abs(d(l))+abs(e(l1)))
        if(b.lt.h) b=h
        do 1100 m=l1,n
          if(abs(e(m)).le.b) go to 1200
1100    continue
        m=n+1
1200    m=m-1
        if(m.eq.l) go to 1800
        ml=m-l
        mn=m*n
1250    g=d(l)
        el=e(l1)
        dl=(d(l1)-g)/(2.*el)
        h=abs(dl)+sqrt(dl**2+1.)
        if(dl.lt.0.) h=-h
        dl=el/h
        d(l)=dl
        h=g-dl
        do 1300 i=l1,n
          d(i)=d(i)-h
1300    continue
        f=f+h
        dl=d(m)
        c=1.
        s=0.
        ind=mn
        do 1600 jk=1,ml
          ind=ind-n
          i=m-jk
          i1=i+1
          el=e(i1)
          g=c*el
          h=c*dl
          if(abs(dl).ge.abs(el)) then
            c=el/dl
            r=sqrt(c**2+1.)
            e(i+2)=s*dl*r
            s=c/r
            c=1./r
          else
            c=dl/el
            r=sqrt(c**2+1.)
            e(i+2)=s*el*r
            s=1./r
            c=c*s
          endif
          dl=c*d(i)-s*g
          d(i1 )=h+s*(c*g+s*d(i))
          do 1550 k=1,n
            iz=k+ind
            izn=iz-n
            h=z(izn)
            g=z(iz)
            z(izn)=c*h-s*g
            z(iz) =s*h+c*g
1550      continue
1600    continue
        el=s*dl
        e(l1)=el
        d(l)=c*dl
        if(abs(el).gt.b) go to 1250
1800    dl=d(l)+f
        jk=l-1
        iz=ln
        do 1850 k=1,jk
          i=l-k
          iz=iz-n
          if(dl.ge.d(i)) go to 1860
          imd=-imd
          do 1820 j=1,n
            ind=iz+j
            inz=ind+n
            p=z(ind)
            z(ind)=z(inz)
            z(inz)=p
1820      continue
1850    d(i+1)=d(i)
        i=0
1860    d(i+1)=dl
2000  continue
      if(imd.lt.0) then
        do 2020 i=1,n
          z(i)=-z(i)
2020    continue
      endif
      iz=1
      do 2500 i=3,n
        l=i-1
        iz=iz+l
        ind=iz+i
        h=a(ind)
        if(h.ne.0.) then
          do 2400 j=1,nb,n
            jk=j-1
            s=0.
            do 2200 k=1,l
              m=iz+k
              ind=jk+k
              s=s+a(m)*z(ind)
2200        continue
            s=s/h
            do 2300 k=1,l
              ind=jk+k
              m=iz+k
              z(ind)=z(ind)-s*a(m)
2300        continue
2400      continue
        endif
2500  continue
      return
      end

!*********************************************
      SUBROUTINE WriteStructure(filename)
!*********************************************
      USE GLOBALDEFINITIONS
      USE GLOBALVARIABLES

      IMPLICIT NONE
      INTEGER, PARAMETER :: clswdth=1,scalesteps=100, korsteps=20
      INTEGER :: i,ii,iii,kori,idel,dMaxNr,CurrentAtoms,sum,attp
      CHARACTER*132 :: filename
      CHARACTER*8 :: AtomName
      INTEGER, ALLOCATABLE, DIMENSION(:,:) :: AtomCount, StepAtNumber
      Type(MaxType) :: DummyMax
      DOUBLE PRECISION :: smin,smax,kmin, kmax, curkor, sc,minsc,kor,minkor,ressig,dummy,minressig
      LOGICAL :: fromcompos=.false.
      
      Allocate(AtomCount(max(ElNr,1),2))
      AtomCount=0
      CurrentAtoms=0
!     Sort the maxima from the highest to the smalest
      DO i=MaxNr,2,-1
        DO ii=1,i-1
          IF (Maxima(ii)%cc<Maxima(ii+1)%cc) THEN
            DummyMax=Maxima(ii)
            Maxima(ii)=Maxima(ii+1)
            Maxima(ii+1)=DummyMax
          ENDIF
        ENDDO
      ENDDO

! First count the number of independent atoms:
      dMaxNr=0
      IF (TotalAtoms<0) THEN !Flag for composition-driven atom counts
        TotalAtoms=0
        fromcompos=.true.
        DO i=1,ElNr
          IF (Compos(i,1)/=1) TotalAtoms=TotalAtoms-Compos(i,2) !Ignore hydrogens
        ENDDO
        ALLOCATE(StepAtNumber(1:ElNr,1:2))
        DO i=1,ElNr
          StepatNumber(i,2)=i
        ENDDO
        !Sort the elements by element number
        DO i=Elnr,2,-1
          DO ii=1,i-1
            IF (compos(StepatNumber(ii,2),1)<compos(StepatNumber(ii+1,2),1)) THEN
              iii=StepatNumber(ii,2)
              StepatNumber(ii,2)=StepatNumber(ii+1,2)
              StepatNumber(ii+1,2)=iii
            ENDIF
          ENDDO
        ENDDO
        ! Sum up the atomcounts for the individual types
        sum=0
        DO i=1,ElNr
          StepAtNumber(i,1)=sum
          sum=sum+compos(StepAtNumber(i,2),2)
        ENDDO
      ENDIF
      IF (TotalAtoms==0) THEN
     !Just count the number of atoms above the threshold
        DO i=1,MaxNr
          IF (Maxima(i)%cc>chlimlist) dMaxNr=dMaxNr+1
        END DO
      ELSE
        DO i=1,MaxNr
           dMaxNr=dMaxNr+1
           CurrentAtoms=currentatoms+Maxima(i)%mult
           if (currentatoms>=abs(TotalAtoms)) EXIT
        END DO
        currentatoms=0
      ENDIF

      IF ((.not.fromcompos).and.(ElNr>0)) THEN
        smin=0.8*Maxima(1)%cc/float(maxZ)
        smax=1.2*Maxima(1)%cc/float(minZ)
        kmin=0.
        kmax=1./Maxima(1)%cc
!         kmax=0.
!        smin=0.99*Maxima(1)%cc/float(maxZ)
!        smax=1.01*Maxima(1)%cc/float(maxZ)
        minressig=1000000.
        DO i=0,scalesteps
          DO kori=0,korsteps
            sc=smin+float(i)*(smax-smin)/scalesteps
            kor=kmin+float(kori)*(kmax-kmin)/korsteps
            ressig=0
            DO ii=1,dMaxNr
             IF (Maxima(ii)%cc<chlimlist) CYCLE
             IF (TotalAtoms==0) THEN !If we guess the number of atoms, throw away al that have too low charge
               dummy=Maxima(ii)%cc**2
             ELSE
               dummy=1000000.
             ENDIF
             DO iii=1,ElNr
               IF (compos(iii,1)==1) CYCLE
               curkor=kor*(Maxima(1)%cc-Maxima(ii)%cc)+1.0
!                   write(*,*) ii,iii,curkor
               dummy=min(dummy,(Maxima(ii)%cc-sc*compos(iii,1)/curkor)**2)
!                 write(*,*) ii, Maxima(ii)%cc, compos(iii,1), curkor, ((Maxima(ii)%cc-sc*compos(iii,1)/curkor)/Maxima(ii)%cc), dummy
!                   write(*,*) ii, Maxima(ii)%cc, compos(iii,1), curkor, (Maxima(ii)%cc-sc*compos(iii,1)/curkor), dummy
             ENDDO
             ressig=ressig+dummy
           ENDDO
            IF (ressig<minressig) THEN
              minressig=ressig
              minsc=sc
              minkor=kor
            ENDIF
!           write(*,*) i, kori, sc, kor, ressig, minsc, minkor, minressig
!           write(*,*)
          ENDDO
        ENDDO
!        write(*,*) minsc, minkor
        DO ii=1,dMaxNr
          Maxima(ii)%an=0
          IF (Maxima(ii)%cc<chlimlist) CYCLE
          IF (TotalAtoms==0) THEN !If we are guessing the number of atoms, throw away all that have too low charge
            dummy=Maxima(ii)%cc**2
          ELSE
            dummy=1000000.
          ENDIF
          DO iii=1,ElNr
            IF (compos(iii,1)==1) CYCLE
            curkor=minkor*(Maxima(1)%cc-Maxima(ii)%cc)+1.
            IF (((Maxima(ii)%cc-minsc*compos(iii,1)/curkor)**2)<dummy) THEN
              dummy=(Maxima(ii)%cc-minsc*compos(iii,1)/curkor)**2
              Maxima(ii)%an=iii
            ENDIF
          ENDDO
          curkor=minkor*(Maxima(1)%cc-Maxima(ii)%cc)+1.
!         write(75,*) ii, Maxima(ii)%an, Maxima(ii)%cc, compos(max(Maxima(ii)%an,1),1), minsc, minkor, &
!                      minsc*compos(max(Maxima(ii)%an,1),1)/curkor
        ENDDO
        IF (TotalAtoms==0) THEN !Exclude from the list the atoms with too low charge
          dMaxNr=0
          DO i=1,MaxNr
            IF ((Maxima(i)%cc>chlimlist).and.(Maxima(i)%an>0)) dMaxNr=dMaxNr+1
          END DO
        ENDIF
      ENDIF

!count once more the real number of independent atoms
      dMaxNr=0
      currentatoms=0
      IF (TotalAtoms>=0) THEN
        DO i=1,MaxNr
         IF (Maxima(i)%an==0) CYCLE
          IF (((TotalAtoms==0).and.(Maxima(i)%cc>chlimlist)).or.((CurrentAtoms<TotalAtoms))) THEN
            dMaxNr=dMaxnr+1
            currentatoms=currentatoms+Maxima(i)%mult
            IF (Currentatoms>=abs(TotalAtoms)) EXIT
          END IF
        END DO
      ELSE
        DO i=1,MaxNr
          DO ii=ElNr,1,-1
            IF (currentatoms>=StepATnumber(ii,1)) THEN
              Attp=StepATnumber(ii,2)
              EXIT
            ENDIF
          ENDDO
          IF (compos(attp,1)==1) EXIT !Ignore hydrogens, and because they are the latest, we can safely exit the whole loop
          dMaxNr=dMaxNr+1
          currentatoms=currentatoms+Maxima(i)%mult
          IF (Currentatoms>=abs(TotalAtoms)) EXIT
        ENDDO
      ENDIF


      open(40,file=filename(1:idel(filename))) 
      if (OutStrType=='ins') open (41,STATUS='SCRATCH')
!Write the header
      call WriteHeader
! Set the default Biso in case none (or zero) was set
      IF (Biso<eps) Biso=DefaultBiso
      currentatoms=0
      IF (TotalAtoms>=0) THEN
        DO i=1,MaxNr
         IF (Maxima(i)%an==0) CYCLE
          IF (((TotalAtoms==0).and.(Maxima(i)%cc>chlimlist)).or.((CurrentAtoms<TotalAtoms))) THEN
            IF (ElNr/=0) THEN
              iii=idel(ElSmb(compos(Maxima(i)%an,1)))
              AtomName(1:iii)=ElSmb(compos(Maxima(i)%an,1))(1:iii)
              AtomCount(Maxima(i)%an,1)=AtomCount(Maxima(i)%an,1)+1
              AtomCount(Maxima(i)%an,2)=AtomCount(Maxima(i)%an,2)+Maxima(i)%mult
              write(AtomName(iii+1:),'(i4)') AtomCount(Maxima(i)%an,1)
              CALL RemoveSpaces(AtomName)
            ELSE
              AtomName(1:iii)='Mx'
              AtomCount(1,1)=AtomCount(1,1)+1
              AtomCount(1,2)=AtomCount(1,2)+Maxima(i)%mult
              write(AtomName(3:),'(i3.3)') AtomCount(1,1)
            END IF
!           write the atom
            CALL WriteAtom(i)
            currentatoms=currentatoms+Maxima(i)%mult
          END IF
        END DO
      ELSE
        DO i=1,MaxNr
          DO ii=ElNr,1,-1
            IF (currentatoms>=StepATnumber(ii,1)) THEN
              Attp=StepATnumber(ii,2)
              EXIT
            ENDIF
          ENDDO
          IF (compos(attp,1)==1) EXIT !Ignore hydrogens, and because they are the latest, we can safely exit the whole loop
          iii=idel(ElSmb(compos(attp,1)))
          AtomName(1:iii)=ElSmb(compos(attp,1))(1:iii)
          AtomCount(attp,1)=AtomCount(attp,1)+1
          Maxima(i)%an=attp
          AtomCount(attp,2)=AtomCount(attp,2)+Maxima(i)%mult
          write(AtomName(iii+1:),'(i3)') AtomCount(attp,1)
          CALL RemoveSpaces(AtomName)
          CALL WriteAtom(i)
          currentatoms=currentatoms+Maxima(i)%mult
          IF (Currentatoms>=abs(TotalAtoms)) EXIT
        ENDDO
      ENDIF
      CALL WriteFooter
      CLOSE(40)
      IF (OutStrType=='ins') close(41)
      Deallocate(AtomCount)

      CONTAINS
        SUBROUTINE WriteHeader
          CHARACTER*(LineLength) :: ModSymStr !Symmetry string modified for CIF or INS purposes
          LOGICAL :: succ,equalop
       REAL :: sum
       DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:,:) :: DummyMat
       DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:) :: DummyVec, DummyVec2
          INTEGER :: LattNr,isum,iabssum,InvOp

       SELECT CASE (OutStrType)
          CASE ('m40')
            WRITE(40,101) dMaxNr
            WRITE(40,102) 1.,(0.,i=1,5)
            WRITE(40,103) 0.
            WRITE(40,102) (0.,i=1,6)
            WRITE(40,102) (0.,i=1,6)
          CASE ('cif')
         write(40,'(a)') 'data_structure_from_EDMA'
            write(40,'(a)') '_symmetry_space_group_name_H-M '//spacegrp(:idel(spacegrp))
         write(40,'(a)') 'loop_'
         write(40,'(a)') ' _symmetry_equiv_pos_site_id'
         write(40,'(a)') ' _symmetry_equiv_pos_as_xyz'
         !symmetry operations come here - generate them from the RMat and TVec
         DO i=1,NSym
           CALL RMat2Str(i,ModSymStr)
              write(40,'(i3,2X,a)') i,ModSymStr(:idel(ModSymStr))
         ENDDO

         write(40,'(a,f10.5)') '_cell_length_a     ',Cell(1)
         write(40,'(a,f10.5)') '_cell_length_b     ',Cell(2)
         write(40,'(a,f10.5)') '_cell_length_c     ',Cell(3)
         write(40,'(a,f10.5)') '_cell_angle_alpha  ',Cell(4)
         write(40,'(a,f10.5)') '_cell_angle_beta   ',Cell(5)
         write(40,'(a,f10.5)') '_cell_angle_gamma  ',Cell(6)
         write(40,'(a,f12.5)') '_cell_volume     ',Cell(7)
         write(40,'(a)') ''
         write(40,'(a)') 'loop_'
         write(40,'(a)') ' _atom_site_label'
         write(40,'(a)') ' _atom_site_type_symbol'
         write(40,'(a)') ' _atom_site_fract_x'
         write(40,'(a)') ' _atom_site_fract_y'
         write(40,'(a)') ' _atom_site_fract_z'
         write(40,'(a)') ' _atom_site_adp_type'
         write(40,'(a)') ' _atom_site_U_iso_or_equiv'
         write(40,'(a)') ' _atom_site_symmetry_multiplicity'
         write(40,'(a)') ' _atom_site_occupancy'
       CASE ('ins')
         write(40,'(a)') 'TITL '//title(:idel(title))
         write(40,'(a,f7.5,6f12.7)') 'CELL ',lambda,(Cell(ii),ii=1,6)
            !Find the proper lattice centering:
         !First make a checksum and control that only 1/2 and 1/3 or 2/3 are present in the centings
            sum=0.0;succ=.true.
         DO i=1,NCenter
           DO ii=1,3
             sum=sum+float(2**(ii-1))*Center(ii,i)
          if ((abs(center(ii,i)-0.0E0)>eps).and.(abs(center(ii,i)-0.5E0)>eps).and. &
                    (abs(center(ii,i)-0.3333333333E0)>eps).and.(abs(center(ii,i)-0.666666666E0)>eps)) succ=.false. 
           ENDDO
         ENDDO
         IF (.not.succ) sum=-100.
         IF (NCenter==1) THEN
           LattNr=-1
         ELSEIF (Ncenter==2) THEN !The only allowed (and possible) is A, B, C and I
            IF (abs(sum-3.5)<eps) THEN
                LattNr=-2 !I-Centering
           ELSEIF (abs(sum-3.0)<eps) THEN
                  LattNr=-5 !A-centering
           ELSEIF (abs(sum-2.5)<eps) THEN
                LattNr=-6 !B-centering
           ELSEIF (abs(sum-1.5)<eps) THEN
                LattNr=-7 !C-centering
           ELSE
             succ=.false.
           ENDIF
         ELSEIF (Ncenter==3) THEN !The only allowed is R obverse
           IF (abs(Center(2,2)-Center(3,2))<eps) THEN
                LattNr=-3 !R-centering obverse
           ELSE
             succ=.false.
           ENDIF
         ELSEIF (Ncenter==4) THEN ! The only allowed is F
           IF (abs(sum-7)<eps) THEN
                LattNr=-4 !F-centering
           ELSE
             succ=.false.
           ENDIF
         ENDIF
         IF(.not.succ) THEN
              write(*,'(a)') 'Sorry, SHELX handles only the standard centerings P, I, F, A, B, C or  R obverse.' 
              write(*,'(a)') 'Your centering is none of them. The list of atoms will still be written, '// &
                          'but the symmetry will not match.'
              write(*,'(a)') 'The lattice will be set to P.' 
           LattNr=-1
            ENDIF
         !Check if we have an inversion center
         DO i=1,NSymPrim
           isum=0;iabssum=0;InvOp=0
           DO ii=1,3
             DO iii=1,3
            isum=isum+RMat(ii,iii,i)
          ENDDO
                iabssum=iabssum+abs(RMat(ii,ii,i))
              ENDDO
              IF ((isum==-3.).and.(iabssum)==3) THEN !Centrosymmetric
                LattNr=-LattNr
             !SHELX accepts inversion only at the origin
             sum=0.0E0
          DO ii=1,3
            sum=sum+abs(TVec(ii,i))
          ENDDO
                IF (sum>eps) THEN
                  write(*,'(a)') 'Sorry, SHELX requires the inversion center to lie at the origin.' 
                  write(*,'(a)') 'Your inversion center does not. The list of atoms will still be written,'// &
                                 'but the symmetry will not match.' 
          ENDIF
             InvOp=i
          EXIT
           ENDIF
         ENDDO
            write(40,'(a,i2)') 'LATT ',LattNr
         ALLOCATE(DummyMat(SDim,SDim))
         ALLOCATE (DummyVec(SDim))
         ALLOCATE (DummyVec2(SDim))
            !Write out the symmetry operations. Without inversion-center related!
         DO i=2,NSymPrim
              succ=.true.
           IF (InvOp>0) THEN !Check, if this one is not related to some of the previous
                CALL ProdOp(RMat(1,1,i),TVec(1,i),RMat(1,1,InvOp),TVec(1,InvOp),DummyMat,DummyVec,SDim)
                DO ii=1,i-1
            DO iii=1,NCenter
              DummyVec2=DummyVec+Center(1:SDim,iii)
              IF (EqualOp(RMat(1,1,ii),TVec(1,ii),DummyMat,DummyVec2,SDim)) succ=.false.
               ENDDO
            IF (.not.succ) EXIT
          ENDDO
              ENDIF
           IF (succ) THEN
             call RMat2Str(i,ModSymStr)
                write(40,'(a)') 'SYMM '//ModSymStr(:idel(ModSymStr)) 
           ENDIF
         ENDDO
         DEALLOCATE(DummyMat)
         DEALLOCATE (DummyVec)
         write(41,'(a)') 'FVAR 1' !Write this to a temporary file
       END SELECT

101       FORMAT(i4,3('    0'))
102       FORMAT(6f9.6)
103       FORMAT(f9.6)
        END SUBROUTINE WriteHeader

     SUBROUTINE WriteFooter
          IMPLICIT NONE
       SELECT CASE (OutStrType)
          CASE ('m40')
          CASE ('cif')
          CASE ('ins')
         ! Composition
         write(40,'(a)',ADVANCE='NO') 'SFAC '
         DO i=1,ElNr
           write(40,'(a)',ADVANCE='NO') ElSmb(Compos(i,1))(1:idel(ElSmb(Compos(i,1))))//' '
         ENDDO
            write(40,'(a)')
         write(40,'(a)',ADVANCE='NO') 'UNIT '
         DO i=1,ElNr
           write(40,'(i4)',ADVANCE='NO') AtomCount(i,2)
         ENDDO
            write(40,'(a)')
         rewind(41)
200         READ(41,'(a)',END=300) Buffer
            WRITE(40,'(a)') Buffer(:idel(Buffer))
            GOTO 200
300         CONTINUE
         write(40,'(a)') 'END'
       END SELECT

     END SUBROUTINE WriteFooter

        SUBROUTINE RMat2Str(i,oText)
       IMPLICIT NONE
       INTEGER :: i
          CHARACTER*(LineLength) :: oText,TrVecStr
          CHARACTER :: cchar,sgnchar
       integer :: tvnom,tvden

           oTExt=''
              DO ii=1,3 !Only the real-space symops make sense here, no modulated cif can yet be produced
          ! First the translation vector
          IF (TVec(ii,i)>eps) THEN
            call DecToFraction(TVec(ii,i),tvnom,tvden)
            write(TrVecStr,'(i2,a,i2)') tvnom,'/',tvden
            oText=oText(:idel(oText))//TrVecStr
          ENDIF
          ! Then the rotational part
          DO iii=1,3
            SELECT CASE (iii)
              CASE (1)
                cchar='x'
              CASE (2)
                cchar='y'
              CASE (3)
                cchar='z'
            END SELECT
            IF (nint(Rmat(ii,iii,i))==1) THEN
              if ((idel(oText)==0).or.(oText(idel(oText):idel(oText))==',')) THEN
                sgnchar=' '
              ELSE
                sgnchar='+'
              ENDIF
              oText=oText(:idel(oText))//sgnchar//cchar
            ELSEIF (nint(Rmat(ii,iii,i))==-1) THEN
              oText=oText(:idel(oText))//'-'//cchar
                  ENDIF
             ENDDO
          if(ii<3) oText=oText(:idel(oText))//','
           ENDDO
              CALL RemoveSpaces(oText)
        END SUBROUTINE RMat2Str 

     SUBROUTINE WriteAtom(Natom)
       INTEGER :: NAtom
          SELECT CASE (OutStrType)
          CASE ('m40')
            if (.not.m40forjana) THEN
               write(40,110) AtomName(1:8),Maxima(NAtom)%an,1,dble(Maxima(NAtom)%mult)/dble(Nsym),(Maxima(NAtom)%Coo(ii),ii=1,RDim)
            write(40,111) Biso/(8.*pi*pi),(0.,ii=1,5)
            else
               write(40,110) AtomName(1:8),Maxima(NAtom)%an,1,dble(Maxima(NAtom)%mult)/dble(Nsym),(Maxima(NAtom)%Coo(ii),ii=1,RDim)
            write(40,'(2f9.4)') Maxima(NAtom)%rho,Maxima(NAtom)%cc
         endif
       CASE ('cif')
         write(40,112) AtomName(1:idel(AtomName)),ElSmb(compos(Maxima(Natom)%an,1)),&
                            (Maxima(NAtom)%Coo(ii),ii=1,RDim),'Uiso',Biso/(8.*pi*pi),Maxima(NAtom)%mult,1.
       CASE ('ins')
         write(41,113) AtomName(1:4),Maxima(Natom)%an,&
                            (Maxima(NAtom)%Coo(ii),ii=1,RDim),10.+dble(Maxima(NAtom)%mult)/dble(Nsym),Biso/(8.*pi*pi)
       END SELECT

110       FORMAT(a8,2i3,'    ',4(f9.6))
111       FORMAT(6f9.6)
112       FORMAT(a,2X,a2,2X,3f10.6,2X,a,f9.6,i3,f4.1)
113       FORMAT(a,i3,5f10.6)

     END SUBROUTINE

      END SUBROUTINE WriteStructure

!Subroutine that makes a product of two symmetry operations
!**************************************************
      SUBROUTINE ProdOp(R1,T1,R2,T2,Rout,Tout,n)
!**************************************************
      real, parameter :: eps=0.00001
      integer :: n,i
      double precision, dimension(1:n,1:n) :: R1,R2,Rout
      double precision, dimension(1:n) :: T1,T2,Tout,temp

      call dmultm(R2,R1,Rout,n,n,n)
      call dmultm(R2,t1,temp,n,n,1)
      Tout=T2+temp
      DO i=1,n
        DO
       IF (Tout(i)<-eps) THEN
         Tout(i)=Tout(i)+1.0E0
       ELSEIF (Tout(i)>(1.E0-eps)) THEN
         Tout(i)=Tout(i)-1.0E0
       ELSE
         EXIT
          ENDIF
        ENDDO
      ENDDO
      END SUBROUTINE ProdOp
      
!Function EqualOp is true if the two input operations are equivalent
!**************************************************
      LOGICAL FUNCTION EqualOp(R1,T1,R2,T2,n)
!**************************************************
      real, parameter :: eps=0.00001
      integer :: i,ii,n
      double precision, dimension(1:n,1:n) :: R1,R2
      double precision, dimension(1:n) :: T1,T2
      logical :: equal
      
      equal=.true.
      DO i=1,n
        IF (abs((abs(T1(i)-T2(i))-nint(abs(T1(i)-T2(i)))))>eps) THEN
       equal=.false.
       EXIT
        ENDIF
     DO ii=1,n
          IF (abs(R1(i,ii)-R2(i,ii))>eps) THEN
         equal=.false.
         EXIT
       ENDIF
     ENDDO
     IF (.not.equal) EXIT
      ENDDO
      EqualOp=equal
      END FUNCTION EqualOp

! A routine that finds a representaton of a number by a fraction
!*********************************************
      SUBROUTINE DecToFraction(InDec,Nom,Den)
        DOUBLE PRECISION :: InDec
     DOUBLE PRECISION, PARAMETER :: Toler=0.00001
     INTEGER :: Nom,Den,LNom,LDen,RNom,RDen

     LNom=0
     LDen=1
     RNom=1
     RDen=1
     DO
       IF (abs(float(LNom)/float(LDen)-InDec)<Toler) THEN
         Nom=LNom
         Den=LDen
         EXIT
       ELSEIF (abs(float(RNom)/float(RDen)-InDec)<Toler) THEN
         Nom=RNom
         Den=RDen
         EXIT
       ELSE
            IF (InDec<(float(LNom+RNom)/float(LDen+RDen))) THEN
           RNom=LNom+RNom
           RDen=LDen+RDen
         ELSE
           LNom=LNom+RNom
           LDen=LDen+RDen
         ENDIF
       ENDIF
     ENDDO
      END SUBROUTINE DecToFraction

!*********************************************

! Take the list of maxima and compare with the atoms in the provided m40
!*********************************************
      SUBROUTINE CompareM40
!*********************************************
      USE GLOBALDEFINITIONS
      USE GLOBALVARIABLES

      IMPLICIT NONE
      REAL, PARAMETER :: CloseDist=0.55, PillowFactor=0.8
      REAL, DIMENSION(MaxDim) :: dhkl,dcoo1,dcoo2,dcoo3
      INTEGER :: i,ii,iii,i1,i2,i3,idel,GMaxNr,CatNr,main,bestgmax,located
      INTEGER :: falseat,goodat
      REAL :: dist, multsc, mindist,sumdist
      Type(MaxType), DIMENSION(MaxMaxNr) :: GMaxima
      Type(MaxType) :: DMaximum
      Type(AtomType),ALLOCATABLE, DIMENSION(:) :: CAtoms
      logical,ALLOCATABLE, DIMENSION(:) :: OccupMax, OccupAt
      GMaxNr=0
      DO i=1,MaxNr
     IF (Maxima(i)%cc>chlimlist) THEN
       GMaxNr=GMaxNr+1
          GMaxima(GMaxNr)=Maxima(i)
     ENDIF
      ENDDO
      ! Bubblesort the maxima by rho
      DO i=GMaxNr,2,-1
        DO ii=i-1,1,-1
       IF (GMaxima(ii)%rho<GMaxima(ii+1)%rho) THEN
         DMaximum=GMaxima(ii)
         GMaxima(ii)=GMaxima(ii+1)
         GMaxima(ii+1)=DMaximum
       ENDIF
     ENDDO
      ENDDO
      !Eliminate the peaks that are too close
      main=0
      DO
        main=main+1
     i=main
     DO
          i=i+1
       ! If current peak is closer than a given limit, kick it out
          mindist=100000.
       DO ii=1,NSym
         DO iii=1,3
           dcoo2(iii)=GMaxima(i)%com(iii)
         ENDDO
         CALL SymOpR(ii,dcoo2,dhkl)
            !Get the distance from the main peak, take into account all possible shfts
         DO i1=-1,1
           DO i2=-1,1
             DO i3=-1,1
            dcoo1(1)=dhkl(1)+float(i1)-GMaxima(main)%com(1)
            dcoo1(2)=dhkl(2)+float(i2)-GMaxima(main)%com(2)
            dcoo1(3)=dhkl(3)+float(i3)-GMaxima(main)%com(3)
                  dist=sqrt(multsc(dcoo1,dcoo1,RDim,.false.))
!            if (dist<0.5) then
!              write(*,*) dcoo2
!              write(*,*) dhkl
!              write(*,*) dcoo1
!              write(*,*) i1,i2,i3
!            endif
!                  write(*,*) (dcoo2(iii),iii=1,3),dist
                  if (dist<mindist) then
                    mindist=dist
                  endif
                ENDDO
           ENDDO
         ENDDO
       ENDDO
       IF (mindist<CloseDist) THEN
         write(*,'(a,3f10.5)') 'Eliminating' ,(GMaxima(i)%com(iii),iii=1,3)
         write(*,'(a,4f10.5)') 'Too close to' ,(GMaxima(main)%com(iii),iii=1,3),mindist
         CALL Shiftlist(i)
       ENDIF
       IF (i>=GMaxNr) EXIT
     ENDDO
     if (main>=(GMaxNr-1)) EXIT
      ENDDO
      ! Now open and read-in the atoms from m40
      OPEN(41,file=compm40file,status='OLD')
      READ(41,*) CAtNr
      !Restrict the number of considered maxima to the number of expected atoms + certain reserve
      dist=pillowfactor*GMaxima(CAtNr)%rho
      DO i=CAtNr,GMaxNr
        if (GMaxima(i)%rho<dist) THEN
       GMaxNr=i-1
       EXIT
     ENDIF
      ENDDO
!      write(*,*) GMaxNr
!      do i=1,GMaxNr
!       write(*,'(5f10.5)') (GMaxima(i)%com(iii),iii=1,3),GMaxima(i)%cc,GMaxima(i)%rho
!      enddo
      ALLOCATE(CAtoms(1:CAtNr))
      ALLOCATE(OccupMax(1:GMaxNr))
      OccupMax=.false.
      DO i=1,4 !Skip header
        READ(41,*)
      ENDDO
      DO i=1,CAtNr
        READ(41,'(a8,19X,3f9.6)') CAtoms(i)%name,(CAtoms(i)%coo(ii),ii=1,3)
     READ(41,*) !Second line contains only ADPs
      ENDDO
      ALLOCATE(OccupAt(1:CAtNr))
      OccupAt=.false.
      !Now take each atom and try to find its counterpart in GMaxima. Take symmetry into account
      located=0; sumdist=0.0
      occupat=.false.
!      write(*,'(a)') 'Located atoms:'
      DO main=1,CAtNr
        dcoo1=CAtoms(main)%coo
     !Shift the coordinates into the unit cell
        DO i=1,3
       DO WHILE (dcoo1(i)>=1.)
            dcoo1(i)=dcoo1(i)-1.0
          END DO
          DO WHILE (dcoo1(i)<0.0)
            dcoo1(i)=dcoo1(i)+1.0
          END DO
        ENDDO
        mindist=1000000.

        DO i=1,GMaxNr
       if (OccupMax(i)) CYCLE !This maximum has been already assigned
          DO ii=1,NSym
         DO iii=1,3
           dcoo2(iii)=GMaxima(i)%com(iii)
         ENDDO
         CALL SymOpR(ii,dcoo2,dhkl)
            !Get the distance from the main peak, take into account all possible shfts
         DO i1=-1,1
           DO i2=-1,1
             DO i3=-1,1
            dcoo2(1)=dhkl(1)+float(i1)-dcoo1(1)
            dcoo2(2)=dhkl(2)+float(i2)-dcoo1(2)
            dcoo2(3)=dhkl(3)+float(i3)-dcoo1(3)
                  dist=sqrt(multsc(dcoo2,dcoo2,RDim,.false.))
                  if (dist<mindist) then
                    mindist=dist
                 bestgmax=i
                  endif
                ENDDO
           ENDDO
         ENDDO
       ENDDO
     ENDDO
        IF (mindist<=CloseDist) THEN !We have found the corresponding atom in the list of maxima
       OccupMax(bestgmax)=.true.
          OccupAt(main)=.true.
       located=located+1
          sumdist=sumdist+mindist
     ENDIF
      ENDDO
!      write(*,*) GMaxNr
!      DO i=1,GMaxNr
!        write(*,'(4f10.5)') (GMaxima(i)%com(ii),ii=1,3),GMaxima(i)%cc
!      ENDDO
      falseat=0
! The old method of counting good and bad atoms
!      do i=GMaxNr,1,-1
!        if(OccupMax(i)) EXIT
!     falseat=falseat+1
!      enddo
!      falseat=GMaxNr-located-falseat
!      write(*,'(a,i4,a,i4,a,f6.3,a)') 'I could locate',located,' out of ',CAtNr,' atoms within the distance of',CloseDist,' A.'
!      write(*,'(a,f8.5,a,i4,a,i4)') 'Average distance:',sumdist/float(located),' A, number of considered maxima:',GMaxNr, &
!            ', false atoms:',falseat

! The new method
      goodat=0;falseat=0;i3=-100000
      do i=1,GMaxNr
        if (OccupMax(i)) THEN
       goodat=goodat+1
     ELSE
       falseat=falseat+1
        ENDIF
     if ((goodat-falseat)>=i3) THEN
          i3=goodat-falseat
          bestgmax=i
       i1=goodat
       i2=falseat
     ENDIF
      enddo
      falseat=i2
      located=i1
      write(*,'(a,i4,a,i4,a,f6.3,a)') 'I could locate',located,' out of ',CAtNr,' atoms within the distance of',CloseDist,' A.'
      write(*,'(a,f8.5,a,i4,a,i4)') 'Average distance:',sumdist/float(located),' A, number of considered maxima:',GMaxNr, &
            ', false atoms:',falseat
!      do i=1,GMaxNr
!        write(*,'(i3,3f10.5,l4)') i,(GMaxima(i)%com(iii),iii=1,3),OccupMax(i)
!      enddo
      write(*,'(a)',ADVANCE='NO') 'Assigned maxima: '
      do i=1,GMaxNr
        if (OccupMax(i)) THEN 
       write(*,'(a1)',ADVANCE='NO') 'O'
        else
       write(*,'(a1)',ADVANCE='NO') 'X'
     endif
      enddo
      write(*,*)
      write(*,'(a)',ADVANCE='NO') 'Located atoms: '
      do i=1,CAtNr
        if (OccupAt(i)) THEN 
       write(*,'(a1)',ADVANCE='NO') 'O'
        else
       write(*,'(a1)',ADVANCE='NO') 'X'
     endif
      enddo
      write(*,*)
      close(41)
      CONTAINS 
        SUBROUTINE Shiftlist(pos)
          integer :: pos, slloop
          DO slloop=pos+1,GMaxNr
            GMaxima(slloop-1)=GMaxima(slloop)
          ENDDO
       GMaxNr=GMaxNr-1
        END SUBROUTINE Shiftlist
      
      END SUBROUTINE CompareM40

! Matrix multiplication of REAL matrices
!***********************************************************
      subroutine multm(a,b,c,n1,n2,n3)
!***********************************************************
      real :: a(1:n1*n2),b(1:n2*n3),c(1:n1*n3)
      real :: p
      integer :: i,j,k,ik,jk,jkp
      do i=1,n1
        jkp=1
        ik=i
        do k=1,n3
          ij=i
          jk=jkp
          p=0.
          do j=1,n2
            p=p+a(ij)*b(jk)
            ij=ij+n1
            jk=jk+1
          enddo
          c(ik)=p
          ik=ik+n1
          jkp=jkp+n2
        enddo
      enddo
      end subroutine multm

! Matrix multiplication of REAL matrices, DOUBLE PRECISION version
!***********************************************************
      subroutine dmultm(a,b,c,n1,n2,n3)
!***********************************************************
      double precision:: a(1:n1*n2),b(1:n2*n3),c(1:n1*n3)
      double precision :: p
      integer :: i,j,k,ik,jk,jkp
      do i=1,n1
        jkp=1
        ik=i
        do k=1,n3
          ij=i
          jk=jkp
          p=0.
          do j=1,n2
            p=p+a(ij)*b(jk)
            ij=ij+n1
            jk=jk+1
          enddo
          c(ik)=p
          ik=ik+n1
          jkp=jkp+n2
        enddo
      enddo
      end subroutine dmultm

! returns the scalar product of two vectors
!******************************************************************
      REAL FUNCTION multsc(iVec1,iVec2,Dim,iRec)
!******************************************************************
      USE GLOBALDEFINITIONS
      USE GLOBALVARIABLES
      
      IMPLICIT NONE
      INTEGER :: Dim !Dimension of the input vectors, allowed only RDim or SDim
      INTEGER :: l1,l2 !loop1, loop2
      REAL, DIMENSION(Dim) :: iVec1,iVec2 ! Input vectors
      REAL, DIMENSION(3,3) :: RG, RGi
      REAL, DIMENSION(MaxDim) :: dvec1,dvec2,dvec3
      REAL, DIMENSION(1:1) :: mul
      LOGICAL :: iRec !Is the vector in reciprocal space?      
      multsc=0.0E0
      DO l1=1,3
        DO l2=1,3
       RG(l1,l2)=G(l1,l2)
       RGi(l1,l2)=Gi(l1,l2)
     ENDDO
      ENDDO
      IF (iRec) THEN
        DO l1=1,Dim
       dvec1(l1)=ivec1(l1)
       dvec2(l1)=ivec2(l1)
     ENDDO
        IF (Dim>RDim) THEN !Reciprocal lattice vector, to determine the scalar product we have to add qvectors
!         add q-vectors
       DO l1=1,Dim-RDim
         DO l2=1,RDim
           dvec1(l2)=dvec1(l2)+dvec1(l1+RDim)*qvec(l2,l1)
           dvec2(l2)=dvec2(l2)+dvec2(l1+RDim)*qvec(l2,l1)
         END DO
       END DO
     ENDIF
     call multm(dvec1,RGi,dvec3,1,RDim,RDim)
        call multm(dvec3,dvec2,mul,1,RDim,1)
      ELSE
        call multm(ivec1,RG,dvec1,1,RDim,RDim)
     call multm(dvec1,ivec2,mul,1,RDim,1)
      END IF
      multsc=mul(1)
      END FUNCTION multsc

      SUBROUTINE RemoveSpaces(ioText)
      implicit none
      character*(*) :: ioText
      integer :: ii,idel
        ii=0
        DO
          ii=ii+1
       IF (ioText(ii:ii)==' ') THEN
         ioText(ii:)=ioText(ii+1:)
         ii=ii-1
          ENDIF
       IF (ii>=idel(ioText)) EXIT
        ENDDO
      END SUBROUTINE RemoveSpaces

      SUBROUTINE FinalCleanUp
      USE GLOBALDEFINITIONS
      USE GLOBALVARIABLES
      implicit none

      IF  (ALLOCATED (Atoms))     DEALLOCATE (Atoms) 
      IF  (ALLOCATED (CPlist))    DEALLOCATE (CPlist)
      IF  (ALLOCATED (QVec))      DEALLOCATE (QVec)  
      IF  (ALLOCATED (Tlist))     DEALLOCATE (Tlist) 
      IF  (ALLOCATED (CO))        DEALLOCATE (CO)    
      IF  (ALLOCATED (Center))    DEALLOCATE (Center)
      IF  (ALLOCATED (RMat) )     DEALLOCATE (RMat) 
      IF  (ALLOCATED (TVec) )     DEALLOCATE (TVec)  
      IF  (ALLOCATED (RMatI))     DEALLOCATE (RMatI) 
      IF  (ALLOCATED (TVecI))     DEALLOCATE (TVecI) 
      IF  (ALLOCATED (Rim)  )     DEALLOCATE (Rim)   
      IF  (ALLOCATED (Aper))      DEALLOCATE (Aper)  
      IF  (ALLOCATED (NVox))      DEALLOCATE (NVox)  
      IF  (ALLOCATED (NVoxR))      DEALLOCATE (NVoxR)  
      IF  (ALLOCATED (NVoxRim))      DEALLOCATE (NVoxRim)  
      IF  (ALLOCATED (Compos))      DEALLOCATE (Compos)  
      IF  (ALLOCATED (rho))      DEALLOCATE (rho)  
      IF  (ALLOCATED (rhogr))      DEALLOCATE (rhogr)
      IF  (ALLOCATED (srho))      DEALLOCATE (srho)  
      IF  (ALLOCATED (rhor))      DEALLOCATE (rhor)  
      IF  (ALLOCATED (DiffVec))      DEALLOCATE (DiffVec)  
      IF  (ALLOCATED (PixList))      DEALLOCATE (PixList)  
      IF  (ALLOCATED (Diff))      DEALLOCATE (Diff)  
      IF  (ALLOCATED (Basin))      DEALLOCATE (Basin)  

      END SUBROUTINE FinalCleanUp
