!     Last change:  LP   16 Mar 2005   10:43 am
!********************************************
      MODULE Input_module
!********************************************
        integer, parameter :: MaxGroups=10 !Maximum number of groups
        integer, parameter :: MaxKwds=100 !Maximum number of keywords
        integer, parameter :: MaxKwdLen=30 !Maximal length of a keyword and groupname
        integer, parameter :: infunit=10 !Unit number the infile will be connected to
        character*2, parameter :: comments='#!'
        INTEGER :: PassNr=2
        CHARACTER*132 :: InBuffer, infile

        TYPE GrpType
          character*(MaxKwdLen) ::name
	  logical :: status !If true, the keywords of this group ale compulsory, otherwise they are optional
	END TYPE GrpType
        TYPE (GrpType), DIMENSION(MaxGroups) :: KwdGroup
        Integer :: KwdGroupNr=0
	CHARACTER*(MaxKwdLen) :: GrpName !Temporary variable storing the name to be passed to the subroutine KwdGroup
        LOGICAL :: GrpStatus 

        Type KwdType
          character*(MaxKwdLen) ::name, endname
          integer :: group
	  integer :: occ
	END TYPE KwdType
        TYPE (KwdType), DIMENSION(1:MaxKwds) :: Kwd
        Integer :: KwdNr=0
        CHARACTER*(MaxKwdLen) :: KN,KEndN,Grp ! Temporary variables to be passed to the subroutine NewKwd
        CHARACTER*(MaxKwdLen) :: CurKwd
        INTEGER :: CurKwdId !Identification number of the current keyword during rading

        INTEGER :: ErrNr=0 !Number of errors occured during reading

        CONTAINS

	SUBROUTINE NewKwdGroup
           KwdGroupNr=KwdGroupNr+1
	   KwdGroup(KwdGroupNr)%name=GrpName
	   KwdGroup(KwdGroupNr)%status=GrpStatus
	END SUBROUTINE NewKwdGroup

	SUBROUTINE FlagKwdGroup(name,stat)
	   character*(*) :: name
	   logical :: stat
	   integer :: i
           DO i=1,KwdGroupNr
	     IF (KwdGroup(i)%name==name) KwdGroup(i)%status=stat
	   END DO
	END SUBROUTINE FlagKwdGroup

	SUBROUTINE NewKwd(kid)
           integer :: i,kid
	   KwdNr=KwdNr+1
	   kid=KwdNr !keyword identification, the keyword number is massed tot he variable, which can be then referred instead of the number itself
	   Kwd(KwdNr)%name=KN !Name of the (first part of the) keyword
	   Kwd(KwdNr)%endname=KEndN !Optional closing part of the keyword
	   Kwd(KwdNr)%occ=0 !Number of occurences in the iput file
           DO i=1,KwdGroupNr
	     if (Grp==KwdGroup(i)%name) Kwd(KwdNr)%group=i
	   END DO
	END SUBROUTINE NewKwd

        SUBROUTINE ErrorCheck
          IMPLICIT NONE
          integer :: i, idel
	  
	  DO i=1,KwdNr

	    IF (Kwd(i)%occ>1) THEN
	      ErrNr=ErrNr+1
	      WRITE(*,'(3a,i2,a)') 'Error, keyword ',Kwd(i)%name(1:idel(Kwd(i)%name)),' occured ',Kwd(i)%occ,' times in the input file.'
            END IF

            IF ((KwdGroup(Kwd(i)%Group)%status.eqv..true.).and.(Kwd(i)%occ==0)) THEN
	      ErrNr=ErrNr+1
	      WRITE(*,'(3a,i2,a)') 'Error, the compulsory keyword ',Kwd(i)%name(1:idel(Kwd(i)%name)),' is missing in the input file.'
	    ENDIF

            Kwd(i)%occ=0 !Make it zero to be "fresh" for next pass
	  END DO 
	  IF (ErrNr>0) THEN
	    WRITE(*,'(a,i4,a)') 'There were ',ErrNr,' errors in the input file.'
	    WRITE(*,'(a)') 'Please check the input file and try again.'
	    STOP
	  END IF

	END SUBROUTINE ErrorCheck

        SUBROUTINE OneInt(Text,oInt,ErrCode)
          CHARACTER*(*) :: Text
          INTEGER :: oInt,ErrCode
          READ(text,*,IOSTAT=ErrCode) oInt
        END SUBROUTINE OneInt

        SUBROUTINE ArrayInt(Text,oArrInt,iCnt,ErrCode)
          CHARACTER*(*) :: text
          INTEGER :: ErrCode,iCnt,i
	  INTEGER, DIMENSION(*) :: oArrInt
	  READ(text,*,IOSTAT=ErrCode) (oArrInt(i),i=1,iCnt)
        END SUBROUTINE ArrayInt

        SUBROUTINE OneReal(text,oReal,ErrCode)
          CHARACTER*(*) :: text
          INTEGER :: ErrCode
	  DOUBLE PRECISION :: oReal
          READ(text,*,IOSTAT=ErrCode) oReal
        END SUBROUTINE OneReal

        SUBROUTINE ArrayReal(text,oArrReal,iCnt,ErrCode)
          CHARACTER*(*) :: text
          INTEGER :: ErrCode,iCnt,i
	  DOUBLE PRECISION, DIMENSION(1:iCnt) :: oArrReal
          READ(text,*,IOSTAT=ErrCode) (oArrReal(i),i=1,iCnt)
	END SUBROUTINE ArrayReal

        SUBROUTINE YesNo(text,oLogic,ErrCode)
          CHARACTER*(*) :: text
          INTEGER :: ErrCode
	  LOGICAL :: oLogic
	  ErrCode=0
	  IF (FirstWord(text,.false.)=='yes') THEN
	    oLogic=.true.
	  ELSEIF (FirstWord(text,.false.)=='no') THEN
	    oLogic=.false.
          ELSE
	    ErrCode=100
	  END IF
	END SUBROUTINE YesNo

        FUNCTION FirstWord(ioText,cutword)
          IMPLICIT NONE
          CHARACTER*(*) :: ioText
          CHARACTER*132 :: FirstWord
          INTEGER :: wbeg,i,wend
          LOGICAL CutWord
          CHARACTER*1 :: Tab
          CHARACTER*1, PARAMETER :: Space=' '
	  
          FirstWord=' '
          Tab=CHAR(9)
          wbeg=0
	  DO wbeg=1,LEN(ioText)!Skip any leading spaces
            IF ((ioText(wbeg:wbeg)/=Space).AND.(ioText(wbeg:wbeg)/=Tab)) EXIT
          END DO
          DO i=wbeg,LEN(ioText)
            IF ((ioText(i:i)==Space).OR.(ioText(i:i)==Tab)) THEN
              wend=i-1
              FirstWord=ioText(wbeg:wend)
              IF (CutWord) ioText=ioText(wend+1:LEN(ioText))
              EXIT
            END IF
          END DO
	  DO wbeg=1,LEN(ioText)!Skip any leading spaces of the new ioText
	    IF ((ioText(wbeg:wbeg)/=Space).AND.(ioText(wbeg:wbeg)/=Tab)) EXIT
          END DO
          if(wbeg<LEN(ioText)) THEN
	    ioText(1:)=ioText(wbeg:)
          ELSE
	    ioText=' '
	  ENDIF
	END FUNCTION FirstWord

        FUNCTION ToLow(iText)
          IMPLICIT NONE
          CHARACTER*(*) :: iText
          CHARACTER*(len(iText)) :: ToLow
          INTEGER :: i
	  ToLow=''
          DO i=1,len_trim(iText)
	    SELECT CASE (iText(i:i))
	      CASE('A')
	        ToLow(i:i)='a'
	      CASE('B')
	        ToLow(i:i)='b'
	      CASE('C')
	        ToLow(i:i)='c'
	      CASE('D')
	        ToLow(i:i)='d'
	      CASE('E')
	        ToLow(i:i)='e'
	      CASE('F')
	        ToLow(i:i)='f'
	      CASE('G')
	        ToLow(i:i)='g'
	      CASE('H')
	        ToLow(i:i)='h'
	      CASE('I')
	        ToLow(i:i)='i'
	      CASE('J')
	        ToLow(i:i)='j'
	      CASE('K')
	        ToLow(i:i)='k'
	      CASE('L')
	        ToLow(i:i)='l'
	      CASE('M')
	        ToLow(i:i)='m'
	      CASE('N')
	        ToLow(i:i)='n'
	      CASE('O')
	        ToLow(i:i)='o'
	      CASE('P')
	        ToLow(i:i)='p'
	      CASE('Q')
	        ToLow(i:i)='q'
	      CASE('R')
	        ToLow(i:i)='r'
	      CASE('S')
	        ToLow(i:i)='s'
	      CASE('T')
	        ToLow(i:i)='t'
	      CASE('U')
	        ToLow(i:i)='u'
	      CASE('V')
	        ToLow(i:i)='v'
	      CASE('W')
	        ToLow(i:i)='w'
	      CASE('X')
	        ToLow(i:i)='x'
	      CASE('Y')
	        ToLow(i:i)='y'
	      CASE('Z')
	        ToLow(i:i)='z'
	      CASE DEFAULT
	        ToLow(i:i)=iText(i:i)
	    END SELECT
	  ENDDO
	END FUNCTION ToLow

        SUBROUTINE DelComments(ioText)
          INTEGER :: i,ii,idel
	  logical :: CommChar
	  CHARACTER*(*) :: ioText
	  DO i=1,idel(ioText)
	    CommChar=.false.
	    DO ii=1,idel(Comments)
	      IF (ioText(i:i)==Comments(ii:ii)) CommChar=.true.
	    END DO
            IF (CommChar) EXIT
	  END DO
          IF (i>0) THEN
	    ioText=ioText(1:i-1)
	  ELSE
	    ioText=''
	  END IF
	END SUBROUTINE DelComments

        SUBROUTINE NextLine(oText,ErrCd)
	  CHARACTER*(*) :: oText
	  INTEGER :: ErrCd, idel
100       READ(infunit,'(a)', IOSTAT=ErrCd) oText
          CALL DelComments(oText)
	  IF ((Idel(oText)==0).and.(ErrCd==0)) GOTO 100 !This line contains only comments
	END SUBROUTINE NextLine
      END MODULE Input_module

!******************************************
      SUBROUTINE INPUT
!******************************************
      USE Input_module
      USE GlobalDefinitions
      USE GlobalVariables
      
      INTEGER :: i,ii,j,NPass, ErrCode
      INTEGER :: idel !Returns the length of a string without the white spaces
      CHARACTER*(LineLength) :: DBuffer,DummyString
      DOUBLE PRECISION, DIMENSION (:,:), ALLOCATABLE :: DMat
      DOUBLE PRECISION, DIMENSION (:), ALLOCATABLE :: Dvec2
      DOUBLE PRECISION, DIMENSION (1:MaxDim) :: Dvec3
      CHARACTER*1, PARAMETER :: Space=" ", Point="."
      CHARACTER*(FileLngth) :: Inputfile        ! Inputfile :-)
      LOGICAL :: ZeroC

      CALL InitInput

!********* Define the groups of keywords here
      GrpName='must';  GrpStatus=.true. ;CALL NewKwdGroup
      GrpName='opt';   GrpStatus=.false.;CALL NewKwdGroup
      GrpName='4D';    GrpStatus=.false.;CALL NewKwdGroup
      GrpName='4Dnp';  GrpStatus=.false.; CALL NewKwdGroup
      GrpName='atoms'; GrpStatus=.false.;CALL NewKwdGroup
      GrpName='cps';   GrpStatus=.false.;CALL NewKwdGroup
      GrpName='export';GrpStatus=.false.;CALL NewKwdGroup

!********* Define the keywords here
! Template: KN=' ';KEndN=' ',Grp=' '; CALL NewKwd(kid_name)
      KN='title';       KEndN='';           Grp='opt' ; CALL NewKwd(kid_title)
      KN='cell';        KEndN='';           Grp='must'; CALL NewKwd(kid_cell)
      KN='inputfile';   KEndN='';           Grp='must'; CALL NewKwd(kid_infile)
      KN='inputformat'; KEndN='';           Grp='opt';  CALL NewKwd(kid_informat)
      KN='outputbase';  KEndN='';           Grp='opt';  CALL NewKwd(kid_outbase)
      KN='spacegroup';  KEndN='';           Grp='opt';  CALL NewKwd(kid_spacegroup)
      KN='symmetry';    KEndN='endsymmetry';Grp='opt';  CALL NewKwd(kid_sym)
      KN='fullcell';    KEndN='';           Grp='opt';  CALL NewKwd(kid_fullcell)
      KN='centers';     KEndN='endcenters'; Grp='opt';  CALL NewKwd(kid_centers)
      KN='qvectors';    KEndN='endqvectors';Grp='4D';   CALL NewKwd(kid_qvec)
      KN='tlist';       KEndN='endtlist';   Grp='4Dnp'; CALL NewKwd(kid_tlist)
      KN='range';       KEndN='';           Grp='opt';  CALL NewKwd(kid_range)
      KN='tolerance';   KEndN='';           Grp='opt';  CALL NewKwd(kid_tol)
      KN='edmaatoms';   KEndN='endedmaatoms';Grp='atoms';CALL NewKwd(kid_atoms)
      KN='axisorder';   KEndN='';           Grp='opt';  CALL NewKwd(kid_axord)
      KN='plimit';      KEndN='';           Grp='opt';  CALL NewKwd(kid_plim)
      KN='chlimit';     KEndN='';           Grp='opt';  CALL NewKwd(kid_chlim)
      KN='chlimlist';   KEndN='';           Grp='opt';  CALL NewKwd(kid_chlimlist)
      KN='tmap';        KEndN='';           Grp='opt';  CALL NewKwd(kid_tmap)
      KN='basins';      KEndN='';           Grp='opt';  CALL NewKwd(kid_basins)
      KN='maxima';      KEndN='';           Grp='opt';  CALL NewKwd(kid_maxima)
      KN='criticalpoints';KEndN='';         Grp='opt';  CALL NewKwd(kid_cps)
      KN='scale';       KEndN='';           Grp='opt';  CALL NewKwd(kid_scale)
      KN='position';    KEndN='';           Grp='opt';  CALL NewKwd(kid_position)
      KN='centerofcharge';KEndN='';         Grp='opt';  CALL NewKwd(kid_coc)
      KN='partitionmode';KEndN='';          Grp='opt';  CALL NewKwd(kid_basinmode)
      KN='addborder';   KEndN='';           Grp='opt';  CALL NewKwd(kid_addborder)
      KN='cplist';      KEndN='endcplist';  Grp='cps';  CALL NewKwd(kid_cplist)
      KN='writem40';    KEndN='';           Grp='opt';  CALL NewKwd(kid_writem40)
      KN='m40forjana';  KEndN='';           Grp='opt';  CALL NewKwd(kid_m40forjana)
      KN='export';      KEndN='';           Grp='opt';  CALL NewKwd(kid_export)
      KN='comparem40';  KEndN='';           Grp='opt';  CALL NewKwd(kid_comparem40)
      KN='projection';  KEndN='';           Grp='opt';  CALL NewKwd(kid_projection)
      KN='composition'; KEndN='';           Grp='export';CALL NewKwd(kid_composition)
      KN='numberofatoms';KEndN='';          Grp='opt';  CALL NewKwd(kid_totalatoms)
      KN='biso';         KEndN='';          Grp='opt';  CALL NewKwd(kid_biso1)
      KN='Biso';         KEndN='';          Grp='opt';  CALL NewKwd(kid_biso2)
      KN='lambda';          KEndN='';           Grp='opt';  CALL NewKwd(kid_lambda)

      DO NPass=1,PassNr
        ErrNr=0
!       read the input line by line
100     CALL NextLine(InBuffer,ErrCode) !Get next non-empty line of the input, strip out the comments
!        write(*,*) NPass, InBuffer,idel(InBuffer)
        IF (ErrCode>0) GOTO 900 !Some reading error
        IF (ErrCode<0) GOTO 200 !Finished reading the file
        CurKwd=FirstWord(InBuffer,.true.) !.true. means function FirstWord returns InBuffer shortened - without the first word
        CurKwdId=0
        DO i=1,KwdNr
	  IF (CurKwd==Kwd(i)%name) THEN
	    CurKwdId=i                       !The first word of the line matches the ith keyword
            Kwd(i)%occ=Kwd(i)%occ+1          !Add one to the number of occurences of the keyword i
	  END IF
	END DO
	IF (CurKwdId==0) GOTO 100 !No valid keyword was found, proceed to next line
        ErrCode=0
!*** Perform keyword-specific operations ***
        IF (CurKwdId==kid_title) THEN
          title=InBuffer
	              
        ELSEIF (CurKwdId==kid_cell) THEN
          IF (NPass==2) THEN
	    !Expression RDim*(RDim+1)/2 maps RDim to number of cell constants: 1->1, 2->3, 3->6
            CALL ArrayReal(InBuffer,Cell,RDim*(RDim+1)/2,ErrCode)
            IF (ErrCode==0) CALL ReciCell (RDim, CellCon, Cell, RCell)
	  END IF
	  
        ELSEIF (CurKwdId==kid_infile) THEN
          InRhoFile=InBuffer

	ELSEIF (CurKwdId==kid_informat) THEN
	  InRhoType=InBuffer	    
	  IF (InBuffer/='jana'.and.InBuffer/='xplor') THEN
            CALL ErrorMsg('Sorry, '//InBuffer(:idel(InBuffer))//' is not a known density format.')
	    ErrCode=100	    
	  END IF
           
        ELSEIF (CurKwdId==kid_spacegroup) THEN
          SpaceGrp=InBuffer

        ELSEIF (CurKwdId==kid_sym) THEN
          IF (NPass==1) THEN !Count symmetry operators
            NSym=0
	    DO
	      CALL NextLine(InBuffer,ErrCode)
	      IF (ErrCode<0) CALL ErrorMsg('Error, end of file detected during reading.')
	      IF (Inbuffer==Kwd(kid_sym)%endname) EXIT
              NSym=NSym+1
	      NSymPoint=NSym
	    END DO
	  ELSE !In pass Nr 2 read the symmetry operators
	    CALL ReadSymmetry(ErrCode)    
	  ENDIF

        ELSEIF (CurKwdId==kid_centers) THEN
          IF (NPass==1) THEN !Count centering vectors
            NCenter=0
	    FlagZeroC=.false.
	    DO
	      CALL NextLine(InBuffer,ErrCode)
	      IF (ErrCode<0) CALL ErrorMsg('Error, end of file detected during reading.')
	      IF (Inbuffer==Kwd(kid_centers)%endname) EXIT
              NCenter=NCenter+1
!             A check for a zero centering vector must be performed
              i=0
	      DO !Count the number of items in the list - we don't know SDim yet!
		IF (idel(FirstWord(InBuffer,.true.))==0) EXIT
	        i=i+1
	      END DO
	      IF ((i<1).or.(i>MaxDim)) CYCLE
	      CALL CenterVec (InBuffer, i, DVec3) !Get the centering vector
	      ZeroC=.true.
	      DO ii=1,i
	        IF (abs(DVec3(ii))>eps) ZeroC=.false.
	      END DO
	      IF (ZeroC) FlagZeroC=.true.
	    END DO
            IF (.not.FlagZeroC) NCenter=NCenter+1
	  ELSE !In pass Nr 2 read the symmetry operators
            cnt1=1
	    IF (.not.FlagZeroC) THEN !Zero centering vector must be added here
	      DO ii=1,SDim
	        Center(ii,1)=0.0D0
	      END DO
	      cnt1=2
	    END IF
            DO i=cnt1,NCenter
	      CALL NextLine(InBuffer,ErrCode)
	      IF (ErrCode<0) CALL ErrorMsg('Error, end of file detected during reading.')
              CALL CenterVec(InBuffer,SDim,DVec2)
              DO ii=1,SDim
                Center(ii,i)=Dvec2(ii)
              END DO
	    END DO
	  ENDIF

        ELSEIF (CurKwdId==kid_outbase) THEN
          OutRhoFile=InBuffer(1:idel(InBuffer))

        ELSEIF (CurKwdId==kid_qvec) THEN
          IF (NPass==2) CALL Qvectors(kid_qvec,ErrCode)

        ELSEIF (CurKwdId==kid_tlist) THEN
          IF (NPass==2) CALL ReadTlist(kid_tlist,ErrCode)
	  
        ELSEIF (CurKwdId==kid_range) THEN
	    CALL OneInt(InBuffer,range,ErrCode)

        ELSEIF (CurKwdId==kid_tol) THEN
          CALL OneReal(InBuffer,Tolerance,ErrCode)
	  
        ELSEIF (CurKwdId==kid_atoms) THEN
            AtomNr=0
	    DO
	      CALL NextLine(InBuffer,ErrCode)
	      IF (ErrCode<0) THEN
	        CALL ErrorMsg('Error, preliminary end of file detected.')
                ErrCode=100
              END IF
	      IF (FirstWord(InBuffer,.false.)==Kwd(kid_atoms)%endname) EXIT
	      AtomNr=AtomNr+1 !Count the lines
	      IF (NPass==2) THEN !Read the atoms
                CALL GetAtomandCP(InBuffer,AtomNr,0,ErrCode)
              ENDIF
	    END DO	  

        ELSEIF (CurKwdId==kid_cplist) THEN
            CPListNr=0
	    DO
	      CALL NextLine(InBuffer,ErrCode)
	      IF (ErrCode<0) THEN
	        CALL ErrorMsg('Error, preliminary end of file detected.')
                ErrCode=100
              END IF
	      IF (FirstWord(InBuffer,.false.)==Kwd(kid_cplist)%endname) EXIT
	      CPListNr=CPListNr+1 !Count the lines
	      IF (NPass==2) THEN !Read the atoms
                CALL GetAtomandCP(InBuffer,CPListNr,1,ErrCode)
              ENDIF
	    END DO	  

        ELSEIF (CurKwdId==kid_axord) THEN
          IF (NPass==2) THEN
            IF (idel(InBuffer)/=SDim) THEN
	      ErrCode=100
	      CALL ErrorMsg('Incorrect number of digits in the keyword AxisOrder.')
            END IF
            DO i=1,SDim
              READ(InBuffer(i:i),'(i1)') CO(i)
            END DO
          END IF

        ELSEIF (CurKwdId==kid_plim) THEN
          CALL OneReal(FirstWord(InBuffer,.true.),plim,ErrCode)
          IF (InBuffer=='relative') THEN
            plimkey='relative'
	  ELSEIF (InBuffer=='absolute') THEN
	    plimkey='absolute'
	  ELSEIF (InBuffer=='sigma') THEN
	    plimkey='sigma'
	  ELSEIF (InBuffer/=' ') THEN
            CALL ErrorMsg('Error: Keyword plimit not properly defined!!!')
	    ErrCode=100
	  ENDIF

        ELSEIF (CurKwdId==kid_chlim) THEN
          CALL OneReal(InBuffer,chlim,ErrCode)

        ELSEIF (CurKwdId==kid_chlimlist) THEN
          CALL OneReal(FirstWord(InBuffer,.true.),chlimlistorig,ErrCode)
          IF (InBuffer=='relative') THEN
            chlimlistkey='relative'
	  ELSEIF (InBuffer=='absolute') THEN
	    chlimlistkey='absolute'
	  ELSEIF (InBuffer/=' ') THEN
            CALL ErrorMsg('Error: Keyword chlimlist not properly defined!!!')
	    ErrCode=100
	  ENDIF

        ELSEIF (CurKwdId==kid_tmap) THEN
          CALL YesNo(InBuffer,writetmap,ErrCode)

        ELSEIF (CurKwdId==kid_fullcell) THEN
          CALL YesNo(InBuffer,fullcellpeaks,ErrCode)

        ELSEIF (CurKwdId==kid_basins) THEN
          CALL YesNo(InBuffer,writebasin,ErrCode)

        ELSEIF (CurKwdId==kid_maxima) THEN
          maxformat=InBuffer
          IF ((maxformat/='atoms').and.(maxformat/='all').and.(maxformat/='none')) THEN
	    Call ErrorMsg('Error, keyword maxima must have value atoms, all or none!')
	    ErrCode=100
	  END IF  
          IF (maxformat=='atoms') CALL FlagKwdGroup('atoms',.true.)

        ELSEIF (CurKwdId==kid_cps) THEN
          cpformat=InBuffer
          IF ((cpformat/='list').and.(cpformat/='all').and.(cpformat/='none')) THEN
	    Call ErrorMsg('Error, keyword criticalpoints must have value list, all or none!')
	    ErrCode=100
	  END IF  
          IF (cpformat=='list') CALL FlagKwdGroup('cps',.true.)

        ELSEIF (CurKwdId==kid_scale) THEN
          scale=InBuffer
          IF ((scale/='fractional').and.(scale/='angstrom')) THEN
	    Call ErrorMsg('Error, keyword scale must have value fractional or angstrom!')
	    ErrCode=100
	  END IF  

        ELSEIF (CurKwdId==kid_position) THEN
          position=InBuffer
          IF ((position/='absolute').and.(position/='relative')) THEN
	    Call ErrorMsg('Error, keyword scale must have value absolute or relative!')
	    ErrCode=100
	  END IF  

        ELSEIF (CurKwdId==kid_coc) THEN
          CALL YesNo(InBuffer,com,ErrCode)

        ELSEIF (CurKwdId==kid_basinmode) THEN
          DBuffer=firstword(InBuffer,.true.)
          IF (DBuffer=='ongrid') THEN
	    basinmode=ongridmode
          ELSEIF (DBuffer=='neargrid') THEN
	    basinmode=neargridmode
	    IF (len_trim(InBuffer)>0) call oneint(InBuffer,EdgeRemovalCycles,ErrCode)	    
          ELSEIF (DBuffer=='pedantic') THEN
	    basinmode=pedanticmode
	  ELSE
	    Call ErrorMsg('Error, keyword partitionmode must have value ongrid, neargrid, neargridfast or pedantic!')
	    ErrCode=100
	  END IF  

        ELSEIF (CurKwdId==kid_addborder) THEN
          CALL OneReal(InBuffer,addborder,ErrCode)

        ELSEIF ((CurKwdId==kid_biso1).or.(CurKwdId==kid_biso2)) THEN
          CALL OneReal(InBuffer,Biso,ErrCode)

        ELSEIF (CurKwdId==kid_lambda) THEN
          CALL OneReal(InBuffer,lambda,ErrCode)

        ELSEIF ((CurKwdId==kid_writem40).or.(CurKwdId==kid_export)) THEN
          OutStrfile=InBuffer
	  IF (OutStrFile(idel(OutStrFile)-3:idel(OutStrFile))=='.m40') THEN
	    OutStrFile=OutStrFile(1:idel(outStrFile)-4) !cut away the m40 extension, will be added later again
	    OutStrType='m40'
	  ELSEIF (OutStrFile(idel(OutStrFile)-3:idel(OutStrFile))=='.cif') THEN
	    OutStrFile=OutStrFile(1:idel(outStrFile)-4) !cut away the cif extension, will be added later again
	    OutStrType='cif'
	  ELSEIF (OutStrFile(idel(OutStrFile)-3:idel(OutStrFile))=='.ins') THEN
	    OutStrFile=OutStrFile(1:idel(outStrFile)-4) !cut away the m40 extension, will be added later again
	    OutStrType='ins'
          ELSE
	    Call ErrorMsg('Error, the file for the exported structure has an unknown extension. Can''t guess the format.')
	    ErrCode=100	    
	  ENDIF
          wantstructure=.true.
	  CALL FlagKwdGroup('export',.true.)
	  com=.true.

        ELSEIF (CurKwdId==kid_m40forjana) THEN
          CALL YesNo(InBuffer,m40forjana,ErrCode)

        ELSEIF (CurKwdId==kid_comparem40) THEN
          Compm40file=InBuffer
          compm40=.true.

        ELSEIF (CurKwdId==kid_projection) THEN
          CALL YesNo(InBuffer,projection,ErrCode)
          if (projection) CALL FlagKwdGroup('4Dnp',.false.) 

        ELSEIF (CurKwdId==kid_composition) THEN
          IF ((NPass==1).and.(ElNr==0)) THEN !The second condition is there to protect again doubled allocation of compos in case of a double occurence of composition in the input file
            IF (Idel(InBuffer)==0) THEN
              ElNr=0
	    ELSE
	      ElNr=1
	    ENDIF
            DO i=1,idel(InBuffer)-1
              if (InBuffer(i:i).eq.' '.and.InBuffer(i+1:i+1).ne.' ') ElNr=ElNr+1
            ENDDO
            IF (ElNr>0) THEN
              ALLOCATE(Compos(1:ElNr,1:2))
              DO i=1,ElNr
                Compos(i,1)=0
                Compos(i,2)=1
                dBuffer=FirstWord(InBuffer,.true.)
                !Find the first number in the string
	        ii=1;iii=0
	        DO WHILE ((ii/=0).and.(iii<=idel(dBuffer)))
	          iii=iii+1
                  CALL CheckAllowedChars(dBuffer(iii:iii),5,ii)
	        ENDDO
		IF ((iii==2).and.(dBuffer(1:1)=='D')) THEN
		  Compos(i,1)=1
                ELSE
                  DO j=1,TotElNr
                     IF (dBuffer(1:iii-1)==ElSmb(j)) Compos(i,1)=j
	          END DO
                ENDIF
		IF (Compos(i,1)==0) THEN
	          dBuffer='Error, symbol '//dBuffer(1:iii-1)//' is not a valid symbol of an element.'
	          CALL ErrorMsg(dBuffer)
	          ErrCode=100
                  CYCLE
                END IF
                MaxZ=max(MaxZ,Compos(i,1))
                if (compos(i,1)>1) MinZ=min(MinZ,Compos(i,1))
		IF (iii<=idel(dBuffer)) THEN
                  DummyString=dBuffer
                  CALL OneInt(dBuffer(iii:),Compos(i,2),ErrCode)
		  IF (ErrCode/=0) THEN
                    dBuffer='Error, string '//DummyString(1:idel(DummyString))// &
		            ' does not have the format "element symbol and number".'
                    CALL ErrorMsg(dBuffer)
                  ENDIF
	        ENDIF	
	      END DO
            ELSE
	      CALL ErrorMsg('Error, no elements found after the keyword composition.')
	      ErrCode=100
	    ENDIF
	  ENDIF

        ELSEIF (CurKwdId==kid_totalatoms) THEN
	  IF (NPass==2) THEN
	    CALL OneInt(InBuffer,totalatoms,ErrCode)
            IF ((TotalAtoms<0).and.(ElNr==0)) THEN
              CALL ErrorMsg('Error, no composition given, the number of atoms cannot be guessed from composition.')		  
              ErrCode=100
	    ENDIF	        	    
	    IF (ErrCode/=0) THEN
	      ErrCode=0
	      IF (InBuffer=='guess') THEN
	        TotalAtoms=0
	      ELSEIF (InBuffer=='composition') THEN
                IF (ElNr>0) THEN
		  TotalAtoms=-100
		ELSE
                  CALL ErrorMsg('Error, no composition given, the number of atoms cannot be guessed from composition.')		  
		  ErrCode=100
		ENDIF	        
	      ELSE
                CALL ErrorMsg('Error, keyword numberofatoms must be followed by "guess", "composition" or an integer.')
                ErrCode=100
	      ENDIF
	    ENDIF
          ENDIF
	END IF !End of the ifblock containing handling of all keywords

        IF (ErrCode/=0) THEN
	  CALL ErrorMsg('Error reading keyword '//Kwd(CurKwdId)%name(1:idel(Kwd(CurKwdId)%name))//'.')
          ErrNr=ErrNr+1
	END IF
!*** End of keyword specific operations ***

        GOTO 100 !Go to reading next line        
200     CONTINUE
        CALL ErrorCheck !Check for errors during reading and take corresponding actions
        CALL AfterPass  !Actions to be taken after the file is read up to the end
        REWIND(infunit)
      END DO
!     Successfully finished reading
      GOTO 9999

!     error messages            
900   CALL ErrorMsg('Error reading the input file '//infile(:idel(infile))//'.')
      STOP
      
9999  CLOSE(infunit)
      CONTINUE      

      CONTAINS

! This routine typically just reads the name of the input file
!****************************************
        SUBROUTINE InitInput
!****************************************
!          USE F90_UNIX_ENV, ONLY : GETARG,IARGC

          INTEGER :: GetReclFac
          Inputfile=' '
	  IF (IARGC()>0) THEN
            CALL GETARG(1,InputFile)
	  ENDIF
11        IF (InputFile(1:1)==' ') THEN
            WRITE (*,'(A)') ''
            WRITE (*,'(A,I4,A)',ADVANCE='NO') &
            'Please insert the name of your inputfile (max. ',Filelngth,' chars): '
            READ (*,'(A)') InputFile
            IF (idel(InputFile)==0) THEN
              WRITE (*,'(A)') ''
              WRITE (*,'(A)') 'No name given! Try again...'
              WRITE (*,'(A)') ''
              GOTO 11
            END IF
            WRITE (*,'(A)') ''
          END IF
          RecLFac=GetReclFac()
!         Open the input file for reading
          OPEN (UNIT=infunit,FILE=Inputfile(:Idel(Inputfile)),IoStat=ErrCode,STATUS='OLD')

	END SUBROUTINE InitInput


! Handles all actions taken after a pass is finished - typically allocation of dynamically allocated arrays
!****************************************
        SUBROUTINE AfterPass
!****************************************
          IF (NPass==1) THEN
    
! Try to guess the types of the density files
            IF (InRhoType=='none') THEN !No type has been input
	      CALL GuessType(InRhoFile,InRhoType)
              IF (InRhoType=='unknown') THEN
	        CALL ErrorMsg('Error, keyword outputformat is missing and I cannot guess the type from the file names!') 
	        CALL ErrorMsg('Please use the keyword outputformat or standard file extensions.') 
                ErrCode=100
	      ENDIF
	    ENDIF
            
            CALL GetNVox !Get the number of voxels dimension etc...
            IF ((projection).and.(SDim==RDim)) THEN
              CALL ErrorMsg('Warning: It makes no sense to make projection of this density (dimension = realdimension). '// &
                             'Ignoring the keyword projection.')
              projection=.false.
            END IF
            if ((SDim>RDim).and..not.projection) CALL FlagKwdGroup('4Dnp',.true.)
            ALLOCATE (Atoms(AtomNr), STAT=AllocError)            ! Type and coordinates of atoms
            IF (AllocError.GT.0) STOP 'Error allocating Atoms!'
            ALLOCATE (CPlist(CPlistNr), STAT=AllocError)         ! Type and coordinates of atoms
            IF (AllocError.GT.0) STOP 'Error allocating Atoms!'
            ALLOCATE (QVec(RDim, SDim-RDim), STAT=AllocError)    ! q-vectors
            IF (AllocError.GT.0) STOP 'Error allocating QVec!'
            ALLOCATE (Tlist(3, SDim-RDim), STAT=AllocError)      ! t-settings
            IF (AllocError.GT.0) STOP 'Error allocating Tlist!'
	    ALLOCATE (CO(SDim), STAT=AllocError)                 ! Axis order
            IF (AllocError.GT.0) STOP 'Error allocating Axis order!'
            ALLOCATE (DMat(SDim,SDim), STAT=AllocError)             
            IF (AllocError.GT.0) STOP 'Error allocating DMat!'
            ALLOCATE (DVec2(SDim), STAT=AllocError)             
            IF (AllocError.GT.0) STOP 'Error allocating DVec!'
            IF (NCenter==0) THEN
	      NCenter=1
	      FlagZeroC=.false.
	    ENDIF
	    ALLOCATE (Center(SDim, NCenter), STAT=AllocError)      ! Centers
            IF (AllocError.GT.0) STOP 'Error allocating Center!'
            Center=0.0D0
            ALLOCATE (RMat(SDim, SDim, NSym*NCenter), STAT=AllocError)! Rotation matrices
	    IF (AllocError.GT.0) STOP 'Error allocating RMat!'
            RMat=0.0D0
            ALLOCATE (TVec(SDim, NSym*NCenter), STAT=AllocError)      ! Translation vectors
            IF (AllocError.GT.0) STOP 'Error allocating TVec!'
            TVec=0.0D0
            ALLOCATE (RMatI(SDim, SDim, NSym*NCenter), STAT=AllocError)! Integer rotation matrices
            IF (AllocError.GT.0) STOP 'Error allocating RMat!'
            RMatI=0
            ALLOCATE (TVecI(SDim, NSym*NCenter), STAT=AllocError)      ! Integertranslation vectors
            IF (AllocError.GT.0) STOP 'Error allocating TVec!'
            TVecI=0
            DO i=1,SDim
              CO(i)=i
            END DO
            ALLOCATE (Rim(RDim), STAT=AllocError)                 ! border of the unit cell in the incommensurate directions
            IF (AllocError.GT.0) STOP 'Error allocating Rim!'
            ALLOCATE (Aper(RDim), STAT=AllocError)                 ! which direction is aperiodic
            IF (AllocError.GT.0) STOP 'Error allocating Aper!'
            IF ((maxformat==' ').and.(AtomNr>0)) maxformat='atoms' 

            if (maxformat=='atoms') CALL FlagKwdGroup('atoms',.true.) 
            if (cpformat=='list') CALL FlagKwdGroup('cps',.true.) 

	  ELSEIF (NPass==2) THEN
            IF (NCenter>0) CALL ExpandSym
            CALL DiscreteSym

          END IF
	END SUBROUTINE AfterPass

!****************************************
        SUBROUTINE ErrorMsg(text)
!****************************************
          CHARACTER*(*) :: text
          write(*,'(3a)') text(1:idel(text))
	END SUBROUTINE ErrorMsg


!        Calculating reciprocal cell parameters
!******************************************************************
      SUBROUTINE ReciCell (iRDim, iCellCon, ioCell, oRCell)
!******************************************************************
      USE GLOBALDEFINITIONS
      USE GLOBALVARIABLES
      
      IMPLICIT NONE
      INTEGER :: iCellCon, iRDim,i,ii
!        Variables with leading r are reciprocal
      DOUBLE PRECISION :: ioCell(iCellCon), oRCell(iCellCon)
      DOUBLE PRECISION :: a, b, c, sina, sinb, sing, cosa, cosb, cosg
      DOUBLE PRECISION :: ra, rb, rc
      DOUBLE PRECISION :: cosra, cosrb, cosrg
      DOUBLE PRECISION :: V, rV
      
!       How much real dimensions
      SELECT CASE (iRDim)
      CASE (1)
!        Declaring auxilliary variables
        a=ioCell(1)
        ra=1/a
        V=a
        rV=ra
        ioCell(7)=V
        oRCell(1)=ra
        oRCell(7)=rV
        ioCell(2)=90.0 !its convenient to have the angle defined, even in one-dimensional space
	oRCell(2)=90.0
      CASE (2)
        a=ioCell(1)
        b=ioCell(2)
        sina=DSIN(toRad*ioCell(3))
!       Areas
        V=a*b*sina
        rV=1./V
!       reciprocal axes
        ra=1/(a*sina)
        rb=1/(b*sina)
!       declaring calculated values
        ioCell(7)=V
        oRCell(1)=ra
        oRCell(2)=rb
!       reciprocal angle
        oRCell(3)=180D0-ioCell(3)
        oRCell(7)=rV
      CASE (3)
!        Declaring auxilliary variables
        a=ioCell(1)
        b=ioCell(2)
        c=ioCell(3)
        sina=DSIN(toRad*ioCell(4))
        sinb=DSIN(toRad*ioCell(5))
        sing=DSIN(toRad*ioCell(6))
        cosa=DCOS(toRad*ioCell(4))
        cosb=DCOS(toRad*ioCell(5))
        cosg=DCOS(toRad*ioCell(6))
!        Volumes
        V=a*b*c*SQRT(1.-cosa**2-cosb**2-cosg**2+2.*cosa*cosb*cosg)
        rV=1./V
!        reciprocal angles
        cosra=(cosb*cosg-cosa)/(sinb*sing)
        cosrb=(cosa*cosg-cosb)/(sina*sing)
        cosrg=(cosa*cosb-cosg)/(sina*sinb)
!        reciprocal axes
        ra=b*c*sina/V
        rb=a*c*sinb/V
        rc=a*b*sing/V
!        declaring calculated values
        ioCell(7)=V
        oRCell(1)=ra
        oRCell(2)=rb
        oRCell(3)=rc
        oRCell(4)=DACOS(cosra)/toRad
        oRCell(5)=DACOS(cosrb)/toRad
        oRCell(6)=DACOS(cosrg)/toRad
        oRCell(7)=rV
      END SELECT
!     fill the direct and reciprocal metric tensor
      DO i=1,iRDim
        DO ii=1,iRDim
	  IF(i==ii) THEN
	    cosa=1.0D0
            cosb=1.0D0
	  ELSE
	    cosa=cos(ioCell(9-i-ii-3*(3-RDim))*torad)
	    cosb=cos(oRCell(9-i-ii-3*(3-RDim))*torad)
	  END IF
	  G(i,ii)=ioCell(i)*ioCell(ii)*cosa
          Gi(i,ii)=oRCell(i)*oRCell(ii)*cosb
        END DO
      END DO
! Fill the conversion matrix E = M*A to the orthonormal system
      Select CASE(iRDim)
      CASE(1)
        M(1,1)=1.0D0/ioCell(1)
      CASE(2)
        M(1,1)=1.0D0/ioCell(1)
        M(1,2)=0.0D0
        M(2,1)=-cosa/(ioCell(1)*sina)
        M(2,2)=1.0D0/(ioCell(2)*sina)
      CASE(3)
        M(1,1)=1.0D0/ioCell(1)
        M(1,2)=0.0D0
        M(1,3)=0.0D0
        M(2,1)=-cosg/(ioCell(1)*sing)
        M(2,2)=1.0D0/(ioCell(2)*sing)
        M(2,3)=0.0D0
        M(3,1)=oRCell(1)*cosrb
        M(3,2)=oRCell(2)*cosra
        M(3,3)=oRCell(3)
      END SELECT
      DO i=1,iRDim
        DO ii=1,iRDim
          Mv(i,ii)=M(i,ii)*DBLE(NVox(i))
        END DO
      END DO
      call invert(M,Minv,RDim)
      call invert(Mv,Mvinv,RDim)
      
12000 END SUBROUTINE ReciCell

! Read the q-vectors
!****************************************
        SUBROUTINE QVectors(kid,ErrCd)
!****************************************
	  INTEGER :: i,ErrCd,kid
	  CHARACTER*50 :: ErrText
          DO i=1,DDim
            CALL NextLine(InBuffer,ErrCd)
            
	    IF (ErrCd/=0) RETURN !Some error occured
            IF (FirstWord(InBuffer,.false.)==Kwd(kid)%endname) THEN !Premature end of the block
              CALL ErrorMsg('Error - '//Kwd(kid)%endname(1:idel(Kwd(kid)%endname))//' detected prematurely.')
              WRITE(ErrText,'(i1,a)') DDIM,' qvectors expected.'
	      CALL ErrorMsg(ErrText)
	      ErrCd=100
	    END IF
            CALL ArrayReal(InBuffer,qvec(1,i),3,ErrCd)
	    IF (ErrCd/=0) RETURN !Some error occured            
	  END DO
          CALL NextLine(InBuffer,ErrCd)
          IF (FirstWord(InBuffer,.false.)/=Kwd(kid)%endname) THEN !The end-keyword is not here!
            CALL ErrorMsg('Too many lines before '//Kwd(kid)%endname(1:idel(Kwd(kid)%endname)))
	    ErrCd=100
	  END IF
	END SUBROUTINE QVectors

! Read the sampling in t-space
!****************************************
        SUBROUTINE ReadTlist(kid,ErrCd)
!****************************************
	  INTEGER :: i,ErrCd,kid
	  CHARACTER*50 :: ErrText
          DO i=1,DDim
            CALL NextLine(InBuffer,ErrCd)
            
	    IF (ErrCd/=0) RETURN !Some error occured
            IF (FirstWord(InBuffer,.false.)==Kwd(kid)%endname) THEN !Premature end of the block
              CALL ErrorMsg('Error - '//Kwd(kid)%endname(1:idel(Kwd(kid)%endname))//' detected prematurely.')
              WRITE(ErrText,'(i1,a)') DDIM,' lines expected.'
	      CALL ErrorMsg(ErrText)
	      ErrCd=100
	    END IF
            CALL ArrayReal(InBuffer,tlist(1,i),3,ErrCd)
	    IF (ErrCd/=0) RETURN !Some error occured            
	  END DO
          CALL NextLine(InBuffer,ErrCd)
          IF (FirstWord(InBuffer,.false.)/=Kwd(kid)%endname) THEN !The end-keyword is not here!
            CALL ErrorMsg('Too many lines before '//Kwd(kid)%endname(1:idel(Kwd(kid)%endname)))
	    ErrCd=100
	  END IF
	END SUBROUTINE ReadTlist

!****************************************
        SUBROUTINE ReadSymmetry(ErrCd)
!****************************************
	  IMPLICIT NONE
	  INTEGER :: ErrCd,Symcnt,iii
          CHARACTER*(LineLength) :: lowbuffer
	  DO Symcnt=1,NSym
            CALL NextLine(InBuffer,ErrCd)
!	    CALL Symmetry (SDim, InBuffer, DMat, DVec2)
            lowbuffer=ToLow(InBuffer)
	    CALL OneSym(SDim, SymCnt, lowbuffer, DMat, DVec2)
	    DO i=1, SDim
              DO ii=1, SDim
                RMat(i,ii,Symcnt)=DMat(i,ii)
                DMat(i,ii)=0.0D0
              END DO
              TVec(i,Symcnt)=DVec2(i)
              DVec2(i)=0.0D0
            END DO
	  END DO
	END SUBROUTINE ReadSymmetry

! Decode one symmetry operation into a matrix and a vector
!****************************************
        SUBROUTINE OneSym(SDim,Truepos, SymString,InRMat,LocTVec)
!****************************************
	  INTEGER :: SDim,ErrCd,i,ii,col,curbeg,curend,jex,jelom,jebod,lgth
	  INTEGER :: indx,indlom,nom,denom,TruePos,SymOpPos
          DOUBLE PRECISION, DIMENSION(SDim,SDim) :: InRmat
          DOUBLE PRECISION, DIMENSION(SDim,SDim) :: LocRmat
          DOUBLE PRECISION, DIMENSION(SDim) :: LocTvec
	  CHARACTER*(*) :: SymString
	  CHARACTER*(linelength) :: Row
	  CHARACTER*(15) :: rowpart
	  CHARACTER*6 :: CooLetter='xyztuv' 
	  !First check if the format is x1, x2, x3... or x, y, z
	  IF (index(SymString,'x1')==0) THEN !x y z format
            DO i=1,SDim
	      DO
	        ii=index(SymString,CooLetter(i:i))
	        IF (ii==0) EXIT
	        write(rowpart,'(a1,i1)')'#',i
	        SymString=SymString(:ii-1)//trim(rowpart)//symString(ii+1:)
	      ENDDO
	    ENDDO
            DO i=1,len_trim(SymString)
	      IF (SymString(i:i)=='#') SymString(i:i)='x'
	    ENDDO
	  ENDIF
	  !Check if comma or space delimited
	  IF (index(Symstring,',')/=0) THEN !Comma delimited
	    CALL RemoveSpaces(SymString)
	    ii=len_trim(SymString)
	    DO i=1,ii
	      if (SymString(i:i)==',') SymString(i:i)=' '
	    ENDDO
	  ENDIF
	  
          DO i=1,SDim
	    LocTVec(i)=0.0E0
	    DO ii=1,SDim
	      LocRMat(i,ii)=0.0E0
	    ENDDO
	    row=FirstWord(SymString,.true.)
	    lgth=len_trim(row)
	    IF (lgth==0) THEN
	      write(row,'(a,i3)') 'Error, not enough components of the symmetry operation number',TruePos
              call ErrorMsg(row)
	      ErrCd=100
	      return
	    ENDIF
            if (lgth==1) goto 9100
	    CALL CheckAllowedChars(row(:lgth),1,ErrCd)
	    IF (ErrCd/=0) goto 9100
	    curbeg=1;jex=0;jelom=0;jebod=0;indx=0;indlom=0
	    DO ii=1,lgth
	      IF (row(ii:ii)=='x') THEN
	        jex=jex+1
	        indx=ii
	      ELSEIF (row(ii:ii)=='/') THEN
	        jelom=jelom+1
	        indlom=ii
	      ELSEIF (row(ii:ii)=='.') THEN
	        jebod=jebod+1
	      ENDIF
	      IF ((((row(ii:ii)=='+').or.(row(ii:ii)=='-')).and.(ii>1)).or.(ii==len_trim(row))) THEN 
                IF (ii==lgth) then
		  curend=ii
                else
		  curend=ii-1
		endif
		IF ((jelom+jebod+jex>1).or.(indlom==curbeg).or.(indlom==curend) &
		    .or.(indx==curend).or.(indx-curbeg>1)) goto 9000
                !Row(curbeg,curend) contains exactly one entry, it can be a +-xi,fraction or a decimal number
		IF (jex>0) then !Part containing entry of the rotation matrix
		  read(row(indx+1:curend),'(i2)',IOSTAT=ErrCd) col
		  if (errcd/=0) goto 9000
		  if (LocRMat(i,col)>eps) goto 9300
		  if (row(curbeg:curbeg)=='-') THEN 
		    LocRMat(i,col)=-1.0E0
		  elseif ((row(curbeg:curbeg)=='+').or.(row(curbeg:curbeg)=='x')) THEN
		    LocRMat(i,col)=1.0E0
		  else
		    goto 9000
		  endif
		elseif(jebod>0) then !decimal number
		  if (LocTVec(i)>eps) goto 9200
		  rowpart=row(curbeg:curend)
		  call onereal(rowpart,LocTVec(i),ErrCd)
		  if (ErrCd/=0) goto 9000
		elseif(jelom>0) then !fraction
		  if (LocTVec(i)>eps) goto 9200
		  rowpart=row(curbeg:indlom-1)               
		  call oneint(rowpart,nom,ErrCd)
		  if (ErrCd/=0) goto 9000
		  rowpart=row(indlom+1:curend)
		  call oneint(rowpart,denom,ErrCd)
		  if (ErrCd/=0) goto 9000
		  LocTVec(i)=float(nom)/float(denom)
                else !no x, no point, no slash -> integer translational part! Pretty bizzare, but possible...
		  if (LocTVec(i)>eps) goto 9200
		  rowpart=row(curbeg:curend)
		  call oneint(rowpart,nom,ErrCd)
		  if (ErrCd/=0) goto 9000
		  LocTVec(i)=float(nom)
	        endif
		jex=0;jelom=0;jebod=0;curbeg=curend+1;indlom=0;indx=0
	      ENDIF
	    ENDDO
	  END DO
          InRMat=LocRMat
	  return
9000      write(row,'(a,i3,a,i2)') 'Syntax error in the symmetry operation number',TruePos,', component number ',i
          call ErrorMsg(row)
	  ErrCd=100
	  return
9100      write(row,'(a,i3)') 'Syntax error in the symmetry operation number',TruePos
          call ErrorMsg(row)
          ErrCd=100
          return
9200      write(row,'(a,i3,a,i2)') 'Error, more than one translational part in the symmetry operation number',TruePos, &
                              ', component number ',i
          call ErrorMsg(row)
          ErrCd=100
          return
9300      write(row,'(a,2i2)') 'Error, more than one definition of the matrix element',i,col
          call ErrorMsg(row)
	  write(row,'(a,i3,a)') 'in symmetry operation number',TruePos,'.'
          call ErrorMsg(row)
          ErrCd=100
          return
	END SUBROUTINE OneSym
	
        SUBROUTINE RemoveSpaces(ioText)
        character*(*) :: ioText
        integer :: ii
          ii=0
          DO
            ii=ii+1
	    IF (ioText(ii:ii)==' ') THEN
	      ioText(ii:)=ioText(ii+1:)
	      ii=ii-1
            ENDIF
	    IF (ii>=len_trim(ioText)) EXIT
          ENDDO
        END SUBROUTINE RemoveSpaces
	

!******************************************************************
      SUBROUTINE CenterVec (iBuffer, iSDim, oVec)
!******************************************************************
      IMPLICIT NONE
      CHARACTER        :: iBuffer*(*)
      CHARACTER        :: DBuffer*(132)
      INTEGER          :: iSDim
      DOUBLE PRECISION :: oVec (iSDim)
      INTEGER          :: Pos, Numerator, Nominator, i
      
      DO i=1, iSDim
        CALL SpaceErase (iBuffer)
        CALL DataLength (iBuffer, DBuffer)
        CALL NextPos (DBuffer, '/', Pos)
        IF (Pos>=0) THEN
          CALL NumberLength (DBuffer, 'l', Pos, Numerator)
          CALL NumberLength (DBuffer, 'r', Pos, Nominator)
          IF (Nominator.EQ.1) STOP &
            'Symmetry data not consistent: Translationvector >= 1?'
          oVec(i)=DBLE(Numerator)/DBLE(Nominator)
        ELSE
          READ (DBuffer, '(F14.0)') oVec(i)
        END IF
      END DO
      
1000  END SUBROUTINE CenterVec

!        Calculate symmetry operations for discrete space
!******************************************************************
      SUBROUTINE DiscreteSym
!******************************************************************
      IMPLICIT NONE
      
      INTEGER :: i, ii, iii,badsmop
!       The diagonals of matrices N and N^-1 holding voxel division
      DOUBLE PRECISION, DIMENSION (SDim) :: Nn
!       Dummy
      DOUBLE PRECISION, DIMENSION (SDim, SDim) :: DRMat
      DOUBLE PRECISION :: TTest                 ! Test
      
!       Preparation
      DO i=1, SDim
        Nn(i)=DBLE(NVox(i))
      END DO
!       All symops
      DO i=1, NSym
!       R*Ninv, N*Tau
!       Line of RMat, column of Ninv
        DO ii=1, SDim
!       N by translation
          TVecI(ii,i)=NINT(Nn(ii)*TVec(ii,i))
!       Test if integer, hence a voxel operation
          TTest=ABS(Nn(ii)*TVec(ii,i)-DBLE(TVecI(ii,i)))
          IF (TTest>100*Eps) THEN
            badsmop=i
            GOTO 900
          END IF
	  
          DO iii=1,SDim
            DRMat(ii,iii)=0.0D0
            TTest=ABS(RMat(ii,iii,i))
            IF (TTest<Eps) THEN
              DRMat(ii,iii)=0.0D0
            ELSE 
              DRMat(ii,iii)=RMat(ii,iii,i)/Nn(iii)
            END IF
          END DO
        END DO
!       N*(R*Ninv)
        DO ii=1, SDim
          DO iii=1,SDim
            RMatI(ii,iii,i)=NINT(Nn(ii)*DRMat(ii,iii))
            TTest=ABS(DBLE(RMatI(ii,iii,i)-(Nn(ii)*DRMat(ii,iii))))
	    IF (TTest>Eps) THEN
	      badsmop=i
	      GOTO 910
            ENDIF
	  END DO
        END DO
      END DO
      GOTO 1000

!       Errors
900   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine DiscreteSym:'
      WRITE (*,'(A)') 'Error: Translationvector multiplied by'
      WRITE (*,'(A)') '       voxelnumber is not an integer!'
      WRITE (*,'(A,i3)') '       Symmetry operation #: ', badsmop
      WRITE (*,'(A)') ''
      STOP
      
910   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine DiscreteSym:'
      WRITE (*,'(A)') 'Error: Translationmatrix divided by'
      WRITE (*,'(A)') '       voxelnumber is not an integer!'
      WRITE (*,'(A,i3)') '       Symmetry operation #: ', badsmop
      WRITE (*,'(A)') ''
      STOP

1000  END SUBROUTINE DiscreteSym

!        Expand symmetrie operations with centers
!******************************************************************
      SUBROUTINE ExpandSym
!******************************************************************
      IMPLICIT NONE
      
      INTEGER :: iES, iiES, iiiES, ivES
      INTEGER :: SymLoc         ! Start point for appending new symops
      
      SymLoc=0
      
All:  DO iES=1, NCenter           ! All centers 
        DO iiES=1, NSym           ! All read symops
          SymLoc=SymLoc+1
!       Copy rotation matrix
          DO iiiES=1,SDim         ! Lines
            DO ivES=1, SDIM       ! Columns
              RMat(iiiES,ivES,SymLoc)=RMat(iiiES,ivES,iiES)
            END DO
!       Calculate new translation vector
            TVec(iiiES,SymLoc)=TVec(iiiES,iiES)+Center(iiiES,iES)
!       Out of unit cell?
!       Too large
Large:      DO WHILE (TVec(iiiES, SymLoc)>=1.0D0)
              TVec(iiiES, Symloc)=TVec(iiiES, Symloc)-1.0D0
            END DO Large
!       Too small
Small:      DO WHILE (TVec(iiiES, SymLoc)<0.0D0)
              TVec(iiiES, Symloc)=TVec(iiiES, Symloc)+1.0D0
            END DO Small
          END DO
        END DO
      END DO All
!       Adjusting number of symops
      NSymPrim=NSym
      NSym=SymLoc
      END SUBROUTINE ExpandSym

! Read values for one atom
!*********************************************************
      SUBROUTINE GetAtomandCP(iBuffer,iAtomNr,key,ErrCd)
!*********************************************************

      USE GlobalVariables

      IMPLICIT NONE

      INTEGER :: iAtomNr,key ! Number of read atom, key to distinguish critical points and atoms
      INTEGER :: ErrCd       ! counter,error code
      CHARACTER*(*) :: iBuffer
      CHARACTER*(LineLength) :: dBuffer  ! One line of input file containing atom information
      ErrCd=0
      IF (key==0) THEN !0 ~ atoms
        Atoms(iAtomNr)%Name=FirstWord(iBuffer,.true.)
        CALL ArrayReal(iBuffer, Atoms(iAtomNr)%Coo(1),3,ErrCd)
      ELSE ! other key ~ critical point
        CPList(iAtomNr)%Name=FirstWord(iBuffer,.true.)
        CALL ArrayReal(iBuffer, CPList(iAtomNr)%Coo(1),3,ErrCd)
      END IF        
      RETURN
      IF (ErrCd/=0) THEN
        write(dBuffer,'(i3)') iAtomNr  
        IF (key==0) THEN
          dBuffer='Error reading line '//dBuffer(1:3)//'of keyword edmaatoms.'
	ELSE
          dBuffer='Error reading line '//dBuffer(1:3)//'of keyword cplist.'
	END IF
	CALL ErrorMsg(dBuffer)
      END IF
      END SUBROUTINE GetAtomandCP

! Get the voxel division and dimension from the input map
!*********************************************************
       SUBROUTINE GETNVOX
!*********************************************************
      USE GlobalVariables

      INTEGER :: nx(6)                  ! Number of steps in direction iorient(i)
      INTEGER :: nxny                   !=nx(1)*nx(2)
      INTEGER :: nmap                   ! Number of sections
      INTEGER :: mapa                   ! Type of map (1=Patterson, 2=Fourier, ?=differece Fourier, ?=checking Fourier)
      REAL    :: dx(6)                  ! Step in iorien(i)
      INTEGER :: iorien(6)              ! Order of axes
      INTEGER :: nsubs                  ! part number of composite crystals (?)
      REAL :: xmin(6), xmax(6)          ! minimum and maximum value of iorien(i)
      INTEGER :: lrec
      INTEGER, DIMENSION(MaxDim) :: NVoxTmp
      CHARACTER*(linelength) :: VoxBuff
      LOGICAL :: Warn

      IF (InRhoType=='jana') THEN
        open(30,file=inrhofile,form='unformatted',status='old', access='direct',recl=34*ReclFac,err=900)
        read(30,rec=1)(lrec,i=1,7)
        close(30)
        open(30,file=inrhofile,form='unformatted',status='unknown', &
               access='direct',recl=max(lrec,34)*ReclFac,err=900)      
!       Read input file header
        READ (30, rec=1, ERR=910) nx, nxny, nmap, (xmin(cnt1),xmax(cnt1),cnt1=1,6), dx, iorien, mapa, nsubs
        CLOSE(30)
!         Guess number of dimensions
        DO i=6,1,-1
          IF (abs(xmax(i))>eps) THEN
            SDim=i
            EXIT
          END IF
        END DO
      ELSEIF (InRhoType=='xplor') THEN
        OPEN (30, ERR=900, FILE=inrhofile,STATUS='OLD')
!       Header of file
!       Empty line
        ErrCode=0
        READ (30, '(a)', ERR=910)
!       number of title lines, skip them
        READ (30, '(a)', ERR=910,END=910) Buffer
        CALL OneInt(Buffer,Lrec,ErrCode)
        IF (ErrCode/=0) GOTO 910
        DO cnt1=1,Lrec
          linecnt=linecnt+1
          READ(30,'(a)',ERR=910)
        ENDDO
!       Voxel spacing
        READ(30,'(a)',ERR=910) Buffer         
	SDim=0
	DO
	  IF (idel(Buffer)>0) THEN
            DO cnt2=1,3
	      VoxBuff=FirstWord(Buffer,.true.)
              CALL OneInt(VoxBuff,nx(cnt2),ErrCode)
	    ENDDO
	    If (errcode/=0) GOTO 910
	    SDim=SDim+1
	  ELSEIF (SDim<3) THEN
	    STOP 'Error, the dimension of the input denisty map is smaller than 3'
	  ELSE
	    EXIT
	  ENDIF
	  NVoxTmp(SDim)=nx(1)
	ENDDO
        CLOSE(30)
        IF (ErrCode/=0) GOTO 910
      ENDIF
      DDim=SDim-RDim
      IF (DDim>0) then
        CALL FlagKwdGroup('4D',.true.)
      END IF        
!     initialize the voxel divisions
      ALLOCATE(NVox(1:SDim))
      IF (AllocError.GT.0) STOP 'Error allocating voxel division!!'
      DO i=1,SDim
        IF (InRhoType=='jana') THEN
	  NVox(i)=nx(i)
        ELSEIF(InRhoType=='xplor') THEN
	  NVox(i)=NVoxTmp(i)
	ENDIF
	NVoxTot=NVoxTot*NVox(i)
        N(i)= NVox(i)*N(i-1)
      END DO
      goto 999
900   STOP 'Error, cannot open the electron density file!'
910   STOP 'Error reading the header of the input file!'
999   continue
      END SUBROUTINE GETNVOX      

!       Determine length of Numbers and sign
!       Input:
!       String
!       Direction to search
!       Position to begin search
!       Output:
!       Number
!******************************************************************
      SUBROUTINE NumberLength (DString, LR, KeyPos, Number)
!******************************************************************
      IMPLICIT NONE
      CHARACTER :: DString*(*)
      INTEGER :: Number, OverAll, KeyPos, i, j, StepL, StepR, Step, BegLoop
      INTEGER :: EnLoop, Begin
      PARAMETER (StepL=-1, StepR=+1, Begin=1)
      CHARACTER*1 :: Nums(10), LR, SignChar(0:1)
      LOGICAL :: IsNum
      DATA Nums /'1','2','3','4','5','6','7','8','9','0'/
      DATA SignChar /'+','-'/
      Number=1
      IsNum=.FALSE.
      OverAll=LEN(DString)
      IF (LR.EQ.'l') THEN
        Step=StepL
        BegLoop=KeyPos-1
        EnLoop=Begin
      END IF
      IF (LR.EQ.'r') THEN
        Step=StepR
        BegLoop=KeyPos+1
        EnLoop=OverAll
      END IF
      DO i=BegLoop, EnLoop, Step
        DO j=1,10
          IF (DString(i:i).EQ.Nums(j)) IsNum=.TRUE.
        END DO
        IF (.NOT.IsNum) THEN
          IF (Step.EQ.StepR) THEN
            READ (DString(BegLoop:i-1),'(I6)') Number
            GOTO 17020
          END IF  
          IF (Step.EQ.StepL) THEN
            IF (DString(i:i).EQ.SignChar(1)) THEN
              IF (i.NE.(KeyPos-1)) THEN
                READ (DString(i:BegLoop),'(I7)') Number
                GOTO 17020
              ELSE
                Number=-1
                GOTO 17020
              END IF
            ELSE
              IF (DString(i:i).EQ.SignChar(0)) THEN
                IF (i.NE.(KeyPos-1)) THEN
                  READ (DString(i+1:BegLoop),'(I7)') Number
                  GOTO 17020
                ELSE
                  Number=1
                  GOTO 17020
                END IF
              END IF
            END IF
          END IF
          ELSE IF((i.EQ.1).AND.(LR.EQ.'l').AND.(IsNum)) THEN
            READ (DString(1:BegLoop),'(I6)') Number
            GOTO 17020
        END IF
        IsNum=.FALSE.
      END DO
17020 CONTINUE
      END SUBROUTINE NumberLength

        SUBROUTINE GuessType(FileName,FileType)
	IMPLICIT NONE
        CHARACTER*(*) :: FileName, FileType
	INTEGER :: i, idel
	! Search for the period
	DO i=idel(FileName),1,-1
	  IF (FileName(i:i)=='.') THEN
	    SELECT CASE (Filename(i+1:idel(Filename)))
	      CASE ('m81')
                FileType='jana'
	      CASE ('xplor')
                FileType='xplor'
              CASE DEFAULT
	        FileType='unknown'
	    END SELECT 
	    EXIT
	  ENDIF
	ENDDO
        END SUBROUTINE GuessType

        SUBROUTINE CheckAllowedChars(String,key,ErrCd)
	integer, parameter :: CharNr=17
	integer :: ErrCd,loop,key
        character*(*) :: String
	character*1, dimension(1:CharNr) :: alchr,alcenchr,alsymchr,justnumchr,reflchr,digits
	logical :: ok
	data alsymchr  /' ',' ','x','1','2','3','4','5','6','7','8','9','0','.','/','+','-'/
	data alcenchr  /' ',' ','1','2','3','4','5','6','7','8','9','0','.','/','+','-',' '/
        data justnumchr/' ',' ','1','2','3','4','5','6','7','8','9','0','.',' ','+','-',' '/
        data reflchr   /' ',' ','1','2','3','4','5','6','7','8','9','0','.',' ','+','-','?'/
        data digits    /'1','2','3','4','5','6','7','8','9','0','0','0','0','0','0','0','0'/
        alsymchr(2)=char(9) !char(9) cannot be in the data statement
        alcenchr(2)=char(9)
        justnumchr(2)=char(9)
	if (key==1) then
	  alchr=alsymchr
	elseif (key==2) then
	  alchr=alcenchr
	elseif (key==3) then
	  alchr=justnumchr
	elseif (key==4) then
	  alchr=reflchr
	elseif (key==5) then
	  alchr=digits
	endif
	DO loop=1,idel(string)
	  ok=.false.
	  DO loop2=1,CharNr
	    IF (string(loop:loop)==alchr(loop2)) THEN
	      ok=.true.
	      exit
	    END IF
	  END DO
          ErrCd=0
	  if (.not.ok) then
	    ErrCd=100
	    RETURN
	  endif
        END DO
        END SUBROUTINE CheckAllowedChars

      END SUBROUTINE INPUT
