!        Read electron density in Jana(98?) format
!******************************************************************
      SUBROUTINE ReadJana (rhofile, FError)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables
      IMPLICIT NONE
      CHARACTER :: rhofile*(FileLngth)  ! Filename
      INTEGER, INTENT (OUT) :: FError   ! Errorhandling
      INTEGER :: nx(6)                  ! Number of steps in direction iorient(i)
      INTEGER :: nxny                   !=nx(1)*nx(2)
      INTEGER :: nmap                   ! Number of sections
      INTEGER :: mapa                   ! Type of map (1=Patterson, 2=Fourier, ?=differece Fourier, ?=checking Fourier)
      REAL    :: dx(6)                  ! Step in iorien(i)
      INTEGER :: iorien(6)              ! Order of axes
      INTEGER :: nsubs                  ! part number of composite crystals (?)
      REAL :: rmin, rmax                ! Minimum and maximum of the whole table
      REAL :: xmin(6), xmax(6)          ! minimum and maximum value of iorien(i)
      REAL, ALLOCATABLE :: table(:)     ! sections of density
      INTEGER :: Loc                    ! Location in rho-array
      LOGICAL :: LStop=.FALSE.          ! Error condition
      INTEGER :: i, lrec       ! Counters

      open(81,file=rhofile,form='unformatted',status='old', access='direct',recl=34*ReclFac,err=900)
      read(81,rec=1,err=900)(lrec,i=1,7)
      close(81)
      open(81,file=rhofile,form='unformatted',status='unknown', &
             access='direct',recl=max(lrec,34)*ReclFac,err=900)      

!       Header of Jana file
      READ (81, rec=1, ERR=910) nx, nxny, nmap, (xmin(cnt1),xmax(cnt1),cnt1=1,6), dx, iorien, mapa, nsubs
      DO Cnt1=1,6
        IF ((iorien(Cnt1)/=Cnt1).AND.(iorien(Cnt1)/=0)) LStop=.TRUE.
      END DO
      IF (LStop) GOTO 930
      IF ((mapa/=4).AND.(mapa/=5).AND.(mapa/=6)) GOTO 950
!       Values of map
      ALLOCATE (table(nxny), STAT=AllocError)
      IF (AllocError.GT.0) STOP 'Error allocating tables of Jana file!'
!       check the voxel division
      loc=fstel-1
      DO Cnt1=1,nmap
        READ (81, rec=Cnt1+1, ERR=910) (table(Cnt2), Cnt2=1,nxny)
!       Mapping into BayMEM rho      
        DO Cnt2=1, nxny
          Loc=Loc+1
          rho(Loc)=dble(table(Cnt2))
        END DO  
      END DO
      READ (81, rec=nmap+2, ERR=800) rmax,rmin
      rhomin=rmin
      rhomax=rmax
800   continue
      CLOSE (81, ERR=920)
      FError=0
     
      GOTO 999
!       Error handling      
900   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadJana:'
      WRITE (*,'(A)') 'Error! Cannot open density file for reading!'
      WRITE (*,'(A)') 'Nothing has been read! Returning...'
      WRITE (*,'(A)') ''
      CLOSE (81, ERR=920)
      FError=1
      GOTO 999

910   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadJana:'
      WRITE (*,'(A)') 'Error! Cannot read density file!'
      WRITE (*,'(A)') 'Something may have been read! Returning...'
      WRITE (*,'(A)') ''
      CLOSE (81, ERR=920)
      FError=2
      GOTO 999

920   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadJana:'
      WRITE (*,'(A)') 'Error! Cannot close density file!'
      WRITE (*,'(A)') 'Density should have been read, but Im not sure! Returning...'
      WRITE (*,'(A)') ''
      FError=3
      GOTO 999

930   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadJana:'
      WRITE (*,'(A)') 'Warning!'
      WRITE (*,'(A)') 'Cannot read this file!'
      WRITE (*,'(A)') 'Order of axes is not a b c ...!'
      WRITE (*,'(A)') 'Please save density with normal order of axes in JANA.'
      WRITE (*,'(A)') ''
      FError=4
      GOTO 999
      
940   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadJana:'
      WRITE (*,'(A)') 'Warning!'
      WRITE (*,'(A)') 'Cannot read this file!'
      WRITE (*,'(A)') 'Total number of voxels is inconsistent with input file!'
      WRITE (*,'(A)') ''
      FError=5
      GOTO 999

950   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadJana:'
      WRITE (*,'(A)') 'Warning!'
      WRITE (*,'(A)') 'Cannot read this file!'
      WRITE (*,'(A)') 'Map type is not Fourier!'
      WRITE (*,'(A)') 'Please save density as Fourier map.'
      WRITE (*,'(A)') ''
      FError=6
      GOTO 999

999   END SUBROUTINE ReadJana


!        Write electron density in Jana(98?) format
!******************************************************************
      SUBROUTINE WriteJana (rhofile, p1, FError)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables
      IMPLICIT NONE
      CHARACTER :: rhofile*(FileLngth)  ! Filename
      INTEGER, INTENT (OUT) :: FError   ! Errorhandling
      INTEGER :: NvoxOut(6)                !Output number of voxels, differs for p1=.true. from the working number
      INTEGER :: NVoxTotOut                !total output number of voxels
      INTEGER :: dxyz(6),Nj(6)           ! dummy vector, jump lengths for calculating the coordinates, when p1=.true.
      INTEGER :: nx(6)                  ! Number of steps in direction iorient(i)
      INTEGER :: nxny                   !=nx(1)*nx(2)
      INTEGER :: nmap                   ! Number of sections
      INTEGER :: mapa                   ! Type of map 
      REAL    :: dx(6)                  ! Step in iorien(i)
      INTEGER :: iorien(6)              ! Order of axes
      INTEGER :: nsubs                  ! part number of composite crystals (?)
      REAL :: xmin(6), xmax(6)          ! minimum and maximum value of iorien(i)
      REAL :: rmin, rmax                ! Minimum and maximum of the whole table
      REAL, ALLOCATABLE :: table(:)     ! sections of density
      INTEGER :: Loc                    ! Location in rho-array
      INTEGER :: Idel                         !Function returning the length of the string
      LOGICAL :: p1                        !Add the last redundant "slice" of voxel at 1.?
      INTEGER :: i,htmp                 !counter,dummy index in calculation of coordinates
!     Prepare Output voxels
      NVoxTotOut=1
      IF (p1) THEN
        DO i=1,SDim
          NVoxOut=1
          NVoxOut(i)=NVox(i)+1
          NVoxTotOut=NVoxTotOut*NVoxOut(i)
        END DO
!       Jump length accross the maps
        Nj=0
        Nj(2)=1
        DO i=3, SDim
          Nj(i)=NVoxOut(i)*Nj(i-1)
        END DO
      ELSE
        NVoxOut=1
        DO i=1,SDim
          NVoxOut(i)=NVox(i)
        END DO
        NVoxTotOut=NVoxTot
      END IF

!       Doing some copy work
!       Voxel spacing
      DO Cnt1=1,6
        nx(Cnt1)=1
      END DO
      DO Cnt1=1,SDim
        nx(Cnt1)=NVoxOut(Cnt1)
      END DO
!       Maplength
      nxny=NVoxOut(1)*NVoxOut(2)
!       Number of maps
      nmap=NVoxTotOut/nxny
!       Kind of map: 4 = Fourier
      mapa=4
      DO Cnt1=1,6
        IF (Cnt1<=SDim) THEN
!       Stepwidth
          dx(Cnt1)=1./NVox(Cnt1) !Here NVox must stay, not replaced by NVoxOut!
!       Maxima
          if (p1) THEN
            xmax(Cnt1)=1.
          ELSE
            xmax(Cnt1)=1.-dx(Cnt1)
          END IF
        ELSE
          dx(Cnt1)=1.0
          xmax(Cnt1)=0.0
        END IF
        xmin(Cnt1)=0.
!       Orientation of axes, ever 1 2 3 4 5 6
        iorien(Cnt1)=Cnt1
      END DO
!       Part of composite Crystal
      nsubs=1
!       Minimum of table
      rmin=100000000.0
!       Maximum of table
      rmax=-100000000.0
            
      open(81,file=rhofile,form='unformatted',status='NEW', &
             access='direct',recl=max(nxny,34)*ReclFac,err=900)      
!       Header of Jana file
      WRITE (81, rec=1, ERR=910) nx, nxny, nmap, (xmin(cnt1),xmax(cnt1),cnt1=1,6), dx, iorien, mapa, nsubs

!       Values of map
      ALLOCATE (table(nxny), STAT=AllocError)
      IF (AllocError.GT.0) STOP 'Error allocating tables of Jana file!'
      Loc=FstEl-1
      DO Cnt1=1,nmap
!       Mapping from BayMEM rho      
        DO Cnt2=1, nxny
!         Part for p1=.true.
          IF (p1) THEN
!         determine coordinates within the cell
            dxyz=0
            dxyz(1)=mod(Cnt2-1,NVoxOut(1))
            dxyz(2)=(Cnt2-1)/NVoxOut(1)
            Loc=Cnt1-1 !because first map has number 1, but the coordinate is 0
!           Reversed step direction
            DO i=SDim, 3, -1
              dxyz(i)=INT(Loc/Nj(i-1))
              Loc=Loc-dxyz(i)*Nj(i-1)
            END DO
!           The rest is a modified subroutine ArrayLocation, with remapping of the voxels to the unit cell
            Loc=0
            DO i=1, SDim
!             Mapping into unit cell
              htmp=dxyz(i)
100           IF (htmp<0) THEN
                htmp=NVox(i)+htmp
                goto 100
              ELSE IF (htmp>(NVox(i)-1)) THEN
                htmp=htmp-NVox(i)
                goto 100
              END IF
              Loc=Loc+htmp*N(i-1)
            END DO
!*********************
!         Part for p1=.false. - how simple :-)
          ELSE
            Loc=Loc+1
          END IF
          table(Cnt2)=rho(Loc)
          IF (table(Cnt2)>rmax) rmax=table(Cnt2)
          IF (table(Cnt2)<rmin) rmin=table(Cnt2)
        END DO  
        WRITE (81, rec=Cnt1+1, ERR=910) (table(Cnt2), Cnt2=1,nxny)
      END DO
      WRITE (81, rec=nmap+2, ERR=910) rmax, rmin
      CLOSE (81, ERR=920)
      FError=0
!     write information about the m81 file
      WRITE(30,'(A)') 'Electron density written to file '//rhofile(1:idel(rhofile))
      GOTO 999
      
!       Error handling      
900   WRITE (30,'(A)') ''
      WRITE (30,'(A)') 'Routine WriteJana:'
      WRITE (30,'(A)') 'Error! Cannot open density file '//rhofile(1:idel(rhofile))//' for writing!'
      WRITE (30,'(A)') 'Output file: '//rhofile(1:idel(rhofile))
      WRITE (30,'(A)') 'Nothing has been written! Returning...'
      WRITE (30,'(A)') ''
      CLOSE (81, ERR=920)
      FError=1
      GOTO 999

910   WRITE (30,'(A)') ''
      WRITE (30,'(A)') 'Routine WriteJana:'
      WRITE (30,'(A)') 'Error! Cannot write density file '//rhofile(1:idel(rhofile))//' !'
      WRITE (30,'(A)') 'Output file: '//rhofile(1:idel(rhofile))
      WRITE (30,'(A)') 'Something may have been written! Returning...'
      WRITE (30,'(A)') ''
      CLOSE (81, ERR=920)
      FError=2
      GOTO 999

920   WRITE (30,'(A)') ''
      WRITE (30,'(A)') 'Routine WriteJana:'
      WRITE (30,'(A)') 'Error! Cannot close density file '//rhofile(1:idel(rhofile))//' !'
      WRITE (30,'(A)') 'Output file: '//rhofile(1:idel(rhofile))
      WRITE (30,'(A)') 'Density should have been written, but Im not sure! Returning...'
      WRITE (30,'(A)') ''
      FError=3
      GOTO 999

999   CONTINUE
      END SUBROUTINE WriteJana

!        Write real-D electron density in Jana(98?) format
!******************************************************************
      SUBROUTINE WriteJanaR (rhofile, p1, FError)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables
      IMPLICIT NONE
      CHARACTER :: rhofile*(*)          ! Filename
      INTEGER, INTENT (OUT) :: FError   ! Errorhandling
      INTEGER :: NvoxOut(6)                !Output number of voxels, differs for p1=.true. from the working number
      INTEGER :: NVoxTotOut                !total output number of voxels
      INTEGER :: dxyz(6),Nj(6)           ! dummy vector, jump lengths for calculating the coordinates, when p1=.true.
      INTEGER :: nx(6)                  ! Number of steps in direction iorient(i)
      INTEGER :: nxny                   !=nx(1)*nx(2)
      INTEGER :: nmap                   ! Number of sections
      INTEGER :: mapa                   ! Type of map 
      REAL    :: dx(6)                  ! Step in iorien(i)
      INTEGER :: iorien(6)              ! Order of axes
      INTEGER :: nsubs                  ! part number of composite crystals (?)
      REAL :: xmin(6), xmax(6)          ! minimum and maximum value of iorien(i)
      REAL :: rmin, rmax                ! Minimum and maximum of the whole table
      REAL, ALLOCATABLE :: table(:)     ! sections of density
      INTEGER :: Loc                    ! Location in rho-array
      INTEGER :: Idel                         !Function returning the length of the string
      LOGICAL :: p1                        !Add the last redundant "slice" of voxel at 1.?
      INTEGER :: i,htmp                 !counter,dummy index in calculation of coordinates
!     Prepare Output voxels
      NVoxTotOut=1
      IF (p1) THEN
        DO i=1,RDim
          NVoxOut=1
          NVoxOut(i)=NVoxRim(i)+1
          NVoxTotOut=NVoxTotOut*NVoxOut(i)
        END DO
!       Jump length accross the maps
        Nj=0
        Nj(2)=1
        DO i=3, RDim
          Nj(i)=NVoxOut(i)*Nj(i-1)
        END DO
      ELSE
        NVoxOut=1
        DO i=1,RDim
          NVoxOut(i)=NVoxRim(i)
        END DO
        NVoxTotOut=NVoxTotRim
      END IF

!       Doing some copy work
!       Voxel spacing
      DO Cnt1=1,6
        nx(Cnt1)=1
      END DO
      DO Cnt1=1,RDim
        nx(Cnt1)=NVoxOut(Cnt1)
      END DO
!       Maplength
      nxny=NVoxOut(1)*NVoxOut(2)
!       Number of maps
      nmap=NVoxTotOut/nxny
!       Kind of map: 4 = Fourier
      mapa=4
      DO Cnt1=1,6
        IF (Cnt1<=RDim) THEN
!       Stepwidth
          dx(Cnt1)=1./NVoxR(Cnt1) !Here NVox must stay, not replaced by NVoxOut!
!       Maxima and minima
          xmin(Cnt1)=-dble(Rim(Cnt1))*dx(Cnt1)
          if (p1) THEN
            xmax(Cnt1)=1.+dble(Rim(Cnt1))*dx(Cnt1)
          ELSE
            xmax(Cnt1)=1.+dble(Rim(Cnt1)-1)*dx(Cnt1)
          END IF
        ELSE
          dx(Cnt1)=1.0
          xmin(Cnt1)=0.
          xmax(Cnt1)=0.0
        END IF
!       Orientation of axes, ever 1 2 3 4 5 6
        iorien(Cnt1)=Cnt1
      END DO
!       Part of composite Crystal
      nsubs=1
!       Minimum of table
      rmin=100000000.0
!       Maximum of table
      rmax=-100000000.0
      open(81,file=rhofile,form='unformatted',status='UNKNOWN', &
             access='direct',recl=max(nxny,34)*ReclFac,err=900)      
!       Header of Jana file
      WRITE (81, rec=1, ERR=910) nx, nxny, nmap, (xmin(cnt1),xmax(cnt1),cnt1=1,6), dx, iorien, mapa, nsubs

!       Values of map
      ALLOCATE (table(nxny), STAT=AllocError)
      IF (AllocError.GT.0) STOP 'Error allocating tables of Jana file!'
      Loc=FstEl-1
      DO Cnt1=1,nmap
!       Mapping from BayMEM rho      
        DO Cnt2=1, nxny
!         Part for p1=.true.
          IF (p1) THEN
!         determine coordinates within the cell
            dxyz=0
            dxyz(1)=mod(Cnt2-1,NVoxOut(1))
            dxyz(2)=(Cnt2-1)/NVoxOut(1)
            Loc=Cnt1-1 !because first map has number 1, but the coordinate is 0
!           Reversed step direction
            DO i=RDim, 3, -1
              dxyz(i)=INT(Loc/Nj(i-1))
              Loc=Loc-dxyz(i)*Nj(i-1)
            END DO
!           The rest is a modified subroutine ArrayLocation, with remapping of the voxels to the unit cell
            Loc=0
            DO i=1, RDim
!             Mapping into unit cell
              htmp=dxyz(i)
100           IF (htmp<0) THEN
                htmp=NVoxR(i)+htmp
                goto 100
              ELSE IF (htmp>(NVoxR(i)-1)) THEN
                htmp=htmp-NVoxR(i)
                goto 100
              END IF
              Loc=Loc+htmp*NRim(i-1)
            END DO
!*********************
!         Part for p1=.false. - how simple :-)
          ELSE
            Loc=Loc+1
          END IF
          table(Cnt2)=rhor(Loc)
          IF (table(Cnt2)>rmax) rmax=table(Cnt2)
          IF (table(Cnt2)<rmin) rmin=table(Cnt2)
        END DO  
        WRITE (81, rec=Cnt1+1, ERR=910) (table(Cnt2), Cnt2=1,nxny)
      END DO
      WRITE (81, rec=nmap+2, ERR=910) rmax, rmin
      CLOSE (81, ERR=920)
      FError=0
!     write information about the m81 file
      WRITE(30,'(A)') 'Electron density written to file '//rhofile(1:idel(rhofile))
      GOTO 999
      
!       Error handling      
900   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine WriteJanaR:'
      WRITE (*,'(A)') 'Error! Cannot open density file '//rhofile(1:idel(rhofile))//' for writing!'
      WRITE (*,'(A)') 'Output file: '//rhofile(1:idel(rhofile))
      WRITE (*,'(A)') 'Nothing has been written! Returning...'
      WRITE (*,'(A)') ''
      CLOSE (81, ERR=920)
      FError=1
      GOTO 999

910   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine WriteJanaR:'
      WRITE (*,'(A)') 'Error! Cannot write density file '//rhofile(1:idel(rhofile))//' !'
      WRITE (*,'(A)') 'Output file: '//rhofile(1:idel(rhofile))
      WRITE (*,'(A)') 'Something may have been written! Returning...'
      WRITE (*,'(A)') ''
      CLOSE (81, ERR=920)
      FError=2
      GOTO 999

920   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine WriteJanaR:'
      WRITE (*,'(A)') 'Error! Cannot close density file '//rhofile(1:idel(rhofile))//' !'
      WRITE (*,'(A)') 'Output file: '//rhofile(1:idel(rhofile))
      WRITE (*,'(A)') 'Density should have been written, but Im not sure! Returning...'
      WRITE (*,'(A)') ''
      FError=3
      GOTO 999

999   CONTINUE
      END SUBROUTINE WriteJanaR


!        Read electron density in XPlor format
!******************************************************************
      SUBROUTINE ReadXPlor(RhoFile,target)
!******************************************************************
      USE GlobalDefinitions
      USE GlobalVariables
      USE input_module
      IMPLICIT NONE
      INTEGER, PARAMETER :: rhu=81
      CHARACTER*(*) :: rhofile  ! Filename
      character*10 :: zyxstr
      character*8 :: efmt='( e  .0)', ffmt='( f  .0)', curfmt 
      INTEGER, DIMENSION(3*SDim) :: NVoxTmp
      DOUBLE PRECISION, DIMENSION(CellCon) :: Celltmp
      INTEGER :: Loc,nmap,ntable,linecnt,ErrCode,target

      linecnt=0
      OPEN (rhu, ERR=900, FILE=rhofile,STATUS='OLD')
! First of all try to determine the format
      READ (rhu, '(a)', ERR=930)
      READ (rhu, '(a)', ERR=930) Buffer
      CALL OneInt(Buffer,Loc,ErrCode)
      IF (ErrCode/=0) GOTO 930
      DO cnt1=1,Loc+6
        linecnt=linecnt+1
        READ(rhu,'(a)',ERR=930)
      ENDDO
      READ (rhu, '(a)', ERR=930) Buffer
      !Buffer contains the first line of pixels
      IF ((index(Buffer,'E')/=0).or.(index(Buffer,'r')/=0)) THEN
        curfmt=efmt
      ELSE
        curfmt=ffmt      
      ENDIF
      loc=0
      DO cnt1=1,len_trim(Buffer)
        if (Buffer(cnt1:cnt1)=='.') loc=loc+1
      ENDDO
      write(curfmt(2:2),'(i1)') loc
      write(curfmt(4:5),'(i2)') len_trim(buffer)/loc
      REWIND(rhu)
!       Header of file
!       Empty line
      linecnt=linecnt+1
      ErrCode=0
      READ (rhu, '(a)', ERR=930)
!       number of title lines, skip them
      linecnt=linecnt+1
      READ (rhu, '(a)', ERR=930,END=930) Buffer
      CALL OneInt(Buffer,Loc,ErrCode)
      IF (ErrCode/=0) GOTO 930
      DO cnt1=1,Loc
        linecnt=linecnt+1
        READ(rhu,'(a)',ERR=930)
      ENDDO
!       Voxel spacing
      linecnt=linecnt+1
      READ(rhu,'(a)',ERR=930) Buffer
      CALL ArrayInt(Buffer,NVoxTmp,3*SDim,ErrCode)
      DO cnt1=1,SDim
        IF (NVoxtmp(3*cnt1-2)/=NVox(cnt1)) THEN
          write(*,'(a,i2,a)') 'Error, the voxel division in the density along dimension',cnt1,&
                           ' is inconsistent with the input file.'
          STOP 'Error during reading of the reference density. Check the log file.'
        ELSEIF((NVoxtmp(3*cnt1-1)/=0).or.(NVoxtmp(3*cnt1)/=NVox(cnt1)-1)) THEN
          write(*,'(a,i2,a)') 'Warning, the start or end of the grid in the density along dimension',cnt1,&
                           'does not lie on the borders of the unit cell.'
        ENDIF
      ENDDO
      IF (ErrCode/=0) GOTO 930
!       Cellparameters
      linecnt=linecnt+1
      READ(rhu,'(a)',ERR=930) Buffer
      CALL ArrayReal(Buffer,Celltmp,CellCon-1,ErrCode)
      IF (ErrCode/=0) GOTO 930
      DO cnt1=1,SDim
        IF (abs(Celltmp(cnt1)-Cell(cnt1))>10.*eps) THEN
          write(*,'(a,i2,a)') 'Warning, cell parameter number',cnt1,' in the density '//&
                           ' is different from what is given in the input file.'
        ENDIF
      ENDDO
!      String ZYX defining the order of axes
      linecnt=linecnt+1
      READ(rhu,'(a)',ERR=930) zyxstr
      if (zyxstr/='ZYX') THEN
        write(*,'(a,i2,a)') 'Warning, string ZYX expected on line',linecnt,' of the density. Instead found '//zyxstr
      ENDIF
!      body of the file
      IF (SDim>1) THEN
        ntable=NVox(1)*NVox(2)
      ELSE
        ntable=NVox(1)
      ENDIF
      nmap=NVoxtot/ntable
      loc=0
      DO cnt1=0,nmap-1
        READ (rhu, '(a)', ERR=950) Buffer
        CALL OneInt(Buffer,cnt3,ErrCode)
        write(*,*) Buffer
        write(*,*) Cnt3,cnt1
        write(*,*) curfmt, ntable
        IF (Cnt3/=cnt1) THEN
          WRITE(*,'(a,i5,a)') 'Error, inconsistency in the density file.'
          WRITE(*,'(a,i4,a,i4,a)') 'Label of the layer nr.',cnt1,' is',cnt3,'.'
          STOP 'Error during reading of the reference density. Check the log file.'
        ENDIF
        IF (ErrCode/=0) GOTO 950
        READ (rhu,curfmt, ERR=950) rho(loc:loc+ntable-1)
        loc=loc+ntable
      END DO
      CLOSE (rhu, ERR=920)
      GOTO 999

!       Error handling      
900   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadXPlor:'
      WRITE (*,'(A)') 'Error! Cannot open density file for reading!'
      WRITE (*,'(A)') 'Input file file: '//trim(rhofile)
      WRITE (*,'(A)') 'Stopping.'
      WRITE (*,'(A)') ''
      STOP 'Error during reading of the density.'
      

920   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadXPlor:'
      WRITE (*,'(A)') 'Error! Cannot close density file.'
      WRITE (*,'(A)') 'Input file: '//trim(rhofile)
      WRITE (*,'(A)') ''
      STOP 'Error during reading of the density.'

930   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadXPlor:'
      STOP 'Error during reading of the density.'

950   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadXPlor:'
      WRITE(*,'(a,i5)') 'Error reading reference density at layer',cnt1+1
      STOP 'Error during reading of the density.'

999   CONTINUE
      END SUBROUTINE ReadXPlor

!        Read electron density in XPlor format
!******************************************************************
      SUBROUTINE ReadXPlorOld(RhoFile,target)
!******************************************************************
      USE GlobalDefinitions
      USE GlobalVariables
      USE input_module
      IMPLICIT NONE
      CHARACTER :: rhofile*(FileLngth)  ! Filename
      character*10 :: zyxstr 
      INTEGER, DIMENSION(3*MaxDim) :: NVoxTmp
      REAL, DIMENSION(CellCon) :: Celltmp
      DOUBLE PRECISION :: dummy
      INTEGER :: Loc,nmap,ntable,linecnt,ErrCode,target
      INTEGER :: Idel                         !Function returning the length of the string

      linecnt=0
      OPEN (81, ERR=900, FILE=rhofile,STATUS='OLD')
!       Header of file
!       Empty line
      linecnt=linecnt+1
      ErrCode=0
      READ (81, '(a)', ERR=950)
!       number of title lines, skip them
      linecnt=linecnt+1
      READ (81, '(a)', ERR=950,END=950) Buffer
      CALL OneInt(Buffer,Loc,ErrCode)
      IF (ErrCode/=0) GOTO 950
      DO cnt1=1,Loc
        linecnt=linecnt+1
        READ(81,'(a)',ERR=950)
      ENDDO
!       Voxel spacing
      linecnt=linecnt+1
      READ(81,'(a)',ERR=950) Buffer
      CALL ArrayInt(Buffer,NVoxTmp,3*SDim,ErrCode)
      DO cnt1=1,SDim
        IF (NVoxtmp(3*cnt1-2)/=NVox(cnt1)) THEN
          write(*,'(a,i2,a)') 'Error, the voxel division in the reference density along dimension',cnt1,&
                           ' is inconsistent with the input file.'
          STOP 'Error during reading of the reference density.'
        ELSEIF((NVoxtmp(3*cnt1-1)/=0).or.(NVoxtmp(3*cnt1)/=NVox(cnt1)-1)) THEN
          write(*,'(a,i2,a)') 'Warning, the start or end of the grid in the density along dimension',cnt1,&
                           ' does not lie on the borders of the unit cell.'
        ENDIF
      ENDDO
      IF (ErrCode/=0) GOTO 950
!       Cellparameters
      linecnt=linecnt+1
      READ(81,'(a)',ERR=950) Buffer
!      String ZYX defining the order of axes
      linecnt=linecnt+1
      READ(81,'(a)',ERR=950) zyxstr
      if (zyxstr/='ZYX') THEN
        write(*,'(a,i2,a)') 'Warning, string ZYX expected on line',linecnt,' of the reference density. Instead found '//zyxstr
      ENDIF
!      body of the file
      loc=-1
      IF (SDim>1) THEN
        ntable=NVox(1)*NVox(2)
      ELSE
        ntable=NVox(1)
      ENDIF
      nmap=NVoxtot/ntable
      DO cnt1=0,nmap-1
        linecnt=linecnt+1
        READ (81, '(a)', ERR=950) Buffer
        CALL OneReal(Buffer,dummy,ErrCode)
        Cnt3=nint(dummy)
        IF (Cnt3/=cnt1) THEN
          WRITE(*,'(a,i5,a)') 'Error, inconsistency in the reference density on line',linecnt,'.'
          WRITE(*,'(a,i4,a,i4,a)') 'Label of the layer nr.',cnt1,' is',cnt3,'.'
          STOP 'Error during reading of the reference density.'
        ENDIF
        IF (ErrCode/=0) GOTO 950
        DO cnt2=0,ntable-1
          loc=loc+1
          linecnt=linecnt+1
          READ (81, '(a)', ERR=950) Buffer
          CALL OneReal(Buffer,rho(loc),ErrCode)
        END DO
      END DO
      CLOSE (81, ERR=920)
      GOTO 999

!       Error handling      
900   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadXPlor:'
      WRITE (*,'(A)') 'Error! Cannot open density file for reading!'
      WRITE (*,'(A)') 'Input file file: '//rhofile(1:idel(rhofile))
      WRITE (*,'(A)') 'Stopping.'
      WRITE (*,'(A)') ''
      STOP 'Error during reading of the reference density.'
      

920   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadXPlor:'
      WRITE (*,'(A)') 'Error! Cannot close density file.'
      WRITE (*,'(A)') 'Input file: '//rhofile(1:idel(rhofile))
      WRITE (*,'(A)') ''
      STOP 'Error during reading of the reference density.'

950   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadXPlor:'
      WRITE (*,'(a,i5)') 'Error reading reference density on line',linecnt
      WRITE (*,'(a)') 'Lately reading: '//Buffer(:idel(Buffer))
      STOP 'Error during reading of the reference density.'

999   CONTINUE
      END SUBROUTINE ReadXPlorOld
