!Library contains subroutines, which serve for spline interpolation of the electron density
!5.7.2002+ Lukas Palatinus

      SUBROUTINE GetPeak(Centr, ExPeak,outrho,fast) !Gets coordinates of the pixel, returns accurate coordinates and density of the maximum
      USE GlobalDefinitions
      USE GlobalVariables
      IMPLICIT NONE
      INTEGER :: Centr(MaxDim) ! Central voxel of the maximum, dummy coords
      INTEGER :: AxOrd(MaxDim),DerOrd(MaxDim),Orig(MaxDim)
      DOUBLE PRECISION :: ExPeak (MaxDim) ! The exact(er) peak position
      DOUBLE PRECISION :: outrho !peak electron density
      DOUBLE PRECISION, DIMENSION(range) :: y,y2 !arrays for subroutine splines
      INTEGER :: ii ! Counters
      LOGICAL :: Succ !flag of success
      LOGICAL :: fast !switches between fast, but possibly less accurate, and slower, more precise calculation. .true. is fast calculation

      ExPeak=dble(Centr)
      call CPRefine(ExPeak,Succ,fast) !returns the coordinates of the maximum, succ is false, if no maximum could be found
      IF (Succ) THEN
        DerOrd=0
        DO ii=1,RDim
          AxOrd(ii)=ii
          Orig(ii)=int(ExPeak(ii)-float(range)/2.+range)+1-range
        END DO
        call FillRhoCut(AxOrd,Orig,RDim) !fill the part of the electron density around the maximum
        IF (fast) THEN
          call RhoAtPointFast(RDim,range,DerOrd,AxOrd,ExPeak,orig,srho,outrho,y,y2) !returns the peak density
        ELSE
          call RhoAtPoint(RDim,range,DerOrd,AxOrd,ExPeak,orig,srho,outrho,y,y2) !returns the peak density
        ENDIF
      ELSE
        ExPeak=-1.0E30
	outrho=-1.0E30
      END IF
      END SUBROUTINE GetPeak

      SUBROUTINE CPRefine(pt,succ,fast) ! find stationary point (derivatives=0)
        USE GLOBALDEFINITIONS
	USE GLOBALVARIABLES
        
	IMPLICIT NONE
        INTEGER, PARAMETER :: ptchangelim=40
        DOUBLE PRECISION, PARAMETER :: MaxGradShift=0.5
        DOUBLE PRECISION, PARAMETER :: MaxPtShift=3.
	DOUBLE PRECISION :: pt(MaxDim),ptor(MaxDim),ptref(MaxDim) !point, where the density is to be determined
			! PtOr to save the starting pt
			!PtRef was added for the cases pt is close to the pixel border. In that case, changing origin can cause oscillations.
        INTEGER, DIMENSION(MaxDim) :: AxOrd !order of the axes in the output rcut
        INTEGER, DIMENSION(MaxDim) :: DerOrd,Orig
        DOUBLE PRECISION :: rout,disc, Grad, GradOld, GradOlder, red, dump !output rho, 1st and 2nd solution of the qadr. equation, discrepancy
        DOUBLE PRECISION, DIMENSION(range) :: y,y2
        DOUBLE PRECISION, ALLOCATABLE :: H(:,:),Hinv(:,:),Ho(:,:),Gr(:), Sh(:) !Hessian, its inversion and gradient
        DOUBLE PRECISION, DIMENSION(6) :: pa !auxiliary matrix
	    LOGICAL :: succ !success
	    LOGICAL :: fast !switches between fast and slow (but more accurate) calculation of the derivatives
        INTEGER :: i,ii,iii, iv, ptchange !counters

        ALLOCATE(H(RDim,RDim),STAT=AllocError)
        ALLOCATE(Hinv(RDim,RDim),STAT=AllocError)
        ALLOCATE(Ho(RDim,RDim),STAT=AllocError)
        ALLOCATE(Gr(RDim),STAT=AllocError)
        ALLOCATE(Sh(RDim),STAT=AllocError)
        IF (AllocError.GT.0) STOP 'Error allocating in H, Hinv and Gr in CpRefine! Too little memory?'
        do i=1,RDim
          Orig(i)=int(pt(i)-dble(range)/2.+range)+1-range !set the origin of the region +- range/2 around pt
          DerOrd(i)=0
        end do
	ptchange=0 !counts the number of cycles
        dump=1.
        DO ii=1,RDim !permute the order of axes to refine each time another direction
          AxOrd(ii)=ii
        END DO
        ptor=pt
	ptref=pt
        DO
          ptchange=ptchange+1
	  succ=.true.
	  IF (ptchange>ptchangelim) THEN !trying too long time
	    Succ=.false.
	    EXIT
	  END IF
	  succ=.true. ! succ here checks, if pt didn't shift out of borders
          DO i=1,RDim
	    if (abs(pt(i)-ptref(i))>0.5) succ=.false.	    
	  ENDDO
	  IF (.not.succ) THEN !We have to update the origin, because the point shifted too much
	    succ=.true.
	    ptref=pt
            DO i=1,RDim !define the origin of the "cube" that serves to interpolate pt
              Orig(i)=int(ptref(i)-dble(range)/2.+range)+1-range
              if (((Orig(i)+(range-1)+Rim(i)>NVoxRim(i)-1).and.(Aper(i)==1)).or.  &
	        ((Orig(i)+Rim(i)<0).and.(Aper(i)==1))) succ=.false.
            END DO
	    IF (.not.succ) THEN
	      EXIT !If Orig runs out of range, subroutine is terminated with flag "unsuccessful"
          ENDIF
          ENDIF
          call FillRhoCut(AxOrd,Orig,RDim) !fill the area of rho with properly permuted axes
!********* Gradient and Hessian
          DO iii=1,3
	    DerOrd=0
	    DerOrd(iii)=1
	    IF (fast) THEN
            call RhoAtPointFast(RDim,range,DerOrd,AxOrd,pt,orig,srho,rout,y,y2) !y and y2 contain parameters of the spline interpolation along the last axis
	    ELSE
            call RhoAtPoint(RDim,range,DerOrd,AxOrd,pt,orig,srho,rout,y,y2) !y and y2 contain parameters of the spline interpolation along the last axis
	    ENDIF
            Gr(iii)=rout
          END DO
	  DO iii=1,RDim
	    DO iv=1,RDim
              DerOrd=0
	      DerOrd(iii)=DerOrd(iii)+1
	      DerOrd(iv)=DerOrd(iv)+1
        IF (fast) THEN
              call RhoAtPointFast(RDim,range,DerOrd,AxOrd,pt,orig,srho,rout,y,y2) !y and y2 contain parameters of the spline interpolation along the last axis
        ELSE
              call RhoAtPoint(RDim,range,DerOrd,AxOrd,pt,orig,srho,rout,y,y2) !y and y2 contain parameters of the spline interpolation along the last axis
        ENDIF
              H(iii,iv)=rout	      
	    END DO
	  END DO
          call invert(H,Hinv,RDim)
          IF (Hinv(1,1)<-1.0D29) THEN
            succ=.false.
	    EXIT
	  END IF
	  Sh=0.0D0
          Grad=0.0D0
	  DO iii=1,RDim
	    DO iv=1,RDim
	      Sh(iii)=Sh(iii)+Gr(iv)*Hinv(iv,iii)
	    END DO
            Grad=Grad+Sh(iii)**2
	  END DO
	  Grad=sqrt(Grad)
          IF (Grad>MaxGradShift) THEN
	    Red=MaxGradShift/Grad
	  ELSE
	    Red=1.0
	  END IF
          disc=0.0D0
	  DO iii=1,RDim
	    pt(iii)=pt(iii)-sh(iii)*red*dump
            disc=disc+(pt(iii)-ptor(iii))**2
	  END DO
	  disc=sqrt(disc)
	  IF (disc>MaxPtShift) Succ=.false.
	  IF (ptchange>2) THEN
	    if ((gradold-grad)>(gradolder-gradold)) dump=dump*0.5
	  END IF
          gradolder=gradold
	  gradold=grad
!          write(*,*) ptchange, grad,(pt(i),i=1,3), dump
	  IF (.not.Succ.or.grad<eps) EXIT !loop until .not.succ or disc is sufficiently small
	END DO
        IF (succ) THEN
!	  write(*,*) pt(1)/NVox(1), pt(2)/NVox(2),pt(3)/NVox(3)
           
          call TrMat(Mv,Ho,RDim)
          call MatMatMul(H,Ho,Hinv,RDim)
          call MatMatMul(Mv,Hinv,Ho,RDim) !Output is the Hessian in the orthonormal system
!	  DO iii=1,RDim
!	    DO iv=1,RDim
!	      write(*,'(f14.6)',ADVANCE='NO') Ho(iii, iv)
!	    END DO
!	    write(*,*)
!	  END DO
!          write(*,*)
! Diagonalize the Hessian
          pa(1)=Ho(1,1)
          pa(2)=Ho(2,1)
          pa(3)=Ho(2,2)
          pa(4)=Ho(3,1)
          pa(5)=Ho(3,2)
          pa(6)=Ho(3,3)
          call ql(pa,Hinv,Evl,RDim)
!Hinv contains eigenvectors as column vectors, Evl contains eigenvalues
          call Trmat(M,Ho,RDim)
	  call MatMatMul(Ho,Hinv,Evc,RDim)
	END IF
      END SUBROUTINE CPRefine
      

      SUBROUTINE FillRhoCut(AxOrd,Org,Dim) !fills part of the electron density from -range/2 to +range/2
        USE GLOBALDEFINITIONS
	USE GLOBALVARIABLES
        IMPLICIT NONE
	INTEGER :: ArrLoc, Dim
        INTEGER :: Dxyz(MaxDim),EndC(MaxDim)     ! output from subrt. Coordinates, input to ArrayLocation
        INTEGER, DIMENSION(*) :: AxOrd,Org !order of the axes in the output rcut, origin in the rho
        LOGICAL ::NextCoo
	
	DO Cnt1=1,RDim !set start and end of the cube
          Dxyz(Cnt1)=Org(Cnt1)
          IF (Cnt1<=Dim) THEN
	    EndC(AxOrd(Cnt1))=Org(AxOrd(Cnt1))+range-1
          ELSE
	    EndC(AxOrd(Cnt1))=Org(AxOrd(Cnt1))
	  END IF
	END DO
        Cnt1=0
        DO
	  Call ArrayLocationR(Dxyz,ArrLoc,1) !get the location of this pixel in rhor
          srho(Cnt1)=rhor(ArrLoc)
          cnt1=cnt1+1
          IF (.not.NextCoo(RDim,Org,EndC,Dxyz,AxOrd)) EXIT
        END DO
      END SUBROUTINE FillRhoCut


      SUBROUTINE RhoAtPoint(Dim,range,DerOrd,AxOrd,pt,orig,inrho,rout,y,y2) !returns electron density at pt and parameters y,y2 od the spline interpolation
        IMPLICIT NONE
	INTEGER :: i,ii,iii,Pcnt,SCnt, rstep ! counters
        INTEGER :: Dim,ADim !Dimension of the rhocut, actual dimension during the spline interpolation
        INTEGER :: range !length of the vertices of the "hypercube"
        DOUBLE PRECISION,DIMENSION(*) :: pt !point, where the density is to be determined
        INTEGER, DIMENSION(*) :: Orig !origin of the rhocut
        DOUBLE PRECISION :: inrho(0:range**Dim-1) !Input array containing the cut of the density around pt
        DOUBLE PRECISION :: rcut(0:range**Dim-1), rcutbck(0:range**Dim-1) !Working array
        DOUBLE PRECISION :: onerow(0:range-1)
        INTEGER :: rowpos(0:range-1)
        DOUBLE PRECISION,DIMENSION(range) :: x,y,y2 !arrays for the routines spline and splint
        INTEGER, DIMENSION(*) :: AxOrd !order of the axes in the output rcut
        INTEGER, DIMENSION(*) :: DerOrd  ! Array defining the derivative of the function returned by RhoAtPoint
        INTEGER, DIMENSION(1:Dim) :: DerOrdTmp, AxOrdTmp,StartC, EndC,RunC, CurC
        DOUBLE PRECISION, DIMENSION(0:2) :: rval ! Density and its derivatives at the point pt, output
        DOUBLE PRECISION :: rout, NumDerivative !output
        LOGICAL :: NextCoo

!        write(*,*) derord(1:Dim)
!        write(*,*) pt(1:Dim)
        Scnt=0
        DO i=1,Dim
          SCnt=SCnt+DerOrd(AxOrd(i)) !How many derivatives must be taken
          DerOrdTmp(i)=DerOrd(AxOrd(i))
          AxOrdTmp(i)=i
        ENDDO
        rcut=inrho
!        write(*,*) 'V Rhoatpoint:',rcut(0:range**Dim-1)
!        write(*,*) 'pt:',pt(1:3)
!        write(*,*) 'Orig:',Orig(1:3)
!        write(*,*) 'Scnt,derord:',Scnt, derord(1:3)
!        goto 1000
        IF (SCnt>0) THEN
          ! fill rcut by a derivative of inrho defined by derord
          DO i=1,SCnt
            DO ii=1,Dim
              IF (DerOrdTmp(ii)>0) THEN
                DerOrdTmp(ii)=DerOrdTmp(ii)-1
                StartC=0
                EndC=range-1
                EndC(ii)=0
                RunC=StartC
                rstep=ii
                EXIT
              ENDIF
            ENDDO
!            write(*,*) 'Jdu do derivace', Dim
            DO
!              write(*,*) RunC
              !Fill a one-dimensional array for calculating the derivative:
              CurC=RunC
!              write(*,*) 'Do vnitrni smycky', RunC, EndC
!              write(*,*) 'Dim, AxOrdTmp:',Dim, AxOrdTmp
              DO ii=0,range-1
                CurC(rstep)=ii
                !get position in rcut
                PCnt=0
                DO iii=1,Dim
                  PCnt=PCnt+CurC(iii)*range**(iii-1)
                ENDDO
!                write(*,*) CurC, PCnt,RCut(PCnt)
                onerow(ii)=rcut(PCnt)
                rowpos(ii)=Pcnt
              ENDDO
!              write(*,*) 'Z vnitrni smycky'
!              write(*,*) 'Pocitam derivace'
              DO iii=0,range-1
                rcutbck(rowpos(iii))=NumDerivative(onerow,range,1.0D0,iii,1)
!                write(*,*) onerow(iii),rcutbck(rowpos(iii))
              ENDDO
!              write(*,*) 'Dopocitano'
!              write(*,*) StartC, EndC
 !             write(*,*)
              IF (.not.NextCoo(Dim,StartC,EndC,RunC,AxOrdTmp)) EXIT
            ENDDO
            rcut=rcutbck
          ENDDO
!          write(*,*) 'Mam derivaci'
!          read(*,*)
        ENDIF
!        write(*,*) 'Po numder:',rcut(0:range**Dim-1)

1000      DO Adim=Dim,1,-1 !Going from Dim to 1, at each cycle interpolating Adim-1 dimensional section of rcut
          Pcnt=1 !counter of the old rho
	  Scnt=0 !counter of the new, Adim-1 dimensional rho
	  DO i=0,range**ADim-1 ! repeat over all pixels of old rcut
            x(Pcnt)=float(Orig(AxOrd(Dim-Adim+1))+Pcnt-1)
	    y(Pcnt)=rcut(i) !fill one row
            PCnt=PCnt+1
	    IF (PCnt==range+1) THEN !if the row is full, interpolate the value at pt
	      PCnt=1
	      CALL spline(x,y,range,y2)
	      CALL splint(x,y,y2,range,pt(AxOrd(Dim-Adim+1)),rval) !rval(0) contains the density, rval(1) the first derivative and rval(2) the second derivative
!              rcut(Scnt)=rval(DerOrd(AxOrd(Dim-Adim+1)))
              rcut(Scnt)=rval(0) !assign the result to rho(Scnt).
	      						!It is perfectly safe to save to the same rho, that point will not be needed in this cycle anymore.
	      SCnt=SCnt+1
	    END IF
!            continue to the next row
          END DO
          ! the ADim-1 dimensional "first slice" of rcut now contains the interpolated section and we can continue with next step
        END DO
	rout=rcut(0)
!	write(*,*) rout
!	STOP
      END SUBROUTINE RhoAtPoint
    
      SUBROUTINE RhoAtPointFast(Dim,range,DerOrd,AxOrd,pt,orig,inrho,rout,y,y2) !returns electron density at pt and parameters y,y2 od the spline interpolation
        IMPLICIT NONE
    INTEGER :: i,Pcnt,SCnt ! counters
        INTEGER :: Dim,ADim !Dimension of the rhocut, actual dimension during the spline interpolation
        INTEGER :: range !length of the vertices of the "hypercube"
        DOUBLE PRECISION,DIMENSION(*) :: pt !point, where the density is to be determined
        INTEGER, DIMENSION(*) :: Orig !origin of the rhocut
        DOUBLE PRECISION :: inrho(0:range**Dim-1) !Input array containing the cut of the density around pt
        DOUBLE PRECISION :: rcut(0:range**Dim-1) !Working array
        DOUBLE PRECISION,DIMENSION(range) :: x,y,y2 !arrays for the routines spline and splint
        INTEGER, DIMENSION(*) :: AxOrd !order of the axes in the output rcut
        INTEGER, DIMENSION(*) :: DerOrd  ! Array defining the derivative of the function returned by RhoAtPoint
        DOUBLE PRECISION, DIMENSION(0:2) :: rval ! Density and its derivatives at the point pt, output
        DOUBLE PRECISION :: rout !output

        rcut=inrho
        DO Adim=Dim,1,-1 !Going from Dim to 1, at each cycle interpolating Adim-1 dimensional section of rcut
          Pcnt=1 !counter of the old rho
      Scnt=0 !counter of the new, Adim-1 dimensional rho
      DO i=0,range**ADim-1 ! repeat over all pixels of ald rcut
            x(Pcnt)=float(Orig(AxOrd(Dim-Adim+1))+Pcnt-1)
        y(Pcnt)=rcut(i) !fill one row
            PCnt=PCnt+1
        IF (PCnt==range+1) THEN !if the row is full, interpolate the value at pt
          PCnt=1
          CALL spline(x,y,range,y2)
          CALL splint(x,y,y2,range,pt(AxOrd(Dim-Adim+1)),rval) !rval(0) contains the density, rval(1) the first derivative and rval(2) the second derivative
              rcut(Scnt)=rval(DerOrd(AxOrd(Dim-Adim+1))) !assign the result to rho(Scnt).
                                !It is perfectly safe to save to the same rho, that point will not be needed in this cycle anymore.
          SCnt=SCnt+1
        END IF
!            continue to the next row
          END DO
          ! the ADim-1 dimensional "first slice" of rcut now contains the interpolated section and we can continue with next step
        END DO
    rout=rcut(0)
      END SUBROUTINE RhoAtPointFast


      SUBROUTINE Spline(x,y,n,y2) !copied from Numerical Recipies
      INTEGER :: i,k,n
      DOUBLE PRECISION :: x(n),y(n),y2(n)
      DOUBLE PRECISION :: p,sig,un,u(n)
      u(1)=0.
      y2(1)=0.
      do i=2,n-1
        sig=(x(i)-x(i-1))/(x(i+1)-x(i-1))
        p=sig*y2(i-1)+2.
        y2(i)=(sig-1.)/p
        u(i)=(6.*((y(i+1)-y(i))/(x(i+1)-x(i))-(y(i)-y(i-1))/(x(i)-x(i-1)))/(x(i+1)-x(i-1))-sig*u(i-1))/p
      end do
      un=0.
      y2(n)=0.
      do k=n-1,1,-1
        y2(k)=y2(k)*y2(k+1)+u(k)
      end do
      return
      END SUBROUTINE Spline

      SUBROUTINE splint(xa,ya,y2a,n,x,y) !copied from Num.Rec., strongly simplified and modified
      INTEGER :: n,khi,klo
      DOUBLE PRECISION :: x,xa(n),y2a(n),ya(n)
      DOUBLE PRECISION, DIMENSION(0:2) :: y ! value and its derivatives
      DOUBLE PRECISION :: a,b
      
!     x lies in our case always exactly in the middle of the range ->
!     no need to bisect, we can calculate klo and khi directly
      klo=int(x-xa(1)+dble(n))-n+1
      khi=klo+1
!     moreover, the spacing between the x-values is always 1 -> we can forget h; h=1
      a=xa(khi)-x
      b=x-xa(klo)
!     value
      y(0)=a*ya(klo)+b*ya(khi)+((a**3-a)*y2a(klo)+(b**3-b)*y2a(khi))/6.
!     the first derivative
      y(1)=(ya(khi)-ya(klo))+((1-3*a**2)*y2a(klo)+(3*b**2-1)*y2a(khi))/6.
!     the second derivative
      y(2)=a*y2a(klo)+b*y2a(khi)
      return
      END

   double precision function NumDerivative(arr,npoints,step,pos,derord)
      IMPLICIT NONE
      integer :: j,k,npoints,derord,pos,a, GetA,factorial
      double precision, DIMENSION(0:npoints-1) :: arr
      double precision:: step
      double precision, dimension(0:npoints-1) :: d

      !Calculate the coefficients d
      DO j=0,npoints-1
        IF (j==pos) CYCLE
        a=GetA(derord,npoints,pos,j)
        d(j)=dble(Factorial(derord)*a)/dble(factorial(j)*factorial(npoints-1-j))
        IF (mod(derord-j,2)/=0) d(j)=-d(j)
      ENDDO
      d(pos)=0
      DO j=0,npoints-1
        IF (j/=pos) d(pos)=d(pos)-d(j)
      ENDDO
      NumDerivative=0.0
      DO j=0,npoints-1
        NumDerivative=NumDerivative+d(j)*arr(j)
      ENDDO
      NumDerivative=NumDerivative/(step**derord)
    END FUNCTION NumDerivative

    integer function Factorial(n)
      integer :: n,k
      Factorial=1
      IF (n>0) THEN
        DO k=1,n
          Factorial=Factorial*k
        ENDDO
      ENDIF
    END FUNCTION Factorial

    integer function GetA(derord,npoints,i,j)
      implicit none
      integer :: derord,npoints,i,j
      integer :: k,l,PartA
      integer, dimension(0:npoints-3) :: plist
      IF (derord>2) STOP 'I am not ready for higher derivatives!'
      IF (i==j) STOP 'Internal error in numerical derivatives.'
      l=-1
      DO k=0,npoints-1
        IF ((k/=i).and.(k/=j)) THEN
          l=l+1
          plist(l)=k-i
        ENDIF
      ENDDO
      GetA=0

      IF (DerOrd==1) THEN
        GetA=1
        DO k=0,npoints-3
          GetA=GetA*plist(k)
        ENDDO
      ELSEIF (DerOrd==2) THEN
        GetA=0
        DO l=0,npoints-3
          PartA=1
          DO k=0,npoints-3
            IF (k/=l) PartA=PartA*plist(k)
          ENDDO
          GetA=GetA+PartA
        ENDDO
      ENDIF
    end function GetA
