!     Last change:  LP   16 Mar 2005   10:17 am
!***********************************************************************
      MODULE GlobalDefinitions
!***********************************************************************
      IMPLICIT NONE
!        Parameters and input variables used by almost all program parts
      
      INTEGER, PARAMETER :: DeltaX=2            ! Distance in rho and F array, must be two in complex arrays
      INTEGER, PARAMETER :: CellCon=7           ! Number of cell parameters
      INTEGER, PARAMETER :: MaxDim=8            ! Maximal dimension of input data
      INTEGER, PARAMETER :: MaxDimp1=MaxDim+1   ! In Refl()%hklFG the place of group
      INTEGER, PARAMETER :: MaxDimp2=MaxDim+2   ! In Refl()%hklFG the place of the corresponding arraylocation of the structure factor
      INTEGER, PARAMETER :: MaxMaxNr=30000        ! Maximal number of maxima in one Rdim map
      INTEGER, PARAMETER :: LineLength=256      ! Maximal line length
      INTEGER, PARAMETER :: FileLngth=264       ! Maximal path and filename length
      INTEGER, PARAMETER :: DPR = KIND(1.0D0)   ! double precision real
 
      INTEGER, PARAMETER :: ongridmode       = 1, &
      			    neargridmode     = 2, &
			    pedanticmode     = 3

      DOUBLE PRECISION, PARAMETER :: pi=3.141592653589793238462643383279502884197169399375D0 !Pi
      DOUBLE PRECISION, PARAMETER :: twopi=2.0D0*pi             ! Two pi
      DOUBLE PRECISION, PARAMETER :: torad=pi/180.0D0           ! Factor for converting degree to radians
      DOUBLE PRECISION, PARAMETER :: Eps=0.000001D0              ! Smaller than Eps is considered to be zero
      DOUBLE PRECISION, PARAMETER :: DEps=0.0000000001D0        ! Smaller than DEps is considered to be zero, for very accurate situations
      INTEGER, PARAMETER :: TotElNr=94                          !Number of elements in the ElSmb array
      CHARACTER*2, DIMENSION(1:TotElNr), PARAMETER :: ElSmb= &       ! Elemental symbols sorted by atomic number
        (/'H ','He','Li','Be','B ','C ','N ','O ','F ','Ne','Na','Mg','Al','Si','P ','S ','Cl','Ar','K ','Ca','Sc','Ti','V ','Cr',&
          'Mn','Fe','Co','Ni','Cu','Zn','Ga','Ge','As','Se','Br','Kr','Rb','Sr','Y ','Zr','Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd',&
          'In','Sn','Sb','Te','I ','Xe','Cs','Ba','La','Ce','Pr','Nd','Pm','Sm','Eu','Gd','Tb','Dy','Ho','Er','Tm','Yb','Lu','Hf',&
          'Ta','W ','Re','Os','Ir','Pt','Au','Hg','Tl','Pb','Bi','Po','At','Rn','Fr','Ra','Ac','Th','Pa','U ','Np','Pu'/)

      DOUBLE PRECISION, PARAMETER :: DefaultBiso=2.0

!       The first and last element of the rho and FFT arrays is:
      INTEGER, PARAMETER :: FstEl=0
      INTEGER, SAVE :: LstRhoEl=0               ! Rho array

      INTEGER, SAVE          :: ReclFac=4 !Factor to multiply recl when opening direct-access file
      INTEGER, SAVE          :: range=9 !  Range, which is taken into account for the spline interpolation
      DOUBLE PRECISION, SAVE :: tolerance=0.15D0 ! Tolerance for the assigning of the maximum to the same atom
      DOUBLE PRECISION, SAVE :: plim=0.0D0 !treshold for the maxima written to the output
      DOUBLE PRECISION, SAVE :: chlim=0.25D0 !treshold for the charge level (relative to maximum) to include in the calculation of the com
      DOUBLE PRECISION, SAVE :: chlimlistorig=0.D0 !treshold for the charge to be included in the listing
      DOUBLE PRECISION, SAVE :: chlimlist !treshold for the charge to be included in the listing
      CHARACTER*8            :: plimkey='absolute', chlimlistkey='absolute' !Can be abslolute or relative to the maximum
      DOUBLE PRECISION, SAVE :: addborder=0.5D0 !how much of the unit cell must be added in the incommensurate direction (for COM and charge)
      LOGICAL, SAVE          :: fullcellpeaks=.false. !Search peaks irrespective of the symmetry
      LOGICAL, SAVE          :: writetmap=.false.
      LOGICAL, SAVE          :: writebasin=.false.
      LOGICAL, SAVE          :: com=.false.
      INTEGER, SAVE          :: basinmode=ongridmode
      INTEGER, SAVE	     :: EdgeRemovalCycles=4
      LOGICAL, SAVE	     :: expandedcp=.false.
      LOGICAL, SAVE	     :: projection=.false. !Project the superspace densitz along the internal coordinates?
      LOGICAL, SAVE	     :: wantstructure=.false.     !Write out the list of maxima as a m40 file
      LOGICAL, SAVE	     :: m40forjana=.false.     !Write out the list of maxima as a m40 file
      LOGICAL, SAVE	     :: compm40=.false.     !Compare the resulting atomic list with another structure
      CHARACTER*10, SAVE     :: CPformat='none'
      CHARACTER*10, SAVE     :: maxformat='all' 
      CHARACTER*10, SAVE     :: position='absolute' 
      CHARACTER*10, SAVE     :: scale='fractional' 
!         Dimensions and voxel spacing
      INTEGER, SAVE :: SDim=3              ! The dimenson of super space=RDim+DDim, defaults to 3
      INTEGER, SAVE :: RDim=3              ! Dimension of the real space      
      INTEGER, SAVE :: DDim=0              ! Additional dimensions
      INTEGER, SAVE :: NVoxTot=1,NVoxTotR=1,NVoxTotRim=1  ! The number of voxels in the unitcell, in the real dimension unit cell
      END MODULE GlobalDefinitions
      
      
!***********************************************************************
      MODULE GlobalVariables
!***********************************************************************
!        Variables used by almost all program parts

      USE GlobalDefinitions                     ! First getting all parameters:

      IMPLICIT NONE

      INTEGER :: DimCount                       ! Counter for already worked dimensions
      INTEGER :: Cnt1, Cnt2, Cnt3, Cnt4, Start, Final   ! General counters
      CHARACTER :: Buffer*(LineLength),Buffer2*(LineLength)          ! String buffer
      INTEGER :: AllocError                     ! Errorcondition while allocating arrays

      DOUBLE PRECISION :: rhomax=0.0D0          ! Maximum in density array
      DOUBLE PRECISION :: rhomin                ! Minimum in density array
      INTEGER, DIMENSION (:), ALLOCATABLE, SAVE :: NVox,NVoxR,NVoxRim  ! Array with number of voxels per axis
      INTEGER, SAVE :: N(0:MaxDim), NRim(0:MaxDim)
      DATA N /MaxDimp1*1/
      DATA NRim /MaxDimp1*1/
      DOUBLE PRECISION, SAVE :: cApVx=1.0D0     ! 'cubic' Angstrom per voxel

!       The electron density
      DOUBLE PRECISION, DIMENSION (:), ALLOCATABLE, SAVE :: rho,rhor,rhogr,srho
!       The structure factor
      DOUBLE PRECISION, DIMENSION (:), ALLOCATABLE, SAVE :: SF
!       The q-vectors
      DOUBLE PRECISION, DIMENSION (:,:), ALLOCATABLE, SAVE :: QVec   ! q-vector
!       The t-setting
      DOUBLE PRECISION, DIMENSION (:,:), ALLOCATABLE, SAVE :: Tlist
!       The order of the axes
      INTEGER, DIMENSION (:), ALLOCATABLE, SAVE :: CO
!  Rim outside the borders of the basic unit-cell
      INTEGER, DIMENSION (:), ALLOCATABLE, SAVE :: Rim 
!  Aper is 0 for periodic directions, 1 for aperiodic
      INTEGER, DIMENSION (:), ALLOCATABLE, SAVE :: Aper

      CHARACTER*(FileLngth), SAVE :: Title=' '     ! Title of the job
!       density files
      CHARACTER*(FileLngth), SAVE :: OutRhoFile=' '     ! Filename of density file to save
      CHARACTER*(FileLngth), SAVE :: OutStrFile=' '     ! Filename of the m40 file
      CHARACTER*(FileLngth), SAVE :: OutStrType=''     ! Format of the exported density
      CHARACTER*(FileLngth), SAVE :: Compm40File=' '    ! Filename of the m40 file to be compared with
      CHARACTER*(FileLngth), SAVE :: InRhoFile=' '      ! File with initial electron density
      CHARACTER*(FileLngth), SAVE :: InRhoType='none'! Typoe of the input density
      CHARACTER*(FileLngth), SAVE :: MapType=' '        ! Content of the output map

!        Cellparas
      DOUBLE PRECISION, DIMENSION (1:CellCon), SAVE :: Cell, RCell        ! Cell parameters a, b, c, alpha, beta, gamma, Volume: Real and reciprocal cell
      DATA Cell /CellCon*0/
      DATA RCell /CellCon*0/
      DOUBLE PRECISION, DIMENSION (3,3), SAVE :: G,Gi,M, Mv, Minv, Mvinv, Evc  ! direct and reciprocal space metric tensor, transformation to orthonormal system...
      DOUBLE PRECISION, DIMENSION (3), SAVE :: Evl !eigenvalues of the critical point

!       Symmetry operations
      CHARACTER*(LineLength) :: SpaceGrp='?'
      INTEGER, SAVE :: NSym, NSymPrim          ! Number of symmetry operations
      INTEGER, SAVE :: NCenter                  ! Number of centers
      LOGICAL, SAVE :: Centro=.FALSE.           ! Centrosymmetric or not
      LOGICAL, SAVE :: FlagZeroC=.FALSE.        ! Zero vector in centerings of input if true
!       Normal
      DOUBLE PRECISION, DIMENSION (:,:,:), ALLOCATABLE, SAVE :: RMat ! Rotation matrices: line, column, #
      DOUBLE PRECISION, DIMENSION (:,:), ALLOCATABLE, SAVE :: TVec   ! Translation vector: line, #
      DOUBLE PRECISION, DIMENSION (:,:), ALLOCATABLE, SAVE :: Center ! Centering vector: line, #
!       Integer for the discretized unit cell
      INTEGER, DIMENSION (:,:,:), ALLOCATABLE, SAVE :: RMatI   ! Rotation matrices
      INTEGER, DIMENSION (:,:), ALLOCATABLE, SAVE :: TVecI     ! Translation vector      

! Atoms and their supposed position
      INTEGER, SAVE :: ElNr !Number of different elements
      INTEGER, SAVE :: AtomNr !Number of atoms
      INTEGER, SAVE :: TotalAtoms=0 !Total expected number of atoms
      TYPE ATOMTYPE
        CHARACTER*8 :: Name     ! Name of the atom
        DOUBLE PRECISION, DIMENSION (MaxDim) :: Coo ! Fractional coordinates
      END TYPE ATOMTYPE
      TYPE (AtomType), DIMENSION (:), ALLOCATABLE, SAVE :: Atoms ! Atoms definition
! Critical points and their supposed position
      INTEGER, SAVE :: CPlistNr !Number of atoms
      TYPE CPlistTYPE
        CHARACTER*8 :: Name     ! Name of the atom
        DOUBLE PRECISION, DIMENSION (MaxDim) :: Coo ! Fractional coordinates
      END TYPE CPlistTYPE
      TYPE (CPlistType), DIMENSION (:), ALLOCATABLE, SAVE :: CPlist ! definition of CPoints

! Array holding the maxima
      TYPE MaxType
        INTEGER, DIMENSION(MaxDim) :: ICoo !Integer coordinates of the nearest pixel
	DOUBLE PRECISION, DIMENSION(MaxDim) :: Coo ! refined position of the maximum
	DOUBLE PRECISION, DIMENSION(MaxDim) :: Com ! ceter of mass of the maximum
	DOUBLE PRECISION :: rho !refined value of density
    !  The Structure Factors, staggered real and imaginary part
	DOUBLE PRECISION :: vol !volume of the atomic basin
	DOUBLE PRECISION :: cc, ulc !cumulative charge of the maximum, "upper level" cc - only charge involved in the calculation of the com
        INTEGER :: al,an !array location of the maximum, atomic number assigned to the maximum
        INTEGER :: mult  !multiplicity of the position
        LOGICAL :: err !flag that indicates that the exact position of the maximum could not be found
      END Type MaxType
      Type(MaxType), DIMENSION(MaxMaxNr) :: Maxima
! Array holding the atomic numbers of atoms and their count expected to be in the structure
       INTEGER, DIMENSION(:,:),ALLOCATABLE, SAVE :: Compos
! Array serving for determination of the borders of the atomic basins
       INTEGER, DIMENSION(:,:),ALLOCATABLE :: Diffvec
! List of pixels found on the way to the maximum, plus the difference vectors
      INTEGER, ALLOCATABLE :: PixList(:), Diff(:,:)
 !Integer array assigning to each pixel the ArrayLocation of the corresponding maximum 
      INTEGER, ALLOCATABLE :: Basin(:), RefBasin(:)
! Array holding the critical points
      TYPE CPType
        INTEGER, DIMENSION(MaxDim) :: ICoo !Integer coordinates of the nearest pixel
	DOUBLE PRECISION, DIMENSION(MaxDim) :: Coo ! refined position of the maximum
	DOUBLE PRECISION, DIMENSION(3) :: Evl !eigenvalues
        DOUBLE PRECISION, DIMENSION(3,3) :: Evc !eigenvectors
	CHARACTER*7 :: type !type of the critical point
	DOUBLE PRECISION :: rho,lapl !refined value of density, laplacian
        LOGICAL :: err !flag that indicates that the exact position of the cp could not be found        
      END Type CPType
      Type(CPType), DIMENSION(MaxMaxNr) :: Cpoint
      INTEGER :: MaxNr, CpNr ! Number of discovered maxima and CPoints
      INTEGER :: MaxZ=1,MinZ=120 !Largest and smallest atomic number in the composition
      DOUBLE PRECISION :: Biso=0.0, maxcc=0.0 !Overall temperature factor to write to m40, maximal charge
      DOUBLE PRECISION, SAVE    :: lambda=0.710 !wavelength of the radiation. Used only for export of the structure!
      END MODULE GlobalVariables

 !***********************************************************************
      MODULE BevLipVariables
 !***********************************************************************
	  !  Variables used by the Beevers Lipson section
	  !  Subroutines BevLipInit and BevLip
	  !
	  USE GLobalDefinitions
	  IMPLICIT NONE
	  !  Contains all declaration statements nescessary to
	  !  compute one Fourier transformation using the
	  !  Beevers Lipson algorithm, except MODULE Global
	  !
	  INTEGER, SAVE :: LstElF                           ! Last element of array F
	  REAL(DPR), ALLOCATABLE, SAVE :: F(:),A(:),B(:)! Array to fill for the FFT
	  REAL(DPR), SAVE :: Arg                     ! Argument for SIN and COS of 'Modulation'
	  REAL(DPR), SAVE :: Norm                    ! Normalization of FFT
	  !       Rod and array statistics
	  INTEGER, ALLOCATABLE, SAVE :: LSteps(:)    ! of large steps
	  INTEGER, ALLOCATABLE, SAVE :: LDist(:)     ! step-size
	  INTEGER, ALLOCATABLE, SAVE :: SSteps(:)    ! of small steps
	  INTEGER, ALLOCATABLE, SAVE :: SDist(:)     ! step-size
	  INTEGER, ALLOCATABLE, SAVE :: RSteps(:)    ! in single rod
	  INTEGER, ALLOCATABLE, SAVE :: RDist(:)     ! Step-size between rod elements
	  !
	  INTEGER :: LstLStep                        ! Last large step was to element LstLStep
	  INTEGER :: LstSStep                        ! small step was to element LstSStep
	  INTEGER :: rhoLoc                          ! Location of currently needed density value
	  INTEGER :: ArrayLoc                        ! Location of currently needed FFT array value
	  INTEGER(8) :: plan1, plan2                 !Variables holding plans for FFT using FFTW library
	  !
	END MODULE BevLipVariables
