!     Last change:  LP   16 Mar 2005   10:17 am
!***********************************************************************
      MODULE GlobalDefinitions
!***********************************************************************
      IMPLICIT NONE
!        Parameters and input variables used by almost all program parts
      
      INTEGER, PARAMETER :: DeltaX=2            ! Distance in rho and F array, must be two in complex arrays
      INTEGER, PARAMETER :: CellCon=7           ! Number of cell parameters
      INTEGER, PARAMETER :: MaxDim=8            ! Maximal dimension of input data
      INTEGER, PARAMETER :: MaxDimp1=MaxDim+1   ! In Refl()%hklFG the place of group
      INTEGER, PARAMETER :: MaxDimp2=MaxDim+2   ! In Refl()%hklFG the place of the corresponding arraylocation of the structure factor
      INTEGER, PARAMETER :: MaxMaxNr=30000        ! Maximal number of maxima in one Rdim map
      INTEGER, PARAMETER :: LineLength=256      ! Maximal line length
      INTEGER, PARAMETER :: FileLngth=264       ! Maximal path and filename length
      INTEGER, PARAMETER :: DPR = KIND(1.0D0)   ! double precision real
 
      INTEGER, PARAMETER :: ongridmode       = 1, &
      			    neargridmode     = 2, &
			    pedanticmode     = 3

      DOUBLE PRECISION, PARAMETER :: pi=3.141592653589793238462643383279502884197169399375D0 !Pi
      DOUBLE PRECISION, PARAMETER :: twopi=2.0D0*pi             ! Two pi
      DOUBLE PRECISION, PARAMETER :: torad=pi/180.0D0           ! Factor for converting degree to radians
      DOUBLE PRECISION, PARAMETER :: Eps=0.000001D0              ! Smaller than Eps is considered to be zero
      DOUBLE PRECISION, PARAMETER :: DEps=0.0000000001D0        ! Smaller than DEps is considered to be zero, for very accurate situations
      INTEGER, PARAMETER :: TotElNr=94                          !Number of elements in the ElSmb array
      CHARACTER*2, DIMENSION(1:TotElNr), PARAMETER :: ElSmb= &       ! Elemental symbols sorted by atomic number
        (/'H ','He','Li','Be','B ','C ','N ','O ','F ','Ne','Na','Mg','Al','Si','P ','S ','Cl','Ar','K ','Ca','Sc','Ti','V ','Cr',&
          'Mn','Fe','Co','Ni','Cu','Zn','Ga','Ge','As','Se','Br','Kr','Rb','Sr','Y ','Zr','Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd',&
          'In','Sn','Sb','Te','I ','Xe','Cs','Ba','La','Ce','Pr','Nd','Pm','Sm','Eu','Gd','Tb','Dy','Ho','Er','Tm','Yb','Lu','Hf',&
          'Ta','W ','Re','Os','Ir','Pt','Au','Hg','Tl','Pb','Bi','Po','At','Rn','Fr','Ra','Ac','Th','Pa','U ','Np','Pu'/)

      DOUBLE PRECISION, PARAMETER :: DefaultBiso=2.0

!       The first and last element of the rho and FFT arrays is:
      INTEGER, PARAMETER :: FstEl=0
      INTEGER, SAVE :: LstRhoEl=0               ! Rho array

      INTEGER, SAVE          :: ReclFac=4 !Factor to multiply recl when opening direct-access file
      INTEGER, SAVE          :: range=9 !  Range, which is taken into account for the spline interpolation
      DOUBLE PRECISION, SAVE :: tolerance=0.15D0 ! Tolerance for the assigning of the maximum to the same atom
      DOUBLE PRECISION, SAVE :: plim=0.0D0 !treshold for the maxima written to the output
      DOUBLE PRECISION, SAVE :: chlim=0.25D0 !treshold for the charge level (relative to maximum) to include in the calculation of the com
      DOUBLE PRECISION, SAVE :: chlimlistorig=0.D0 !treshold for the charge to be included in the listing
      DOUBLE PRECISION, SAVE :: chlimlist !treshold for the charge to be included in the listing
      CHARACTER*8            :: plimkey='absolute', chlimlistkey='absolute' !Can be abslolute or relative to the maximum
      DOUBLE PRECISION, SAVE :: addborder=0.5D0 !how much of the unit cell must be added in the incommensurate direction (for COM and charge)
      LOGICAL, SAVE          :: fullcellpeaks=.false. !Search peaks irrespective of the symmetry
      LOGICAL, SAVE          :: writetmap=.false.
      LOGICAL, SAVE          :: writebasin=.false.
      LOGICAL, SAVE          :: com=.false.
      INTEGER, SAVE          :: basinmode=ongridmode
      INTEGER, SAVE	     :: EdgeRemovalCycles=4
      LOGICAL, SAVE	     :: expandedcp=.false.
      LOGICAL, SAVE	     :: projection=.false. !Project the superspace densitz along the internal coordinates?
      LOGICAL, SAVE	     :: wantstructure=.false.     !Write out the list of maxima as a m40 file
      LOGICAL, SAVE	     :: m40forjana=.false.     !Write out the list of maxima as a m40 file
      LOGICAL, SAVE	     :: compm40=.false.     !Compare the resulting atomic list with another structure
      CHARACTER*10, SAVE     :: CPformat='none'
      CHARACTER*10, SAVE     :: maxformat='all' 
      CHARACTER*10, SAVE     :: position='absolute' 
      CHARACTER*10, SAVE     :: scale='fractional' 
!         Dimensions and voxel spacing
      INTEGER, SAVE :: SDim=3              ! The dimenson of super space=RDim+DDim, defaults to 3
      INTEGER, SAVE :: RDim=3              ! Dimension of the real space      
      INTEGER, SAVE :: DDim=0              ! Additional dimensions
      INTEGER, SAVE :: NVoxTot=1,NVoxTotR=1,NVoxTotRim=1  ! The number of voxels in the unitcell, in the real dimension unit cell
      END MODULE GlobalDefinitions
      
      
!***********************************************************************
      MODULE GlobalVariables
!***********************************************************************
!        Variables used by almost all program parts

      USE GlobalDefinitions                     ! First getting all parameters:

      IMPLICIT NONE

      INTEGER :: DimCount                       ! Counter for already worked dimensions
      INTEGER :: Cnt1, Cnt2, Cnt3, Cnt4, Start, Final   ! General counters
      CHARACTER :: Buffer*(LineLength),Buffer2*(LineLength)          ! String buffer
      INTEGER :: AllocError                     ! Errorcondition while allocating arrays

      DOUBLE PRECISION :: rhomax=0.0D0          ! Maximum in density array
      DOUBLE PRECISION :: rhomin                ! Minimum in density array
      INTEGER, DIMENSION (:), ALLOCATABLE, SAVE :: NVox,NVoxR,NVoxRim  ! Array with number of voxels per axis
      INTEGER, SAVE :: N(0:MaxDim), NRim(0:MaxDim)
      DATA N /MaxDimp1*1/
      DATA NRim /MaxDimp1*1/
      DOUBLE PRECISION, SAVE :: cApVx=1.0D0     ! 'cubic' Angstrom per voxel

!       The electron density
      DOUBLE PRECISION, DIMENSION (:), ALLOCATABLE, SAVE :: rho,rhor,rhogr,srho
!       The structure factor
      DOUBLE PRECISION, DIMENSION (:), ALLOCATABLE, SAVE :: SF
!       The q-vectors
      DOUBLE PRECISION, DIMENSION (:,:), ALLOCATABLE, SAVE :: QVec   ! q-vector
!       The t-setting
      DOUBLE PRECISION, DIMENSION (:,:), ALLOCATABLE, SAVE :: Tlist
!       The order of the axes
      INTEGER, DIMENSION (:), ALLOCATABLE, SAVE :: CO
!  Rim outside the borders of the basic unit-cell
      INTEGER, DIMENSION (:), ALLOCATABLE, SAVE :: Rim 
!  Aper is 0 for periodic directions, 1 for aperiodic
      INTEGER, DIMENSION (:), ALLOCATABLE, SAVE :: Aper

      CHARACTER*(FileLngth), SAVE :: Title=' '     ! Title of the job
!       density files
      CHARACTER*(FileLngth), SAVE :: OutRhoFile=' '     ! Filename of density file to save
      CHARACTER*(FileLngth), SAVE :: OutStrFile=' '     ! Filename of the m40 file
      CHARACTER*(FileLngth), SAVE :: OutStrType=''     ! Format of the exported density
      CHARACTER*(FileLngth), SAVE :: Compm40File=' '    ! Filename of the m40 file to be compared with
      CHARACTER*(FileLngth), SAVE :: InRhoFile=' '      ! File with initial electron density
      CHARACTER*(FileLngth), SAVE :: InRhoType='none'! Typoe of the input density
      CHARACTER*(FileLngth), SAVE :: MapType=' '        ! Content of the output map

!        Cellparas
      DOUBLE PRECISION, DIMENSION (1:CellCon), SAVE :: Cell, RCell        ! Cell parameters a, b, c, alpha, beta, gamma, Volume: Real and reciprocal cell
      DATA Cell /CellCon*0/
      DATA RCell /CellCon*0/
      DOUBLE PRECISION, DIMENSION (3,3), SAVE :: G,Gi,M, Mv, Minv, Mvinv, Evc  ! direct and reciprocal space metric tensor, transformation to orthonormal system...
      DOUBLE PRECISION, DIMENSION (3), SAVE :: Evl !eigenvalues of the critical point

!       Symmetry operations
      CHARACTER*(LineLength) :: SpaceGrp='?'
      INTEGER, SAVE :: NSym, NSymPrim          ! Number of symmetry operations
      INTEGER, SAVE :: NCenter                  ! Number of centers
      LOGICAL, SAVE :: Centro=.FALSE.           ! Centrosymmetric or not
      LOGICAL, SAVE :: FlagZeroC=.FALSE.        ! Zero vector in centerings of input if true
!       Normal
      DOUBLE PRECISION, DIMENSION (:,:,:), ALLOCATABLE, SAVE :: RMat ! Rotation matrices: line, column, #
      DOUBLE PRECISION, DIMENSION (:,:), ALLOCATABLE, SAVE :: TVec   ! Translation vector: line, #
      DOUBLE PRECISION, DIMENSION (:,:), ALLOCATABLE, SAVE :: Center ! Centering vector: line, #
!       Integer for the discretized unit cell
      INTEGER, DIMENSION (:,:,:), ALLOCATABLE, SAVE :: RMatI   ! Rotation matrices
      INTEGER, DIMENSION (:,:), ALLOCATABLE, SAVE :: TVecI     ! Translation vector      

! Atoms and their supposed position
      INTEGER, SAVE :: ElNr !Number of different elements
      INTEGER, SAVE :: AtomNr !Number of atoms
      INTEGER, SAVE :: TotalAtoms=0 !Total expected number of atoms
      TYPE ATOMTYPE
        CHARACTER*8 :: Name     ! Name of the atom
        DOUBLE PRECISION, DIMENSION (MaxDim) :: Coo ! Fractional coordinates
      END TYPE ATOMTYPE
      TYPE (AtomType), DIMENSION (:), ALLOCATABLE, SAVE :: Atoms ! Atoms definition
! Critical points and their supposed position
      INTEGER, SAVE :: CPlistNr !Number of atoms
      TYPE CPlistTYPE
        CHARACTER*8 :: Name     ! Name of the atom
        DOUBLE PRECISION, DIMENSION (MaxDim) :: Coo ! Fractional coordinates
      END TYPE CPlistTYPE
      TYPE (CPlistType), DIMENSION (:), ALLOCATABLE, SAVE :: CPlist ! definition of CPoints

! Array holding the maxima
      TYPE MaxType
        INTEGER, DIMENSION(MaxDim) :: ICoo !Integer coordinates of the nearest pixel
	DOUBLE PRECISION, DIMENSION(MaxDim) :: Coo ! refined position of the maximum
	DOUBLE PRECISION, DIMENSION(MaxDim) :: Com ! ceter of mass of the maximum
	DOUBLE PRECISION :: rho !refined value of density
    !  The Structure Factors, staggered real and imaginary part
	DOUBLE PRECISION :: vol !volume of the atomic basin
	DOUBLE PRECISION :: cc, ulc !cumulative charge of the maximum, "upper level" cc - only charge involved in the calculation of the com
        INTEGER :: al,an !array location of the maximum, atomic number assigned to the maximum
        INTEGER :: mult  !multiplicity of the position
        LOGICAL :: err !flag that indicates that the exact position of the maximum could not be found
      END Type MaxType
      Type(MaxType), DIMENSION(MaxMaxNr) :: Maxima
! Array holding the atomic numbers of atoms and their count expected to be in the structure
       INTEGER, DIMENSION(:,:),ALLOCATABLE, SAVE :: Compos
! Array serving for determination of the borders of the atomic basins
       INTEGER, DIMENSION(:,:),ALLOCATABLE :: Diffvec
! List of pixels found on the way to the maximum, plus the difference vectors
      INTEGER, ALLOCATABLE :: PixList(:), Diff(:,:)
 !Integer array assigning to each pixel the ArrayLocation of the corresponding maximum 
      INTEGER, ALLOCATABLE :: Basin(:), RefBasin(:)
! Array holding the critical points
      TYPE CPType
        INTEGER, DIMENSION(MaxDim) :: ICoo !Integer coordinates of the nearest pixel
	DOUBLE PRECISION, DIMENSION(MaxDim) :: Coo ! refined position of the maximum
	DOUBLE PRECISION, DIMENSION(3) :: Evl !eigenvalues
        DOUBLE PRECISION, DIMENSION(3,3) :: Evc !eigenvectors
	CHARACTER*7 :: type !type of the critical point
	DOUBLE PRECISION :: rho,lapl !refined value of density, laplacian
        LOGICAL :: err !flag that indicates that the exact position of the cp could not be found        
      END Type CPType
      Type(CPType), DIMENSION(MaxMaxNr) :: Cpoint
      INTEGER :: MaxNr, CpNr ! Number of discovered maxima and CPoints
      INTEGER :: MaxZ=1,MinZ=120 !Largest and smallest atomic number in the composition
      DOUBLE PRECISION :: Biso=0.0, maxcc=0.0 !Overall temperature factor to write to m40, maximal charge
      DOUBLE PRECISION, SAVE    :: lambda=0.710 !wavelength of the radiation. Used only for export of the structure!
      END MODULE GlobalVariables

 !***********************************************************************
      MODULE BevLipVariables
 !***********************************************************************
	  !  Variables used by the Beevers Lipson section
	  !  Subroutines BevLipInit and BevLip
	  !
	  USE GLobalDefinitions
	  IMPLICIT NONE
	  !  Contains all declaration statements nescessary to
	  !  compute one Fourier transformation using the
	  !  Beevers Lipson algorithm, except MODULE Global
	  !
	  INTEGER, SAVE :: LstElF                           ! Last element of array F
	  REAL(DPR), ALLOCATABLE, SAVE :: F(:),A(:),B(:)! Array to fill for the FFT
	  REAL(DPR), SAVE :: Arg                     ! Argument for SIN and COS of 'Modulation'
	  REAL(DPR), SAVE :: Norm                    ! Normalization of FFT
	  !       Rod and array statistics
	  INTEGER, ALLOCATABLE, SAVE :: LSteps(:)    ! of large steps
	  INTEGER, ALLOCATABLE, SAVE :: LDist(:)     ! step-size
	  INTEGER, ALLOCATABLE, SAVE :: SSteps(:)    ! of small steps
	  INTEGER, ALLOCATABLE, SAVE :: SDist(:)     ! step-size
	  INTEGER, ALLOCATABLE, SAVE :: RSteps(:)    ! in single rod
	  INTEGER, ALLOCATABLE, SAVE :: RDist(:)     ! Step-size between rod elements
	  !
	  INTEGER :: LstLStep                        ! Last large step was to element LstLStep
	  INTEGER :: LstSStep                        ! small step was to element LstSStep
	  INTEGER :: rhoLoc                          ! Location of currently needed density value
	  INTEGER :: ArrayLoc                        ! Location of currently needed FFT array value
	  INTEGER(8) :: plan1, plan2                 !Variables holding plans for FFT using FFTW library
	  !
	END MODULE BevLipVariables
!     Last change:  LP   16 Mar 2005   10:43 am
!********************************************
      MODULE Input_module
!********************************************
        integer, parameter :: MaxGroups=10 !Maximum number of groups
        integer, parameter :: MaxKwds=100 !Maximum number of keywords
        integer, parameter :: MaxKwdLen=30 !Maximal length of a keyword and groupname
        integer, parameter :: infunit=10 !Unit number the infile will be connected to
        character*2, parameter :: comments='#!'
        INTEGER :: PassNr=2
        CHARACTER*132 :: InBuffer, infile

        TYPE GrpType
          character*(MaxKwdLen) ::name
	  logical :: status !If true, the keywords of this group ale compulsory, otherwise they are optional
	END TYPE GrpType
        TYPE (GrpType), DIMENSION(MaxGroups) :: KwdGroup
        Integer :: KwdGroupNr=0
	CHARACTER*(MaxKwdLen) :: GrpName !Temporary variable storing the name to be passed to the subroutine KwdGroup
        LOGICAL :: GrpStatus 

        Type KwdType
          character*(MaxKwdLen) ::name, endname
          integer :: group
	  integer :: occ
	END TYPE KwdType
        TYPE (KwdType), DIMENSION(1:MaxKwds) :: Kwd
        Integer :: KwdNr=0
        CHARACTER*(MaxKwdLen) :: KN,KEndN,Grp ! Temporary variables to be passed to the subroutine NewKwd
        CHARACTER*(MaxKwdLen) :: CurKwd
        INTEGER :: CurKwdId !Identification number of the current keyword during rading

        INTEGER :: ErrNr=0 !Number of errors occured during reading

        CONTAINS

	SUBROUTINE NewKwdGroup
           KwdGroupNr=KwdGroupNr+1
	   KwdGroup(KwdGroupNr)%name=GrpName
	   KwdGroup(KwdGroupNr)%status=GrpStatus
	END SUBROUTINE NewKwdGroup

	SUBROUTINE FlagKwdGroup(name,stat)
	   character*(*) :: name
	   logical :: stat
	   integer :: i
           DO i=1,KwdGroupNr
	     IF (KwdGroup(i)%name==name) KwdGroup(i)%status=stat
	   END DO
	END SUBROUTINE FlagKwdGroup

	SUBROUTINE NewKwd(kid)
           integer :: i,kid
	   KwdNr=KwdNr+1
	   kid=KwdNr !keyword identification, the keyword number is massed tot he variable, which can be then referred instead of the number itself
	   Kwd(KwdNr)%name=KN !Name of the (first part of the) keyword
	   Kwd(KwdNr)%endname=KEndN !Optional closing part of the keyword
	   Kwd(KwdNr)%occ=0 !Number of occurences in the iput file
           DO i=1,KwdGroupNr
	     if (Grp==KwdGroup(i)%name) Kwd(KwdNr)%group=i
	   END DO
	END SUBROUTINE NewKwd

        SUBROUTINE ErrorCheck
          IMPLICIT NONE
          integer :: i, idel
	  
	  DO i=1,KwdNr

	    IF (Kwd(i)%occ>1) THEN
	      ErrNr=ErrNr+1
	      WRITE(*,'(3a,i2,a)') 'Error, keyword ',Kwd(i)%name(1:idel(Kwd(i)%name)),' occured ',Kwd(i)%occ,' times in the input file.'
            END IF

            IF ((KwdGroup(Kwd(i)%Group)%status.eqv..true.).and.(Kwd(i)%occ==0)) THEN
	      ErrNr=ErrNr+1
	      WRITE(*,'(3a,i2,a)') 'Error, the compulsory keyword ',Kwd(i)%name(1:idel(Kwd(i)%name)),' is missing in the input file.'
	    ENDIF

            Kwd(i)%occ=0 !Make it zero to be "fresh" for next pass
	  END DO 
	  IF (ErrNr>0) THEN
	    WRITE(*,'(a,i4,a)') 'There were ',ErrNr,' errors in the input file.'
	    WRITE(*,'(a)') 'Please check the input file and try again.'
	    STOP
	  END IF

	END SUBROUTINE ErrorCheck

        SUBROUTINE OneInt(Text,oInt,ErrCode)
          CHARACTER*(*) :: Text
          INTEGER :: oInt,ErrCode
          READ(text,*,IOSTAT=ErrCode) oInt
        END SUBROUTINE OneInt

        SUBROUTINE ArrayInt(Text,oArrInt,iCnt,ErrCode)
          CHARACTER*(*) :: text
          INTEGER :: ErrCode,iCnt,i
	  INTEGER, DIMENSION(*) :: oArrInt
	  READ(text,*,IOSTAT=ErrCode) (oArrInt(i),i=1,iCnt)
        END SUBROUTINE ArrayInt

        SUBROUTINE OneReal(text,oReal,ErrCode)
          CHARACTER*(*) :: text
          INTEGER :: ErrCode
	  DOUBLE PRECISION :: oReal
          READ(text,*,IOSTAT=ErrCode) oReal
        END SUBROUTINE OneReal

        SUBROUTINE ArrayReal(text,oArrReal,iCnt,ErrCode)
          CHARACTER*(*) :: text
          INTEGER :: ErrCode,iCnt,i
	  DOUBLE PRECISION, DIMENSION(1:iCnt) :: oArrReal
          READ(text,*,IOSTAT=ErrCode) (oArrReal(i),i=1,iCnt)
	END SUBROUTINE ArrayReal

        SUBROUTINE YesNo(text,oLogic,ErrCode)
          CHARACTER*(*) :: text
          INTEGER :: ErrCode
	  LOGICAL :: oLogic
	  ErrCode=0
	  IF (FirstWord(text,.false.)=='yes') THEN
	    oLogic=.true.
	  ELSEIF (FirstWord(text,.false.)=='no') THEN
	    oLogic=.false.
          ELSE
	    ErrCode=100
	  END IF
	END SUBROUTINE YesNo

        FUNCTION FirstWord(ioText,cutword)
          IMPLICIT NONE
          CHARACTER*(*) :: ioText
          CHARACTER*132 :: FirstWord
          INTEGER :: wbeg,i,wend
          LOGICAL CutWord
          CHARACTER*1 :: Tab
          CHARACTER*1, PARAMETER :: Space=' '
	  
          FirstWord=' '
          Tab=CHAR(9)
          wbeg=0
	  DO wbeg=1,LEN(ioText)!Skip any leading spaces
            IF ((ioText(wbeg:wbeg)/=Space).AND.(ioText(wbeg:wbeg)/=Tab)) EXIT
          END DO
          DO i=wbeg,LEN(ioText)
            IF ((ioText(i:i)==Space).OR.(ioText(i:i)==Tab)) THEN
              wend=i-1
              FirstWord=ioText(wbeg:wend)
              IF (CutWord) ioText=ioText(wend+1:LEN(ioText))
              EXIT
            END IF
          END DO
	  DO wbeg=1,LEN(ioText)!Skip any leading spaces of the new ioText
	    IF ((ioText(wbeg:wbeg)/=Space).AND.(ioText(wbeg:wbeg)/=Tab)) EXIT
          END DO
          if(wbeg<LEN(ioText)) THEN
	    ioText(1:)=ioText(wbeg:)
          ELSE
	    ioText=' '
	  ENDIF
	END FUNCTION FirstWord

        FUNCTION ToLow(iText)
          IMPLICIT NONE
          CHARACTER*(*) :: iText
          CHARACTER*(len(iText)) :: ToLow
          INTEGER :: i
	  ToLow=''
          DO i=1,len_trim(iText)
	    SELECT CASE (iText(i:i))
	      CASE('A')
	        ToLow(i:i)='a'
	      CASE('B')
	        ToLow(i:i)='b'
	      CASE('C')
	        ToLow(i:i)='c'
	      CASE('D')
	        ToLow(i:i)='d'
	      CASE('E')
	        ToLow(i:i)='e'
	      CASE('F')
	        ToLow(i:i)='f'
	      CASE('G')
	        ToLow(i:i)='g'
	      CASE('H')
	        ToLow(i:i)='h'
	      CASE('I')
	        ToLow(i:i)='i'
	      CASE('J')
	        ToLow(i:i)='j'
	      CASE('K')
	        ToLow(i:i)='k'
	      CASE('L')
	        ToLow(i:i)='l'
	      CASE('M')
	        ToLow(i:i)='m'
	      CASE('N')
	        ToLow(i:i)='n'
	      CASE('O')
	        ToLow(i:i)='o'
	      CASE('P')
	        ToLow(i:i)='p'
	      CASE('Q')
	        ToLow(i:i)='q'
	      CASE('R')
	        ToLow(i:i)='r'
	      CASE('S')
	        ToLow(i:i)='s'
	      CASE('T')
	        ToLow(i:i)='t'
	      CASE('U')
	        ToLow(i:i)='u'
	      CASE('V')
	        ToLow(i:i)='v'
	      CASE('W')
	        ToLow(i:i)='w'
	      CASE('X')
	        ToLow(i:i)='x'
	      CASE('Y')
	        ToLow(i:i)='y'
	      CASE('Z')
	        ToLow(i:i)='z'
	      CASE DEFAULT
	        ToLow(i:i)=iText(i:i)
	    END SELECT
	  ENDDO
	END FUNCTION ToLow

        SUBROUTINE DelComments(ioText)
          INTEGER :: i,ii,idel
	  logical :: CommChar
	  CHARACTER*(*) :: ioText
	  DO i=1,idel(ioText)
	    CommChar=.false.
	    DO ii=1,idel(Comments)
	      IF (ioText(i:i)==Comments(ii:ii)) CommChar=.true.
	    END DO
            IF (CommChar) EXIT
	  END DO
          IF (i>0) THEN
	    ioText=ioText(1:i-1)
	  ELSE
	    ioText=''
	  END IF
	END SUBROUTINE DelComments

        SUBROUTINE NextLine(oText,ErrCd)
	  CHARACTER*(*) :: oText
	  INTEGER :: ErrCd, idel
100       READ(infunit,'(a)', IOSTAT=ErrCd) oText
          CALL DelComments(oText)
	  IF ((Idel(oText)==0).and.(ErrCd==0)) GOTO 100 !This line contains only comments
	END SUBROUTINE NextLine
      END MODULE Input_module

!******************************************
      SUBROUTINE INPUT
!******************************************
      USE Input_module
      USE GlobalDefinitions
      USE GlobalVariables
      
      INTEGER :: i,ii,j,NPass, ErrCode
      INTEGER :: idel !Returns the length of a string without the white spaces
      CHARACTER*(LineLength) :: DBuffer,DummyString
      DOUBLE PRECISION, DIMENSION (:,:), ALLOCATABLE :: DMat
      DOUBLE PRECISION, DIMENSION (:), ALLOCATABLE :: Dvec2
      DOUBLE PRECISION, DIMENSION (1:MaxDim) :: Dvec3
      CHARACTER*1, PARAMETER :: Space=" ", Point="."
      CHARACTER*(FileLngth) :: Inputfile        ! Inputfile :-)
      LOGICAL :: ZeroC

      CALL InitInput

!********* Define the groups of keywords here
      GrpName='must';  GrpStatus=.true. ;CALL NewKwdGroup
      GrpName='opt';   GrpStatus=.false.;CALL NewKwdGroup
      GrpName='4D';    GrpStatus=.false.;CALL NewKwdGroup
      GrpName='4Dnp';  GrpStatus=.false.; CALL NewKwdGroup
      GrpName='atoms'; GrpStatus=.false.;CALL NewKwdGroup
      GrpName='cps';   GrpStatus=.false.;CALL NewKwdGroup
      GrpName='export';GrpStatus=.false.;CALL NewKwdGroup

!********* Define the keywords here
! Template: KN=' ';KEndN=' ',Grp=' '; CALL NewKwd(kid_name)
      KN='title';       KEndN='';           Grp='opt' ; CALL NewKwd(kid_title)
      KN='cell';        KEndN='';           Grp='must'; CALL NewKwd(kid_cell)
      KN='inputfile';   KEndN='';           Grp='must'; CALL NewKwd(kid_infile)
      KN='inputformat'; KEndN='';           Grp='opt';  CALL NewKwd(kid_informat)
      KN='outputbase';  KEndN='';           Grp='opt';  CALL NewKwd(kid_outbase)
      KN='spacegroup';  KEndN='';           Grp='opt';  CALL NewKwd(kid_spacegroup)
      KN='symmetry';    KEndN='endsymmetry';Grp='opt';  CALL NewKwd(kid_sym)
      KN='fullcell';    KEndN='';           Grp='opt';  CALL NewKwd(kid_fullcell)
      KN='centers';     KEndN='endcenters'; Grp='opt';  CALL NewKwd(kid_centers)
      KN='qvectors';    KEndN='endqvectors';Grp='4D';   CALL NewKwd(kid_qvec)
      KN='tlist';       KEndN='endtlist';   Grp='4Dnp'; CALL NewKwd(kid_tlist)
      KN='range';       KEndN='';           Grp='opt';  CALL NewKwd(kid_range)
      KN='tolerance';   KEndN='';           Grp='opt';  CALL NewKwd(kid_tol)
      KN='edmaatoms';   KEndN='endedmaatoms';Grp='atoms';CALL NewKwd(kid_atoms)
      KN='axisorder';   KEndN='';           Grp='opt';  CALL NewKwd(kid_axord)
      KN='plimit';      KEndN='';           Grp='opt';  CALL NewKwd(kid_plim)
      KN='chlimit';     KEndN='';           Grp='opt';  CALL NewKwd(kid_chlim)
      KN='chlimlist';   KEndN='';           Grp='opt';  CALL NewKwd(kid_chlimlist)
      KN='tmap';        KEndN='';           Grp='opt';  CALL NewKwd(kid_tmap)
      KN='basins';      KEndN='';           Grp='opt';  CALL NewKwd(kid_basins)
      KN='maxima';      KEndN='';           Grp='opt';  CALL NewKwd(kid_maxima)
      KN='criticalpoints';KEndN='';         Grp='opt';  CALL NewKwd(kid_cps)
      KN='scale';       KEndN='';           Grp='opt';  CALL NewKwd(kid_scale)
      KN='position';    KEndN='';           Grp='opt';  CALL NewKwd(kid_position)
      KN='centerofcharge';KEndN='';         Grp='opt';  CALL NewKwd(kid_coc)
      KN='partitionmode';KEndN='';          Grp='opt';  CALL NewKwd(kid_basinmode)
      KN='addborder';   KEndN='';           Grp='opt';  CALL NewKwd(kid_addborder)
      KN='cplist';      KEndN='endcplist';  Grp='cps';  CALL NewKwd(kid_cplist)
      KN='writem40';    KEndN='';           Grp='opt';  CALL NewKwd(kid_writem40)
      KN='m40forjana';  KEndN='';           Grp='opt';  CALL NewKwd(kid_m40forjana)
      KN='export';      KEndN='';           Grp='opt';  CALL NewKwd(kid_export)
      KN='comparem40';  KEndN='';           Grp='opt';  CALL NewKwd(kid_comparem40)
      KN='projection';  KEndN='';           Grp='opt';  CALL NewKwd(kid_projection)
      KN='composition'; KEndN='';           Grp='export';CALL NewKwd(kid_composition)
      KN='numberofatoms';KEndN='';          Grp='opt';  CALL NewKwd(kid_totalatoms)
      KN='biso';         KEndN='';          Grp='opt';  CALL NewKwd(kid_biso1)
      KN='Biso';         KEndN='';          Grp='opt';  CALL NewKwd(kid_biso2)
      KN='lambda';          KEndN='';           Grp='opt';  CALL NewKwd(kid_lambda)

      DO NPass=1,PassNr
        ErrNr=0
!       read the input line by line
100     CALL NextLine(InBuffer,ErrCode) !Get next non-empty line of the input, strip out the comments
!        write(*,*) NPass, InBuffer,idel(InBuffer)
        IF (ErrCode>0) GOTO 900 !Some reading error
        IF (ErrCode<0) GOTO 200 !Finished reading the file
        CurKwd=FirstWord(InBuffer,.true.) !.true. means function FirstWord returns InBuffer shortened - without the first word
        CurKwdId=0
        DO i=1,KwdNr
	  IF (CurKwd==Kwd(i)%name) THEN
	    CurKwdId=i                       !The first word of the line matches the ith keyword
            Kwd(i)%occ=Kwd(i)%occ+1          !Add one to the number of occurences of the keyword i
	  END IF
	END DO
	IF (CurKwdId==0) GOTO 100 !No valid keyword was found, proceed to next line
        ErrCode=0
!*** Perform keyword-specific operations ***
        IF (CurKwdId==kid_title) THEN
          title=InBuffer
	              
        ELSEIF (CurKwdId==kid_cell) THEN
          IF (NPass==2) THEN
	    !Expression RDim*(RDim+1)/2 maps RDim to number of cell constants: 1->1, 2->3, 3->6
            CALL ArrayReal(InBuffer,Cell,RDim*(RDim+1)/2,ErrCode)
            IF (ErrCode==0) CALL ReciCell (RDim, CellCon, Cell, RCell)
	  END IF
	  
        ELSEIF (CurKwdId==kid_infile) THEN
          InRhoFile=InBuffer

	ELSEIF (CurKwdId==kid_informat) THEN
	  InRhoType=InBuffer	    
	  IF (InBuffer/='jana'.and.InBuffer/='xplor') THEN
            CALL ErrorMsg('Sorry, '//InBuffer(:idel(InBuffer))//' is not a known density format.')
	    ErrCode=100	    
	  END IF
           
        ELSEIF (CurKwdId==kid_spacegroup) THEN
          SpaceGrp=InBuffer

        ELSEIF (CurKwdId==kid_sym) THEN
          IF (NPass==1) THEN !Count symmetry operators
            NSym=0
	    DO
	      CALL NextLine(InBuffer,ErrCode)
	      IF (ErrCode<0) CALL ErrorMsg('Error, end of file detected during reading.')
	      IF (Inbuffer==Kwd(kid_sym)%endname) EXIT
              NSym=NSym+1
	      NSymPoint=NSym
	    END DO
	  ELSE !In pass Nr 2 read the symmetry operators
	    CALL ReadSymmetry(ErrCode)    
	  ENDIF

        ELSEIF (CurKwdId==kid_centers) THEN
          IF (NPass==1) THEN !Count centering vectors
            NCenter=0
	    FlagZeroC=.false.
	    DO
	      CALL NextLine(InBuffer,ErrCode)
	      IF (ErrCode<0) CALL ErrorMsg('Error, end of file detected during reading.')
	      IF (Inbuffer==Kwd(kid_centers)%endname) EXIT
              NCenter=NCenter+1
!             A check for a zero centering vector must be performed
              i=0
	      DO !Count the number of items in the list - we don't know SDim yet!
		IF (idel(FirstWord(InBuffer,.true.))==0) EXIT
	        i=i+1
	      END DO
	      IF ((i<1).or.(i>MaxDim)) CYCLE
	      CALL CenterVec (InBuffer, i, DVec3) !Get the centering vector
	      ZeroC=.true.
	      DO ii=1,i
	        IF (abs(DVec3(ii))>eps) ZeroC=.false.
	      END DO
	      IF (ZeroC) FlagZeroC=.true.
	    END DO
            IF (.not.FlagZeroC) NCenter=NCenter+1
	  ELSE !In pass Nr 2 read the symmetry operators
            cnt1=1
	    IF (.not.FlagZeroC) THEN !Zero centering vector must be added here
	      DO ii=1,SDim
	        Center(ii,1)=0.0D0
	      END DO
	      cnt1=2
	    END IF
            DO i=cnt1,NCenter
	      CALL NextLine(InBuffer,ErrCode)
	      IF (ErrCode<0) CALL ErrorMsg('Error, end of file detected during reading.')
              CALL CenterVec(InBuffer,SDim,DVec2)
              DO ii=1,SDim
                Center(ii,i)=Dvec2(ii)
              END DO
	    END DO
	  ENDIF

        ELSEIF (CurKwdId==kid_outbase) THEN
          OutRhoFile=InBuffer(1:idel(InBuffer))

        ELSEIF (CurKwdId==kid_qvec) THEN
          IF (NPass==2) CALL Qvectors(kid_qvec,ErrCode)

        ELSEIF (CurKwdId==kid_tlist) THEN
          IF (NPass==2) CALL ReadTlist(kid_tlist,ErrCode)
	  
        ELSEIF (CurKwdId==kid_range) THEN
	    CALL OneInt(InBuffer,range,ErrCode)

        ELSEIF (CurKwdId==kid_tol) THEN
          CALL OneReal(InBuffer,Tolerance,ErrCode)
	  
        ELSEIF (CurKwdId==kid_atoms) THEN
            AtomNr=0
	    DO
	      CALL NextLine(InBuffer,ErrCode)
	      IF (ErrCode<0) THEN
	        CALL ErrorMsg('Error, preliminary end of file detected.')
                ErrCode=100
              END IF
	      IF (FirstWord(InBuffer,.false.)==Kwd(kid_atoms)%endname) EXIT
	      AtomNr=AtomNr+1 !Count the lines
	      IF (NPass==2) THEN !Read the atoms
                CALL GetAtomandCP(InBuffer,AtomNr,0,ErrCode)
              ENDIF
	    END DO	  

        ELSEIF (CurKwdId==kid_cplist) THEN
            CPListNr=0
	    DO
	      CALL NextLine(InBuffer,ErrCode)
	      IF (ErrCode<0) THEN
	        CALL ErrorMsg('Error, preliminary end of file detected.')
                ErrCode=100
              END IF
	      IF (FirstWord(InBuffer,.false.)==Kwd(kid_cplist)%endname) EXIT
	      CPListNr=CPListNr+1 !Count the lines
	      IF (NPass==2) THEN !Read the atoms
                CALL GetAtomandCP(InBuffer,CPListNr,1,ErrCode)
              ENDIF
	    END DO	  

        ELSEIF (CurKwdId==kid_axord) THEN
          IF (NPass==2) THEN
            IF (idel(InBuffer)/=SDim) THEN
	      ErrCode=100
	      CALL ErrorMsg('Incorrect number of digits in the keyword AxisOrder.')
            END IF
            DO i=1,SDim
              READ(InBuffer(i:i),'(i1)') CO(i)
            END DO
          END IF

        ELSEIF (CurKwdId==kid_plim) THEN
          CALL OneReal(FirstWord(InBuffer,.true.),plim,ErrCode)
          IF (InBuffer=='relative') THEN
            plimkey='relative'
	  ELSEIF (InBuffer=='absolute') THEN
	    plimkey='absolute'
	  ELSEIF (InBuffer=='sigma') THEN
	    plimkey='sigma'
	  ELSEIF (InBuffer/=' ') THEN
            CALL ErrorMsg('Error: Keyword plimit not properly defined!!!')
	    ErrCode=100
	  ENDIF

        ELSEIF (CurKwdId==kid_chlim) THEN
          CALL OneReal(InBuffer,chlim,ErrCode)

        ELSEIF (CurKwdId==kid_chlimlist) THEN
          CALL OneReal(FirstWord(InBuffer,.true.),chlimlistorig,ErrCode)
          IF (InBuffer=='relative') THEN
            chlimlistkey='relative'
	  ELSEIF (InBuffer=='absolute') THEN
	    chlimlistkey='absolute'
	  ELSEIF (InBuffer/=' ') THEN
            CALL ErrorMsg('Error: Keyword chlimlist not properly defined!!!')
	    ErrCode=100
	  ENDIF

        ELSEIF (CurKwdId==kid_tmap) THEN
          CALL YesNo(InBuffer,writetmap,ErrCode)

        ELSEIF (CurKwdId==kid_fullcell) THEN
          CALL YesNo(InBuffer,fullcellpeaks,ErrCode)

        ELSEIF (CurKwdId==kid_basins) THEN
          CALL YesNo(InBuffer,writebasin,ErrCode)

        ELSEIF (CurKwdId==kid_maxima) THEN
          maxformat=InBuffer
          IF ((maxformat/='atoms').and.(maxformat/='all').and.(maxformat/='none')) THEN
	    Call ErrorMsg('Error, keyword maxima must have value atoms, all or none!')
	    ErrCode=100
	  END IF  
          IF (maxformat=='atoms') CALL FlagKwdGroup('atoms',.true.)

        ELSEIF (CurKwdId==kid_cps) THEN
          cpformat=InBuffer
          IF ((cpformat/='list').and.(cpformat/='all').and.(cpformat/='none')) THEN
	    Call ErrorMsg('Error, keyword criticalpoints must have value list, all or none!')
	    ErrCode=100
	  END IF  
          IF (cpformat=='list') CALL FlagKwdGroup('cps',.true.)

        ELSEIF (CurKwdId==kid_scale) THEN
          scale=InBuffer
          IF ((scale/='fractional').and.(scale/='angstrom')) THEN
	    Call ErrorMsg('Error, keyword scale must have value fractional or angstrom!')
	    ErrCode=100
	  END IF  

        ELSEIF (CurKwdId==kid_position) THEN
          position=InBuffer
          IF ((position/='absolute').and.(position/='relative')) THEN
	    Call ErrorMsg('Error, keyword scale must have value absolute or relative!')
	    ErrCode=100
	  END IF  

        ELSEIF (CurKwdId==kid_coc) THEN
          CALL YesNo(InBuffer,com,ErrCode)

        ELSEIF (CurKwdId==kid_basinmode) THEN
          DBuffer=firstword(InBuffer,.true.)
          IF (DBuffer=='ongrid') THEN
	    basinmode=ongridmode
          ELSEIF (DBuffer=='neargrid') THEN
	    basinmode=neargridmode
	    IF (len_trim(InBuffer)>0) call oneint(InBuffer,EdgeRemovalCycles,ErrCode)	    
          ELSEIF (DBuffer=='pedantic') THEN
	    basinmode=pedanticmode
	  ELSE
	    Call ErrorMsg('Error, keyword partitionmode must have value ongrid, neargrid, neargridfast or pedantic!')
	    ErrCode=100
	  END IF  

        ELSEIF (CurKwdId==kid_addborder) THEN
          CALL OneReal(InBuffer,addborder,ErrCode)

        ELSEIF ((CurKwdId==kid_biso1).or.(CurKwdId==kid_biso2)) THEN
          CALL OneReal(InBuffer,Biso,ErrCode)

        ELSEIF (CurKwdId==kid_lambda) THEN
          CALL OneReal(InBuffer,lambda,ErrCode)

        ELSEIF ((CurKwdId==kid_writem40).or.(CurKwdId==kid_export)) THEN
          OutStrfile=InBuffer
	  IF (OutStrFile(idel(OutStrFile)-3:idel(OutStrFile))=='.m40') THEN
	    OutStrFile=OutStrFile(1:idel(outStrFile)-4) !cut away the m40 extension, will be added later again
	    OutStrType='m40'
	  ELSEIF (OutStrFile(idel(OutStrFile)-3:idel(OutStrFile))=='.cif') THEN
	    OutStrFile=OutStrFile(1:idel(outStrFile)-4) !cut away the cif extension, will be added later again
	    OutStrType='cif'
	  ELSEIF (OutStrFile(idel(OutStrFile)-3:idel(OutStrFile))=='.ins') THEN
	    OutStrFile=OutStrFile(1:idel(outStrFile)-4) !cut away the m40 extension, will be added later again
	    OutStrType='ins'
          ELSE
	    Call ErrorMsg('Error, the file for the exported structure has an unknown extension. Can''t guess the format.')
	    ErrCode=100	    
	  ENDIF
          wantstructure=.true.
	  CALL FlagKwdGroup('export',.true.)
	  com=.true.

        ELSEIF (CurKwdId==kid_m40forjana) THEN
          CALL YesNo(InBuffer,m40forjana,ErrCode)

        ELSEIF (CurKwdId==kid_comparem40) THEN
          Compm40file=InBuffer
          compm40=.true.

        ELSEIF (CurKwdId==kid_projection) THEN
          CALL YesNo(InBuffer,projection,ErrCode)
          if (projection) CALL FlagKwdGroup('4Dnp',.false.) 

        ELSEIF (CurKwdId==kid_composition) THEN
          IF ((NPass==1).and.(ElNr==0)) THEN !The second condition is there to protect again doubled allocation of compos in case of a double occurence of composition in the input file
            IF (Idel(InBuffer)==0) THEN
              ElNr=0
	    ELSE
	      ElNr=1
	    ENDIF
            DO i=1,idel(InBuffer)-1
              if (InBuffer(i:i).eq.' '.and.InBuffer(i+1:i+1).ne.' ') ElNr=ElNr+1
            ENDDO
            IF (ElNr>0) THEN
              ALLOCATE(Compos(1:ElNr,1:2))
              DO i=1,ElNr
                Compos(i,1)=0
                Compos(i,2)=1
                dBuffer=FirstWord(InBuffer,.true.)
                !Find the first number in the string
	        ii=1;iii=0
	        DO WHILE ((ii/=0).and.(iii<=idel(dBuffer)))
	          iii=iii+1
                  CALL CheckAllowedChars(dBuffer(iii:iii),5,ii)
	        ENDDO
		IF ((iii==2).and.(dBuffer(1:1)=='D')) THEN
		  Compos(i,1)=1
                ELSE
                  DO j=1,TotElNr
                     IF (dBuffer(1:iii-1)==ElSmb(j)) Compos(i,1)=j
	          END DO
                ENDIF
		IF (Compos(i,1)==0) THEN
	          dBuffer='Error, symbol '//dBuffer(1:iii-1)//' is not a valid symbol of an element.'
	          CALL ErrorMsg(dBuffer)
	          ErrCode=100
                  CYCLE
                END IF
                MaxZ=max(MaxZ,Compos(i,1))
                if (compos(i,1)>1) MinZ=min(MinZ,Compos(i,1))
		IF (iii<=idel(dBuffer)) THEN
                  DummyString=dBuffer
                  CALL OneInt(dBuffer(iii:),Compos(i,2),ErrCode)
		  IF (ErrCode/=0) THEN
                    dBuffer='Error, string '//DummyString(1:idel(DummyString))// &
		            ' does not have the format "element symbol and number".'
                    CALL ErrorMsg(dBuffer)
                  ENDIF
	        ENDIF	
	      END DO
            ELSE
	      CALL ErrorMsg('Error, no elements found after the keyword composition.')
	      ErrCode=100
	    ENDIF
	  ENDIF

        ELSEIF (CurKwdId==kid_totalatoms) THEN
	  IF (NPass==2) THEN
	    CALL OneInt(InBuffer,totalatoms,ErrCode)
            IF ((TotalAtoms<0).and.(ElNr==0)) THEN
              CALL ErrorMsg('Error, no composition given, the number of atoms cannot be guessed from composition.')		  
              ErrCode=100
	    ENDIF	        	    
	    IF (ErrCode/=0) THEN
	      ErrCode=0
	      IF (InBuffer=='guess') THEN
	        TotalAtoms=0
	      ELSEIF (InBuffer=='composition') THEN
                IF (ElNr>0) THEN
		  TotalAtoms=-100
		ELSE
                  CALL ErrorMsg('Error, no composition given, the number of atoms cannot be guessed from composition.')		  
		  ErrCode=100
		ENDIF	        
	      ELSE
                CALL ErrorMsg('Error, keyword numberofatoms must be followed by "guess", "composition" or an integer.')
                ErrCode=100
	      ENDIF
	    ENDIF
          ENDIF
	END IF !End of the ifblock containing handling of all keywords

        IF (ErrCode/=0) THEN
	  CALL ErrorMsg('Error reading keyword '//Kwd(CurKwdId)%name(1:idel(Kwd(CurKwdId)%name))//'.')
          ErrNr=ErrNr+1
	END IF
!*** End of keyword specific operations ***

        GOTO 100 !Go to reading next line        
200     CONTINUE
        CALL ErrorCheck !Check for errors during reading and take corresponding actions
        CALL AfterPass  !Actions to be taken after the file is read up to the end
        REWIND(infunit)
      END DO
!     Successfully finished reading
      GOTO 9999

!     error messages            
900   CALL ErrorMsg('Error reading the input file '//infile(:idel(infile))//'.')
      STOP
      
9999  CLOSE(infunit)
      CONTINUE      

      CONTAINS

! This routine typically just reads the name of the input file
!****************************************
        SUBROUTINE InitInput
!****************************************
!          USE F90_UNIX_ENV, ONLY : GETARG,IARGC

          INTEGER :: GetReclFac
          Inputfile=' '
	  IF (IARGC()>0) THEN
            CALL GETARG(1,InputFile)
	  ENDIF
11        IF (InputFile(1:1)==' ') THEN
            WRITE (*,'(A)') ''
            WRITE (*,'(A,I4,A)',ADVANCE='NO') &
            'Please insert the name of your inputfile (max. ',Filelngth,' chars): '
            READ (*,'(A)') InputFile
            IF (idel(InputFile)==0) THEN
              WRITE (*,'(A)') ''
              WRITE (*,'(A)') 'No name given! Try again...'
              WRITE (*,'(A)') ''
              GOTO 11
            END IF
            WRITE (*,'(A)') ''
          END IF
          RecLFac=GetReclFac()
!         Open the input file for reading
          OPEN (UNIT=infunit,FILE=Inputfile(:Idel(Inputfile)),IoStat=ErrCode,STATUS='OLD')

	END SUBROUTINE InitInput


! Handles all actions taken after a pass is finished - typically allocation of dynamically allocated arrays
!****************************************
        SUBROUTINE AfterPass
!****************************************
          IF (NPass==1) THEN
    
! Try to guess the types of the density files
            IF (InRhoType=='none') THEN !No type has been input
	      CALL GuessType(InRhoFile,InRhoType)
              IF (InRhoType=='unknown') THEN
	        CALL ErrorMsg('Error, keyword outputformat is missing and I cannot guess the type from the file names!') 
	        CALL ErrorMsg('Please use the keyword outputformat or standard file extensions.') 
                ErrCode=100
	      ENDIF
	    ENDIF
            
            CALL GetNVox !Get the number of voxels dimension etc...
            IF ((projection).and.(SDim==RDim)) THEN
              CALL ErrorMsg('Warning: It makes no sense to make projection of this density (dimension = realdimension). '// &
                             'Ignoring the keyword projection.')
              projection=.false.
            END IF
            if ((SDim>RDim).and..not.projection) CALL FlagKwdGroup('4Dnp',.true.)
            ALLOCATE (Atoms(AtomNr), STAT=AllocError)            ! Type and coordinates of atoms
            IF (AllocError.GT.0) STOP 'Error allocating Atoms!'
            ALLOCATE (CPlist(CPlistNr), STAT=AllocError)         ! Type and coordinates of atoms
            IF (AllocError.GT.0) STOP 'Error allocating Atoms!'
            ALLOCATE (QVec(RDim, SDim-RDim), STAT=AllocError)    ! q-vectors
            IF (AllocError.GT.0) STOP 'Error allocating QVec!'
            ALLOCATE (Tlist(3, SDim-RDim), STAT=AllocError)      ! t-settings
            IF (AllocError.GT.0) STOP 'Error allocating Tlist!'
	    ALLOCATE (CO(SDim), STAT=AllocError)                 ! Axis order
            IF (AllocError.GT.0) STOP 'Error allocating Axis order!'
            ALLOCATE (DMat(SDim,SDim), STAT=AllocError)             
            IF (AllocError.GT.0) STOP 'Error allocating DMat!'
            ALLOCATE (DVec2(SDim), STAT=AllocError)             
            IF (AllocError.GT.0) STOP 'Error allocating DVec!'
            IF (NCenter==0) THEN
	      NCenter=1
	      FlagZeroC=.false.
	    ENDIF
	    ALLOCATE (Center(SDim, NCenter), STAT=AllocError)      ! Centers
            IF (AllocError.GT.0) STOP 'Error allocating Center!'
            Center=0.0D0
            ALLOCATE (RMat(SDim, SDim, NSym*NCenter), STAT=AllocError)! Rotation matrices
	    IF (AllocError.GT.0) STOP 'Error allocating RMat!'
            RMat=0.0D0
            ALLOCATE (TVec(SDim, NSym*NCenter), STAT=AllocError)      ! Translation vectors
            IF (AllocError.GT.0) STOP 'Error allocating TVec!'
            TVec=0.0D0
            ALLOCATE (RMatI(SDim, SDim, NSym*NCenter), STAT=AllocError)! Integer rotation matrices
            IF (AllocError.GT.0) STOP 'Error allocating RMat!'
            RMatI=0
            ALLOCATE (TVecI(SDim, NSym*NCenter), STAT=AllocError)      ! Integertranslation vectors
            IF (AllocError.GT.0) STOP 'Error allocating TVec!'
            TVecI=0
            DO i=1,SDim
              CO(i)=i
            END DO
            ALLOCATE (Rim(RDim), STAT=AllocError)                 ! border of the unit cell in the incommensurate directions
            IF (AllocError.GT.0) STOP 'Error allocating Rim!'
            ALLOCATE (Aper(RDim), STAT=AllocError)                 ! which direction is aperiodic
            IF (AllocError.GT.0) STOP 'Error allocating Aper!'
            IF ((maxformat==' ').and.(AtomNr>0)) maxformat='atoms' 

            if (maxformat=='atoms') CALL FlagKwdGroup('atoms',.true.) 
            if (cpformat=='list') CALL FlagKwdGroup('cps',.true.) 

	  ELSEIF (NPass==2) THEN
            IF (NCenter>0) CALL ExpandSym
            CALL DiscreteSym

          END IF
	END SUBROUTINE AfterPass

!****************************************
        SUBROUTINE ErrorMsg(text)
!****************************************
          CHARACTER*(*) :: text
          write(*,'(3a)') text(1:idel(text))
	END SUBROUTINE ErrorMsg


!        Calculating reciprocal cell parameters
!******************************************************************
      SUBROUTINE ReciCell (iRDim, iCellCon, ioCell, oRCell)
!******************************************************************
      USE GLOBALDEFINITIONS
      USE GLOBALVARIABLES
      
      IMPLICIT NONE
      INTEGER :: iCellCon, iRDim,i,ii
!        Variables with leading r are reciprocal
      DOUBLE PRECISION :: ioCell(iCellCon), oRCell(iCellCon)
      DOUBLE PRECISION :: a, b, c, sina, sinb, sing, cosa, cosb, cosg
      DOUBLE PRECISION :: ra, rb, rc
      DOUBLE PRECISION :: cosra, cosrb, cosrg
      DOUBLE PRECISION :: V, rV
      
!       How much real dimensions
      SELECT CASE (iRDim)
      CASE (1)
!        Declaring auxilliary variables
        a=ioCell(1)
        ra=1/a
        V=a
        rV=ra
        ioCell(7)=V
        oRCell(1)=ra
        oRCell(7)=rV
        ioCell(2)=90.0 !its convenient to have the angle defined, even in one-dimensional space
	oRCell(2)=90.0
      CASE (2)
        a=ioCell(1)
        b=ioCell(2)
        sina=DSIN(toRad*ioCell(3))
!       Areas
        V=a*b*sina
        rV=1./V
!       reciprocal axes
        ra=1/(a*sina)
        rb=1/(b*sina)
!       declaring calculated values
        ioCell(7)=V
        oRCell(1)=ra
        oRCell(2)=rb
!       reciprocal angle
        oRCell(3)=180D0-ioCell(3)
        oRCell(7)=rV
      CASE (3)
!        Declaring auxilliary variables
        a=ioCell(1)
        b=ioCell(2)
        c=ioCell(3)
        sina=DSIN(toRad*ioCell(4))
        sinb=DSIN(toRad*ioCell(5))
        sing=DSIN(toRad*ioCell(6))
        cosa=DCOS(toRad*ioCell(4))
        cosb=DCOS(toRad*ioCell(5))
        cosg=DCOS(toRad*ioCell(6))
!        Volumes
        V=a*b*c*SQRT(1.-cosa**2-cosb**2-cosg**2+2.*cosa*cosb*cosg)
        rV=1./V
!        reciprocal angles
        cosra=(cosb*cosg-cosa)/(sinb*sing)
        cosrb=(cosa*cosg-cosb)/(sina*sing)
        cosrg=(cosa*cosb-cosg)/(sina*sinb)
!        reciprocal axes
        ra=b*c*sina/V
        rb=a*c*sinb/V
        rc=a*b*sing/V
!        declaring calculated values
        ioCell(7)=V
        oRCell(1)=ra
        oRCell(2)=rb
        oRCell(3)=rc
        oRCell(4)=DACOS(cosra)/toRad
        oRCell(5)=DACOS(cosrb)/toRad
        oRCell(6)=DACOS(cosrg)/toRad
        oRCell(7)=rV
      END SELECT
!     fill the direct and reciprocal metric tensor
      DO i=1,iRDim
        DO ii=1,iRDim
	  IF(i==ii) THEN
	    cosa=1.0D0
            cosb=1.0D0
	  ELSE
	    cosa=cos(ioCell(9-i-ii-3*(3-RDim))*torad)
	    cosb=cos(oRCell(9-i-ii-3*(3-RDim))*torad)
	  END IF
	  G(i,ii)=ioCell(i)*ioCell(ii)*cosa
          Gi(i,ii)=oRCell(i)*oRCell(ii)*cosb
        END DO
      END DO
! Fill the conversion matrix E = M*A to the orthonormal system
      Select CASE(iRDim)
      CASE(1)
        M(1,1)=1.0D0/ioCell(1)
      CASE(2)
        M(1,1)=1.0D0/ioCell(1)
        M(1,2)=0.0D0
        M(2,1)=-cosa/(ioCell(1)*sina)
        M(2,2)=1.0D0/(ioCell(2)*sina)
      CASE(3)
        M(1,1)=1.0D0/ioCell(1)
        M(1,2)=0.0D0
        M(1,3)=0.0D0
        M(2,1)=-cosg/(ioCell(1)*sing)
        M(2,2)=1.0D0/(ioCell(2)*sing)
        M(2,3)=0.0D0
        M(3,1)=oRCell(1)*cosrb
        M(3,2)=oRCell(2)*cosra
        M(3,3)=oRCell(3)
      END SELECT
      DO i=1,iRDim
        DO ii=1,iRDim
          Mv(i,ii)=M(i,ii)*DBLE(NVox(i))
        END DO
      END DO
      call invert(M,Minv,RDim)
      call invert(Mv,Mvinv,RDim)
      
12000 END SUBROUTINE ReciCell

! Read the q-vectors
!****************************************
        SUBROUTINE QVectors(kid,ErrCd)
!****************************************
	  INTEGER :: i,ErrCd,kid
	  CHARACTER*50 :: ErrText
          DO i=1,DDim
            CALL NextLine(InBuffer,ErrCd)
            
	    IF (ErrCd/=0) RETURN !Some error occured
            IF (FirstWord(InBuffer,.false.)==Kwd(kid)%endname) THEN !Premature end of the block
              CALL ErrorMsg('Error - '//Kwd(kid)%endname(1:idel(Kwd(kid)%endname))//' detected prematurely.')
              WRITE(ErrText,'(i1,a)') DDIM,' qvectors expected.'
	      CALL ErrorMsg(ErrText)
	      ErrCd=100
	    END IF
            CALL ArrayReal(InBuffer,qvec(1,i),3,ErrCd)
	    IF (ErrCd/=0) RETURN !Some error occured            
	  END DO
          CALL NextLine(InBuffer,ErrCd)
          IF (FirstWord(InBuffer,.false.)/=Kwd(kid)%endname) THEN !The end-keyword is not here!
            CALL ErrorMsg('Too many lines before '//Kwd(kid)%endname(1:idel(Kwd(kid)%endname)))
	    ErrCd=100
	  END IF
	END SUBROUTINE QVectors

! Read the sampling in t-space
!****************************************
        SUBROUTINE ReadTlist(kid,ErrCd)
!****************************************
	  INTEGER :: i,ErrCd,kid
	  CHARACTER*50 :: ErrText
          DO i=1,DDim
            CALL NextLine(InBuffer,ErrCd)
            
	    IF (ErrCd/=0) RETURN !Some error occured
            IF (FirstWord(InBuffer,.false.)==Kwd(kid)%endname) THEN !Premature end of the block
              CALL ErrorMsg('Error - '//Kwd(kid)%endname(1:idel(Kwd(kid)%endname))//' detected prematurely.')
              WRITE(ErrText,'(i1,a)') DDIM,' lines expected.'
	      CALL ErrorMsg(ErrText)
	      ErrCd=100
	    END IF
            CALL ArrayReal(InBuffer,tlist(1,i),3,ErrCd)
	    IF (ErrCd/=0) RETURN !Some error occured            
	  END DO
          CALL NextLine(InBuffer,ErrCd)
          IF (FirstWord(InBuffer,.false.)/=Kwd(kid)%endname) THEN !The end-keyword is not here!
            CALL ErrorMsg('Too many lines before '//Kwd(kid)%endname(1:idel(Kwd(kid)%endname)))
	    ErrCd=100
	  END IF
	END SUBROUTINE ReadTlist

!****************************************
        SUBROUTINE ReadSymmetry(ErrCd)
!****************************************
	  IMPLICIT NONE
	  INTEGER :: ErrCd,Symcnt,iii
          CHARACTER*(LineLength) :: lowbuffer
	  DO Symcnt=1,NSym
            CALL NextLine(InBuffer,ErrCd)
!	    CALL Symmetry (SDim, InBuffer, DMat, DVec2)
            lowbuffer=ToLow(InBuffer)
	    CALL OneSym(SDim, SymCnt, lowbuffer, DMat, DVec2)
	    DO i=1, SDim
              DO ii=1, SDim
                RMat(i,ii,Symcnt)=DMat(i,ii)
                DMat(i,ii)=0.0D0
              END DO
              TVec(i,Symcnt)=DVec2(i)
              DVec2(i)=0.0D0
            END DO
	  END DO
	END SUBROUTINE ReadSymmetry

! Decode one symmetry operation into a matrix and a vector
!****************************************
        SUBROUTINE OneSym(SDim,Truepos, SymString,InRMat,LocTVec)
!****************************************
	  INTEGER :: SDim,ErrCd,i,ii,col,curbeg,curend,jex,jelom,jebod,lgth
	  INTEGER :: indx,indlom,nom,denom,TruePos,SymOpPos
          DOUBLE PRECISION, DIMENSION(SDim,SDim) :: InRmat
          DOUBLE PRECISION, DIMENSION(SDim,SDim) :: LocRmat
          DOUBLE PRECISION, DIMENSION(SDim) :: LocTvec
	  CHARACTER*(*) :: SymString
	  CHARACTER*(linelength) :: Row
	  CHARACTER*(15) :: rowpart
	  CHARACTER*6 :: CooLetter='xyztuv' 
	  !First check if the format is x1, x2, x3... or x, y, z
	  IF (index(SymString,'x1')==0) THEN !x y z format
            DO i=1,SDim
	      DO
	        ii=index(SymString,CooLetter(i:i))
	        IF (ii==0) EXIT
	        write(rowpart,'(a1,i1)')'#',i
	        SymString=SymString(:ii-1)//trim(rowpart)//symString(ii+1:)
	      ENDDO
	    ENDDO
            DO i=1,len_trim(SymString)
	      IF (SymString(i:i)=='#') SymString(i:i)='x'
	    ENDDO
	  ENDIF
	  !Check if comma or space delimited
	  IF (index(Symstring,',')/=0) THEN !Comma delimited
	    CALL RemoveSpaces(SymString)
	    ii=len_trim(SymString)
	    DO i=1,ii
	      if (SymString(i:i)==',') SymString(i:i)=' '
	    ENDDO
	  ENDIF
	  
          DO i=1,SDim
	    LocTVec(i)=0.0E0
	    DO ii=1,SDim
	      LocRMat(i,ii)=0.0E0
	    ENDDO
	    row=FirstWord(SymString,.true.)
	    lgth=len_trim(row)
	    IF (lgth==0) THEN
	      write(row,'(a,i3)') 'Error, not enough components of the symmetry operation number',TruePos
              call ErrorMsg(row)
	      ErrCd=100
	      return
	    ENDIF
            if (lgth==1) goto 9100
	    CALL CheckAllowedChars(row(:lgth),1,ErrCd)
	    IF (ErrCd/=0) goto 9100
	    curbeg=1;jex=0;jelom=0;jebod=0;indx=0;indlom=0
	    DO ii=1,lgth
	      IF (row(ii:ii)=='x') THEN
	        jex=jex+1
	        indx=ii
	      ELSEIF (row(ii:ii)=='/') THEN
	        jelom=jelom+1
	        indlom=ii
	      ELSEIF (row(ii:ii)=='.') THEN
	        jebod=jebod+1
	      ENDIF
	      IF ((((row(ii:ii)=='+').or.(row(ii:ii)=='-')).and.(ii>1)).or.(ii==len_trim(row))) THEN 
                IF (ii==lgth) then
		  curend=ii
                else
		  curend=ii-1
		endif
		IF ((jelom+jebod+jex>1).or.(indlom==curbeg).or.(indlom==curend) &
		    .or.(indx==curend).or.(indx-curbeg>1)) goto 9000
                !Row(curbeg,curend) contains exactly one entry, it can be a +-xi,fraction or a decimal number
		IF (jex>0) then !Part containing entry of the rotation matrix
		  read(row(indx+1:curend),'(i2)',IOSTAT=ErrCd) col
		  if (errcd/=0) goto 9000
		  if (LocRMat(i,col)>eps) goto 9300
		  if (row(curbeg:curbeg)=='-') THEN 
		    LocRMat(i,col)=-1.0E0
		  elseif ((row(curbeg:curbeg)=='+').or.(row(curbeg:curbeg)=='x')) THEN
		    LocRMat(i,col)=1.0E0
		  else
		    goto 9000
		  endif
		elseif(jebod>0) then !decimal number
		  if (LocTVec(i)>eps) goto 9200
		  rowpart=row(curbeg:curend)
		  call onereal(rowpart,LocTVec(i),ErrCd)
		  if (ErrCd/=0) goto 9000
		elseif(jelom>0) then !fraction
		  if (LocTVec(i)>eps) goto 9200
		  rowpart=row(curbeg:indlom-1)               
		  call oneint(rowpart,nom,ErrCd)
		  if (ErrCd/=0) goto 9000
		  rowpart=row(indlom+1:curend)
		  call oneint(rowpart,denom,ErrCd)
		  if (ErrCd/=0) goto 9000
		  LocTVec(i)=float(nom)/float(denom)
                else !no x, no point, no slash -> integer translational part! Pretty bizzare, but possible...
		  if (LocTVec(i)>eps) goto 9200
		  rowpart=row(curbeg:curend)
		  call oneint(rowpart,nom,ErrCd)
		  if (ErrCd/=0) goto 9000
		  LocTVec(i)=float(nom)
	        endif
		jex=0;jelom=0;jebod=0;curbeg=curend+1;indlom=0;indx=0
	      ENDIF
	    ENDDO
	  END DO
          InRMat=LocRMat
	  return
9000      write(row,'(a,i3,a,i2)') 'Syntax error in the symmetry operation number',TruePos,', component number ',i
          call ErrorMsg(row)
	  ErrCd=100
	  return
9100      write(row,'(a,i3)') 'Syntax error in the symmetry operation number',TruePos
          call ErrorMsg(row)
          ErrCd=100
          return
9200      write(row,'(a,i3,a,i2)') 'Error, more than one translational part in the symmetry operation number',TruePos, &
                              ', component number ',i
          call ErrorMsg(row)
          ErrCd=100
          return
9300      write(row,'(a,2i2)') 'Error, more than one definition of the matrix element',i,col
          call ErrorMsg(row)
	  write(row,'(a,i3,a)') 'in symmetry operation number',TruePos,'.'
          call ErrorMsg(row)
          ErrCd=100
          return
	END SUBROUTINE OneSym
	
        SUBROUTINE RemoveSpaces(ioText)
        character*(*) :: ioText
        integer :: ii
          ii=0
          DO
            ii=ii+1
	    IF (ioText(ii:ii)==' ') THEN
	      ioText(ii:)=ioText(ii+1:)
	      ii=ii-1
            ENDIF
	    IF (ii>=len_trim(ioText)) EXIT
          ENDDO
        END SUBROUTINE RemoveSpaces
	

!******************************************************************
      SUBROUTINE CenterVec (iBuffer, iSDim, oVec)
!******************************************************************
      IMPLICIT NONE
      CHARACTER        :: iBuffer*(*)
      CHARACTER        :: DBuffer*(132)
      INTEGER          :: iSDim
      DOUBLE PRECISION :: oVec (iSDim)
      INTEGER          :: Pos, Numerator, Nominator, i
      
      DO i=1, iSDim
        CALL SpaceErase (iBuffer)
        CALL DataLength (iBuffer, DBuffer)
        CALL NextPos (DBuffer, '/', Pos)
        IF (Pos>=0) THEN
          CALL NumberLength (DBuffer, 'l', Pos, Numerator)
          CALL NumberLength (DBuffer, 'r', Pos, Nominator)
          IF (Nominator.EQ.1) STOP &
            'Symmetry data not consistent: Translationvector >= 1?'
          oVec(i)=DBLE(Numerator)/DBLE(Nominator)
        ELSE
          READ (DBuffer, '(F14.0)') oVec(i)
        END IF
      END DO
      
1000  END SUBROUTINE CenterVec

!        Calculate symmetry operations for discrete space
!******************************************************************
      SUBROUTINE DiscreteSym
!******************************************************************
      IMPLICIT NONE
      
      INTEGER :: i, ii, iii,badsmop
!       The diagonals of matrices N and N^-1 holding voxel division
      DOUBLE PRECISION, DIMENSION (SDim) :: Nn
!       Dummy
      DOUBLE PRECISION, DIMENSION (SDim, SDim) :: DRMat
      DOUBLE PRECISION :: TTest                 ! Test
      
!       Preparation
      DO i=1, SDim
        Nn(i)=DBLE(NVox(i))
      END DO
!       All symops
      DO i=1, NSym
!       R*Ninv, N*Tau
!       Line of RMat, column of Ninv
        DO ii=1, SDim
!       N by translation
          TVecI(ii,i)=NINT(Nn(ii)*TVec(ii,i))
!       Test if integer, hence a voxel operation
          TTest=ABS(Nn(ii)*TVec(ii,i)-DBLE(TVecI(ii,i)))
          IF (TTest>100*Eps) THEN
            badsmop=i
            GOTO 900
          END IF
	  
          DO iii=1,SDim
            DRMat(ii,iii)=0.0D0
            TTest=ABS(RMat(ii,iii,i))
            IF (TTest<Eps) THEN
              DRMat(ii,iii)=0.0D0
            ELSE 
              DRMat(ii,iii)=RMat(ii,iii,i)/Nn(iii)
            END IF
          END DO
        END DO
!       N*(R*Ninv)
        DO ii=1, SDim
          DO iii=1,SDim
            RMatI(ii,iii,i)=NINT(Nn(ii)*DRMat(ii,iii))
            TTest=ABS(DBLE(RMatI(ii,iii,i)-(Nn(ii)*DRMat(ii,iii))))
	    IF (TTest>Eps) THEN
	      badsmop=i
	      GOTO 910
            ENDIF
	  END DO
        END DO
      END DO
      GOTO 1000

!       Errors
900   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine DiscreteSym:'
      WRITE (*,'(A)') 'Error: Translationvector multiplied by'
      WRITE (*,'(A)') '       voxelnumber is not an integer!'
      WRITE (*,'(A,i3)') '       Symmetry operation #: ', badsmop
      WRITE (*,'(A)') ''
      STOP
      
910   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine DiscreteSym:'
      WRITE (*,'(A)') 'Error: Translationmatrix divided by'
      WRITE (*,'(A)') '       voxelnumber is not an integer!'
      WRITE (*,'(A,i3)') '       Symmetry operation #: ', badsmop
      WRITE (*,'(A)') ''
      STOP

1000  END SUBROUTINE DiscreteSym

!        Expand symmetrie operations with centers
!******************************************************************
      SUBROUTINE ExpandSym
!******************************************************************
      IMPLICIT NONE
      
      INTEGER :: iES, iiES, iiiES, ivES
      INTEGER :: SymLoc         ! Start point for appending new symops
      
      SymLoc=0
      
All:  DO iES=1, NCenter           ! All centers 
        DO iiES=1, NSym           ! All read symops
          SymLoc=SymLoc+1
!       Copy rotation matrix
          DO iiiES=1,SDim         ! Lines
            DO ivES=1, SDIM       ! Columns
              RMat(iiiES,ivES,SymLoc)=RMat(iiiES,ivES,iiES)
            END DO
!       Calculate new translation vector
            TVec(iiiES,SymLoc)=TVec(iiiES,iiES)+Center(iiiES,iES)
!       Out of unit cell?
!       Too large
Large:      DO WHILE (TVec(iiiES, SymLoc)>=1.0D0)
              TVec(iiiES, Symloc)=TVec(iiiES, Symloc)-1.0D0
            END DO Large
!       Too small
Small:      DO WHILE (TVec(iiiES, SymLoc)<0.0D0)
              TVec(iiiES, Symloc)=TVec(iiiES, Symloc)+1.0D0
            END DO Small
          END DO
        END DO
      END DO All
!       Adjusting number of symops
      NSymPrim=NSym
      NSym=SymLoc
      END SUBROUTINE ExpandSym

! Read values for one atom
!*********************************************************
      SUBROUTINE GetAtomandCP(iBuffer,iAtomNr,key,ErrCd)
!*********************************************************

      USE GlobalVariables

      IMPLICIT NONE

      INTEGER :: iAtomNr,key ! Number of read atom, key to distinguish critical points and atoms
      INTEGER :: ErrCd       ! counter,error code
      CHARACTER*(*) :: iBuffer
      CHARACTER*(LineLength) :: dBuffer  ! One line of input file containing atom information
      ErrCd=0
      IF (key==0) THEN !0 ~ atoms
        Atoms(iAtomNr)%Name=FirstWord(iBuffer,.true.)
        CALL ArrayReal(iBuffer, Atoms(iAtomNr)%Coo(1),3,ErrCd)
      ELSE ! other key ~ critical point
        CPList(iAtomNr)%Name=FirstWord(iBuffer,.true.)
        CALL ArrayReal(iBuffer, CPList(iAtomNr)%Coo(1),3,ErrCd)
      END IF        
      RETURN
      IF (ErrCd/=0) THEN
        write(dBuffer,'(i3)') iAtomNr  
        IF (key==0) THEN
          dBuffer='Error reading line '//dBuffer(1:3)//'of keyword edmaatoms.'
	ELSE
          dBuffer='Error reading line '//dBuffer(1:3)//'of keyword cplist.'
	END IF
	CALL ErrorMsg(dBuffer)
      END IF
      END SUBROUTINE GetAtomandCP

! Get the voxel division and dimension from the input map
!*********************************************************
       SUBROUTINE GETNVOX
!*********************************************************
      USE GlobalVariables

      INTEGER :: nx(6)                  ! Number of steps in direction iorient(i)
      INTEGER :: nxny                   !=nx(1)*nx(2)
      INTEGER :: nmap                   ! Number of sections
      INTEGER :: mapa                   ! Type of map (1=Patterson, 2=Fourier, ?=differece Fourier, ?=checking Fourier)
      REAL    :: dx(6)                  ! Step in iorien(i)
      INTEGER :: iorien(6)              ! Order of axes
      INTEGER :: nsubs                  ! part number of composite crystals (?)
      REAL :: xmin(6), xmax(6)          ! minimum and maximum value of iorien(i)
      INTEGER :: lrec
      INTEGER, DIMENSION(MaxDim) :: NVoxTmp
      CHARACTER*(linelength) :: VoxBuff
      LOGICAL :: Warn

      IF (InRhoType=='jana') THEN
        open(30,file=inrhofile,form='unformatted',status='old', access='direct',recl=34*ReclFac,err=900)
        read(30,rec=1)(lrec,i=1,7)
        close(30)
        open(30,file=inrhofile,form='unformatted',status='unknown', &
               access='direct',recl=max(lrec,34)*ReclFac,err=900)      
!       Read input file header
        READ (30, rec=1, ERR=910) nx, nxny, nmap, (xmin(cnt1),xmax(cnt1),cnt1=1,6), dx, iorien, mapa, nsubs
        CLOSE(30)
!         Guess number of dimensions
        DO i=6,1,-1
          IF (abs(xmax(i))>eps) THEN
            SDim=i
            EXIT
          END IF
        END DO
      ELSEIF (InRhoType=='xplor') THEN
        OPEN (30, ERR=900, FILE=inrhofile,STATUS='OLD')
!       Header of file
!       Empty line
        ErrCode=0
        READ (30, '(a)', ERR=910)
!       number of title lines, skip them
        READ (30, '(a)', ERR=910,END=910) Buffer
        CALL OneInt(Buffer,Lrec,ErrCode)
        IF (ErrCode/=0) GOTO 910
        DO cnt1=1,Lrec
          linecnt=linecnt+1
          READ(30,'(a)',ERR=910)
        ENDDO
!       Voxel spacing
        READ(30,'(a)',ERR=910) Buffer         
	SDim=0
	DO
	  IF (idel(Buffer)>0) THEN
            DO cnt2=1,3
	      VoxBuff=FirstWord(Buffer,.true.)
              CALL OneInt(VoxBuff,nx(cnt2),ErrCode)
	    ENDDO
	    If (errcode/=0) GOTO 910
	    SDim=SDim+1
	  ELSEIF (SDim<3) THEN
	    STOP 'Error, the dimension of the input denisty map is smaller than 3'
	  ELSE
	    EXIT
	  ENDIF
	  NVoxTmp(SDim)=nx(1)
	ENDDO
        CLOSE(30)
        IF (ErrCode/=0) GOTO 910
      ENDIF
      DDim=SDim-RDim
      IF (DDim>0) then
        CALL FlagKwdGroup('4D',.true.)
      END IF        
!     initialize the voxel divisions
      ALLOCATE(NVox(1:SDim))
      IF (AllocError.GT.0) STOP 'Error allocating voxel division!!'
      DO i=1,SDim
        IF (InRhoType=='jana') THEN
	  NVox(i)=nx(i)
        ELSEIF(InRhoType=='xplor') THEN
	  NVox(i)=NVoxTmp(i)
	ENDIF
	NVoxTot=NVoxTot*NVox(i)
        N(i)= NVox(i)*N(i-1)
      END DO
      goto 999
900   STOP 'Error, cannot open the electron density file!'
910   STOP 'Error reading the header of the input file!'
999   continue
      END SUBROUTINE GETNVOX      

!       Determine length of Numbers and sign
!       Input:
!       String
!       Direction to search
!       Position to begin search
!       Output:
!       Number
!******************************************************************
      SUBROUTINE NumberLength (DString, LR, KeyPos, Number)
!******************************************************************
      IMPLICIT NONE
      CHARACTER :: DString*(*)
      INTEGER :: Number, OverAll, KeyPos, i, j, StepL, StepR, Step, BegLoop
      INTEGER :: EnLoop, Begin
      PARAMETER (StepL=-1, StepR=+1, Begin=1)
      CHARACTER*1 :: Nums(10), LR, SignChar(0:1)
      LOGICAL :: IsNum
      DATA Nums /'1','2','3','4','5','6','7','8','9','0'/
      DATA SignChar /'+','-'/
      Number=1
      IsNum=.FALSE.
      OverAll=LEN(DString)
      IF (LR.EQ.'l') THEN
        Step=StepL
        BegLoop=KeyPos-1
        EnLoop=Begin
      END IF
      IF (LR.EQ.'r') THEN
        Step=StepR
        BegLoop=KeyPos+1
        EnLoop=OverAll
      END IF
      DO i=BegLoop, EnLoop, Step
        DO j=1,10
          IF (DString(i:i).EQ.Nums(j)) IsNum=.TRUE.
        END DO
        IF (.NOT.IsNum) THEN
          IF (Step.EQ.StepR) THEN
            READ (DString(BegLoop:i-1),'(I6)') Number
            GOTO 17020
          END IF  
          IF (Step.EQ.StepL) THEN
            IF (DString(i:i).EQ.SignChar(1)) THEN
              IF (i.NE.(KeyPos-1)) THEN
                READ (DString(i:BegLoop),'(I7)') Number
                GOTO 17020
              ELSE
                Number=-1
                GOTO 17020
              END IF
            ELSE
              IF (DString(i:i).EQ.SignChar(0)) THEN
                IF (i.NE.(KeyPos-1)) THEN
                  READ (DString(i+1:BegLoop),'(I7)') Number
                  GOTO 17020
                ELSE
                  Number=1
                  GOTO 17020
                END IF
              END IF
            END IF
          END IF
          ELSE IF((i.EQ.1).AND.(LR.EQ.'l').AND.(IsNum)) THEN
            READ (DString(1:BegLoop),'(I6)') Number
            GOTO 17020
        END IF
        IsNum=.FALSE.
      END DO
17020 CONTINUE
      END SUBROUTINE NumberLength

        SUBROUTINE GuessType(FileName,FileType)
	IMPLICIT NONE
        CHARACTER*(*) :: FileName, FileType
	INTEGER :: i, idel
	! Search for the period
	DO i=idel(FileName),1,-1
	  IF (FileName(i:i)=='.') THEN
	    SELECT CASE (Filename(i+1:idel(Filename)))
	      CASE ('m81')
                FileType='jana'
	      CASE ('xplor')
                FileType='xplor'
              CASE DEFAULT
	        FileType='unknown'
	    END SELECT 
	    EXIT
	  ENDIF
	ENDDO
        END SUBROUTINE GuessType

        SUBROUTINE CheckAllowedChars(String,key,ErrCd)
	integer, parameter :: CharNr=17
	integer :: ErrCd,loop,key
        character*(*) :: String
	character*1, dimension(1:CharNr) :: alchr,alcenchr,alsymchr,justnumchr,reflchr,digits
	logical :: ok
	data alsymchr  /' ',' ','x','1','2','3','4','5','6','7','8','9','0','.','/','+','-'/
	data alcenchr  /' ',' ','1','2','3','4','5','6','7','8','9','0','.','/','+','-',' '/
        data justnumchr/' ',' ','1','2','3','4','5','6','7','8','9','0','.',' ','+','-',' '/
        data reflchr   /' ',' ','1','2','3','4','5','6','7','8','9','0','.',' ','+','-','?'/
        data digits    /'1','2','3','4','5','6','7','8','9','0','0','0','0','0','0','0','0'/
        alsymchr(2)=char(9) !char(9) cannot be in the data statement
        alcenchr(2)=char(9)
        justnumchr(2)=char(9)
	if (key==1) then
	  alchr=alsymchr
	elseif (key==2) then
	  alchr=alcenchr
	elseif (key==3) then
	  alchr=justnumchr
	elseif (key==4) then
	  alchr=reflchr
	elseif (key==5) then
	  alchr=digits
	endif
	DO loop=1,idel(string)
	  ok=.false.
	  DO loop2=1,CharNr
	    IF (string(loop:loop)==alchr(loop2)) THEN
	      ok=.true.
	      exit
	    END IF
	  END DO
          ErrCd=0
	  if (.not.ok) then
	    ErrCd=100
	    RETURN
	  endif
        END DO
        END SUBROUTINE CheckAllowedChars

      END SUBROUTINE INPUT
!     Last change:  LP   16 Mar 2005   10:09 am
!****************************************************************
!* PROGRAM EDMA - for topological analysis of discrete          *
!* electron density maps                                        *
!* Lukas Palatinus                                              *
!* Sander van Smaalen                                           *
!* version 05/01/2005                                           *
!****************************************************************
      program EDMA
      
      USE GLOBALDEFINITIONS
      USE GLOBALVARIABLES
      
      IMPLICIT NONE
      
      INTEGER :: i,ii,iii,pcnt ! counters
      INTEGER :: ipc ! Cljf
      INTEGER :: GetRecLFac !Function returning RecLFac for direct-acces files
      INTEGER :: ArrLoc,ArrLocR !Positions in rho arrays. Any name ending with R applies to the real dimension density
      INTEGER :: Dxyz(MaxDim),DxyzCur(MaxDim),Dxyz2(MaxDim),Dxyz3(MaxDim)  !dummy coordinates
      INTEGER :: StartC(MaxDim),StartCR(MaxDim),EndC(MaxDim),EndCR(MaxDim) !starting and ending coordinates of the "hypercubes" of electron density
      INTEGER :: Orig(MaxDim)  !Origin of the density section for spline interpolation
      INTEGER :: Idel !function returning length of the string
      DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: y,y2 !dummy variables, need in the Splines subroutine, but not used here
      DOUBLE PRECISION,DIMENSION(MaxDim) :: t,x4 !vectors of t a x in additional dimensions
      DOUBLE PRECISION,DIMENSION(MaxDim) :: pt !precise coordinates returned by peak serching routine
      DOUBLE PRECISION :: grad, rout,disc,dpos,dsc(MaxDim),Gr !"rho out", discrepancy, dummy position, scaling of the output coordinates
      INTEGER, DIMENSION(MaxDim) :: DerOrd,AxOrd !Order of the derivatives(not used here), ored of the axes in the input map
      INTEGER, DIMENSION(MaxDim) :: CumDifVec !Cummulative difference vector - calculation of the center of charge 
      CHARACTER*(24) :: filefmt='( (''_'',f4.2),''.m81'')' !format template for the output t-maps
      TYPE EVstr_type
        character*(linelength) :: line
      END TYPE EVstr_type
      Type(EVstr_type), DIMENSION(3) :: EVstr
      DOUBLE PRECISION :: rhorsigma
      LOGICAL :: cpyes, succ, succ2, NextCoo, ongridstep, defaultongridstep, hadloop !flag, funciton returning next coordinate in the density array
 
!      open(2,status='scratch') !Cljf
      RecLFac=GetRecLFac()
      CALL Input !read in all the parameters
      WRITE(filefmt(2:2),'(i1)') SDim-RDim !Serves to create output filenames
      ALLOCATE(NVoxR(1:RDim),STAT=AllocError) !pixel division of the real space density
      IF (AllocError.GT.0) STOP 'Error allocating NVoxR! Too little memory?'
      ALLOCATE(NVoxRim(1:RDim),STAT=AllocError) !pixel division with additional areas of the neighboring cells
      IF (AllocError.GT.0) STOP 'Error allocating NVoxR! Too little memory?'
      Aper=0
      DO i=1,SDim-RDim
        DO ii=1,RDim
          IF (.not.projection) THEN !Otherwise we project the superspace density along the internal coordinates and that is periodic
            if (abs(Qvec(ii,i))>eps) Aper(ii)=1 !! indicates particular direction being an aperiodic one
	  END IF
	END DO
      END DO
      Rim=0
      DO i=1,RDim
          Rim(i)=(int(tolerance*NVox(CO(i))/Cell(i))+1+range/2+1)       ! add range and tolerance to the unit cell 
            !Muptiplication by Aper assures periodic directions have Rim=0
          Rim(i)=Aper(i)*max(Rim(i),int(addborder*NVox(CO(i)))+1) ! if com is computed, take maximum of what is needed
      END DO    
      DO i=1,RDim ! Permute the superspace pixel division acording to the order of the axes, fill all different NVox* arrays
	NVoxr(i)=NVox(CO(i))	
	NVoxRim(i)=NVoxR(i)+2*Rim(i)
	NVoxTotR=NVoxTotR*NVoxR(i)
	NVoxTotRim=NVoxTotRim*NVoxRim(i)
        NRim(i)= NVoxRim(i)*NRim(i-1)
      END DO
      cApvx=Cell(7)/dble(NVoxTotR) !cubic Angstrom per voxel
      ALLOCATE(rho(0:NVoxTot-1),STAT=AllocError) !Superspace density
      IF (AllocError.GT.0) STOP 'Error allocating rho! Too little memory?'
      ALLOCATE(rhor(0:NVoxTotRim-1),STAT=AllocError) !real dimension density
      IF (AllocError.GT.0) STOP 'Error allocating rhor! Too little memory?'
      IF (cpformat/='none') THEN ! gradient of the density to quickly locate the CP's
        ALLOCATE(rhogr(0:NVoxTotRim-1),STAT=AllocError) !real dimension density
        IF (AllocError.GT.0) STOP 'Error allocating rhor! Too little memory?'
      END IF
      ALLOCATE(srho(0:range**(max(SDim-RDim,RDim))-1),STAT=AllocError) !Part of the density for the spline interpolation
      IF (AllocError.GT.0) STOP 'Error allocating srho! Too little memory?'
!     Arrays, which are not in use, but must be supplied for the Splines subroutine
      ALLOCATE(y(range),STAT=AllocError)
      IF (AllocError.GT.0) STOP 'Error allocating dummy spline variables! Too little memory?'
      ALLOCATE(y2(range),STAT=AllocError)
      IF (AllocError.GT.0) STOP 'Error allocating dummy spline variables! Too little memory?'
      DO i=1,SDim
        DerOrd(i)=0
	AxOrd(i)=i
      END DO
      write(*,*) 'Reading the density map ',Inrhofile(1:Idel(inrhofile))
      SELECT CASE(InRhoType)
        CASE ('jana')
	  CALL ReadJana(InRhoFile,AllocError) !read in the density
        CASE ('xplor') 
	  CALL ReadXplor(InRhoFile,AllocError) !read in the density
      END SELECT
!     If fullcell no is given, check if the map really has the symmetry given in the unit cell
      IF (.not.fullcellpeaks) CALL CheckSymmetry
      IF (AllocError/=0) STOP 'Error reading input electron density'
! If only projection of the density onto the realspace is to be analyzed, then
      IF ((projection).and.(SDim>RDim)) THEN
! 1/ project the density onto realspace
        Dxyz=0
        DO i=1,RDim
          StartCR(i)=0 !Fill the "lower front left" and "upper back right" corner of the electron density section considered in the calculation of the maxima and charge
          EndCR(i)=NVoxR(i)-1
          Dxyz(i)=StartCR(i)
	END DO
        write(*,'(a)') ' Projecting the electron density along the internal coordinates:'
	DO
	  Call ArrayLocationR(Dxyz,ArrLocR,1) !get array location for current coordinates
!         Permute accvording to the order of axes in the input file
          DO i=1,SDim
	    Dxyz2(i)=Dxyz(CO(i))
	  END DO
!         Now cycle over all internal dimensions and sum the density
          rhor(ArrlocR)=0.0D0
121       Call ArrayLocation(Dxyz2,ArrLoc,1)
          rhor(ArrlocR)=rhor(ArrlocR)+rho(ArrLoc)
!           move one point further
          Cnt2=RDim+1
141       Dxyz2(CO(cnt2))=Dxyz2(CO(cnt2))+1
          IF (Dxyz2(CO(cnt2))>NVox(CO(Cnt2))-1) THEN
            Dxyz2(CO(cnt2))=0
            cnt2=cnt2+1
            if (cnt2==SDim+1) goto 161
	    goto 141
          END IF      
          goto 121
161       continue        
          DO i=RDim+1,SDim
            rhor(ArrLocR)=rhor(ArrLocR)/NVox(CO(i))
          END DO
!           move one point further
          IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz,AxOrd)) EXIT
        END DO
! 2/ change the dimension accordingly
        SDim=RDim
! And that's all what has to be be done!
      END IF

!       Write the header of the output file, if needed
      IF ((maxformat/='none').or.(cpformat/='none')) THEN
        OPEN(UNIT=90,file=outrhofile(1:idel(outrhofile))//'.coo')
        WRITE(90,'(2a)')   '# Analysis of the density map ',inrhofile(1:idel(inrhofile))
        WRITE(90,'(a,i3)') '# Dimension of the map  : ',SDim
        Buffer='(a, i5)'
	WRITE(Buffer(4:4),'(i1)') SDim
	WRITE(90,Buffer)   '# Division of the map   : ',(NVox(i),i=1,SDim)
        Buffer='(a, i1)'
	WRITE(Buffer(4:4),'(i1)') SDim
        WRITE(90,Buffer)   '# Order of the axes     : ',(CO(i),i=1,SDim)
        WRITE(90,'(a,i3)') '# Range for spline interpolation: ',range
        IF (SDim/=RDim) THEN!applies only to the more-dimensional case
          Buffer='(a, f9.4)'
          WRITE(Buffer(4:4),'(i1)') 2*RDim
          WRITE(90,Buffer) '# Cell: ',(Cell(i),i=1,2*RDim)
          WRITE(90,'(a)') '# q-vectors:'
          Buffer='(a2, f9.5)'
	  WRITE(Buffer(5:5),'(i1)') RDim
          DO i=1,SDim-RDim
	    write(90,Buffer) '# ',(qvec(ii,i),ii=1,RDim)
	  END DO
          IF (.not.projection) THEN
	    WRITE(90,'(a)') '# limits and steps of t:'
	    WRITE(90,'(a)') '#   start     end     step'
            DO i=1,SDim-RDim
	      write(90,'(''# '',3f9.5)') (TList(ii,i),ii=1,3)
	    END DO
          ELSE
	    WRITE(90,'(a)') '# The density will be projected onto 3D before analysis,\.'
	  END IF
	END IF
	IF (com) THEN
	  WRITE(90,'(a)', ADVANCE='NO') '# The center of charge and total charge of atomic basins will be computed using '
	  SELECT CASE (basinmode)
	    CASE(ongridmode)
              WRITE(90,'(a)') 'an on-grid algorithm (fast, but possibly inaccurate)'
	    CASE(neargridmode)
              WRITE(90,'(a,i2,a)') 'a near-grid algorithm with ',EdgeRemovalCycles,' cycles of edge removal.'
	    CASE(pedanticmode)
              WRITE(90,'(a,i2,a)') 'a pedantic algorithm (very slow, but the most accurate).'
	  END SELECT
	  WRITE(90,'(2(a,f7.3))') '# The density in the incommensurate directions is calculated from',-addborder,' to',1.+addborder
	  WRITE(90,'(a,f7.3,a)') '# The density limit for the pixel to be accepted in the com calculation is ',chlim,'*Rho(max)'
	ELSE
	  WRITE(90,'(a)') '# The center of charge and total charge of atomic basins will not be computed.'
	END IF
        WRITE(90,'(2a)') '# The scale of the output coordinates will be ',scale
        IF (maxformat(1:idel(maxformat))=='atoms') THEN
	  write(90,'(a,f6.3,a,f9.3,a)') '# Positions of following atoms +- ',tolerance, &
	                       ' and with density over', plim, ' will be searched for maxima at each t'
          IF (position=='relative') THEN
            write(90,'(a)') '# Coordinates are related to the input coordinates of the atoms:'
	  ELSE IF (position=='absolute') THEN
            write(90,'(a)') '# Coordinates are related to the origin of the unit cell.'
	  END IF
          Buffer='(a2,a8, f9.5)'
          WRITE(Buffer(8:8),'(i1)') RDim
          DO i=1,AtomNr
            write(90,Buffer) '#   ',Atoms(i)%Name,(Atoms(i)%Coo(ii),ii=1,RDim)
	  END DO
          write(90,'()')
	ELSE IF (maxformat/='none') THEN
	  write(90,'(a,f9.3,a)') '# All maxima with density above ',plim,' will be printed for each t.'
          write(90,'()')
        END IF
        IF (cpformat(1:idel(cpformat))=='list') THEN
	  write(90,'(a,f6.3,a,f9.3,a)') '# Following positions +- ',tolerance, &
	                       ' and with density over', plim, ' will be searched for crtitical points at each t'
          IF (position=='relative') THEN
            write(90,'(a)') '# Coordinates are related to the input coordinates of the critical points:'
	  ELSE IF (position=='absolute') THEN
            write(90,'(a)') '# Coordinates are related to the origin of the unit cell.'
	  END IF
          Buffer='(a2,a8, f9.5)'
          WRITE(Buffer(8:8),'(i1)') RDim
          DO i=1,CPlistNr
            write(90,Buffer) '#   ',CPlist(i)%Name,(Cplist(i)%Coo(ii),ii=1,RDim)
	  END DO
	ELSE IF (cpformat/='none') THEN
	  write(90,'(a,f9.3,a)') '# All critical points with density above ',plim,' will be printed for each t.'
        END IF
        IF (cpformat/='none') then
	  IF (expandedcp) THEN
	    write(90,'(a)') '# Eigenvectors of the critical points will be listed.'
            write(90,'()')
	  ELSE
	    write(90,'(a)') '# Eigenvectors of the critical points will not be listed.'
            write(90,'()')
	  END IF
	END IF
	
        CLOSE (90)
      END IF
      IF ((maxformat=='all').or.(cpformat=='all')) THEN
        OPEN(UNIT=91,file=outrhofile(1:idel(outrhofile))//'.coo',POSITION='APPEND') !we can write data directly to the output file
      ELSE IF (maxformat=='atoms') THEN
        OPEN(UNIT=91,file=outrhofile(1:idel(outrhofile))//'.tmp') !we need to write to the temporary file and then rewrite the data into convenient format later
        fullcellpeaks=.true. !We don't know where are the atoms located -> search the whole unit cell
      END IF
      IF (cpformat=='list')THEN
        OPEN(UNIT=92,file=outrhofile(1:idel(outrhofile))//'.tmc') !we need to write to the temporary file and then rewrite the data into convenient format later
      END IF
!     cycle over the additional dimensions according to the t-settings
      DO i=1,SDim-RDim
        t(i)=Tlist(1,i) !Fill actual t with the start point of the intervals
      END DO
50    continue            
        IF (SDim==RDim) THEN !periodic "3D" case, we just need to assign rho to rhor
          !just if we haven't made projection, otherwise we have filled rhor already
	  if (.not.projection) THEN
            rhor=rho
            deallocate(rho)
          ENDIF
          if (maxformat=='atoms') write(91,*) !maintain the format of the temporary file
          if (cpformat=='list') write(92,*) !maintain the format of the temporary file
	  goto 170 ! and we can skip to the determination of the t-section
        ELSE
	  fullcellpeaks=.true. !The symmetry does not work in the t-sections of the modulated density 
	END IF
        write(*,'(a)',ADVANCE='NO') 't= '
        IF ((maxformat=='all').or.(cpformat=='all')) write(91,'(a)',ADVANCE='NO') '# t= '
        DO i=1,SDim-RDim
          write(*,'(f7.3)',ADVANCE='NO') t(i) !list the actual t's on the screen...
          IF ((maxformat/='none').or.(cpformat=='all'))  write(91,'(f7.3)',ADVANCE='NO') t(i) !...and in the output file
          IF (cpformat=='list')  write(92,'(f7.3)',ADVANCE='NO') t(i) !...and in the output file
	END DO
        write(*,*)
        IF ((maxformat/='none').or.(cpformat=='all')) write(91,*)
        IF (cpformat=='list') write(92,*)
!       cycle over the R-Dimensional real space:              
	DO i=1,RDim
          StartCR(i)=-Rim(i) !Fill the "lower front left" and "upper back right" corner of the electron density section considered in the calculation of the maxima and charge
          EndCR(i)=NVoxR(i)+Rim(i)-1
          Dxyz(i)=StartCR(i)
	END DO
        pcnt=0 !counts the pixel which were already searched - just for writing out the progress
        write(*,'(a)') ' Extracting the electron density in real dimensions:'
        call send(-1) !Cljf
	DO
	  pcnt=pcnt+1
!Cljf
          IF (mod(pcnt,int(float(NVoxTotRim)/100.))==0) THEN
	    write(*,'(i3,2a)',ADVANCE='NO') nint(100.*float(pcnt)/float(NVoxTotRim)),'%',char(13) !write out progress
            ipc =  nint(100.*float(pcnt)/float(NVoxTotRim))
            write(*,'(i3,2a)',ADVANCE='NO') ipc,'%',char(13) !write out progress
            call fplivc
            call send(ipc)
	  END IF
	  Call ArrayLocationR(Dxyz,ArrLocR,1) !get array location for current coordinates
!         calculate x4 from t's
          x4=0.0D0
          DO i=1,SDim-RDim
            DO ii=1,RDim
              x4(i)=x4(i)+qvec(ii,i)*DBLE(Dxyz(ii))/DBLE(NVoxR(ii))
            END DO
            x4(i)=(x4(i)+t(i))*NVox(CO(i+RDim))
          END DO
!           fill the srho array
          DO ii=1,SDim-RDim
            Dxyz(ii+RDim)=int(x4(ii)+float(range)/2.)+1-range !This construct assures proper setting of the origin in additional dimensions
          END DO
!         Dxyz now contains correct real space coordinates and origin of the "hypercube" in additional dimensions.
!         The indices must be permuted according to CO to reach the correct points in the superspace density
          DO i=1,SDim
	    Dxyz2(i)=Dxyz(CO(i))
	  END DO
          Cnt1=0
!         And now fill the srho ("spline rho") with corresponding section of the censity in additional dimension(s)
120       Call ArrayLocation(Dxyz2,ArrLoc,1)
          srho(Cnt1)=rho(ArrLoc)
!           move one point further
          Cnt2=RDim+1
140       Dxyz2(CO(cnt2))=Dxyz2(CO(cnt2))+1
          IF (Dxyz2(CO(cnt2))>Dxyz(Cnt2)+range-1) THEN
            Dxyz2(CO(cnt2))=Dxyz(Cnt2)
            cnt2=cnt2+1
            if (cnt2==SDim+1) goto 160
	    goto 140
          END IF      
          Cnt1=Cnt1+1
          goto 120    
160       continue        
            !RhoAtPoint returns the electron density at x4 interpolated using bicubic splines
          CALL RhoAtPoint(SDim-RDim,range,DerOrd,AxOrd,x4,Dxyz(RDim+1),srho,rout,y,y2)
          rhor(ArrLocR)=rout
!           move one point further
          IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz,AxOrd)) EXIT
        END DO
        write(*,*)
!       well, we have the density, now extract the maxima
170     CONTINUE
!       if plimkey='relative', we must find the peak density in the map
        IF (plimkey=='relative') THEN
          dpos=0.0D0
          DO i=1,NVoxTotR
	    dpos=max(dpos,rhor(i))
	  END DO
          plim=plim*dpos !Plim was a relative number up to now, let's make it absolute
	ELSEIF (plimkey=='sigma') THEN
          rhorsigma=0.0
          DO i=0,NVoxTotR-1
            rhorsigma=rhorsigma+(rhor(i))**2
          ENDDO
	  rhorsigma=sqrt(rhorsigma/NVoxTotR)
	  plim=plim*rhorsigma	
	END IF
        IF (maxformat=='none') goto 180 !skip writing maxima
	CALL SearchMaxima
        IF (scale=='angstrom') THEN 
          DO i=1,RDim
            dsc(i)=Cell(i) !set scaling of the output coordinates to angstroms
          END DO
         ELSE
           dsc=1.0D0 !no scaling
        END IF
        IF (maxformat=='atoms') THEN
          IF (position=='relative') THEN 
	    dpos=1.0D0 !allow subtraction of the input atomic coordinates
	  ELSE
	    dpos=0.0D0 !absolute coords in the unit cell
	  END IF	    
	  DO i=1,AtomNr ! repeat over all atoms in the list
! Map the periodic directions onto the unit cell. Aperiodic can't be simply mapped
            DO Cnt1=1,RDim !find out the largest difference between current maximum and current atom
              DO WHILE (Atoms(i)%Coo(Cnt1)>=1.0D0)
                Atoms(i)%Coo(Cnt1)=Atoms(i)%Coo(Cnt1)-1.0D0
              END DO
              DO WHILE (Atoms(i)%Coo(Cnt1)<0.0D0)
                Atoms(i)%Coo(Cnt1)=Atoms(i)%Coo(Cnt1)+1.0D0
              END DO
	    END DO
	    DO ii=1,MaxNr !repeat over all maxima           
              disc=0.0D0
	      cpyes=.false.
	      DO Cnt1=1,RDim !find out the largest difference between current maximum and current atom
	        disc=max(disc,Cell(Cnt1)*abs(Maxima(ii)%Coo(Cnt1)-Atoms(i)%Coo(Cnt1)))
	      END DO
              IF (disc<=tolerance) THEN
                IF (com.and.(Maxima(ii)%cc<chlimlist)) CYCLE
                  IF (maxformat=='atoms')  THEN
                  DO cnt2=1,SDim-RDim
                    x4(cnt2)=0.0D0
		    DO iii=1,RDim
                      x4(cnt2)=x4(cnt2)+qvec(iii,cnt2)*Atoms(i)%Coo(iii)
                    END DO
                    x4(cnt2)=x4(cnt2)+t(cnt2)
                    write(91,'(f7.3)',ADVANCE='NO') x4(cnt2)
	          END DO
		END IF
                IF (com) THEN !write coordinates of peak, center of charge, total charge and peak density
                  write(91,'(4X,3f12.6,''   '',3f12.6,''   '',f11.5,2f12.4)',ADVANCE='NO') &
                   (dsc(Cnt1)*(Maxima(ii)%Coo(Cnt1)-Atoms(i)%Coo(Cnt1)*dpos),Cnt1=1,RDim), & 
                   (dsc(Cnt1)*(Maxima(ii)%Com(Cnt1)-Atoms(i)%Coo(Cnt1)*dpos),Cnt1=1,RDim), & 
		   Maxima(ii)%cc, Maxima(ii)%rho, Maxima(ii)%vol
                ELSE !write only coordinates and peak density
                  write(91,'(4X,3f12.6,''   '',f12.4)',ADVANCE='NO') &
                   (dsc(Cnt1)*(Maxima(ii)%Coo(Cnt1)-Atoms(i)%Coo(Cnt1)*dpos),Cnt1=1,RDim), &
		   Maxima(ii)%rho
 	        END IF
                IF (Maxima(ii)%err) THEN
	          write(91,'(a)') '  !'
	        ELSE
	          write(91,'(a)')
                END IF
		cpyes=.true. !flag that the maximum for given atom has been found
		EXIT
	      END IF
	    END DO
            IF (.not.cpyes) write(91,'(a47)') '                    no maximum found           '
	  END DO
	ELSE !we simply list all maxima
          write(91,'(a)') '#  maxima:'
          IF (com) THEN
	    write(91,'(2a)') '#     x           y           z              x_coc       y_coc       z_coc        ',&
	                    'charge     density        vol.'
          ELSE
            write(91,'(a)') '#     x           y           z             density'
	  END IF
	  DO i=1,MaxNr
            IF (com.and.(Maxima(i)%cc<chlimlist)) CYCLE
            IF (com) THEN !write coordinates of peak, center of charge, total charge and peak density
              write(91,'(3f12.6,''   '',3f12.6,''   '',f11.5,2f12.4)',ADVANCE='NO') &
               (dsc(ii)*Maxima(i)%Coo(ii),ii=1,RDim), &
               (dsc(ii)*Maxima(i)%Com(ii),ii=1,RDim),Maxima(i)%cc,Maxima(i)%rho,Maxima(i)%vol 
	    ELSE !write only coordinates and peak density
              write(91,'(3f12.6,''   '',f12.4)',ADVANCE='NO') &
               (dsc(ii)*Maxima(i)%Coo(ii),ii=1,RDim), Maxima(i)%rho
            END IF
            IF (Maxima(i)%err) THEN
	      write(91,'(a)') '  !'
	    ELSE
	      write(91,'(a)')
            END IF
	  END DO
          write(91,*)
          write(91,*)
	END IF
180     IF (cpformat=='none') goto 190 !skip the search for critical points
!       Search for critical points
        CALL SearchCP
        IF (scale=='angstrom') THEN 
          DO i=1,RDim
            dsc(i)=Cell(i) !set scaling of the output coordinates to angstroms
          END DO
         ELSE
           dsc=1.0D0 !no scaling
        END IF
        IF (cpformat=='list') THEN
          IF (position=='relative') THEN 
	    dpos=1.0D0 !allow subtraction of the input atomic coordinates
	  ELSE
	    dpos=0.0D0 !absolute coords in the unit cell
	  END IF	    
	  DO i=1,CPlistNr ! repeat over all cpoints in the list
	    DO ii=1,CPNr !repeat over all found CP
              disc=0.0D0
	      cpyes=.false.
	      DO Cnt1=1,RDim !find out the largest difference between current cpoint and current atom
	        disc=max(disc,Cell(Cnt1)*abs(Cpoint(ii)%Coo(Cnt1)-CPlist(i)%Coo(Cnt1)))
	      END DO
              IF (disc<=tolerance) THEN
                write(92,'(4X,3f12.6,2X,a9,f12.4,3f14.4)') & 
                 (dsc(Cnt1)*(Cpoint(ii)%Coo(Cnt1)-CPlist(i)%Coo(Cnt1)*dpos),Cnt1=1,RDim), &
                  Cpoint(ii)%type, Cpoint(ii)%rho, (Cpoint(ii)%Evl(Cnt1),Cnt1=1,RDim)
	          cpyes=.true. !flag that the maximum for given atom has been found
                IF (expandedcp) THEN
	          DO Cnt1=1,RDim
	            write(92,'(20X,3f12.6)') (CPoint(ii)%Evc(Cnt1,Cnt2),Cnt2=1,RDim)
	          END DO
	        END IF
                EXIT
	      END IF
	    END DO
            IF (.not.cpyes) write(92,'(a47)') '                no critical point found       '
	  END DO
	ELSE !we simply list all critical points
          write(91,'(a)') '#  critical points:'
          write(91,'(a)') '#     x           y           z         type        density       lambda1       lambda2       lambda3'
	  DO i=1,CPNr                        
            write(91,'(3f12.6,2X,a9,f12.4,3f14.4)') &
             (dsc(ii)*Cpoint(i)%Coo(ii),ii=1,RDim), Cpoint(i)%type, Cpoint(i)%rho,(Cpoint(i)%Evl(ii),ii=1,RDim)
            IF (expandedcp) THEN
	      DO Cnt1=1,RDim
	        write(91,'(20X,3f12.6)') (CPoint(i)%Evc(Cnt1,Cnt2),Cnt2=1,RDim)
	      END DO
	    END IF
	  END DO
          write(91,*)
          write(91,*)
	END IF
190     IF ((writetmap).and.((SDim/=RDim).or.(projection))) THEN !write out current tmap, only for more-dimensional densities
          IF (projection) THEN
            Buffer='_proj.m81'
	  ELSE
            WRITE(Buffer,FMT=filefmt) (t(i),i=1,SDim-RDim) !add values of t to the output filename
          END IF
	  Buffer=outrhofile(1:idel(outrhofile))//Buffer(1:Idel(Buffer))
          write(*,'(2a)') 'Saving the current density to ',Buffer(1:Idel(Buffer)) !show the output filename
          CALL WriteJanaR(Buffer(1:idel(Buffer)),.false.,AllocError) !write the map
        END IF
        IF ((writebasin).and.(com)) THEN
          IF (SDim/=RDim) THEN !nD case has different format then 3D case
            WRITE(Buffer,FMT=filefmt) (t(i),i=1,SDim-RDim)
            Buffer=outrhofile(1:idel(outrhofile))//'_basins'//Buffer(1:Idel(Buffer))
          ELSE ! 3D case
            Buffer=outrhofile(1:idel(outrhofile))//'_basins.m81'
          END IF
          write(*,*) Buffer(1:Idel(Buffer))
          DO i=0,NVoxTotRim-1
            rhor(i)=float(basin(i)) !write values from the array basin (containing numbers of maxima) to rhor
          END DO
          CALL WriteJanaR(Buffer(1:idel(Buffer)),.false.,AllocError) !write the map
        END IF
        IF (wantstructure) THEN
          IF (RDim==SDim) THEN
	    Buffer='.'//OutstrType(1:idel(outstrtype))
	  ELSE
	    WRITE(Buffer,'(''_'',f4.2,a)') t(1),'.'//OutstrType(1:idel(outstrtype))
          END IF
	  Buffer=outstrfile(1:idel(outstrfile))//Buffer(1:Idel(Buffer))
          SELECT CASE (OutStrType)
	  CASE ('m40')
	    write(*,'(2a)') ' Saving the maxima in jana m40 format to file ',Buffer(1:Idel(Buffer)) !show the output filename
	  CASE ('cif')
	    write(*,'(2a)') ' Saving the maxima in cif format to file ',Buffer(1:Idel(Buffer)) !show the output filename
	  CASE ('ins')
	    write(*,'(2a)') ' Saving the maxima in SHELX ins format to file ',Buffer(1:Idel(Buffer)) !show the output filename
          END SELECT
	  CALL WriteStructure(buffer)
          call send(-4) !Cljf	  
        END IF
        IF (Compm40.and.(SDim==RDim)) CALL Comparem40
        IF (SDim==RDim) goto 500
!       move one t further
        Cnt2=1
400     t(cnt2)=t(cnt2)+TList(3,Cnt2)
        IF (t(cnt2)>Tlist(2,Cnt2)+eps) THEN
          t(Cnt2)=TList(1,Cnt2)
          cnt2=cnt2+1
          if (cnt2==SDim-RDim+1) goto 500
          goto 400
        END IF      
      goto 50      
500   CONTINUE
      IF (maxformat=='atoms') THEN ! we have the data in the temporary file, we must rearange them!
        OPEN(UNIT=90,file=outrhofile(1:idel(outrhofile))//'.coo',POSITION='APPEND') !we will append the data in convenient format to the main data file
	DO cnt1=1,AtomNr
          write(90,*)
          write(90,*)
	  write(90,'(2a)') '# ',atoms(cnt1)%name(1:idel(atoms(cnt1)%name))
	  rewind(91) !return back to the beginning of the file
          read(91,'(a)',END=700) buffer ! read t
          write(90,'(a)',ADVANCE='NO') '#'
	  DO i=1,2*idel(Buffer)
	    write(90,'(a1)',ADVANCE='NO') ' '
	  END DO
          IF (com) THEN
	    write(90,'(2a)') '         x           y           z              x_coc       y_coc       z_coc        charge     ',&
	                     'density        vol.'
          ELSE
            write(90,'(a)') '         x           y           z             density'
	  END IF
	  rewind(91) !return back to the beginning of the file
600       read(91,'(a)',END=700) buffer ! read t
          write(90,'(a)',ADVANCE='NO') buffer(1:idel(buffer)) !write t
	  DO cnt2=1,AtomNr !read sequentially the records for atoms...
	    read(91,'(a)') Buffer
	    if (cnt2==cnt1) write(90,'(a)') buffer(1:idel(buffer)) !...and write that one that matches
	  END DO
          GOTO 600
700       continue
        END DO
      END IF
      IF (maxformat/='none') THEN 
        CLOSE (90)
        CLOSE (91)
      END IF
!     And do the same with the critical points
      IF (cpformat=='list') THEN ! we have the data in the temporary file, we must rearange them!
        OPEN(UNIT=90,file=outrhofile(1:idel(outrhofile))//'.coo',POSITION='APPEND') !we will append the data in convenient format to the main data file
	DO cnt1=1,CPlistNr
          write(90,*)
          write(90,*)
	  write(90,'(2a)') '# ',CPlist(cnt1)%name(1:idel(CPlist(cnt1)%name))
          rewind(92)
	  read(92,'(a)',END=701) buffer ! read t
          write(90,'(a)',ADVANCE='NO') '#'
          DO i=1,idel(Buffer)
	    write(90,'(a1)',ADVANCE='NO') ' '
	  END DO
	  write(90,'(a)') '         x           y           z         type        density       lambda1       lambda2       lambda3'
	  rewind(92) !return back to the beginning of the file
651       read(92,'(a)',END=701) buffer ! read t
          write(90,'(a)',ADVANCE='NO') buffer(1:idel(buffer)) !write t
	  DO cnt2=1,CPlistNr !read sequentially the records for atoms...
	    read(92,'(a)') Buffer
            IF (expandedcp) THEN
              DO i=1,RDim
                read(92,'(a)') Evstr(i)%line
              END DO
	    END IF
	    if (cnt2==cnt1) THEN
	      write(90,'(a)') buffer(1:idel(buffer)) !...and write that one that matches
              IF (expandedcp) THEN
	        DO i=1,RDim
	          write(90,'(a)') Evstr(i)%line(1:idel(Evstr(i)%line))
	        END DO
	      END IF
	    end if
	  END DO
          GOTO 651
701       continue
        END DO
      END IF
      IF (cpformat=='list') THEN 
        CLOSE (90)
	CLOSE (92)
      END IF
  
      CALL FinalCleanUp      
      call send(-5) !Cljf
                     
      CONTAINS
        SUBROUTINE SearchMaxima
          IMPLICIT NONE
	  !Sobel operator in 3D: http://www.aravind.ca/cs788h_Final_Project/gradient_estimators.htm
          INTEGER, PARAMETER :: wghtpower=2
	  INTEGER :: ListNr,ActMax,MaxAl, PixelCnt, NbCount,HalfNbCount !number of item in the PixList, number of actual maximum, array location for actual maximum
          INTEGER :: ERLoop, EdgeRefFlag, ErcLoop, curbasin, currefbasin, MaxPixCnt
          DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:) :: voxdist, rhonbhood, DeltaRho
          DOUBLE PRECISION, DIMENSION(1:MaxDim) :: StepxG, Gradvec,GradCor, CurGradCor
	  DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:,:) :: MMatrix, Minverse, RMatrix, MRW
          DOUBLE PRECISION :: MaxRho, MaxGrad, dummy
	  LOGICAL :: pedantic

!	  write(*,*) 'Entering Laplacian'
!	  CALL Laplacian
!zero everything what must be zero
          MaxNr=0
          IF (Com) THEN
            IF (.not.allocated(Diffvec)) THEN !We come here for the first time)
              cnt1=0
              MaxPixCnt=0
              DO i=1,RDim
	        cnt1=max(cnt1,NVoxRim(i))
	        MaxPixCnt=MaxPixCnt+NVoxRim(i) !maximum possible length of the path to maximum
	      END DO
              ALLOCATE (DiffVec(0:NVoxTotRim-1,RDim),STAT=AllocError) !allows to determine borders of particular atomic basin
              IF (AllocError.GT.0) STOP 'Error allocating Diffvec! Too little memory?'
              ALLOCATE (Pixlist(1:MaxPixCnt),STAT=AllocError) ! Stores all pixels found on the way to the maximum
              IF (AllocError.GT.0) STOP 'Error allocating PixList! Too little memory?'
              ALLOCATE (Diff(1:MaxPixCnt,1:RDim),STAT=AllocError) ! Stores differences along the gradient
              IF (AllocError.GT.0) STOP 'Error allocating Diff! Too little memory?'
              ALLOCATE (Basin(0:NVoxTotRim-1),STAT=AllocError) !array containing assignment of the pixel to the atomic basin
              IF (AllocError.GT.0) STOP 'Error allocating Basin! Too little memory?'
	    ENDIF
            Basin=0
            DO i=1,MaxMaxNr
              Maxima(i)%cc=0.0
              Maxima(i)%ulc=0.0
              Maxima(i)%err=.false.
              Maxima(i)%vol=0.0D0
              DO ii=1,MaxDim
                Maxima(i)%com(ii)=0.0
              END DO
	    END DO
	  END IF  
          pixelCnt=1
	  DO i=1,RDim !set the starting point of the "cube" (unit cell +- tolerance)
            StartC(i)=-(int(tolerance*NVoxR(i)/Cell(i))+1)
	    EndC(i)=NVoxR(i)+int(tolerance*NVoxR(i)/Cell(i))+1
            Dxyz(i)=StartC(i)
	    AxOrd(i)=i
            PixelCnt=PixelCnt*(NVoxR(i)+int(tolerance*NVoxR(i)/Cell(i))+int(tolerance*NVoxR(i)/Cell(i))+3)
          END DO
          pcnt=0
	  write(*,*) 'Searching for maxima:'
!Cljf
          call send(-2)
	  DO !repeat over all pixels in the "cube"
            pcnt=pcnt+1
	    IF (mod(pcnt,int(float(PixelCnt)/100.))==0) THEN
!Cljf start change
              ipc = nint(100.*float(pcnt)/float(PixelCnt))
              write(*,'(i3,2a)',ADVANCE='NO') ipc,'%',char(13) !write progress
              call fplivc
              call send(ipc)
!Cljf end change
	    END IF
	    Call ArrayLocationR(Dxyz,ArrLoc,1)
            IF (rhor(ArrLoc)<plim) GOTO 1000 !ignore too low (potential) maxima
            cpyes=.true.
!             Cycle over the cube 3x3x3... to find a maximum
            DO i=1,RDim
              StartCR(i)=Dxyz(i)-1
	      EndCR(i)=Dxyz(i)+1
              Dxyz2(i)=StartCR(i)
            END DO
            DO
              Call ArrayLocationR(Dxyz2,Cnt2,1)
	      IF (rhor(ArrLoc)<rhor(Cnt2)) THEN
                cpyes=.false.
                EXIT
              ELSEIF ((rhor(ArrLoc)==rhor(Cnt2)).and.(Arrloc>Cnt2)) THEN !If equal values of different pixels, take only the one with lower ArrLoc
                cpyes=.false.
                EXIT
	      END IF
!              move one point further            
              IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz2,AxOrd)) EXIT
            END DO
	    IF (cpyes) THEN !we have found a maximum, let's save it
              IF (.not.fullcellpeaks) THEN !We want to exclude symmetry equivalent peaks
		succ=.true.
                DO i=1,RDim
		  IF ((Dxyz(i)>=NVox(CO(i))).or.(Dxyz(i)<0)) succ=.false.
                END DO
                IF (.not.succ) goto 1000 !Outside the unit cell, ignore such maximum
                succ2=.true.
                DO i=1,MaxNr
		  DO ii=2,NSym
		    CALL SymOp(ii,Maxima(i)%ICoo,Dxyz2)
                    succ=.false.
                    DO iii=1,RDim
                      DO WHILE (Dxyz2(iii)>=NVox(CO(iii)))
                        Dxyz2(iii)=Dxyz2(iii)-NVox(CO(iii))
                      END DO
                      DO WHILE (Dxyz2(iii)<0)
                        Dxyz2(iii)=Dxyz2(iii)+NVox(CO(iii))
                      END DO
                      IF (abs(Dxyz(iii)-Dxyz2(iii))>1) succ=.true.
                    END DO
		    IF (.not.succ) EXIT
		  END DO
		  IF (.not.succ) THEN !Another equivalent maximum already exists
		    Maxima(i)%mult=Maxima(i)%mult+1
		    succ2=.false.
		  END IF
		END DO
                IF (.not.succ2) GOTO 1000
	      ENDIF
	      !We come here only if the maximum is really new and not symmetricaly equivalent to another maximum
	      CALL GetPeak(Dxyz,pt,rout,.true.) !get the exact position and density of the maximum
              IF ((pt(1)>-1.0E29).and.(rout>plim)) THEN
                MaxNr=MaxNr+1
	        IF (MaxNr>MaxMaxNr-1) THEN !
	          write(*,'(A,I4,A)') 'Warning: The maximum number of ',MaxMaxNr,' maxima in one t-map exceeded.'
	          write(*,'(A)')      'Some maxima have been ommited. Try increasing the density treshold (keyword plim)'
	          write(*,'(A)')      'or increasing the parameter MaxMaxNr in module GlobalDefinition and recompiling the program.'
                  IF (com) THEN 
		    write(*,'(A)')      'Charges and centers of charge will not be calculated.'
                    com=.false.
                  END IF
                  MaxNr=MaxNr-1
                  RETURN
	        END IF
	        DO i=1,RDim
	          Maxima(MaxNr)%ICoo(i)=Dxyz(i)
	        END DO
                  DO i=1,RDim
	            Maxima(MaxNr)%Coo(i)=pt(i)/DBLE(NVoxR(i))
	          END DO
                Maxima(MaxNr)%rho=rout
                Maxima(MaxNr)%err=.false.
	        Maxima(MaxNr)%al=ArrLoc
                Maxima(MaxNr)%mult=1
                IF (com) THEN !assign the number of maximum to the pixel of maximum in the Basin array
	        ! only if not assigned yet, otherwise we have two equivalent maxima (e.g. 0,0,0 and 1,0,0)
		  IF (Basin(Arrloc)==0) THEN
		    DO Cnt1=1,RDim
		      Diffvec(ArrLoc,Cnt1)=0 !Difference of the maximum to itself is zero
		    END DO
		    Basin(Arrloc)=MaxNr
                    !We have to check, if a neighboring pixel does not have the same value. In that case assign the same basin number
                    StartCR=Maxima(MaxNr)%Icoo-1
                    EndCR=Maxima(MaxNr)%Icoo+1
                    Dxyz2=StartCR
                    DO
                      Call ArrayLocationR(Dxyz2,Cnt2,1)
                      IF (rhor(ArrLoc)==rhor(Cnt2)) Basin(Cnt2)=MaxNr
                      IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz2,AxOrd)) EXIT
                    END DO
		  END IF
	        END IF
	      END IF
	    END IF
1000        IF (.not.NextCoo(RDim,StartC,EndC,Dxyz,AxOrd)) EXIT !if .true., returns next Dxyz 
          END DO
          write(*,*)
1010      IF (com) THEN !we want to calculate the center of charge and total charge
	    !Fill the voxdist array and the MRW matrix to prepare the gradient calculation
            NBCount=3**RDim
	    HalfNBCount=3**RDim/2
            allocate(VoxDist(1:NBCount))
	    allocate(RhoNBHood(1:NBCount))
	    allocate(DeltaRho(1:HalfNbCount))
	    allocate(MRW(1:RDim,1:HalfNbCount))
	    allocate(Mmatrix(1:RDim,1:RDim))
	    allocate(Minverse(1:RDim,1:RDim))
            IF (((basinmode==neargridmode).and.(edgeremovalcycles>0)).or.(basinmode==pedanticmode)) THEN
              ALLOCATE (RefBasin(0:NVoxTotRim-1),STAT=AllocError)
              IF (AllocError.GT.0) STOP 'Error allocating RefBasin! Too little memory?'
	    ENDIF

            StartCR=-1; EndCR=1; Dxyz=StartCR
	    i=0
	    DO
              i=i+1
	      if (all(Dxyz(1:3)==0)) THEN
                voxdist(i)=1.0
	      else
	        StepxG(1:RDim)=matmul(dble(Dxyz(1:RDim))/NVox,G)
                voxdist(i)=sqrt(sum((dble(Dxyz(1:RDim))/NVox)*StepxG(1:RDim)))
	      endif
              IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz,AxOrd)) EXIT
            END DO
            
	    !Calculate the matrix G_grid^-1M^-1
	    MMatrix=0.0
	    StartCR=-1; EndCR=1; Dxyz=StartCR
	    iii=0
	    DO
	      iii=iii+1
	      DO i=1,RDim
	        DO ii=1,RDim
		  Mmatrix(i,ii)=MMatrix(i,ii)+dble(Dxyz(i)*Dxyz(ii))/(NVox(i)*NVox(ii)*voxdist(iii)**wghtpower)
	        ENDDO
	      ENDDO
              IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz,AxOrd)) EXIT
            ENDDO	    
	    call invert(MMatrix, Minverse,RDim)	    
	    Minverse=matmul(Gi,Minverse) !Minverse now contains G_grid^-1 M^-1
            !Calculate The matrix W'
	    StartCR=-1; EndCR=1; Dxyz=0; Dxyz(1)=1 !We start from the center of the cube, not from the corner, to obtain only the "positive" vectors
	    iii=0; 
	    DO
	      iii=iii+1
              MRW(1:3,iii)=Dxyz(1:3)/(voxdist(iii+HalfNbCount+1)**wghtpower) !MRW now contains W'
              IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz,AxOrd)) EXIT
            ENDDO	    
	    MRW=matmul(Minverse,MRW)	    

            write(*,'(a)') ' Integrating charge of the atoms: '
!Cljf
            call send(-3)
	    IF (basinmode==ongridmode) THEN
	      ongridstep=.true.; defaultongridstep=.true.; EdgeRefFlag=1; pedantic=.false.
	    ELSEIF (basinmode==neargridmode) THEN
	      ongridstep=.false.; defaultongridstep=.false.; pedantic=.false.
	      IF (EdgeRemovalCycles==0) THEN
	        EdgeRefFlag=1
	      ELSE
	        EdgeRefFlag=2
	      ENDIF
	    ELSEif (basinmode==pedanticmode) THEN
	      ongridstep=.false.; defaultongridstep=.false.; EdgeRefFlag=1; pedantic=.true.
              RefBasin=Basin
	    ENDIF
            DO erloop=1,EdgeRefFlag              
              StartC(1:RDim)=-Rim(1:RDim)+Aper(1:RDim)
	      EndC(1:RDim)=NVoxR(1:RDim)-1+Rim(1:RDim)-Aper(1:RDim)
              DxyzCur(1:RDim)=StartC(1:RDim)
              HadLoop=.false.
              pcnt=0
	      DO !repeat over all pixels
                pcnt=pcnt+1
	        IF (mod(pcnt,int(float(NVoxTotRim)/100.))==0) THEN
                  ipc = nint(100.*float(pcnt)/float(NVoxTotRim))
                  write(*,'(i3,2a)',ADVANCE='NO') ipc,'%',char(13) !write progress
                  call fplivc
                  call send(ipc)
                END IF
                Call ArrayLocationR(DxyzCur,ArrLoc,1) !get location in rhor
                if ((Basin(ArrLoc)>0).or.(Basin(ArrLoc)==-1)) GOTO 1300 !this pixel has been already checked -> go to next pixel
	        PixList=0; ListNr=0
                Dxyz=DxyzCur
	        GradCor=0.0
1100            Call ArrayLocationR(Dxyz,ArrLoc,1) !Now follow the gradient from the Dxyzcur to maximum
		DO i=1,RDim
	           IF ((Aper(i)==1).and.((Dxyz(i)<=-Rim(i)).or.(Dxyz(i)>=NVoxR(i)+Rim(i)-1))) THEN
	             ! we have come to the edge of the known density -> discard these pixels from the analysis
                     actmax=-1		     
		     goto 1200
	           END IF
	        END DO
		IF (.not.pedantic) THEN
		  curbasin=Basin(ArrLoc)
		  currefbasin=Basin(Arrloc)
		ELSE
		  curbasin=Basin(ArrLoc)
		  currefbasin=RefBasin(Arrloc)
		ENDIF
                IF (((currefbasin>0).and.(currefbasin<99000)).or.(currefbasin==-1)) THEN !this pixel has been already assigned
	          actmax=currefbasin !make its maximum the active maximum
		  goto 1200
	        ELSEIF (curbasin>99000) THEN !We have been here already before -> we are in a loop!
		  DO i=1,ListNr-1
		    IF (ArrLoc==PixList(i)) THEN
		      ListNr=i-1
		      ongridstep=.true.
                      hadloop=.true.
		      EXIT
		    ENDIF 
                  ENDDO
	        END IF
	        ListNr=ListNr+1 !add the pixel to the PixList
	        IF (Basin(Arrloc)<99000) Basin(Arrloc)=Basin(Arrloc)+100000 !Flag the pixel as part of the currently analyzed trace
                IF (ListNR>MaxPixCnt) THEN
	          write(*,*) 'Error, internal inconsistency! Please contact the author.'
	          STOP
	        ENDIF
                PixList(ListNr)=ArrLoc
!               Cycle over the hypercube 3x3x3... to fill the list of neighboring density values
                StartCR(1:RDim)=Dxyz(1:RDim)-1; EndCR(1:RDim)=Dxyz(1:RDim)+1; Dxyz2(1:RDim)=StartCR(1:RDim)
                ii=0
	        DO
	          ii=ii+1
                  Call ArrayLocationR(Dxyz2,Cnt2,1)
	          RhoNBHood(ii)=rhor(Cnt2)
                  IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz2,AxOrd)) EXIT
                END DO	      
	        IF (maxval(RhoNBhood)==RhoNBhood(14)) THEN
	          GradVec=0.0
		  MaxAl=ArrLoc
	        ELSEIF (ongridstep) THEN
	          MaxGrad=-1000000.
                  StartCR=Dxyz-1; EndCR=Dxyz+1; Dxyz2=StartCR
                  MaxGrad=-100000.0
	          ii=0
	          DO
	            ii=ii+1
                    Call ArrayLocationR(Dxyz2,Cnt2,1)
	            grad=((rhor(Cnt2)-rhor(ArrLoc))/voxdist(ii))
                    IF (grad>maxgrad) THEN
                      MaxGrad=grad !Remember the highest gradient
                      Dxyz3=Dxyz2 !Remember the coordinates of the highest density
                      MaxAl=Cnt2
                    END IF
!                    move one point further            
                    IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz2,AxOrd)) EXIT
                  ENDDO		
		  GradCor=0.0
		  IF (Basin(MaxAl)<99000) ongridstep=defaultongridstep !Return to the default mode if we are not trapped in a loop
	        ELSE
		  !Prepare DeltaRho
		  DO i=1,HalfNbCount
		    DeltaRho(i)=RhoNBHood(HalfNbCount+1+i)-RhoNBHood(HalfNbCount+1-i)
		  ENDDO
                  GradVec(1:RDim)=matmul(MRW,DeltaRho)	!Calculate the gradient
                  dummy=maxval(abs(GradVec(1:RDim)))	!Get the largest component
                  if (dummy<eps) THEN !This is a minimum in  special position. Disgusting! We have to get out of it in an arbitrary direction
                    GradVec=0.0
                    GRadVec(1)=1.
                  ELSE
	            GradVec(1:3)=GradVec(1:3)/dummy		!Normalize so that the latgest component is 1
	          ENDIF
                  CurGradCor(1:3)=GradCor(1:3)+GradVec(1:3) !Add the current gradient to the sum of previous corrections
                  Dxyz3(1:3)=nint(dble(Dxyz(1:3))+CurGradCor(1:3)) !Find the voxel nearest to the tip of the corrected normalized gradient
		  GradCor(1:3)=dble(Dxyz(1:3)-Dxyz3(1:3))+CurGradCor(1:3) !Calculate the new correction
	          Call ArrayLocation(Dxyz3,MaxAl, 1)
	        ENDIF

                IF (MaxAl==ArrLoc) THEN ! The current pixel is a maximum
		  IF (Basin(ArrLoc)-100000<=0) THEN
	            ! we have found an "uninteresting" maximum, one that is not inside the basic unit cell or bellow plim
                    actmax=-1
		    !Add it to refbasin, if necessary
                  ELSE
		     actmax=Basin(ArrLoc)-100000 !actmax will be asigned to all pixels in PixList
                     Diff(ListNr,1:RDim)=0
		  END IF
	        ELSE !move to the pixel with higher density and repeat the search
		  DO Cnt1=1,RDim
                    Diff(ListNr,Cnt1)=Dxyz3(Cnt1)-Dxyz(Cnt1)
                  END DO
                  Dxyz=Dxyz3
                  goto 1100
	        END IF
1200            continue ! We have found pixel with non-zero ActMax
                IF (actmax>0) CumDifVec(1:RDim)=DiffVec(Arrloc,1:RDim)		
		DO i=ListNr,1,-1
                  IF (.not.pedantic) THEN
		    basin(PixList(i))=actmax !fill the Basin array
                  ELSE
		    basin(PixList(i))=basin(PixList(i))-100000 !Erase the "on-the-current-track" flag
		  ENDIF
		  IF (actmax>0) THEN !if we have some "accepted" maximum, fill the Mxmbox array appropriately 
	            CALL CoordinatesR(PixList(i),1, Dxyz3)
                    CumDifVec(1:RDim)=CumDifVec(1:RDim)+Diff(i,1:RDim) 
                    IF (.not.pedantic) DiffVec(Pixlist(i),1:RDim)=CumDifVec(1:RDim)
                  END IF
	        END DO
                IF (pedantic) THEN
		  basin(Pixlist(1))=actmax
		  IF (actmax>0) DiffVec(Pixlist(1),1:RDim)=CumDifVec(1:RDim)
		ENDIF
		IF (HadLoop) THEN !Set to zero all remaining labels of touched pixels
                  where (Basin>99000) Basin=Basin-100000
		  HadLoop=.false.
	        ENDIF
!		IF (any(Basin>99000)) THEN
!		  write(*,*) 'Neco nevynulovaneho:'
!		  DO i=1,ListNr
!		    write(*,*) PixList(i),Basin(PixList(i)), pedantic
!		  ENDDO
!		  read(*,*)
!		ENDIF
1300            IF (.not.NextCoo(RDim,StartC,EndC,DxyzCur,AxOrd)) EXIT  !cycle to the next non-tested pixel
	      END DO !End of the scan over all pixels
         
              IF ((EdgeRefFlag==2).and.(erloop==1)) THEN !Set to zero all pixels on the border between basins
                write(*,*) 'Erasing the voxel assignment at the borders between the basins:'
		RefBasin=Basin
                pcnt=0		
		DO ercloop=1,EdgeRemovalCycles
		  StartC(1:RDim)=-Rim(1:RDim)+Aper(1:RDim)
	          EndC(1:RDim)=NVoxR(1:RDim)-1+Rim(1:RDim)-Aper(1:RDim)
                  Dxyz=StartC		
		  DO
                    pcnt=pcnt+1
	            IF (mod(pcnt,int(float(NVoxTotRim*EdgeRemovalCycles)/100.))==0) THEN
                      ipc = nint(100.*float(pcnt)/float(NVoxTotRim*EdgeRemovalCycles))
                      write(*,'(i3,2a)',ADVANCE='NO') ipc,'%',char(13) !write progress
                      call fplivc
                      call send(ipc)
                    END IF
                    Call ArrayLocationR(Dxyz,ArrLoc,1) !get location in rhor
		    !For every pixel check the neighborhood, and if any surrounding pixel is different, set this to zero
                    StartCR=Dxyz-1; EndCR=Dxyz+1; Dxyz2=StartCR
                    ii=0
                    DO
	              ii=ii+1
                      Call ArrayLocationR(Dxyz2,Cnt2,1)
	              IF (RefBasin(Cnt2)/=RefBasin(Arrloc)) Basin(ArrLoc)=0
                      IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz2,AxOrd)) EXIT
                    ENDDO
                    IF (.not.NextCoo(RDim,StartC,EndC,Dxyz,AxOrd)) EXIT
		  ENDDO
		  RefBasin=Basin
                  pedantic=.true.
                ENDDO
                !Now assing the positions of the maxima again, to make sure that they were not erased
		DO i=1,MaxNr
                  IF (Basin(Maxima(i)%al)==0) THEN
		    Basin(Maxima(i)%al)=i
                    !We have to check, if a neighboring pixel does not have the same value. In that case assign the same basin number
                    StartCR=Maxima(MaxNr)%Icoo-1
                    EndCR=Maxima(MaxNr)%Icoo+1
                    Dxyz2=StartCR
                    DO
                      Call ArrayLocationR(Dxyz2,Cnt2,1)
                      IF (rhor(ArrLoc)==rhor(Cnt2)) Basin(Cnt2)=i
                      IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz2,AxOrd)) EXIT
                    END DO
		  END IF
	        ENDDO
		RefBasin=Basin
		write(*,*) 'Assigning the voxels at the borders:'
	      ENDIF
	    ENDDO !End of the "edge-refine" loop
!	    !Check the symmetry
!		write(*,*) 'Checking the symmetry:'
!		DO i=1,MaxNr-1
!                  DO ii=i+1,MaxNr
!                    DO iii=2,NSym
!                      CALL SymOp(iii,Maxima(i)%ICoo,Dxyz2)
!	              CALL ArrayLocation(Dxyz2,cnt2,1)
!		      IF (cnt2==Maxima(ii)%al) THEN
!		        where (Basin==ii) Basin=i
!		      ENDIF
!                    ENDDO
!                  ENDDO		  
!		ENDDO
!		StartC(1:RDim)=-Rim(1:RDim)+Aper(1:RDim)
!	        EndC(1:RDim)=NVoxR(1:RDim)-1+Rim(1:RDim)-Aper(1:RDim)
!                Dxyz=StartC
!		pcnt=0
!		DO
!                  Call ArrayLocationR(Dxyz,ArrLoc,1) !get location in rhor
!                  DO ii=2,NSym
!                    CALL SymOp(ii,Dxyz,Dxyz2)
!	            CALL ArrayLocation(Dxyz2,cnt2,1)
!		    IF (Basin(ArrLoc)/=Basin(cnt2)) THEN
!		      write(*,*) Dxyz(1:3), Arrloc, Basin(Arrloc), ':', Dxyz2(1:3), Cnt2, Basin(Cnt2)
!		    ENDIF
!                  ENDDO
!                  IF (.not.NextCoo(RDim,StartC,EndC,Dxyz,AxOrd)) EXIT
!		ENDDO


            deallocate(VoxDist)
	    deallocate(RhoNBHood)
	    deallocate(DeltaRho)
	    deallocate(MRW)
	    deallocate(Mmatrix)
	    deallocate(Minverse)
	    if (allocated(RefBasin)) deallocate(RefBasin)

	    ! OK, we have filled the basin array, now we know exactly what pixel belongs to what basin, we can sum up the charge and calculate the center of charge
            IF (maxformat=='none') RETURN !we are interested only in the atomic basin, not in the maxima
!            Sum up the weighted average of coordinates and at the same time sum up the cummulative charge
            DO cnt1=0,NVoxTotRim-1
              IF (basin(cnt1)<=0) cycle !belongs to "uninteresting maximum"
              CALL CoordinatesR(Maxima(basin(cnt1))%al,1,Dxyz2)
              Maxima(basin(cnt1))%cc=Maxima(basin(cnt1))%cc+max(rhor(Cnt1),0.0D0)
              Maxima(basin(cnt1))%vol=Maxima(basin(cnt1))%vol+1.0D0
              IF (rhor(Cnt1)>chlim*rhor(Maxima(basin(cnt1))%al)) THEN !density is high enough to include it in the com calculation
                ! transform the coordinates to neighboiring cell, if necessary
                DO i=1,RDim
                  Dxyz(i)=Dxyz2(i)-DiffVec(cnt1,i)
                  Maxima(basin(cnt1))%com(i)=Maxima(basin(cnt1))%com(i)+rhor(Cnt1)*DBLE(Dxyz(i))/DBLE(NVoxR(i))
                END DO
                Maxima(basin(cnt1))%ulc=Maxima(basin(cnt1))%ulc+rhor(Cnt1) !sum the charge included in com calculation
              END IF
            END DO            
            dpos=0.0D0
            DO cnt1=1,MaxNr
              IF (Maxima(cnt1)%ulc>eps) THEN
                DO i=1,RDim
                  Maxima(Cnt1)%com(i)=Maxima(Cnt1)%com(i)/Maxima(Cnt1)%ulc
                END DO
	          ELSE
                DO i=1,RDim
                  Maxima(Cnt1)%com(i)=Maxima(Cnt1)%coo(i)
                END DO
	          ENDIF
              Maxima(Cnt1)%cc=Maxima(Cnt1)%cc*cApvx !renormalize it to the real charge
              dpos=max(dpos,Maxima(Cnt1)%cc) !dpos stores now the maximal charge
              Maxima(Cnt1)%ulc=Maxima(Cnt1)%ulc*cApvx
              Maxima(Cnt1)%vol=Maxima(Cnt1)%vol*cApvx
	        END DO
            IF (chlimlistkey=='relative') then
              chlimlist=chlimlistorig*dpos
            ELSE
              chlimlist=chlimlistorig
            ENDIF
            DO cnt1=1,MaxNr
              IF (basin(Maxima(cnt1)%al)/=cnt1) THEN !maximum equivalent to another maximum
	            !everything is the same, only the coordinates must be correspondingly transformed
                cnt2=basin(Maxima(cnt1)%al)
 	            Maxima(cnt1)%rho=Maxima(cnt2)%rho
	            Maxima(cnt1)%cc=Maxima(cnt2)%cc
	            Maxima(cnt1)%ulc=Maxima(cnt2)%ulc
	            Maxima(cnt1)%err=Maxima(cnt2)%err
                Maxima(cnt1)%vol=Maxima(cnt2)%vol
                DO i=1,RDim
		          Maxima(cnt1)%com(i)=Maxima(cnt2)%com(i)+Maxima(cnt1)%coo(i)-Maxima(cnt2)%coo(i)
		        END DO
	          END IF
	        END DO
!         assign an atomic type, if composition provided
!	    DO i=1,MaxNr
!	      maxcc=max(maxcc,Maxima(i)%cc)
!            END DO
!	    DO i=1,MaxNr
!              disc=10000.
!              if (ElNr>0) THEN
!                DO ii=1,ElNr
!		  if (compos(ii,1)==1) CYCLE !Ignore hydrogens in the composition
!                  IF (abs(1.-(Maxima(i)%cc/maxcc)/(dble(Compos(ii,1))/dble(MaxZ)))<disc) then
!                    disc=abs(1.-(Maxima(i)%cc/maxcc)/(dble(Compos(ii,1))/dble(MaxZ)))
!	            Maxima(i)%an=ii
!                  END IF
!                END DO	  
!              else
!                Maxima(i)%an=1
!              END IF
!            END DO
          END IF
	END SUBROUTINE SearchMaxima

        SUBROUTINE SearchCP ! Locate critical points of the density
        INTEGER :: PixelCnt,LoopCnt,Loop
!       Create a map that contains the length of the gradient of rhor
! Combined handling of creating the whole map and only sections around the points from the list causes a bit of mess
        IF (CPformat=='list') THEN
	  LoopCnt=CPlistNr
	ELSE
	  LoopCnt=1
	END IF
	write(*,*) 'Calculating the gradient map:'
        DO Loop=1,LoopCnt
          PixelCnt=LoopCnt
	  DO i=1,RDim !set the starting point of the "cube" (unit cell +- tolerance)
            IF (CPformat=='list') THEN
              StartC(i)=int(CPlist(Loop)%Coo(i)*dble(NVoxR(i)))-(int(tolerance*NVoxR(i)/Cell(i))+1)
              EndC(i)=int(CPlist(Loop)%Coo(i)*dble(NVoxR(i)))+int(tolerance*NVoxR(i)/Cell(i))+2	    
              PixelCnt=PixelCnt*(int(tolerance*NVoxR(i)/Cell(i))+int(tolerance*NVoxR(i)/Cell(i))+4)
	    ELSE          
              StartC(i)=-(int(tolerance*NVoxR(i)/Cell(i))+1)
              EndC(i)=NVoxR(i)+int(tolerance*NVoxR(i)/Cell(i))+1
              PixelCnt=PixelCnt*(NVoxR(i)+int(tolerance*NVoxR(i)/Cell(i))+int(tolerance*NVoxR(i)/Cell(i))+3)
            END IF
            Dxyz(i)=StartC(i)
            AxOrd(i)=i
	  END DO
          pcnt=0
	  DO !repeat over all pixels in the "cube"
            pcnt=pcnt+1
!Cljf
	    IF (mod(pcnt,int(float(PixelCnt)/100.))==0) THEN
              ipc = nint(100.*float(pcnt)/float(PixelCnt)) 
              write(*,'(i3,2a)',ADVANCE='NO') ipc,'%',char(13) !write progress
              call fplivc
              call send(ipc)
	    END IF
	    
	    Call ArrayLocationR(Dxyz,ArrLoc,1)
            DO i=1,3
	      pt(i)=DBLE(Dxyz(i))
	    END DO
	    succ=.true.
	    IF (.not.succ) cycle
!       we calculate the gradient at the point ArrLoc
            Gr=0.0D0
!            call FillRhoCut(AxOrd,Orig,RDim) !fill just one row of rhoalong the actual axis to get the first derivative
            DO i=1,3
              DO cnt1=1,RDim
	        AxOrd(cnt1)=mod(i+cnt1-1,RDim)
		IF (AxOrd(cnt1)==0) AxOrd(cnt1)=RDim
                Orig(cnt1)=Dxyz(cnt1)
	      END DO
              Orig(AxOrd(1))=int(pt(AxOrd(1))-dble(range)/2.+range)+1-range
              call FillRhoCut(AxOrd,Orig,1) !fill just one row of rho along the actual axis to get the first derivative
!	      DerOrd=0
!             DerOrd(i)=1
	      DerOrd=0
	      DerOrd(AxOrd(1))=1
!	      write(*,*) srho(0:range-1)
              call RhoAtPoint(1,range,DerOrd,AxOrd,pt,orig,srho,rout,y,y2) !y and y2 contain parameters of the spline interpolation along the last axis
!              write(*,*) 'Derivace:',Orig(1:3),AxOrd(1:3),rout
              Gr=Gr+rout**2 !gr will contain square of the gradient length
	    END DO
!	    read(*,*)
            rhogr(ArrLoc)=sqrt(Gr)
	    DO i=1,RDim
              AxOrd(i)=i
	    END DO
            IF (.not.NextCoo(RDim,StartC,EndC,Dxyz,AxOrd)) EXIT !if .true., returns next Dxyz 
          END DO            
        END DO
!       Now search for minima in the RhoGr map and check presence of the critical point at each minimum
        write(*,*) 'Searching for critical points:'
        CPNr=0
        DO Loop=1,LoopCnt	
          PixelCnt=LoopCnt
          pcnt=0
	  DO i=1,RDim !set the starting point of the "cube" (unit cell +- tolerance)
            IF (CPformat=='list') THEN
              StartC(i)=int(CPlist(Loop)%Coo(i)*dble(NVoxR(i)))-int(tolerance*NVoxR(i)/Cell(i))
              EndC(i)=int(CPlist(Loop)%Coo(i)*dble(NVoxR(i)))+int(tolerance*NVoxR(i)/Cell(i))+1
              PixelCnt=PixelCnt*(int(tolerance*NVoxR(i)/Cell(i))+int(tolerance*NVoxR(i)/Cell(i))+2)
	    ELSE          
              StartC(i)=-int(tolerance*NVoxR(i)/Cell(i))
              EndC(i)=NVoxR(i)+int(tolerance*NVoxR(i)/Cell(i))
              PixelCnt=PixelCnt*(NVoxR(i)+int(tolerance*NVoxR(i)/Cell(i))+int(tolerance*NVoxR(i)/Cell(i))+3)
            END IF
            Dxyz(i)=StartC(i)
            AxOrd(i)=i
	  END DO
	  DO !repeat over all pixels in the "cube"
!Cljf
        Call ArrayLocationR(Dxyz,ArrLoc,1)
        IF (rhor(ArrLoc)<plim) GOTO 9989
        pcnt=pcnt+1
	    IF (mod(pcnt,max(int(float(PixelCnt)/100.),1))==0) THEN
              ipc =  nint(100.*float(pcnt)/float(PixelCnt))
              write(*,'(i3,2a)',ADVANCE='NO') ipc,'%',char(13) !write progress
              call fplivc
              call send(ipc)
	    END IF
	    cpyes=.true.
!           Cycle over the cube 3x3x3... to find a minimum
            DO i=1,RDim
              StartCR(i)=Dxyz(i)-1
	      EndCR(i)=Dxyz(i)+1
              Dxyz2(i)=StartCR(i)
            END DO
            DO
              Call ArrayLocationR(Dxyz2,Cnt2,1)
	      IF (rhogr(ArrLoc)>rhogr(Cnt2)) THEN
                cpyes=.false.
                EXIT
              END IF
!              move one point further            
              IF (.not.NextCoo(RDim,StartCR,EndCR,Dxyz2,AxOrd)) EXIT
            END DO
            IF (cpyes) THEN
	      CALL GetPeak(Dxyz,pt,rout,.false.) !get the exact position and density of the critical point, using spline interpolation
              IF (pt(1)>-1.0E29) THEN
	        IF (MaxNr>MaxMaxNr-1) THEN !
	          write(*,'(A,I4,A)') 'Warning: The maximum number of ',MaxMaxNr,' critical points in one t-map exceeded.'
	          write(*,'(A)')      'Some critical points have been ommited. Try increasing the parameter MaxMaxNr'
	          write(*,'(A)')      'in module GlobalDefinition and recompiling the program.'
                  RETURN
	        END IF
                CPNr=CPNr+1
		Cnt1=0
	        DO i=1,RDim
	          Cpoint(CPNr)%ICoo(i)=Dxyz(i)
                  Cpoint(CPNr)%Coo(i)=pt(i)/DBLE(NVoxR(i))
                  CPoint(CPNr)%Evl(i)=Evl(i)
		  DO ii=1,RDim
		    CPoint(CPNr)%Evc(i,ii)=Evc(i,ii)
		  END DO
                  IF (Evl(i)>0) Cnt1=Cnt1+1
                END DO
		SELECT CASE (Cnt1)
		CASE (0)
		  Cpoint(CPNr)%type='(3,-3)'
		CASE (1)
		  Cpoint(CPNr)%type='(3,-1)'
		CASE (2)
		  Cpoint(CPNr)%type='(3,+1)'
		CASE (3)
		  Cpoint(CPNr)%type='(3,+3)'
		END SELECT
                CPoint(CPNr)%rho=rout
	      END IF
            END IF
9989        IF (.not.NextCoo(RDim,StartC,EndC,Dxyz,AxOrd)) EXIT !if .true., returns next Dxyz
          END DO
	END DO
	END SUBROUTINE SearchCP
      END

! Working subroutine that prints neighborhood of the point cpt
      SUBROUTINE PrintRho(cpt,spn)
      USE GLOBALDEFINITIONS
      USE GLOBALVARIABLES

      IMPLICIT NONE
      integer :: cpt(3),tmp(MaxDim),spn,ArrLoc,i,ii,iii
      
      write(*,*) cpt
      DO i=-spn,spn
        DO ii=-spn,spn
	  DO iii=-spn,spn
            tmp(1)=cpt(1)+iii
            tmp(2)=cpt(2)+ii
            tmp(3)=cpt(3)+i
            Call ArrayLocationR(tmp,ArrLoc,1)	    
	    write(*,'(f8.4)',ADVANCE='NO') rhor(Arrloc)
	  END DO
	  write(*,*)
	END DO
	write(*,*)
      END DO
      END SUBROUTINE PrintRho

!Cljf
      SUBROUTINE send(i)
! dummy subroutine
      integer :: i

      END SUBROUTINE send
!        Read electron density in Jana(98?) format
!******************************************************************
      SUBROUTINE ReadJana (rhofile, FError)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables
      IMPLICIT NONE
      CHARACTER :: rhofile*(FileLngth)  ! Filename
      INTEGER, INTENT (OUT) :: FError   ! Errorhandling
      INTEGER :: nx(6)                  ! Number of steps in direction iorient(i)
      INTEGER :: nxny                   !=nx(1)*nx(2)
      INTEGER :: nmap                   ! Number of sections
      INTEGER :: mapa                   ! Type of map (1=Patterson, 2=Fourier, ?=differece Fourier, ?=checking Fourier)
      REAL    :: dx(6)                  ! Step in iorien(i)
      INTEGER :: iorien(6)              ! Order of axes
      INTEGER :: nsubs                  ! part number of composite crystals (?)
      REAL :: rmin, rmax                ! Minimum and maximum of the whole table
      REAL :: xmin(6), xmax(6)          ! minimum and maximum value of iorien(i)
      REAL, ALLOCATABLE :: table(:)     ! sections of density
      INTEGER :: Loc                    ! Location in rho-array
      LOGICAL :: LStop=.FALSE.          ! Error condition
      INTEGER :: i, lrec       ! Counters

      open(81,file=rhofile,form='unformatted',status='old', access='direct',recl=34*ReclFac,err=900)
      read(81,rec=1,err=900)(lrec,i=1,7)
      close(81)
      open(81,file=rhofile,form='unformatted',status='unknown', &
             access='direct',recl=max(lrec,34)*ReclFac,err=900)      

!       Header of Jana file
      READ (81, rec=1, ERR=910) nx, nxny, nmap, (xmin(cnt1),xmax(cnt1),cnt1=1,6), dx, iorien, mapa, nsubs
      DO Cnt1=1,6
        IF ((iorien(Cnt1)/=Cnt1).AND.(iorien(Cnt1)/=0)) LStop=.TRUE.
      END DO
      IF (LStop) GOTO 930
      IF ((mapa/=4).AND.(mapa/=5).AND.(mapa/=6)) GOTO 950
!       Values of map
      ALLOCATE (table(nxny), STAT=AllocError)
      IF (AllocError.GT.0) STOP 'Error allocating tables of Jana file!'
!       check the voxel division
      loc=fstel-1
      DO Cnt1=1,nmap
        READ (81, rec=Cnt1+1, ERR=910) (table(Cnt2), Cnt2=1,nxny)
!       Mapping into BayMEM rho      
        DO Cnt2=1, nxny
          Loc=Loc+1
          rho(Loc)=dble(table(Cnt2))
        END DO  
      END DO
      READ (81, rec=nmap+2, ERR=800) rmax,rmin
      rhomin=rmin
      rhomax=rmax
800   continue
      CLOSE (81, ERR=920)
      FError=0
     
      GOTO 999
!       Error handling      
900   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadJana:'
      WRITE (*,'(A)') 'Error! Cannot open density file for reading!'
      WRITE (*,'(A)') 'Nothing has been read! Returning...'
      WRITE (*,'(A)') ''
      CLOSE (81, ERR=920)
      FError=1
      GOTO 999

910   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadJana:'
      WRITE (*,'(A)') 'Error! Cannot read density file!'
      WRITE (*,'(A)') 'Something may have been read! Returning...'
      WRITE (*,'(A)') ''
      CLOSE (81, ERR=920)
      FError=2
      GOTO 999

920   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadJana:'
      WRITE (*,'(A)') 'Error! Cannot close density file!'
      WRITE (*,'(A)') 'Density should have been read, but Im not sure! Returning...'
      WRITE (*,'(A)') ''
      FError=3
      GOTO 999

930   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadJana:'
      WRITE (*,'(A)') 'Warning!'
      WRITE (*,'(A)') 'Cannot read this file!'
      WRITE (*,'(A)') 'Order of axes is not a b c ...!'
      WRITE (*,'(A)') 'Please save density with normal order of axes in JANA.'
      WRITE (*,'(A)') ''
      FError=4
      GOTO 999
      
940   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadJana:'
      WRITE (*,'(A)') 'Warning!'
      WRITE (*,'(A)') 'Cannot read this file!'
      WRITE (*,'(A)') 'Total number of voxels is inconsistent with input file!'
      WRITE (*,'(A)') ''
      FError=5
      GOTO 999

950   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadJana:'
      WRITE (*,'(A)') 'Warning!'
      WRITE (*,'(A)') 'Cannot read this file!'
      WRITE (*,'(A)') 'Map type is not Fourier!'
      WRITE (*,'(A)') 'Please save density as Fourier map.'
      WRITE (*,'(A)') ''
      FError=6
      GOTO 999

999   END SUBROUTINE ReadJana


!        Write electron density in Jana(98?) format
!******************************************************************
      SUBROUTINE WriteJana (rhofile, p1, FError)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables
      IMPLICIT NONE
      CHARACTER :: rhofile*(FileLngth)  ! Filename
      INTEGER, INTENT (OUT) :: FError   ! Errorhandling
      INTEGER :: NvoxOut(6)                !Output number of voxels, differs for p1=.true. from the working number
      INTEGER :: NVoxTotOut                !total output number of voxels
      INTEGER :: dxyz(6),Nj(6)           ! dummy vector, jump lengths for calculating the coordinates, when p1=.true.
      INTEGER :: nx(6)                  ! Number of steps in direction iorient(i)
      INTEGER :: nxny                   !=nx(1)*nx(2)
      INTEGER :: nmap                   ! Number of sections
      INTEGER :: mapa                   ! Type of map 
      REAL    :: dx(6)                  ! Step in iorien(i)
      INTEGER :: iorien(6)              ! Order of axes
      INTEGER :: nsubs                  ! part number of composite crystals (?)
      REAL :: xmin(6), xmax(6)          ! minimum and maximum value of iorien(i)
      REAL :: rmin, rmax                ! Minimum and maximum of the whole table
      REAL, ALLOCATABLE :: table(:)     ! sections of density
      INTEGER :: Loc                    ! Location in rho-array
      INTEGER :: Idel                         !Function returning the length of the string
      LOGICAL :: p1                        !Add the last redundant "slice" of voxel at 1.?
      INTEGER :: i,htmp                 !counter,dummy index in calculation of coordinates
!     Prepare Output voxels
      NVoxTotOut=1
      IF (p1) THEN
        DO i=1,SDim
          NVoxOut=1
          NVoxOut(i)=NVox(i)+1
          NVoxTotOut=NVoxTotOut*NVoxOut(i)
        END DO
!       Jump length accross the maps
        Nj=0
        Nj(2)=1
        DO i=3, SDim
          Nj(i)=NVoxOut(i)*Nj(i-1)
        END DO
      ELSE
        NVoxOut=1
        DO i=1,SDim
          NVoxOut(i)=NVox(i)
        END DO
        NVoxTotOut=NVoxTot
      END IF

!       Doing some copy work
!       Voxel spacing
      DO Cnt1=1,6
        nx(Cnt1)=1
      END DO
      DO Cnt1=1,SDim
        nx(Cnt1)=NVoxOut(Cnt1)
      END DO
!       Maplength
      nxny=NVoxOut(1)*NVoxOut(2)
!       Number of maps
      nmap=NVoxTotOut/nxny
!       Kind of map: 4 = Fourier
      mapa=4
      DO Cnt1=1,6
        IF (Cnt1<=SDim) THEN
!       Stepwidth
          dx(Cnt1)=1./NVox(Cnt1) !Here NVox must stay, not replaced by NVoxOut!
!       Maxima
          if (p1) THEN
            xmax(Cnt1)=1.
          ELSE
            xmax(Cnt1)=1.-dx(Cnt1)
          END IF
        ELSE
          dx(Cnt1)=1.0
          xmax(Cnt1)=0.0
        END IF
        xmin(Cnt1)=0.
!       Orientation of axes, ever 1 2 3 4 5 6
        iorien(Cnt1)=Cnt1
      END DO
!       Part of composite Crystal
      nsubs=1
!       Minimum of table
      rmin=100000000.0
!       Maximum of table
      rmax=-100000000.0
            
      open(81,file=rhofile,form='unformatted',status='NEW', &
             access='direct',recl=max(nxny,34)*ReclFac,err=900)      
!       Header of Jana file
      WRITE (81, rec=1, ERR=910) nx, nxny, nmap, (xmin(cnt1),xmax(cnt1),cnt1=1,6), dx, iorien, mapa, nsubs

!       Values of map
      ALLOCATE (table(nxny), STAT=AllocError)
      IF (AllocError.GT.0) STOP 'Error allocating tables of Jana file!'
      Loc=FstEl-1
      DO Cnt1=1,nmap
!       Mapping from BayMEM rho      
        DO Cnt2=1, nxny
!         Part for p1=.true.
          IF (p1) THEN
!         determine coordinates within the cell
            dxyz=0
            dxyz(1)=mod(Cnt2-1,NVoxOut(1))
            dxyz(2)=(Cnt2-1)/NVoxOut(1)
            Loc=Cnt1-1 !because first map has number 1, but the coordinate is 0
!           Reversed step direction
            DO i=SDim, 3, -1
              dxyz(i)=INT(Loc/Nj(i-1))
              Loc=Loc-dxyz(i)*Nj(i-1)
            END DO
!           The rest is a modified subroutine ArrayLocation, with remapping of the voxels to the unit cell
            Loc=0
            DO i=1, SDim
!             Mapping into unit cell
              htmp=dxyz(i)
100           IF (htmp<0) THEN
                htmp=NVox(i)+htmp
                goto 100
              ELSE IF (htmp>(NVox(i)-1)) THEN
                htmp=htmp-NVox(i)
                goto 100
              END IF
              Loc=Loc+htmp*N(i-1)
            END DO
!*********************
!         Part for p1=.false. - how simple :-)
          ELSE
            Loc=Loc+1
          END IF
          table(Cnt2)=rho(Loc)
          IF (table(Cnt2)>rmax) rmax=table(Cnt2)
          IF (table(Cnt2)<rmin) rmin=table(Cnt2)
        END DO  
        WRITE (81, rec=Cnt1+1, ERR=910) (table(Cnt2), Cnt2=1,nxny)
      END DO
      WRITE (81, rec=nmap+2, ERR=910) rmax, rmin
      CLOSE (81, ERR=920)
      FError=0
!     write information about the m81 file
      WRITE(30,'(A)') 'Electron density written to file '//rhofile(1:idel(rhofile))
      GOTO 999
      
!       Error handling      
900   WRITE (30,'(A)') ''
      WRITE (30,'(A)') 'Routine WriteJana:'
      WRITE (30,'(A)') 'Error! Cannot open density file '//rhofile(1:idel(rhofile))//' for writing!'
      WRITE (30,'(A)') 'Output file: '//rhofile(1:idel(rhofile))
      WRITE (30,'(A)') 'Nothing has been written! Returning...'
      WRITE (30,'(A)') ''
      CLOSE (81, ERR=920)
      FError=1
      GOTO 999

910   WRITE (30,'(A)') ''
      WRITE (30,'(A)') 'Routine WriteJana:'
      WRITE (30,'(A)') 'Error! Cannot write density file '//rhofile(1:idel(rhofile))//' !'
      WRITE (30,'(A)') 'Output file: '//rhofile(1:idel(rhofile))
      WRITE (30,'(A)') 'Something may have been written! Returning...'
      WRITE (30,'(A)') ''
      CLOSE (81, ERR=920)
      FError=2
      GOTO 999

920   WRITE (30,'(A)') ''
      WRITE (30,'(A)') 'Routine WriteJana:'
      WRITE (30,'(A)') 'Error! Cannot close density file '//rhofile(1:idel(rhofile))//' !'
      WRITE (30,'(A)') 'Output file: '//rhofile(1:idel(rhofile))
      WRITE (30,'(A)') 'Density should have been written, but Im not sure! Returning...'
      WRITE (30,'(A)') ''
      FError=3
      GOTO 999

999   CONTINUE
      END SUBROUTINE WriteJana

!        Write real-D electron density in Jana(98?) format
!******************************************************************
      SUBROUTINE WriteJanaR (rhofile, p1, FError)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables
      IMPLICIT NONE
      CHARACTER :: rhofile*(*)          ! Filename
      INTEGER, INTENT (OUT) :: FError   ! Errorhandling
      INTEGER :: NvoxOut(6)                !Output number of voxels, differs for p1=.true. from the working number
      INTEGER :: NVoxTotOut                !total output number of voxels
      INTEGER :: dxyz(6),Nj(6)           ! dummy vector, jump lengths for calculating the coordinates, when p1=.true.
      INTEGER :: nx(6)                  ! Number of steps in direction iorient(i)
      INTEGER :: nxny                   !=nx(1)*nx(2)
      INTEGER :: nmap                   ! Number of sections
      INTEGER :: mapa                   ! Type of map 
      REAL    :: dx(6)                  ! Step in iorien(i)
      INTEGER :: iorien(6)              ! Order of axes
      INTEGER :: nsubs                  ! part number of composite crystals (?)
      REAL :: xmin(6), xmax(6)          ! minimum and maximum value of iorien(i)
      REAL :: rmin, rmax                ! Minimum and maximum of the whole table
      REAL, ALLOCATABLE :: table(:)     ! sections of density
      INTEGER :: Loc                    ! Location in rho-array
      INTEGER :: Idel                         !Function returning the length of the string
      LOGICAL :: p1                        !Add the last redundant "slice" of voxel at 1.?
      INTEGER :: i,htmp                 !counter,dummy index in calculation of coordinates
!     Prepare Output voxels
      NVoxTotOut=1
      IF (p1) THEN
        DO i=1,RDim
          NVoxOut=1
          NVoxOut(i)=NVoxRim(i)+1
          NVoxTotOut=NVoxTotOut*NVoxOut(i)
        END DO
!       Jump length accross the maps
        Nj=0
        Nj(2)=1
        DO i=3, RDim
          Nj(i)=NVoxOut(i)*Nj(i-1)
        END DO
      ELSE
        NVoxOut=1
        DO i=1,RDim
          NVoxOut(i)=NVoxRim(i)
        END DO
        NVoxTotOut=NVoxTotRim
      END IF

!       Doing some copy work
!       Voxel spacing
      DO Cnt1=1,6
        nx(Cnt1)=1
      END DO
      DO Cnt1=1,RDim
        nx(Cnt1)=NVoxOut(Cnt1)
      END DO
!       Maplength
      nxny=NVoxOut(1)*NVoxOut(2)
!       Number of maps
      nmap=NVoxTotOut/nxny
!       Kind of map: 4 = Fourier
      mapa=4
      DO Cnt1=1,6
        IF (Cnt1<=RDim) THEN
!       Stepwidth
          dx(Cnt1)=1./NVoxR(Cnt1) !Here NVox must stay, not replaced by NVoxOut!
!       Maxima and minima
          xmin(Cnt1)=-dble(Rim(Cnt1))*dx(Cnt1)
          if (p1) THEN
            xmax(Cnt1)=1.+dble(Rim(Cnt1))*dx(Cnt1)
          ELSE
            xmax(Cnt1)=1.+dble(Rim(Cnt1)-1)*dx(Cnt1)
          END IF
        ELSE
          dx(Cnt1)=1.0
          xmin(Cnt1)=0.
          xmax(Cnt1)=0.0
        END IF
!       Orientation of axes, ever 1 2 3 4 5 6
        iorien(Cnt1)=Cnt1
      END DO
!       Part of composite Crystal
      nsubs=1
!       Minimum of table
      rmin=100000000.0
!       Maximum of table
      rmax=-100000000.0
      open(81,file=rhofile,form='unformatted',status='UNKNOWN', &
             access='direct',recl=max(nxny,34)*ReclFac,err=900)      
!       Header of Jana file
      WRITE (81, rec=1, ERR=910) nx, nxny, nmap, (xmin(cnt1),xmax(cnt1),cnt1=1,6), dx, iorien, mapa, nsubs

!       Values of map
      ALLOCATE (table(nxny), STAT=AllocError)
      IF (AllocError.GT.0) STOP 'Error allocating tables of Jana file!'
      Loc=FstEl-1
      DO Cnt1=1,nmap
!       Mapping from BayMEM rho      
        DO Cnt2=1, nxny
!         Part for p1=.true.
          IF (p1) THEN
!         determine coordinates within the cell
            dxyz=0
            dxyz(1)=mod(Cnt2-1,NVoxOut(1))
            dxyz(2)=(Cnt2-1)/NVoxOut(1)
            Loc=Cnt1-1 !because first map has number 1, but the coordinate is 0
!           Reversed step direction
            DO i=RDim, 3, -1
              dxyz(i)=INT(Loc/Nj(i-1))
              Loc=Loc-dxyz(i)*Nj(i-1)
            END DO
!           The rest is a modified subroutine ArrayLocation, with remapping of the voxels to the unit cell
            Loc=0
            DO i=1, RDim
!             Mapping into unit cell
              htmp=dxyz(i)
100           IF (htmp<0) THEN
                htmp=NVoxR(i)+htmp
                goto 100
              ELSE IF (htmp>(NVoxR(i)-1)) THEN
                htmp=htmp-NVoxR(i)
                goto 100
              END IF
              Loc=Loc+htmp*NRim(i-1)
            END DO
!*********************
!         Part for p1=.false. - how simple :-)
          ELSE
            Loc=Loc+1
          END IF
          table(Cnt2)=rhor(Loc)
          IF (table(Cnt2)>rmax) rmax=table(Cnt2)
          IF (table(Cnt2)<rmin) rmin=table(Cnt2)
        END DO  
        WRITE (81, rec=Cnt1+1, ERR=910) (table(Cnt2), Cnt2=1,nxny)
      END DO
      WRITE (81, rec=nmap+2, ERR=910) rmax, rmin
      CLOSE (81, ERR=920)
      FError=0
!     write information about the m81 file
      WRITE(30,'(A)') 'Electron density written to file '//rhofile(1:idel(rhofile))
      GOTO 999
      
!       Error handling      
900   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine WriteJanaR:'
      WRITE (*,'(A)') 'Error! Cannot open density file '//rhofile(1:idel(rhofile))//' for writing!'
      WRITE (*,'(A)') 'Output file: '//rhofile(1:idel(rhofile))
      WRITE (*,'(A)') 'Nothing has been written! Returning...'
      WRITE (*,'(A)') ''
      CLOSE (81, ERR=920)
      FError=1
      GOTO 999

910   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine WriteJanaR:'
      WRITE (*,'(A)') 'Error! Cannot write density file '//rhofile(1:idel(rhofile))//' !'
      WRITE (*,'(A)') 'Output file: '//rhofile(1:idel(rhofile))
      WRITE (*,'(A)') 'Something may have been written! Returning...'
      WRITE (*,'(A)') ''
      CLOSE (81, ERR=920)
      FError=2
      GOTO 999

920   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine WriteJanaR:'
      WRITE (*,'(A)') 'Error! Cannot close density file '//rhofile(1:idel(rhofile))//' !'
      WRITE (*,'(A)') 'Output file: '//rhofile(1:idel(rhofile))
      WRITE (*,'(A)') 'Density should have been written, but Im not sure! Returning...'
      WRITE (*,'(A)') ''
      FError=3
      GOTO 999

999   CONTINUE
      END SUBROUTINE WriteJanaR


!        Read electron density in XPlor format
!******************************************************************
      SUBROUTINE ReadXPlor(RhoFile,target)
!******************************************************************
      USE GlobalDefinitions
      USE GlobalVariables
      USE input_module
      IMPLICIT NONE
      INTEGER, PARAMETER :: rhu=81
      CHARACTER*(*) :: rhofile  ! Filename
      character*10 :: zyxstr
      character*8 :: efmt='( e  .0)', ffmt='( f  .0)', curfmt 
      INTEGER, DIMENSION(3*SDim) :: NVoxTmp
      DOUBLE PRECISION, DIMENSION(CellCon) :: Celltmp
      INTEGER :: Loc,nmap,ntable,linecnt,ErrCode,target

      linecnt=0
      OPEN (rhu, ERR=900, FILE=rhofile,STATUS='OLD')
! First of all try to determine the format
      READ (rhu, '(a)', ERR=930)
      READ (rhu, '(a)', ERR=930) Buffer
      CALL OneInt(Buffer,Loc,ErrCode)
      IF (ErrCode/=0) GOTO 930
      DO cnt1=1,Loc+6
        linecnt=linecnt+1
        READ(rhu,'(a)',ERR=930)
      ENDDO
      READ (rhu, '(a)', ERR=930) Buffer
      !Buffer contains the first line of pixels
      IF ((index(Buffer,'E')/=0).or.(index(Buffer,'r')/=0)) THEN
        curfmt=efmt
      ELSE
        curfmt=ffmt      
      ENDIF
      loc=0
      DO cnt1=1,len_trim(Buffer)
        if (Buffer(cnt1:cnt1)=='.') loc=loc+1
      ENDDO
      write(curfmt(2:2),'(i1)') loc
      write(curfmt(4:5),'(i2)') len_trim(buffer)/loc
      REWIND(rhu)
!       Header of file
!       Empty line
      linecnt=linecnt+1
      ErrCode=0
      READ (rhu, '(a)', ERR=930)
!       number of title lines, skip them
      linecnt=linecnt+1
      READ (rhu, '(a)', ERR=930,END=930) Buffer
      CALL OneInt(Buffer,Loc,ErrCode)
      IF (ErrCode/=0) GOTO 930
      DO cnt1=1,Loc
        linecnt=linecnt+1
        READ(rhu,'(a)',ERR=930)
      ENDDO
!       Voxel spacing
      linecnt=linecnt+1
      READ(rhu,'(a)',ERR=930) Buffer
      CALL ArrayInt(Buffer,NVoxTmp,3*SDim,ErrCode)
      DO cnt1=1,SDim
        IF (NVoxtmp(3*cnt1-2)/=NVox(cnt1)) THEN
          write(*,'(a,i2,a)') 'Error, the voxel division in the density along dimension',cnt1,&
                           ' is inconsistent with the input file.'
          STOP 'Error during reading of the reference density. Check the log file.'
        ELSEIF((NVoxtmp(3*cnt1-1)/=0).or.(NVoxtmp(3*cnt1)/=NVox(cnt1)-1)) THEN
          write(*,'(a,i2,a)') 'Warning, the start or end of the grid in the density along dimension',cnt1,&
                           'does not lie on the borders of the unit cell.'
        ENDIF
      ENDDO
      IF (ErrCode/=0) GOTO 930
!       Cellparameters
      linecnt=linecnt+1
      READ(rhu,'(a)',ERR=930) Buffer
      CALL ArrayReal(Buffer,Celltmp,CellCon-1,ErrCode)
      IF (ErrCode/=0) GOTO 930
      DO cnt1=1,SDim
        IF (abs(Celltmp(cnt1)-Cell(cnt1))>10.*eps) THEN
          write(*,'(a,i2,a)') 'Warning, cell parameter number',cnt1,' in the density '//&
                           ' is different from what is given in the input file.'
        ENDIF
      ENDDO
!      String ZYX defining the order of axes
      linecnt=linecnt+1
      READ(rhu,'(a)',ERR=930) zyxstr
      if (zyxstr/='ZYX') THEN
        write(*,'(a,i2,a)') 'Warning, string ZYX expected on line',linecnt,' of the density. Instead found '//zyxstr
      ENDIF
!      body of the file
      IF (SDim>1) THEN
        ntable=NVox(1)*NVox(2)
      ELSE
        ntable=NVox(1)
      ENDIF
      nmap=NVoxtot/ntable
      loc=0
      DO cnt1=0,nmap-1
        READ (rhu, '(a)', ERR=950) Buffer
        CALL OneInt(Buffer,cnt3,ErrCode)
        write(*,*) Buffer
        write(*,*) Cnt3,cnt1
        write(*,*) curfmt, ntable
        IF (Cnt3/=cnt1) THEN
          WRITE(*,'(a,i5,a)') 'Error, inconsistency in the density file.'
          WRITE(*,'(a,i4,a,i4,a)') 'Label of the layer nr.',cnt1,' is',cnt3,'.'
          STOP 'Error during reading of the reference density. Check the log file.'
        ENDIF
        IF (ErrCode/=0) GOTO 950
        READ (rhu,curfmt, ERR=950) rho(loc:loc+ntable-1)
        loc=loc+ntable
      END DO
      CLOSE (rhu, ERR=920)
      GOTO 999

!       Error handling      
900   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadXPlor:'
      WRITE (*,'(A)') 'Error! Cannot open density file for reading!'
      WRITE (*,'(A)') 'Input file file: '//trim(rhofile)
      WRITE (*,'(A)') 'Stopping.'
      WRITE (*,'(A)') ''
      STOP 'Error during reading of the density.'
      

920   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadXPlor:'
      WRITE (*,'(A)') 'Error! Cannot close density file.'
      WRITE (*,'(A)') 'Input file: '//trim(rhofile)
      WRITE (*,'(A)') ''
      STOP 'Error during reading of the density.'

930   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadXPlor:'
      STOP 'Error during reading of the density.'

950   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadXPlor:'
      WRITE(*,'(a,i5)') 'Error reading reference density at layer',cnt1+1
      STOP 'Error during reading of the density.'

999   CONTINUE
      END SUBROUTINE ReadXPlor

!        Read electron density in XPlor format
!******************************************************************
      SUBROUTINE ReadXPlorOld(RhoFile,target)
!******************************************************************
      USE GlobalDefinitions
      USE GlobalVariables
      USE input_module
      IMPLICIT NONE
      CHARACTER :: rhofile*(FileLngth)  ! Filename
      character*10 :: zyxstr 
      INTEGER, DIMENSION(3*MaxDim) :: NVoxTmp
      REAL, DIMENSION(CellCon) :: Celltmp
      DOUBLE PRECISION :: dummy
      INTEGER :: Loc,nmap,ntable,linecnt,ErrCode,target
      INTEGER :: Idel                         !Function returning the length of the string

      linecnt=0
      OPEN (81, ERR=900, FILE=rhofile,STATUS='OLD')
!       Header of file
!       Empty line
      linecnt=linecnt+1
      ErrCode=0
      READ (81, '(a)', ERR=950)
!       number of title lines, skip them
      linecnt=linecnt+1
      READ (81, '(a)', ERR=950,END=950) Buffer
      CALL OneInt(Buffer,Loc,ErrCode)
      IF (ErrCode/=0) GOTO 950
      DO cnt1=1,Loc
        linecnt=linecnt+1
        READ(81,'(a)',ERR=950)
      ENDDO
!       Voxel spacing
      linecnt=linecnt+1
      READ(81,'(a)',ERR=950) Buffer
      CALL ArrayInt(Buffer,NVoxTmp,3*SDim,ErrCode)
      DO cnt1=1,SDim
        IF (NVoxtmp(3*cnt1-2)/=NVox(cnt1)) THEN
          write(*,'(a,i2,a)') 'Error, the voxel division in the reference density along dimension',cnt1,&
                           ' is inconsistent with the input file.'
          STOP 'Error during reading of the reference density.'
        ELSEIF((NVoxtmp(3*cnt1-1)/=0).or.(NVoxtmp(3*cnt1)/=NVox(cnt1)-1)) THEN
          write(*,'(a,i2,a)') 'Warning, the start or end of the grid in the density along dimension',cnt1,&
                           ' does not lie on the borders of the unit cell.'
        ENDIF
      ENDDO
      IF (ErrCode/=0) GOTO 950
!       Cellparameters
      linecnt=linecnt+1
      READ(81,'(a)',ERR=950) Buffer
!      String ZYX defining the order of axes
      linecnt=linecnt+1
      READ(81,'(a)',ERR=950) zyxstr
      if (zyxstr/='ZYX') THEN
        write(*,'(a,i2,a)') 'Warning, string ZYX expected on line',linecnt,' of the reference density. Instead found '//zyxstr
      ENDIF
!      body of the file
      loc=-1
      IF (SDim>1) THEN
        ntable=NVox(1)*NVox(2)
      ELSE
        ntable=NVox(1)
      ENDIF
      nmap=NVoxtot/ntable
      DO cnt1=0,nmap-1
        linecnt=linecnt+1
        READ (81, '(a)', ERR=950) Buffer
        CALL OneReal(Buffer,dummy,ErrCode)
        Cnt3=nint(dummy)
        IF (Cnt3/=cnt1) THEN
          WRITE(*,'(a,i5,a)') 'Error, inconsistency in the reference density on line',linecnt,'.'
          WRITE(*,'(a,i4,a,i4,a)') 'Label of the layer nr.',cnt1,' is',cnt3,'.'
          STOP 'Error during reading of the reference density.'
        ENDIF
        IF (ErrCode/=0) GOTO 950
        DO cnt2=0,ntable-1
          loc=loc+1
          linecnt=linecnt+1
          READ (81, '(a)', ERR=950) Buffer
          CALL OneReal(Buffer,rho(loc),ErrCode)
        END DO
      END DO
      CLOSE (81, ERR=920)
      GOTO 999

!       Error handling      
900   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadXPlor:'
      WRITE (*,'(A)') 'Error! Cannot open density file for reading!'
      WRITE (*,'(A)') 'Input file file: '//rhofile(1:idel(rhofile))
      WRITE (*,'(A)') 'Stopping.'
      WRITE (*,'(A)') ''
      STOP 'Error during reading of the reference density.'
      

920   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadXPlor:'
      WRITE (*,'(A)') 'Error! Cannot close density file.'
      WRITE (*,'(A)') 'Input file: '//rhofile(1:idel(rhofile))
      WRITE (*,'(A)') ''
      STOP 'Error during reading of the reference density.'

950   WRITE (*,'(A)') ''
      WRITE (*,'(A)') 'Routine ReadXPlor:'
      WRITE (*,'(a,i5)') 'Error reading reference density on line',linecnt
      WRITE (*,'(a)') 'Lately reading: '//Buffer(:idel(Buffer))
      STOP 'Error during reading of the reference density.'

999   CONTINUE
      END SUBROUTINE ReadXPlorOld
!     Last change:  LP   16 Mar 2005   10:12 am
!****
!****   SUBROUTINES used in the program EDMA and identical with BayMEM
!****
!****
!****   Made on 29.10.98 by Martin Schneider
!****   Modified on 19.02.99 by Martin Schneider
!****   and then many times by Lukas Palatinus

!        Determine length of first non space string in DString and data
!******************************************************************
      SUBROUTINE DataLength (ioDString,oDData)
!******************************************************************
      IMPLICIT NONE
      CHARACTER ioDString*(*)
      CHARACTER oDData*132
      INTEGER DLength, OverAll, i
      CHARACTER*1 Space, Tab
      PARAMETER (Space=' ')
      OverAll=LEN(ioDString)
      Tab=CHAR(9)
      DO i=1,OverAll
        IF ((ioDString(i:i).EQ.Space).OR.(ioDString(i:i).EQ.Tab)) THEN
          DLength=i-1
          oDData=ioDString(1:DLength)
          ioDString=ioDString(DLength+1:OverAll)
          GOTO 2010
        END IF
      END DO
2010  END SUBROUTINE DataLength

!       Finding the fist position of occurance of 'char'
!******************************************************************
      SUBROUTINE NextPos (iString, iChar, oPos)
!******************************************************************
      IMPLICIT NONE
      CHARACTER*(*) :: iString
      CHARACTER*1 :: iChar
      INTEGER StringLength, i, oPos
      StringLength=LEN(iString)
      DO i=1,StringLength
        IF (iString(i:i)==iChar) THEN
          oPos=i
          GOTO 1010
        END IF
      END DO 
      oPos=-1
1010  END SUBROUTINE NextPos

!       Cutting the first spaces or tabs if any off the string
!******************************************************************
      SUBROUTINE SpaceErase (ioString)
!******************************************************************
      IMPLICIT NONE
      CHARACTER*(*) ioString
      INTEGER StringLength, i
      CHARACTER*1 Space, Tab
      PARAMETER (Space=' ')
      Tab=CHAR(9)
      StringLength=LEN(ioString)
      DO i=1,StringLength
        IF ((ioString(i:i).NE.Space).AND.(ioString(i:i).NE.Tab)) THEN
          ioString=ioString(i:StringLength)
          GOTO 1010
        END IF
      END DO 
1010  END SUBROUTINE SpaceErase

! Function counting the true length of the string
!*******************************
      FUNCTION Idel(l)
!*******************************
      character*(*) l
      integer idel
      
      Idel=len(l)
1000  IF(Idel/=0) THEN
        if(l(Idel:Idel)==' ') THEN
          Idel=Idel-1
          GO TO 1000
        END IF
      END IF
      END FUNCTION Idel

!       Applying a symmetrie operation to a vector, integer version
!******************************************************************
      SUBROUTINE SymOp (iSymN, ixyz, oSxyz)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      
      INTEGER, INTENT (IN) :: iSymN              ! Number of symmetry operation
      INTEGER, INTENT (IN) :: ixyz(MaxDim)       ! Vector
      INTEGER, INTENT (OUT) :: oSxyz(MaxDim)     ! Symmetrie related vector to xyz
      
      INTEGER :: i, ii                          ! Counters      

      oSxyz=0
      DO i=1, SDim
        DO ii=1, SDim
          oSxyz(i)=oSxyz(i)+ixyz(ii)*RMatI(i,ii,iSymN)
        END DO
        oSxyz(i)=oSxyz(i)+TVecI(i,iSymN)
      END DO
      END SUBROUTINE SymOp

!       Applying a symmetrie operation to a vector, integer version
!******************************************************************
      SUBROUTINE SymOpR (iSymN, ixyz, oSxyz)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      
      INTEGER, INTENT (IN) :: iSymN              ! Number of symmetry operation
      REAL, INTENT (IN) :: ixyz(MaxDim)       ! Vector
      REAL, INTENT (OUT) :: oSxyz(MaxDim)     ! Symmetrie related vector to xyz
      
      INTEGER :: i, ii                          ! Counters      

      oSxyz=0
      DO i=1, SDim
        DO ii=1, SDim
          oSxyz(i)=oSxyz(i)+ixyz(ii)*RMat(i,ii,iSymN)
        END DO
        oSxyz(i)=oSxyz(i)+TVec(i,iSymN)
      END DO
      END SUBROUTINE SymOpR

!******************************************************************
      SUBROUTINE Arraylocation (iDhkl, oArrLoc, iDelta)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      
      INTEGER, INTENT (IN) :: iDhkl (MaxDim)   ! hkl to compute from
      INTEGER, INTENT (IN) :: iDelta           ! Difference between elements, 2 in case of complex array, 1 else  
      INTEGER :: htmp                   ! Temporary h,k,l... buffer
      INTEGER, INTENT (OUT) :: oArrLoc   ! Computed array location of structure factor
      LOGICAL, ALLOCATABLE :: Warning(:)! Warning about too small divisions
      LOGICAL :: Warn                   ! dto.
      INTEGER :: i, ii                  !  Internal loop counters
      
      ALLOCATE (Warning(SDim))
      DO i=1, SDim
        Warning(i)=.FALSE.
      END DO
      Warn=.FALSE.
      oArrLoc=0

      DO ii=1, SDim
!       Mapping into positive
        htmp=iDhkl(ii)
        DO WHILE (htmp>=NVox(ii))
          htmp=htmp-NVox(ii)
        END DO
        DO WHILE (htmp<0)
          htmp=htmp+NVox(ii)
        END DO
        oArrLoc=oArrLoc+htmp*N(ii-1)*iDelta
      END DO
      END SUBROUTINE Arraylocation

!       Calculate Coordinates from array location
!******************************************************************
      SUBROUTINE Coordinates (iArrLoc, iDelta, oDxyz)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      
      INTEGER, INTENT (Out) :: oDxyz (MaxDim)    ! xyz to compute
      INTEGER, INTENT (IN) :: iDelta             ! Difference between elements, 2 in case of complex array, 1 else  
      INTEGER, INTENT (IN) :: iArrLoc    ! Array location of location
      INTEGER :: LocRest                ! Rest of ArrLoc for calculation of next coordinate
      INTEGER :: i                      !  Internal loop counters

      oDxyz=0
      LocRest=iArrLoc/iDelta
!       Reversed step direction
      DO i=SDim, 1, -1
        oDxyz(i)=INT(LocRest/N(i-1))
        LocRest=LocRest-oDxyz(i)*N(i-1)
      END DO
      END SUBROUTINE Coordinates


      
!       Conversion of integers to double precision
!       without artefacts from binary conversion
!*******************************************************************************
      FUNCTION INT2DBL (Int)
!*******************************************************************************

      DOUBLE PRECISION :: INT2DBL
      DOUBLE PRECISION :: DBuffer
      INTEGER, INTENT (IN) :: Int
      CHARACTER :: Buffer*20
      
      WRITE (Buffer(1:16), '(I15)') Int
      Buffer(17:20)='.0D0'
      READ (Buffer, '(ES20.6)') DBuffer
      INT2DBL=DBuffer
      END FUNCTION INT2DBL


!       Remapping a vector into the unit cell
!******************************************************************
      SUBROUTINE Remap (xyz)
!******************************************************************
      
      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      
      INTEGER :: xyz (MaxDim)
      INTEGER :: i
      INTEGER :: sgn
      
      DO i=1, SDim
        sgn=SIGN(1, xyz(i))
        SELECT CASE (sgn)
          CASE (-1)
            xyz(i)=xyz(i)+(1+INT(xyz(i)/NVox(i)))*NVox(i)
          CASE DEFAULT
            xyz(i)=MOD(xyz(i),NVox(i))
        END SELECT
      END DO

      END SUBROUTINE Remap

! Function that determines the value of the Recl factor / differs between DEC and other systems
!******************************************************
       FUNCTION GetRecLFac()
!******************************************************
       IMPLICIT NONE
       character*13 :: filename
       integer GetRecLFac
       integer :: size,status, nameln,i,orecl
       orecl=4
       filename='getreclen.tmp'
       nameln=13
       open (99,access='DIRECT', file=filename, form='UNFORMATTED',STATUS='UNKNOWN',recl=orecl)
       i=132
       write(99,rec=1) i
       close(99)
       call fstatc(filename, nameln, size, status)
       GetRecLFac=orecl**2/size
       open(99,file='getreclen.tmp')
       close(99,status='delete')
       end FUNCTION GetRecLFac
!     Last change:  LP   16 Mar 2005    9:02 am
      ! Library of routines calling the c-routines from BayMEM_wlib.lib - Windows DLL
      subroutine fplivc
      call plivc
!       call flush(30)
      end subroutine fplivc

      subroutine fstatc(filename, nameln, size, status)
      character*13 :: filename
      integer :: size,status, nameln
      call statc(filename, nameln, size, status)
      end subroutine
!****
!****   SUBROUTINES used in the program EDMA and differing from BayMEM
!****
!****   Made on 27.1.03 by Lukas Palatinus
!        Compute location of structure factors in the REAL-dimension unit-cell

!     ! Calculate a Fourier transform of the density
!     !******************************************************************
!           SUBROUTINE Laplacian
!     !******************************************************************
!
!           USE GlobalDefinitions
!           USE GlobalVariables
!
!           IMPLICIT NONE
!           integer :: i,FDir, Err
!           integer, dimension(1:RDim) :: hkl
!           double precision :: laplacefact
!           character*132 :: tmprho
!
!           ! Allocate SF
!           IF (ALLOCATED(SF)) DEALLOCATE(SF)
!           write(*,*) 'NVoxTotRim:',NVoxTotRim
!           ALLOCATE(SF(0:2*NVoxTotrim-1))
!           write(*,*) 'Initializing FFT'
!           CALL FFTInit
!           write(*,*) capvx
!           ! Copy rho to SF
!           SF=0.0D0
!           DO i=0,NVoxTotrim-1
!             SF(2*i)=rhor(i)*capvx
!           ENDDO
!           ! Call FFT
!           FDir=1
!           CALL FFT(FDir)
!           write(*,*) NVoxRim
!           !Modify structure factors
!           do i=0,NVoxTotRim-1
!             call CoordinatesRhkl(i,1,hkl)
!             laplacefact=4*pi*pi*sum((hkl/Cell(1:3))**2)
!             SF(2*i)=SF(2*i)*laplacefact
!             SF(2*i+1)=SF(2*i+1)*laplacefact
!     !        IF (any(hkl==NVox/2)) THEN
!     !          SF(2*i)=0.0
!     !          SF(2*i+1)=0.0
!     !        ENDIF
!      !       write(*,*) hkl,SF(2*i),SF(i+1)
!           enddo
!           ! CALL inverse FFT
!           FDir=-1
!           CALL FFT(FDir)
!           DO i=0,NVoxTotRim-1
!             rhor(i)=SF(2*i)/capvx
!           ENDDO
!           !Save the new density
!           tmprho='test.m81'
!           CALL WriteJanaR(trim(tmprho),.false.,Err)
!           DEALLOCATE(SF)
!           write(*,*) 'Finished'
!           STOP
!           END SUBROUTINE Laplacian

! Check if the symmetry of the density corresponds to that given in the input file
!******************************************************************
      SUBROUTINE CheckSymmetry
!******************************************************************
      
      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      LOGICAL :: succ
      INTEGER :: i,ii,iii
      INTEGER :: Dxyz(MaxDim),Dxyz2(MaxDim)
      
        DO i=1,126 ! 127 is a prime number picked on my own choice
          cnt1=int(float(NVoxTot)*float(i)/127.)
       CALL Coordinates(cnt1,1,Dxyz)
          succ=.true.
       DO ii=2,NSym
            CALL SymOp(ii,Dxyz,Dxyz2)
         CALL ArrayLocation(Dxyz2,cnt2,1)
         if ((rho(cnt1)>0.05).and.(abs(rho(cnt1)-rho(cnt2))/abs(rho(cnt1)))>0.01) THEN
!           write(*,*) cnt1,cnt2,rho(cnt1),rho(cnt2)
           succ=.false.
         endif
       ENDDO
       if (.not.succ) EXIT
     ENDDO
     IF (.not.succ) THEN
       write(*,'(a)') 'Warning!!!! The density does not match the symmetry given in the input file!'
       write(*,'(a)') 'The use of the "fullcell no" option might fail!'
     ENDIF
      END SUBROUTINE CheckSymmetry


!******************************************************************
      SUBROUTINE ArraylocationR (iDhkl, oArrLoc, iDelta)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      
      INTEGER :: iDhkl (MaxDim)   ! hkl to compute from
      INTEGER, INTENT (IN) :: iDelta           ! Difference between elements, 2 in case of complex array, 1 else  
      INTEGER :: htmp                   ! Temporary h,k,l... buffer
      INTEGER, INTENT (OUT) :: oArrLoc  ! Computed array location of structure factor
      INTEGER :: ii                     !  Internal loop counters
      
      oArrLoc=0
    
      DO ii=1, RDim
!       Check, if inside the unit cell wih rims - we cannot use the periodicity!
        htmp=iDhkl(ii)+Rim(ii)
        IF (Aper(ii)==1) THEN
          IF ((htmp<0).or.(htmp>NVoxRim(ii)-1)) THEN
            write(*,*) 'NVoxRim:',NVoxRim
         write(*,*) 'Error ArrLoc:',ii, rim,iDhkl
         STOP 'Error in ArrayLocationR - pixel outside the unit cell in an aperiodic direction!'
          END IF
     ELSE
!         Mapping into the unit cell
          htmp=iDhkl(ii)
          DO WHILE (htmp>=NVoxRim(ii))
            htmp=htmp-NVoxRim(ii)
          END DO
          DO WHILE (htmp<0)
            htmp=htmp+NVoxRim(ii)
          END DO
        END IF
     oArrLoc=oArrLoc+htmp*NRim(ii-1)*iDelta
      END DO
      END SUBROUTINE ArraylocationR

!       Calculate Coordinates from array location
!******************************************************************
      SUBROUTINE CoordinatesR (iArrLoc, iDelta, oDxyz)
!******************************************************************

      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE
      
      INTEGER, INTENT (Out) :: oDxyz (MaxDim)    ! xyz to compute
      INTEGER, INTENT (IN) :: iDelta             ! Difference between elements, 2 in case of complex array, 1 else  
      INTEGER, INTENT (IN) :: iArrLoc    ! Array location of location
      INTEGER :: LocRest                 ! Rest of ArrLoc for calculation of next coordinate
      INTEGER :: i                       !  Internal loop counters

      oDxyz=0
      LocRest=iArrLoc/iDelta
!       Reversed step direction
      DO i=RDim, 1, -1
        oDxyz(i)=INT(LocRest/NRim(i-1))
        LocRest=LocRest-oDxyz(i)*NRim(i-1)
        oDxyz(i)=oDxyz(i)-Rim(i)
      END DO
      END SUBROUTINE CoordinatesR

!Subroutine returning the coordinates in the grid for given arraylocation in SF array
!**********************************************
      SUBROUTINE CoordinatesRhkl(iArrLoc, iDelta, ohkl)
!**********************************************
      USE GlobalDefinitions
      USE GlobalVariables

      IMPLICIT NONE

      INTEGER, INTENT (Out) :: ohkl (MaxDim)    ! hkl to compute
      INTEGER, INTENT (IN) :: iDelta             ! Difference between elements, 2 in case of complex array, 1 else
      INTEGER, INTENT (IN) :: iArrLoc    ! Array location of location
      INTEGER :: LocRest                 ! Rest of ArrLoc for calculation of next coordinate
      INTEGER :: i                       !  Internal loop counters

      ohkl=0
      LocRest=iArrLoc/iDelta
!       Reversed step direction
      DO i=RDim, 1, -1
        ohkl(i)=INT(LocRest/NRim(i-1))
        LocRest=LocRest-ohkl(i)*NRim(i-1)
        IF (ohkl(i)>NVoxRim(i)/2) ohkl(i)=ohkl(i)-NVoxRim(i)
      END DO
      END SUBROUTINE CoordinatesRhkl

! Shift coordinates in the hypercube between StartC and EndC
!*****************************************************************************
      FUNCTION NextCoo(Dim,StartC,EndC,Coo,AxOrd)
!*****************************************************************************
      LOGICAL :: NextCoo
      INTEGER :: Dim,i
      INTEGER, DIMENSION(*) :: StartC,EndC,Coo,AxOrd
        NextCoo=.true.
        i=1
100     Coo(AxOrd(i))=Coo(AxOrd(i))+1
        IF (Coo(AxOrd(i))>EndC(AxOrd(i))) THEN
          Coo(AxOrd(i))=StartC(AxOrd(i))
          i=i+1
          IF (i==Dim+1) THEN
         NextCoo=.false.
         goto 200
       END IF
       goto 100
        END IF      
200     RETURN
      END FUNCTION NextCoo

!*****************************************************************************
      subroutine MatMatMul(M1,M2,Mout,n)
!*****************************************************************************
      IMPLICIT NONE
      INTEGER ::i,ii,iii,n
      DOUBLE PRECISION,DIMENSION(n,n) :: M1,M2,Mout  !matrices
      DO i=1,n
        DO ii=1,n
       Mout(i,ii)=0.0D0
       DO iii=1,n
         Mout(i,ii)=Mout(i,ii)+M1(i,iii)*M2(iii,ii)
       END DO
     END DO
      END DO
      END SUBROUTINE MatMatMul

!*****************************************************************************
      subroutine TrMat(M1,Mout,n)
!*****************************************************************************
      IMPLICIT NONE
      INTEGER ::i,ii,n
      DOUBLE PRECISION,DIMENSION(n,n) :: M1,Mout  !matrices
      DO i=1,n
        DO ii=1,n
          Mout(i,ii)=M1(ii,i)
     END DO
      END DO
      END SUBROUTINE TrMat

!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!c http://www.nauticom.net/www/jdtaft/FortranMatrix.htm
!c    routine to invert a matrix by Gaussian elimination
!c     Ainv=inverse(A)    A(n,n)   Ainv(n,n)
!c     Note: sizes maxed at 64 x 64 - re dimension for larger matrices;
!c           D matrix is extended.
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
       subroutine invert( A, Ainv, n)
       integer :: n,i,j
       double precision, DIMENSION(1:n,1:n) :: A,Ainv
       double precision :: D(1:n,1:2*n), ALPHA, BETA

!  initialize the reduction matrix
       n2 = 2*n
       do i = 1,n
         do j = 1,n
           D(i,j) = A(i,j)
           d(i,n+j) = 0.
         end do
         D(i,n+i) = 1.
       end do

!  do the reduction 
       do i = 1,n
         alpha = D(i,i)
         if(alpha==0.) go to 300
         do j = 1,n2
           D(i,j) = D(i,j)/alpha
         end do
         do k = 1,n
           if((k-i)==0) cycle
           beta = D(k,i)
           do j = 1,n2
             D(k,j) = D(k,j) - beta*D(i,j)
           end do
         end do
       end do

!  copy result into output matrix
       do i = 1,n
         do j = 1,n
           Ainv(i,j) = D(i,j+n)
         end do
       end do

       return
 300   Ainv(1,1)=-1.0D30
       return
       end


! QL diagonalization of matrix / taken from Jana, in Jana probably from Numerical Recipes
!*****************************************************************************
      subroutine ql(a,z,d,n)
!*****************************************************************************
      INTEGER :: n
      DOUBLE PRECISION :: e(20),d(n),a(*),z(*)
      DOUBLE PRECISION, PARAMETER :: tol=.0000001, eps=.0000001
      imd=1-2*mod(n,2)
      nb=n*(n-1)+1
      k=0
      do 20 i=1,n
        do 10 j=1,n
          k=k+1
          if(i.eq.j) then
            z(k)=1.
          else
            z(k)=0.
          endif
10      continue
20    continue
      iz=n*(n+1)/2
      l=n
      do 900 ip=3,n
        i=l
        l=l-1
        h=0.
        iz=iz-i
        do 100 k=1,l
          ind=iz+k
          f=a(ind)
          d(k)=f
          h=h+f**2
100     continue
        if(h.le.tol) then
          e(i)=0.
          h=0.
          go to 890
        endif
        g=-sqrt(h)
        if(f.lt.0.) g=-g
        e(i)=g
        h=h-f*g
        rh=1./h
        dl=f-g
        d(l)=dl
        ind=iz+l
        a(ind)=dl
        f=0.
        ind=1
        do 300 j=1,l
          g=0.
          jk=ind
          do 200 k=1,l
            g=g+a(jk)*d(k)
            id=1
            if(j.le.k) id=k
            jk=jk+id
200       continue
          g=g*rh
          e(j)=g
          f=f+g*d(j)
          ind=ind+j
300     continue
        hh=f*.5*rh
        jk=0
        do 500 j=1,l
          f=d(j)
          g=e(j)-hh*f
          e(j)=g
          do 400 k=1,j
            jk=jk+1
            a(jk)=a(jk)-f*e(k)-g*d(k)
400       continue
500     continue
890     jk=iz+i
        d(i)=a(jk)
        a(jk)=h
900   continue
      d(1)=a(1)
      e(2)=a(2)
      d(2)=a(3)
      b=0.
      f=0.
      do 2000 l=1,n
        ln=l*n-n
        if(l.eq.n) go to 1800
        l1=l+1
        h=eps*(abs(d(l))+abs(e(l1)))
        if(b.lt.h) b=h
        do 1100 m=l1,n
          if(abs(e(m)).le.b) go to 1200
1100    continue
        m=n+1
1200    m=m-1
        if(m.eq.l) go to 1800
        ml=m-l
        mn=m*n
1250    g=d(l)
        el=e(l1)
        dl=(d(l1)-g)/(2.*el)
        h=abs(dl)+sqrt(dl**2+1.)
        if(dl.lt.0.) h=-h
        dl=el/h
        d(l)=dl
        h=g-dl
        do 1300 i=l1,n
          d(i)=d(i)-h
1300    continue
        f=f+h
        dl=d(m)
        c=1.
        s=0.
        ind=mn
        do 1600 jk=1,ml
          ind=ind-n
          i=m-jk
          i1=i+1
          el=e(i1)
          g=c*el
          h=c*dl
          if(abs(dl).ge.abs(el)) then
            c=el/dl
            r=sqrt(c**2+1.)
            e(i+2)=s*dl*r
            s=c/r
            c=1./r
          else
            c=dl/el
            r=sqrt(c**2+1.)
            e(i+2)=s*el*r
            s=1./r
            c=c*s
          endif
          dl=c*d(i)-s*g
          d(i1 )=h+s*(c*g+s*d(i))
          do 1550 k=1,n
            iz=k+ind
            izn=iz-n
            h=z(izn)
            g=z(iz)
            z(izn)=c*h-s*g
            z(iz) =s*h+c*g
1550      continue
1600    continue
        el=s*dl
        e(l1)=el
        d(l)=c*dl
        if(abs(el).gt.b) go to 1250
1800    dl=d(l)+f
        jk=l-1
        iz=ln
        do 1850 k=1,jk
          i=l-k
          iz=iz-n
          if(dl.ge.d(i)) go to 1860
          imd=-imd
          do 1820 j=1,n
            ind=iz+j
            inz=ind+n
            p=z(ind)
            z(ind)=z(inz)
            z(inz)=p
1820      continue
1850    d(i+1)=d(i)
        i=0
1860    d(i+1)=dl
2000  continue
      if(imd.lt.0) then
        do 2020 i=1,n
          z(i)=-z(i)
2020    continue
      endif
      iz=1
      do 2500 i=3,n
        l=i-1
        iz=iz+l
        ind=iz+i
        h=a(ind)
        if(h.ne.0.) then
          do 2400 j=1,nb,n
            jk=j-1
            s=0.
            do 2200 k=1,l
              m=iz+k
              ind=jk+k
              s=s+a(m)*z(ind)
2200        continue
            s=s/h
            do 2300 k=1,l
              ind=jk+k
              m=iz+k
              z(ind)=z(ind)-s*a(m)
2300        continue
2400      continue
        endif
2500  continue
      return
      end

!*********************************************
      SUBROUTINE WriteStructure(filename)
!*********************************************
      USE GLOBALDEFINITIONS
      USE GLOBALVARIABLES

      IMPLICIT NONE
      INTEGER, PARAMETER :: clswdth=1,scalesteps=100, korsteps=20
      INTEGER :: i,ii,iii,kori,idel,dMaxNr,CurrentAtoms,sum,attp
      CHARACTER*132 :: filename
      CHARACTER*8 :: AtomName
      INTEGER, ALLOCATABLE, DIMENSION(:,:) :: AtomCount, StepAtNumber
      Type(MaxType) :: DummyMax
      DOUBLE PRECISION :: smin,smax,kmin, kmax, curkor, sc,minsc,kor,minkor,ressig,dummy,minressig
      LOGICAL :: fromcompos=.false.
      
      Allocate(AtomCount(max(ElNr,1),2))
      AtomCount=0
      CurrentAtoms=0
!     Sort the maxima from the highest to the smalest
      DO i=MaxNr,2,-1
        DO ii=1,i-1
          IF (Maxima(ii)%cc<Maxima(ii+1)%cc) THEN
            DummyMax=Maxima(ii)
            Maxima(ii)=Maxima(ii+1)
            Maxima(ii+1)=DummyMax
          ENDIF
        ENDDO
      ENDDO

! First count the number of independent atoms:
      dMaxNr=0
      IF (TotalAtoms<0) THEN !Flag for composition-driven atom counts
        TotalAtoms=0
        fromcompos=.true.
        DO i=1,ElNr
          IF (Compos(i,1)/=1) TotalAtoms=TotalAtoms-Compos(i,2) !Ignore hydrogens
        ENDDO
        ALLOCATE(StepAtNumber(1:ElNr,1:2))
        DO i=1,ElNr
          StepatNumber(i,2)=i
        ENDDO
        !Sort the elements by element number
        DO i=Elnr,2,-1
          DO ii=1,i-1
            IF (compos(StepatNumber(ii,2),1)<compos(StepatNumber(ii+1,2),1)) THEN
              iii=StepatNumber(ii,2)
              StepatNumber(ii,2)=StepatNumber(ii+1,2)
              StepatNumber(ii+1,2)=iii
            ENDIF
          ENDDO
        ENDDO
        ! Sum up the atomcounts for the individual types
        sum=0
        DO i=1,ElNr
          StepAtNumber(i,1)=sum
          sum=sum+compos(StepAtNumber(i,2),2)
        ENDDO
      ENDIF
      IF (TotalAtoms==0) THEN
     !Just count the number of atoms above the threshold
        DO i=1,MaxNr
          IF (Maxima(i)%cc>chlimlist) dMaxNr=dMaxNr+1
        END DO
      ELSE
        DO i=1,MaxNr
           dMaxNr=dMaxNr+1
           CurrentAtoms=currentatoms+Maxima(i)%mult
           if (currentatoms>=abs(TotalAtoms)) EXIT
        END DO
        currentatoms=0
      ENDIF

      IF ((.not.fromcompos).and.(ElNr>0)) THEN
        smin=0.8*Maxima(1)%cc/float(maxZ)
        smax=1.2*Maxima(1)%cc/float(minZ)
        kmin=0.
        kmax=1./Maxima(1)%cc
!         kmax=0.
!        smin=0.99*Maxima(1)%cc/float(maxZ)
!        smax=1.01*Maxima(1)%cc/float(maxZ)
        minressig=1000000.
        DO i=0,scalesteps
          DO kori=0,korsteps
            sc=smin+float(i)*(smax-smin)/scalesteps
            kor=kmin+float(kori)*(kmax-kmin)/korsteps
            ressig=0
            DO ii=1,dMaxNr
             IF (Maxima(ii)%cc<chlimlist) CYCLE
             IF (TotalAtoms==0) THEN !If we guess the number of atoms, throw away al that have too low charge
               dummy=Maxima(ii)%cc**2
             ELSE
               dummy=1000000.
             ENDIF
             DO iii=1,ElNr
               IF (compos(iii,1)==1) CYCLE
               curkor=kor*(Maxima(1)%cc-Maxima(ii)%cc)+1.0
!                   write(*,*) ii,iii,curkor
               dummy=min(dummy,(Maxima(ii)%cc-sc*compos(iii,1)/curkor)**2)
!                 write(*,*) ii, Maxima(ii)%cc, compos(iii,1), curkor, ((Maxima(ii)%cc-sc*compos(iii,1)/curkor)/Maxima(ii)%cc), dummy
!                   write(*,*) ii, Maxima(ii)%cc, compos(iii,1), curkor, (Maxima(ii)%cc-sc*compos(iii,1)/curkor), dummy
             ENDDO
             ressig=ressig+dummy
           ENDDO
            IF (ressig<minressig) THEN
              minressig=ressig
              minsc=sc
              minkor=kor
            ENDIF
!           write(*,*) i, kori, sc, kor, ressig, minsc, minkor, minressig
!           write(*,*)
          ENDDO
        ENDDO
!        write(*,*) minsc, minkor
        DO ii=1,dMaxNr
          Maxima(ii)%an=0
          IF (Maxima(ii)%cc<chlimlist) CYCLE
          IF (TotalAtoms==0) THEN !If we are guessing the number of atoms, throw away all that have too low charge
            dummy=Maxima(ii)%cc**2
          ELSE
            dummy=1000000.
          ENDIF
          DO iii=1,ElNr
            IF (compos(iii,1)==1) CYCLE
            curkor=minkor*(Maxima(1)%cc-Maxima(ii)%cc)+1.
            IF (((Maxima(ii)%cc-minsc*compos(iii,1)/curkor)**2)<dummy) THEN
              dummy=(Maxima(ii)%cc-minsc*compos(iii,1)/curkor)**2
              Maxima(ii)%an=iii
            ENDIF
          ENDDO
          curkor=minkor*(Maxima(1)%cc-Maxima(ii)%cc)+1.
!         write(75,*) ii, Maxima(ii)%an, Maxima(ii)%cc, compos(max(Maxima(ii)%an,1),1), minsc, minkor, &
!                      minsc*compos(max(Maxima(ii)%an,1),1)/curkor
        ENDDO
        IF (TotalAtoms==0) THEN !Exclude from the list the atoms with too low charge
          dMaxNr=0
          DO i=1,MaxNr
            IF ((Maxima(i)%cc>chlimlist).and.(Maxima(i)%an>0)) dMaxNr=dMaxNr+1
          END DO
        ENDIF
      ENDIF

!count once more the real number of independent atoms
      dMaxNr=0
      currentatoms=0
      IF (TotalAtoms>=0) THEN
        DO i=1,MaxNr
         IF (Maxima(i)%an==0) CYCLE
          IF (((TotalAtoms==0).and.(Maxima(i)%cc>chlimlist)).or.((CurrentAtoms<TotalAtoms))) THEN
            dMaxNr=dMaxnr+1
            currentatoms=currentatoms+Maxima(i)%mult
            IF (Currentatoms>=abs(TotalAtoms)) EXIT
          END IF
        END DO
      ELSE
        DO i=1,MaxNr
          DO ii=ElNr,1,-1
            IF (currentatoms>=StepATnumber(ii,1)) THEN
              Attp=StepATnumber(ii,2)
              EXIT
            ENDIF
          ENDDO
          IF (compos(attp,1)==1) EXIT !Ignore hydrogens, and because they are the latest, we can safely exit the whole loop
          dMaxNr=dMaxNr+1
          currentatoms=currentatoms+Maxima(i)%mult
          IF (Currentatoms>=abs(TotalAtoms)) EXIT
        ENDDO
      ENDIF


      open(40,file=filename(1:idel(filename))) 
      if (OutStrType=='ins') open (41,STATUS='SCRATCH')
!Write the header
      call WriteHeader
! Set the default Biso in case none (or zero) was set
      IF (Biso<eps) Biso=DefaultBiso
      currentatoms=0
      IF (TotalAtoms>=0) THEN
        DO i=1,MaxNr
         IF (Maxima(i)%an==0) CYCLE
          IF (((TotalAtoms==0).and.(Maxima(i)%cc>chlimlist)).or.((CurrentAtoms<TotalAtoms))) THEN
            IF (ElNr/=0) THEN
              iii=idel(ElSmb(compos(Maxima(i)%an,1)))
              AtomName(1:iii)=ElSmb(compos(Maxima(i)%an,1))(1:iii)
              AtomCount(Maxima(i)%an,1)=AtomCount(Maxima(i)%an,1)+1
              AtomCount(Maxima(i)%an,2)=AtomCount(Maxima(i)%an,2)+Maxima(i)%mult
              write(AtomName(iii+1:),'(i4)') AtomCount(Maxima(i)%an,1)
              CALL RemoveSpaces(AtomName)
            ELSE
              AtomName(1:iii)='Mx'
              AtomCount(1,1)=AtomCount(1,1)+1
              AtomCount(1,2)=AtomCount(1,2)+Maxima(i)%mult
              write(AtomName(3:),'(i3.3)') AtomCount(1,1)
            END IF
!           write the atom
            CALL WriteAtom(i)
            currentatoms=currentatoms+Maxima(i)%mult
          END IF
        END DO
      ELSE
        DO i=1,MaxNr
          DO ii=ElNr,1,-1
            IF (currentatoms>=StepATnumber(ii,1)) THEN
              Attp=StepATnumber(ii,2)
              EXIT
            ENDIF
          ENDDO
          IF (compos(attp,1)==1) EXIT !Ignore hydrogens, and because they are the latest, we can safely exit the whole loop
          iii=idel(ElSmb(compos(attp,1)))
          AtomName(1:iii)=ElSmb(compos(attp,1))(1:iii)
          AtomCount(attp,1)=AtomCount(attp,1)+1
          Maxima(i)%an=attp
          AtomCount(attp,2)=AtomCount(attp,2)+Maxima(i)%mult
          write(AtomName(iii+1:),'(i3)') AtomCount(attp,1)
          CALL RemoveSpaces(AtomName)
          CALL WriteAtom(i)
          currentatoms=currentatoms+Maxima(i)%mult
          IF (Currentatoms>=abs(TotalAtoms)) EXIT
        ENDDO
      ENDIF
      CALL WriteFooter
      CLOSE(40)
      IF (OutStrType=='ins') close(41)
      Deallocate(AtomCount)

      CONTAINS
        SUBROUTINE WriteHeader
          CHARACTER*(LineLength) :: ModSymStr !Symmetry string modified for CIF or INS purposes
          LOGICAL :: succ,equalop
       REAL :: sum
       DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:,:) :: DummyMat
       DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:) :: DummyVec, DummyVec2
          INTEGER :: LattNr,isum,iabssum,InvOp

       SELECT CASE (OutStrType)
          CASE ('m40')
            WRITE(40,101) dMaxNr
            WRITE(40,102) 1.,(0.,i=1,5)
            WRITE(40,103) 0.
            WRITE(40,102) (0.,i=1,6)
            WRITE(40,102) (0.,i=1,6)
          CASE ('cif')
         write(40,'(a)') 'data_structure_from_EDMA'
            write(40,'(a)') '_symmetry_space_group_name_H-M '//spacegrp(:idel(spacegrp))
         write(40,'(a)') 'loop_'
         write(40,'(a)') ' _symmetry_equiv_pos_site_id'
         write(40,'(a)') ' _symmetry_equiv_pos_as_xyz'
         !symmetry operations come here - generate them from the RMat and TVec
         DO i=1,NSym
           CALL RMat2Str(i,ModSymStr)
              write(40,'(i3,2X,a)') i,ModSymStr(:idel(ModSymStr))
         ENDDO

         write(40,'(a,f10.5)') '_cell_length_a     ',Cell(1)
         write(40,'(a,f10.5)') '_cell_length_b     ',Cell(2)
         write(40,'(a,f10.5)') '_cell_length_c     ',Cell(3)
         write(40,'(a,f10.5)') '_cell_angle_alpha  ',Cell(4)
         write(40,'(a,f10.5)') '_cell_angle_beta   ',Cell(5)
         write(40,'(a,f10.5)') '_cell_angle_gamma  ',Cell(6)
         write(40,'(a,f12.5)') '_cell_volume     ',Cell(7)
         write(40,'(a)') ''
         write(40,'(a)') 'loop_'
         write(40,'(a)') ' _atom_site_label'
         write(40,'(a)') ' _atom_site_type_symbol'
         write(40,'(a)') ' _atom_site_fract_x'
         write(40,'(a)') ' _atom_site_fract_y'
         write(40,'(a)') ' _atom_site_fract_z'
         write(40,'(a)') ' _atom_site_adp_type'
         write(40,'(a)') ' _atom_site_U_iso_or_equiv'
         write(40,'(a)') ' _atom_site_symmetry_multiplicity'
         write(40,'(a)') ' _atom_site_occupancy'
       CASE ('ins')
         write(40,'(a)') 'TITL '//title(:idel(title))
         write(40,'(a,f7.5,6f12.7)') 'CELL ',lambda,(Cell(ii),ii=1,6)
            !Find the proper lattice centering:
         !First make a checksum and control that only 1/2 and 1/3 or 2/3 are present in the centings
            sum=0.0;succ=.true.
         DO i=1,NCenter
           DO ii=1,3
             sum=sum+float(2**(ii-1))*Center(ii,i)
          if ((abs(center(ii,i)-0.0E0)>eps).and.(abs(center(ii,i)-0.5E0)>eps).and. &
                    (abs(center(ii,i)-0.3333333333E0)>eps).and.(abs(center(ii,i)-0.666666666E0)>eps)) succ=.false. 
           ENDDO
         ENDDO
         IF (.not.succ) sum=-100.
         IF (NCenter==1) THEN
           LattNr=-1
         ELSEIF (Ncenter==2) THEN !The only allowed (and possible) is A, B, C and I
            IF (abs(sum-3.5)<eps) THEN
                LattNr=-2 !I-Centering
           ELSEIF (abs(sum-3.0)<eps) THEN
                  LattNr=-5 !A-centering
           ELSEIF (abs(sum-2.5)<eps) THEN
                LattNr=-6 !B-centering
           ELSEIF (abs(sum-1.5)<eps) THEN
                LattNr=-7 !C-centering
           ELSE
             succ=.false.
           ENDIF
         ELSEIF (Ncenter==3) THEN !The only allowed is R obverse
           IF (abs(Center(2,2)-Center(3,2))<eps) THEN
                LattNr=-3 !R-centering obverse
           ELSE
             succ=.false.
           ENDIF
         ELSEIF (Ncenter==4) THEN ! The only allowed is F
           IF (abs(sum-7)<eps) THEN
                LattNr=-4 !F-centering
           ELSE
             succ=.false.
           ENDIF
         ENDIF
         IF(.not.succ) THEN
              write(*,'(a)') 'Sorry, SHELX handles only the standard centerings P, I, F, A, B, C or  R obverse.' 
              write(*,'(a)') 'Your centering is none of them. The list of atoms will still be written, '// &
                          'but the symmetry will not match.'
              write(*,'(a)') 'The lattice will be set to P.' 
           LattNr=-1
            ENDIF
         !Check if we have an inversion center
         DO i=1,NSymPrim
           isum=0;iabssum=0;InvOp=0
           DO ii=1,3
             DO iii=1,3
            isum=isum+RMat(ii,iii,i)
          ENDDO
                iabssum=iabssum+abs(RMat(ii,ii,i))
              ENDDO
              IF ((isum==-3.).and.(iabssum)==3) THEN !Centrosymmetric
                LattNr=-LattNr
             !SHELX accepts inversion only at the origin
             sum=0.0E0
          DO ii=1,3
            sum=sum+abs(TVec(ii,i))
          ENDDO
                IF (sum>eps) THEN
                  write(*,'(a)') 'Sorry, SHELX requires the inversion center to lie at the origin.' 
                  write(*,'(a)') 'Your inversion center does not. The list of atoms will still be written,'// &
                                 'but the symmetry will not match.' 
          ENDIF
             InvOp=i
          EXIT
           ENDIF
         ENDDO
            write(40,'(a,i2)') 'LATT ',LattNr
         ALLOCATE(DummyMat(SDim,SDim))
         ALLOCATE (DummyVec(SDim))
         ALLOCATE (DummyVec2(SDim))
            !Write out the symmetry operations. Without inversion-center related!
         DO i=2,NSymPrim
              succ=.true.
           IF (InvOp>0) THEN !Check, if this one is not related to some of the previous
                CALL ProdOp(RMat(1,1,i),TVec(1,i),RMat(1,1,InvOp),TVec(1,InvOp),DummyMat,DummyVec,SDim)
                DO ii=1,i-1
            DO iii=1,NCenter
              DummyVec2=DummyVec+Center(1:SDim,iii)
              IF (EqualOp(RMat(1,1,ii),TVec(1,ii),DummyMat,DummyVec2,SDim)) succ=.false.
               ENDDO
            IF (.not.succ) EXIT
          ENDDO
              ENDIF
           IF (succ) THEN
             call RMat2Str(i,ModSymStr)
                write(40,'(a)') 'SYMM '//ModSymStr(:idel(ModSymStr)) 
           ENDIF
         ENDDO
         DEALLOCATE(DummyMat)
         DEALLOCATE (DummyVec)
         write(41,'(a)') 'FVAR 1' !Write this to a temporary file
       END SELECT

101       FORMAT(i4,3('    0'))
102       FORMAT(6f9.6)
103       FORMAT(f9.6)
        END SUBROUTINE WriteHeader

     SUBROUTINE WriteFooter
          IMPLICIT NONE
       SELECT CASE (OutStrType)
          CASE ('m40')
          CASE ('cif')
          CASE ('ins')
         ! Composition
         write(40,'(a)',ADVANCE='NO') 'SFAC '
         DO i=1,ElNr
           write(40,'(a)',ADVANCE='NO') ElSmb(Compos(i,1))(1:idel(ElSmb(Compos(i,1))))//' '
         ENDDO
            write(40,'(a)')
         write(40,'(a)',ADVANCE='NO') 'UNIT '
         DO i=1,ElNr
           write(40,'(i4)',ADVANCE='NO') AtomCount(i,2)
         ENDDO
            write(40,'(a)')
         rewind(41)
200         READ(41,'(a)',END=300) Buffer
            WRITE(40,'(a)') Buffer(:idel(Buffer))
            GOTO 200
300         CONTINUE
         write(40,'(a)') 'END'
       END SELECT

     END SUBROUTINE WriteFooter

        SUBROUTINE RMat2Str(i,oText)
       IMPLICIT NONE
       INTEGER :: i
          CHARACTER*(LineLength) :: oText,TrVecStr
          CHARACTER :: cchar,sgnchar
       integer :: tvnom,tvden

           oTExt=''
              DO ii=1,3 !Only the real-space symops make sense here, no modulated cif can yet be produced
          ! First the translation vector
          IF (TVec(ii,i)>eps) THEN
            call DecToFraction(TVec(ii,i),tvnom,tvden)
            write(TrVecStr,'(i2,a,i2)') tvnom,'/',tvden
            oText=oText(:idel(oText))//TrVecStr
          ENDIF
          ! Then the rotational part
          DO iii=1,3
            SELECT CASE (iii)
              CASE (1)
                cchar='x'
              CASE (2)
                cchar='y'
              CASE (3)
                cchar='z'
            END SELECT
            IF (nint(Rmat(ii,iii,i))==1) THEN
              if ((idel(oText)==0).or.(oText(idel(oText):idel(oText))==',')) THEN
                sgnchar=' '
              ELSE
                sgnchar='+'
              ENDIF
              oText=oText(:idel(oText))//sgnchar//cchar
            ELSEIF (nint(Rmat(ii,iii,i))==-1) THEN
              oText=oText(:idel(oText))//'-'//cchar
                  ENDIF
             ENDDO
          if(ii<3) oText=oText(:idel(oText))//','
           ENDDO
              CALL RemoveSpaces(oText)
        END SUBROUTINE RMat2Str 

     SUBROUTINE WriteAtom(Natom)
       INTEGER :: NAtom
          SELECT CASE (OutStrType)
          CASE ('m40')
            if (.not.m40forjana) THEN
               write(40,110) AtomName(1:8),Maxima(NAtom)%an,1,dble(Maxima(NAtom)%mult)/dble(Nsym),(Maxima(NAtom)%Coo(ii),ii=1,RDim)
            write(40,111) Biso/(8.*pi*pi),(0.,ii=1,5)
            else
               write(40,110) AtomName(1:8),Maxima(NAtom)%an,1,dble(Maxima(NAtom)%mult)/dble(Nsym),(Maxima(NAtom)%Coo(ii),ii=1,RDim)
            write(40,'(2f9.4)') Maxima(NAtom)%rho,Maxima(NAtom)%cc
         endif
       CASE ('cif')
         write(40,112) AtomName(1:idel(AtomName)),ElSmb(compos(Maxima(Natom)%an,1)),&
                            (Maxima(NAtom)%Coo(ii),ii=1,RDim),'Uiso',Biso/(8.*pi*pi),Maxima(NAtom)%mult,1.
       CASE ('ins')
         write(41,113) AtomName(1:4),Maxima(Natom)%an,&
                            (Maxima(NAtom)%Coo(ii),ii=1,RDim),10.+dble(Maxima(NAtom)%mult)/dble(Nsym),Biso/(8.*pi*pi)
       END SELECT

110       FORMAT(a8,2i3,'    ',4(f9.6))
111       FORMAT(6f9.6)
112       FORMAT(a,2X,a2,2X,3f10.6,2X,a,f9.6,i3,f4.1)
113       FORMAT(a,i3,5f10.6)

     END SUBROUTINE

      END SUBROUTINE WriteStructure

!Subroutine that makes a product of two symmetry operations
!**************************************************
      SUBROUTINE ProdOp(R1,T1,R2,T2,Rout,Tout,n)
!**************************************************
      real, parameter :: eps=0.00001
      integer :: n,i
      double precision, dimension(1:n,1:n) :: R1,R2,Rout
      double precision, dimension(1:n) :: T1,T2,Tout,temp

      call dmultm(R2,R1,Rout,n,n,n)
      call dmultm(R2,t1,temp,n,n,1)
      Tout=T2+temp
      DO i=1,n
        DO
       IF (Tout(i)<-eps) THEN
         Tout(i)=Tout(i)+1.0E0
       ELSEIF (Tout(i)>(1.E0-eps)) THEN
         Tout(i)=Tout(i)-1.0E0
       ELSE
         EXIT
          ENDIF
        ENDDO
      ENDDO
      END SUBROUTINE ProdOp
      
!Function EqualOp is true if the two input operations are equivalent
!**************************************************
      LOGICAL FUNCTION EqualOp(R1,T1,R2,T2,n)
!**************************************************
      real, parameter :: eps=0.00001
      integer :: i,ii,n
      double precision, dimension(1:n,1:n) :: R1,R2
      double precision, dimension(1:n) :: T1,T2
      logical :: equal
      
      equal=.true.
      DO i=1,n
        IF (abs((abs(T1(i)-T2(i))-nint(abs(T1(i)-T2(i)))))>eps) THEN
       equal=.false.
       EXIT
        ENDIF
     DO ii=1,n
          IF (abs(R1(i,ii)-R2(i,ii))>eps) THEN
         equal=.false.
         EXIT
       ENDIF
     ENDDO
     IF (.not.equal) EXIT
      ENDDO
      EqualOp=equal
      END FUNCTION EqualOp

! A routine that finds a representaton of a number by a fraction
!*********************************************
      SUBROUTINE DecToFraction(InDec,Nom,Den)
        DOUBLE PRECISION :: InDec
     DOUBLE PRECISION, PARAMETER :: Toler=0.00001
     INTEGER :: Nom,Den,LNom,LDen,RNom,RDen

     LNom=0
     LDen=1
     RNom=1
     RDen=1
     DO
       IF (abs(float(LNom)/float(LDen)-InDec)<Toler) THEN
         Nom=LNom
         Den=LDen
         EXIT
       ELSEIF (abs(float(RNom)/float(RDen)-InDec)<Toler) THEN
         Nom=RNom
         Den=RDen
         EXIT
       ELSE
            IF (InDec<(float(LNom+RNom)/float(LDen+RDen))) THEN
           RNom=LNom+RNom
           RDen=LDen+RDen
         ELSE
           LNom=LNom+RNom
           LDen=LDen+RDen
         ENDIF
       ENDIF
     ENDDO
      END SUBROUTINE DecToFraction

!*********************************************

! Take the list of maxima and compare with the atoms in the provided m40
!*********************************************
      SUBROUTINE CompareM40
!*********************************************
      USE GLOBALDEFINITIONS
      USE GLOBALVARIABLES

      IMPLICIT NONE
      REAL, PARAMETER :: CloseDist=0.55, PillowFactor=0.8
      REAL, DIMENSION(MaxDim) :: dhkl,dcoo1,dcoo2,dcoo3
      INTEGER :: i,ii,iii,i1,i2,i3,idel,GMaxNr,CatNr,main,bestgmax,located
      INTEGER :: falseat,goodat
      REAL :: dist, multsc, mindist,sumdist
      Type(MaxType), DIMENSION(MaxMaxNr) :: GMaxima
      Type(MaxType) :: DMaximum
      Type(AtomType),ALLOCATABLE, DIMENSION(:) :: CAtoms
      logical,ALLOCATABLE, DIMENSION(:) :: OccupMax, OccupAt
      GMaxNr=0
      DO i=1,MaxNr
     IF (Maxima(i)%cc>chlimlist) THEN
       GMaxNr=GMaxNr+1
          GMaxima(GMaxNr)=Maxima(i)
     ENDIF
      ENDDO
      ! Bubblesort the maxima by rho
      DO i=GMaxNr,2,-1
        DO ii=i-1,1,-1
       IF (GMaxima(ii)%rho<GMaxima(ii+1)%rho) THEN
         DMaximum=GMaxima(ii)
         GMaxima(ii)=GMaxima(ii+1)
         GMaxima(ii+1)=DMaximum
       ENDIF
     ENDDO
      ENDDO
      !Eliminate the peaks that are too close
      main=0
      DO
        main=main+1
     i=main
     DO
          i=i+1
       ! If current peak is closer than a given limit, kick it out
          mindist=100000.
       DO ii=1,NSym
         DO iii=1,3
           dcoo2(iii)=GMaxima(i)%com(iii)
         ENDDO
         CALL SymOpR(ii,dcoo2,dhkl)
            !Get the distance from the main peak, take into account all possible shfts
         DO i1=-1,1
           DO i2=-1,1
             DO i3=-1,1
            dcoo1(1)=dhkl(1)+float(i1)-GMaxima(main)%com(1)
            dcoo1(2)=dhkl(2)+float(i2)-GMaxima(main)%com(2)
            dcoo1(3)=dhkl(3)+float(i3)-GMaxima(main)%com(3)
                  dist=sqrt(multsc(dcoo1,dcoo1,RDim,.false.))
!            if (dist<0.5) then
!              write(*,*) dcoo2
!              write(*,*) dhkl
!              write(*,*) dcoo1
!              write(*,*) i1,i2,i3
!            endif
!                  write(*,*) (dcoo2(iii),iii=1,3),dist
                  if (dist<mindist) then
                    mindist=dist
                  endif
                ENDDO
           ENDDO
         ENDDO
       ENDDO
       IF (mindist<CloseDist) THEN
         write(*,'(a,3f10.5)') 'Eliminating' ,(GMaxima(i)%com(iii),iii=1,3)
         write(*,'(a,4f10.5)') 'Too close to' ,(GMaxima(main)%com(iii),iii=1,3),mindist
         CALL Shiftlist(i)
       ENDIF
       IF (i>=GMaxNr) EXIT
     ENDDO
     if (main>=(GMaxNr-1)) EXIT
      ENDDO
      ! Now open and read-in the atoms from m40
      OPEN(41,file=compm40file,status='OLD')
      READ(41,*) CAtNr
      !Restrict the number of considered maxima to the number of expected atoms + certain reserve
      dist=pillowfactor*GMaxima(CAtNr)%rho
      DO i=CAtNr,GMaxNr
        if (GMaxima(i)%rho<dist) THEN
       GMaxNr=i-1
       EXIT
     ENDIF
      ENDDO
!      write(*,*) GMaxNr
!      do i=1,GMaxNr
!       write(*,'(5f10.5)') (GMaxima(i)%com(iii),iii=1,3),GMaxima(i)%cc,GMaxima(i)%rho
!      enddo
      ALLOCATE(CAtoms(1:CAtNr))
      ALLOCATE(OccupMax(1:GMaxNr))
      OccupMax=.false.
      DO i=1,4 !Skip header
        READ(41,*)
      ENDDO
      DO i=1,CAtNr
        READ(41,'(a8,19X,3f9.6)') CAtoms(i)%name,(CAtoms(i)%coo(ii),ii=1,3)
     READ(41,*) !Second line contains only ADPs
      ENDDO
      ALLOCATE(OccupAt(1:CAtNr))
      OccupAt=.false.
      !Now take each atom and try to find its counterpart in GMaxima. Take symmetry into account
      located=0; sumdist=0.0
      occupat=.false.
!      write(*,'(a)') 'Located atoms:'
      DO main=1,CAtNr
        dcoo1=CAtoms(main)%coo
     !Shift the coordinates into the unit cell
        DO i=1,3
       DO WHILE (dcoo1(i)>=1.)
            dcoo1(i)=dcoo1(i)-1.0
          END DO
          DO WHILE (dcoo1(i)<0.0)
            dcoo1(i)=dcoo1(i)+1.0
          END DO
        ENDDO
        mindist=1000000.

        DO i=1,GMaxNr
       if (OccupMax(i)) CYCLE !This maximum has been already assigned
          DO ii=1,NSym
         DO iii=1,3
           dcoo2(iii)=GMaxima(i)%com(iii)
         ENDDO
         CALL SymOpR(ii,dcoo2,dhkl)
            !Get the distance from the main peak, take into account all possible shfts
         DO i1=-1,1
           DO i2=-1,1
             DO i3=-1,1
            dcoo2(1)=dhkl(1)+float(i1)-dcoo1(1)
            dcoo2(2)=dhkl(2)+float(i2)-dcoo1(2)
            dcoo2(3)=dhkl(3)+float(i3)-dcoo1(3)
                  dist=sqrt(multsc(dcoo2,dcoo2,RDim,.false.))
                  if (dist<mindist) then
                    mindist=dist
                 bestgmax=i
                  endif
                ENDDO
           ENDDO
         ENDDO
       ENDDO
     ENDDO
        IF (mindist<=CloseDist) THEN !We have found the corresponding atom in the list of maxima
       OccupMax(bestgmax)=.true.
          OccupAt(main)=.true.
       located=located+1
          sumdist=sumdist+mindist
     ENDIF
      ENDDO
!      write(*,*) GMaxNr
!      DO i=1,GMaxNr
!        write(*,'(4f10.5)') (GMaxima(i)%com(ii),ii=1,3),GMaxima(i)%cc
!      ENDDO
      falseat=0
! The old method of counting good and bad atoms
!      do i=GMaxNr,1,-1
!        if(OccupMax(i)) EXIT
!     falseat=falseat+1
!      enddo
!      falseat=GMaxNr-located-falseat
!      write(*,'(a,i4,a,i4,a,f6.3,a)') 'I could locate',located,' out of ',CAtNr,' atoms within the distance of',CloseDist,' A.'
!      write(*,'(a,f8.5,a,i4,a,i4)') 'Average distance:',sumdist/float(located),' A, number of considered maxima:',GMaxNr, &
!            ', false atoms:',falseat

! The new method
      goodat=0;falseat=0;i3=-100000
      do i=1,GMaxNr
        if (OccupMax(i)) THEN
       goodat=goodat+1
     ELSE
       falseat=falseat+1
        ENDIF
     if ((goodat-falseat)>=i3) THEN
          i3=goodat-falseat
          bestgmax=i
       i1=goodat
       i2=falseat
     ENDIF
      enddo
      falseat=i2
      located=i1
      write(*,'(a,i4,a,i4,a,f6.3,a)') 'I could locate',located,' out of ',CAtNr,' atoms within the distance of',CloseDist,' A.'
      write(*,'(a,f8.5,a,i4,a,i4)') 'Average distance:',sumdist/float(located),' A, number of considered maxima:',GMaxNr, &
            ', false atoms:',falseat
!      do i=1,GMaxNr
!        write(*,'(i3,3f10.5,l4)') i,(GMaxima(i)%com(iii),iii=1,3),OccupMax(i)
!      enddo
      write(*,'(a)',ADVANCE='NO') 'Assigned maxima: '
      do i=1,GMaxNr
        if (OccupMax(i)) THEN 
       write(*,'(a1)',ADVANCE='NO') 'O'
        else
       write(*,'(a1)',ADVANCE='NO') 'X'
     endif
      enddo
      write(*,*)
      write(*,'(a)',ADVANCE='NO') 'Located atoms: '
      do i=1,CAtNr
        if (OccupAt(i)) THEN 
       write(*,'(a1)',ADVANCE='NO') 'O'
        else
       write(*,'(a1)',ADVANCE='NO') 'X'
     endif
      enddo
      write(*,*)
      close(41)
      CONTAINS 
        SUBROUTINE Shiftlist(pos)
          integer :: pos, slloop
          DO slloop=pos+1,GMaxNr
            GMaxima(slloop-1)=GMaxima(slloop)
          ENDDO
       GMaxNr=GMaxNr-1
        END SUBROUTINE Shiftlist
      
      END SUBROUTINE CompareM40

! Matrix multiplication of REAL matrices
!***********************************************************
      subroutine multm(a,b,c,n1,n2,n3)
!***********************************************************
      real :: a(1:n1*n2),b(1:n2*n3),c(1:n1*n3)
      real :: p
      integer :: i,j,k,ik,jk,jkp
      do i=1,n1
        jkp=1
        ik=i
        do k=1,n3
          ij=i
          jk=jkp
          p=0.
          do j=1,n2
            p=p+a(ij)*b(jk)
            ij=ij+n1
            jk=jk+1
          enddo
          c(ik)=p
          ik=ik+n1
          jkp=jkp+n2
        enddo
      enddo
      end subroutine multm

! Matrix multiplication of REAL matrices, DOUBLE PRECISION version
!***********************************************************
      subroutine dmultm(a,b,c,n1,n2,n3)
!***********************************************************
      double precision:: a(1:n1*n2),b(1:n2*n3),c(1:n1*n3)
      double precision :: p
      integer :: i,j,k,ik,jk,jkp
      do i=1,n1
        jkp=1
        ik=i
        do k=1,n3
          ij=i
          jk=jkp
          p=0.
          do j=1,n2
            p=p+a(ij)*b(jk)
            ij=ij+n1
            jk=jk+1
          enddo
          c(ik)=p
          ik=ik+n1
          jkp=jkp+n2
        enddo
      enddo
      end subroutine dmultm

! returns the scalar product of two vectors
!******************************************************************
      REAL FUNCTION multsc(iVec1,iVec2,Dim,iRec)
!******************************************************************
      USE GLOBALDEFINITIONS
      USE GLOBALVARIABLES
      
      IMPLICIT NONE
      INTEGER :: Dim !Dimension of the input vectors, allowed only RDim or SDim
      INTEGER :: l1,l2 !loop1, loop2
      REAL, DIMENSION(Dim) :: iVec1,iVec2 ! Input vectors
      REAL, DIMENSION(3,3) :: RG, RGi
      REAL, DIMENSION(MaxDim) :: dvec1,dvec2,dvec3
      REAL, DIMENSION(1:1) :: mul
      LOGICAL :: iRec !Is the vector in reciprocal space?      
      multsc=0.0E0
      DO l1=1,3
        DO l2=1,3
       RG(l1,l2)=G(l1,l2)
       RGi(l1,l2)=Gi(l1,l2)
     ENDDO
      ENDDO
      IF (iRec) THEN
        DO l1=1,Dim
       dvec1(l1)=ivec1(l1)
       dvec2(l1)=ivec2(l1)
     ENDDO
        IF (Dim>RDim) THEN !Reciprocal lattice vector, to determine the scalar product we have to add qvectors
!         add q-vectors
       DO l1=1,Dim-RDim
         DO l2=1,RDim
           dvec1(l2)=dvec1(l2)+dvec1(l1+RDim)*qvec(l2,l1)
           dvec2(l2)=dvec2(l2)+dvec2(l1+RDim)*qvec(l2,l1)
         END DO
       END DO
     ENDIF
     call multm(dvec1,RGi,dvec3,1,RDim,RDim)
        call multm(dvec3,dvec2,mul,1,RDim,1)
      ELSE
        call multm(ivec1,RG,dvec1,1,RDim,RDim)
     call multm(dvec1,ivec2,mul,1,RDim,1)
      END IF
      multsc=mul(1)
      END FUNCTION multsc

      SUBROUTINE RemoveSpaces(ioText)
      implicit none
      character*(*) :: ioText
      integer :: ii,idel
        ii=0
        DO
          ii=ii+1
       IF (ioText(ii:ii)==' ') THEN
         ioText(ii:)=ioText(ii+1:)
         ii=ii-1
          ENDIF
       IF (ii>=idel(ioText)) EXIT
        ENDDO
      END SUBROUTINE RemoveSpaces

      SUBROUTINE FinalCleanUp
      USE GLOBALDEFINITIONS
      USE GLOBALVARIABLES
      implicit none

      IF  (ALLOCATED (Atoms))     DEALLOCATE (Atoms) 
      IF  (ALLOCATED (CPlist))    DEALLOCATE (CPlist)
      IF  (ALLOCATED (QVec))      DEALLOCATE (QVec)  
      IF  (ALLOCATED (Tlist))     DEALLOCATE (Tlist) 
      IF  (ALLOCATED (CO))        DEALLOCATE (CO)    
      IF  (ALLOCATED (Center))    DEALLOCATE (Center)
      IF  (ALLOCATED (RMat) )     DEALLOCATE (RMat) 
      IF  (ALLOCATED (TVec) )     DEALLOCATE (TVec)  
      IF  (ALLOCATED (RMatI))     DEALLOCATE (RMatI) 
      IF  (ALLOCATED (TVecI))     DEALLOCATE (TVecI) 
      IF  (ALLOCATED (Rim)  )     DEALLOCATE (Rim)   
      IF  (ALLOCATED (Aper))      DEALLOCATE (Aper)  
      IF  (ALLOCATED (NVox))      DEALLOCATE (NVox)  
      IF  (ALLOCATED (NVoxR))      DEALLOCATE (NVoxR)  
      IF  (ALLOCATED (NVoxRim))      DEALLOCATE (NVoxRim)  
      IF  (ALLOCATED (Compos))      DEALLOCATE (Compos)  
      IF  (ALLOCATED (rho))      DEALLOCATE (rho)  
      IF  (ALLOCATED (rhogr))      DEALLOCATE (rhogr)
      IF  (ALLOCATED (srho))      DEALLOCATE (srho)  
      IF  (ALLOCATED (rhor))      DEALLOCATE (rhor)  
      IF  (ALLOCATED (DiffVec))      DEALLOCATE (DiffVec)  
      IF  (ALLOCATED (PixList))      DEALLOCATE (PixList)  
      IF  (ALLOCATED (Diff))      DEALLOCATE (Diff)  
      IF  (ALLOCATED (Basin))      DEALLOCATE (Basin)  

      END SUBROUTINE FinalCleanUp
!Library contains subroutines, which serve for spline interpolation of the electron density
!5.7.2002+ Lukas Palatinus

      SUBROUTINE GetPeak(Centr, ExPeak,outrho,fast) !Gets coordinates of the pixel, returns accurate coordinates and density of the maximum
      USE GlobalDefinitions
      USE GlobalVariables
      IMPLICIT NONE
      INTEGER :: Centr(MaxDim) ! Central voxel of the maximum, dummy coords
      INTEGER :: AxOrd(MaxDim),DerOrd(MaxDim),Orig(MaxDim)
      DOUBLE PRECISION :: ExPeak (MaxDim) ! The exact(er) peak position
      DOUBLE PRECISION :: outrho !peak electron density
      DOUBLE PRECISION, DIMENSION(range) :: y,y2 !arrays for subroutine splines
      INTEGER :: ii ! Counters
      LOGICAL :: Succ !flag of success
      LOGICAL :: fast !switches between fast, but possibly less accurate, and slower, more precise calculation. .true. is fast calculation

      ExPeak=dble(Centr)
      call CPRefine(ExPeak,Succ,fast) !returns the coordinates of the maximum, succ is false, if no maximum could be found
      IF (Succ) THEN
        DerOrd=0
        DO ii=1,RDim
          AxOrd(ii)=ii
          Orig(ii)=int(ExPeak(ii)-float(range)/2.+range)+1-range
        END DO
        call FillRhoCut(AxOrd,Orig,RDim) !fill the part of the electron density around the maximum
        IF (fast) THEN
          call RhoAtPointFast(RDim,range,DerOrd,AxOrd,ExPeak,orig,srho,outrho,y,y2) !returns the peak density
        ELSE
          call RhoAtPoint(RDim,range,DerOrd,AxOrd,ExPeak,orig,srho,outrho,y,y2) !returns the peak density
        ENDIF
      ELSE
        ExPeak=-1.0E30
	outrho=-1.0E30
      END IF
      END SUBROUTINE GetPeak

      SUBROUTINE CPRefine(pt,succ,fast) ! find stationary point (derivatives=0)
        USE GLOBALDEFINITIONS
	USE GLOBALVARIABLES
        
	IMPLICIT NONE
        INTEGER, PARAMETER :: ptchangelim=40
        DOUBLE PRECISION, PARAMETER :: MaxGradShift=0.5
        DOUBLE PRECISION, PARAMETER :: MaxPtShift=3.
	DOUBLE PRECISION :: pt(MaxDim),ptor(MaxDim),ptref(MaxDim) !point, where the density is to be determined
			! PtOr to save the starting pt
			!PtRef was added for the cases pt is close to the pixel border. In that case, changing origin can cause oscillations.
        INTEGER, DIMENSION(MaxDim) :: AxOrd !order of the axes in the output rcut
        INTEGER, DIMENSION(MaxDim) :: DerOrd,Orig
        DOUBLE PRECISION :: rout,disc, Grad, GradOld, GradOlder, red, dump !output rho, 1st and 2nd solution of the qadr. equation, discrepancy
        DOUBLE PRECISION, DIMENSION(range) :: y,y2
        DOUBLE PRECISION, ALLOCATABLE :: H(:,:),Hinv(:,:),Ho(:,:),Gr(:), Sh(:) !Hessian, its inversion and gradient
        DOUBLE PRECISION, DIMENSION(6) :: pa !auxiliary matrix
	    LOGICAL :: succ !success
	    LOGICAL :: fast !switches between fast and slow (but more accurate) calculation of the derivatives
        INTEGER :: i,ii,iii, iv, ptchange !counters

        ALLOCATE(H(RDim,RDim),STAT=AllocError)
        ALLOCATE(Hinv(RDim,RDim),STAT=AllocError)
        ALLOCATE(Ho(RDim,RDim),STAT=AllocError)
        ALLOCATE(Gr(RDim),STAT=AllocError)
        ALLOCATE(Sh(RDim),STAT=AllocError)
        IF (AllocError.GT.0) STOP 'Error allocating in H, Hinv and Gr in CpRefine! Too little memory?'
        do i=1,RDim
          Orig(i)=int(pt(i)-dble(range)/2.+range)+1-range !set the origin of the region +- range/2 around pt
          DerOrd(i)=0
        end do
	ptchange=0 !counts the number of cycles
        dump=1.
        DO ii=1,RDim !permute the order of axes to refine each time another direction
          AxOrd(ii)=ii
        END DO
        ptor=pt
	ptref=pt
        DO
          ptchange=ptchange+1
	  succ=.true.
	  IF (ptchange>ptchangelim) THEN !trying too long time
	    Succ=.false.
	    EXIT
	  END IF
	  succ=.true. ! succ here checks, if pt didn't shift out of borders
          DO i=1,RDim
	    if (abs(pt(i)-ptref(i))>0.5) succ=.false.	    
	  ENDDO
	  IF (.not.succ) THEN !We have to update the origin, because the point shifted too much
	    succ=.true.
	    ptref=pt
            DO i=1,RDim !define the origin of the "cube" that serves to interpolate pt
              Orig(i)=int(ptref(i)-dble(range)/2.+range)+1-range
              if (((Orig(i)+(range-1)+Rim(i)>NVoxRim(i)-1).and.(Aper(i)==1)).or.  &
	        ((Orig(i)+Rim(i)<0).and.(Aper(i)==1))) succ=.false.
            END DO
	    IF (.not.succ) THEN
	      EXIT !If Orig runs out of range, subroutine is terminated with flag "unsuccessful"
          ENDIF
          ENDIF
          call FillRhoCut(AxOrd,Orig,RDim) !fill the area of rho with properly permuted axes
!********* Gradient and Hessian
          DO iii=1,3
	    DerOrd=0
	    DerOrd(iii)=1
	    IF (fast) THEN
            call RhoAtPointFast(RDim,range,DerOrd,AxOrd,pt,orig,srho,rout,y,y2) !y and y2 contain parameters of the spline interpolation along the last axis
	    ELSE
            call RhoAtPoint(RDim,range,DerOrd,AxOrd,pt,orig,srho,rout,y,y2) !y and y2 contain parameters of the spline interpolation along the last axis
	    ENDIF
            Gr(iii)=rout
          END DO
	  DO iii=1,RDim
	    DO iv=1,RDim
              DerOrd=0
	      DerOrd(iii)=DerOrd(iii)+1
	      DerOrd(iv)=DerOrd(iv)+1
        IF (fast) THEN
              call RhoAtPointFast(RDim,range,DerOrd,AxOrd,pt,orig,srho,rout,y,y2) !y and y2 contain parameters of the spline interpolation along the last axis
        ELSE
              call RhoAtPoint(RDim,range,DerOrd,AxOrd,pt,orig,srho,rout,y,y2) !y and y2 contain parameters of the spline interpolation along the last axis
        ENDIF
              H(iii,iv)=rout	      
	    END DO
	  END DO
          call invert(H,Hinv,RDim)
          IF (Hinv(1,1)<-1.0D29) THEN
            succ=.false.
	    EXIT
	  END IF
	  Sh=0.0D0
          Grad=0.0D0
	  DO iii=1,RDim
	    DO iv=1,RDim
	      Sh(iii)=Sh(iii)+Gr(iv)*Hinv(iv,iii)
	    END DO
            Grad=Grad+Sh(iii)**2
	  END DO
	  Grad=sqrt(Grad)
          IF (Grad>MaxGradShift) THEN
	    Red=MaxGradShift/Grad
	  ELSE
	    Red=1.0
	  END IF
          disc=0.0D0
	  DO iii=1,RDim
	    pt(iii)=pt(iii)-sh(iii)*red*dump
            disc=disc+(pt(iii)-ptor(iii))**2
	  END DO
	  disc=sqrt(disc)
	  IF (disc>MaxPtShift) Succ=.false.
	  IF (ptchange>2) THEN
	    if ((gradold-grad)>(gradolder-gradold)) dump=dump*0.5
	  END IF
          gradolder=gradold
	  gradold=grad
!          write(*,*) ptchange, grad,(pt(i),i=1,3), dump
	  IF (.not.Succ.or.grad<eps) EXIT !loop until .not.succ or disc is sufficiently small
	END DO
        IF (succ) THEN
!	  write(*,*) pt(1)/NVox(1), pt(2)/NVox(2),pt(3)/NVox(3)
           
          call TrMat(Mv,Ho,RDim)
          call MatMatMul(H,Ho,Hinv,RDim)
          call MatMatMul(Mv,Hinv,Ho,RDim) !Output is the Hessian in the orthonormal system
!	  DO iii=1,RDim
!	    DO iv=1,RDim
!	      write(*,'(f14.6)',ADVANCE='NO') Ho(iii, iv)
!	    END DO
!	    write(*,*)
!	  END DO
!          write(*,*)
! Diagonalize the Hessian
          pa(1)=Ho(1,1)
          pa(2)=Ho(2,1)
          pa(3)=Ho(2,2)
          pa(4)=Ho(3,1)
          pa(5)=Ho(3,2)
          pa(6)=Ho(3,3)
          call ql(pa,Hinv,Evl,RDim)
!Hinv contains eigenvectors as column vectors, Evl contains eigenvalues
          call Trmat(M,Ho,RDim)
	  call MatMatMul(Ho,Hinv,Evc,RDim)
	END IF
      END SUBROUTINE CPRefine
      

      SUBROUTINE FillRhoCut(AxOrd,Org,Dim) !fills part of the electron density from -range/2 to +range/2
        USE GLOBALDEFINITIONS
	USE GLOBALVARIABLES
        IMPLICIT NONE
	INTEGER :: ArrLoc, Dim
        INTEGER :: Dxyz(MaxDim),EndC(MaxDim)     ! output from subrt. Coordinates, input to ArrayLocation
        INTEGER, DIMENSION(*) :: AxOrd,Org !order of the axes in the output rcut, origin in the rho
        LOGICAL ::NextCoo
	
	DO Cnt1=1,RDim !set start and end of the cube
          Dxyz(Cnt1)=Org(Cnt1)
          IF (Cnt1<=Dim) THEN
	    EndC(AxOrd(Cnt1))=Org(AxOrd(Cnt1))+range-1
          ELSE
	    EndC(AxOrd(Cnt1))=Org(AxOrd(Cnt1))
	  END IF
	END DO
        Cnt1=0
        DO
	  Call ArrayLocationR(Dxyz,ArrLoc,1) !get the location of this pixel in rhor
          srho(Cnt1)=rhor(ArrLoc)
          cnt1=cnt1+1
          IF (.not.NextCoo(RDim,Org,EndC,Dxyz,AxOrd)) EXIT
        END DO
      END SUBROUTINE FillRhoCut


      SUBROUTINE RhoAtPoint(Dim,range,DerOrd,AxOrd,pt,orig,inrho,rout,y,y2) !returns electron density at pt and parameters y,y2 od the spline interpolation
        IMPLICIT NONE
	INTEGER :: i,ii,iii,Pcnt,SCnt, rstep ! counters
        INTEGER :: Dim,ADim !Dimension of the rhocut, actual dimension during the spline interpolation
        INTEGER :: range !length of the vertices of the "hypercube"
        DOUBLE PRECISION,DIMENSION(*) :: pt !point, where the density is to be determined
        INTEGER, DIMENSION(*) :: Orig !origin of the rhocut
        DOUBLE PRECISION :: inrho(0:range**Dim-1) !Input array containing the cut of the density around pt
        DOUBLE PRECISION :: rcut(0:range**Dim-1), rcutbck(0:range**Dim-1) !Working array
        DOUBLE PRECISION :: onerow(0:range-1)
        INTEGER :: rowpos(0:range-1)
        DOUBLE PRECISION,DIMENSION(range) :: x,y,y2 !arrays for the routines spline and splint
        INTEGER, DIMENSION(*) :: AxOrd !order of the axes in the output rcut
        INTEGER, DIMENSION(*) :: DerOrd  ! Array defining the derivative of the function returned by RhoAtPoint
        INTEGER, DIMENSION(1:Dim) :: DerOrdTmp, AxOrdTmp,StartC, EndC,RunC, CurC
        DOUBLE PRECISION, DIMENSION(0:2) :: rval ! Density and its derivatives at the point pt, output
        DOUBLE PRECISION :: rout, NumDerivative !output
        LOGICAL :: NextCoo

!        write(*,*) derord(1:Dim)
!        write(*,*) pt(1:Dim)
        Scnt=0
        DO i=1,Dim
          SCnt=SCnt+DerOrd(AxOrd(i)) !How many derivatives must be taken
          DerOrdTmp(i)=DerOrd(AxOrd(i))
          AxOrdTmp(i)=i
        ENDDO
        rcut=inrho
!        write(*,*) 'V Rhoatpoint:',rcut(0:range**Dim-1)
!        write(*,*) 'pt:',pt(1:3)
!        write(*,*) 'Orig:',Orig(1:3)
!        write(*,*) 'Scnt,derord:',Scnt, derord(1:3)
!        goto 1000
        IF (SCnt>0) THEN
          ! fill rcut by a derivative of inrho defined by derord
          DO i=1,SCnt
            DO ii=1,Dim
              IF (DerOrdTmp(ii)>0) THEN
                DerOrdTmp(ii)=DerOrdTmp(ii)-1
                StartC=0
                EndC=range-1
                EndC(ii)=0
                RunC=StartC
                rstep=ii
                EXIT
              ENDIF
            ENDDO
!            write(*,*) 'Jdu do derivace', Dim
            DO
!              write(*,*) RunC
              !Fill a one-dimensional array for calculating the derivative:
              CurC=RunC
!              write(*,*) 'Do vnitrni smycky', RunC, EndC
!              write(*,*) 'Dim, AxOrdTmp:',Dim, AxOrdTmp
              DO ii=0,range-1
                CurC(rstep)=ii
                !get position in rcut
                PCnt=0
                DO iii=1,Dim
                  PCnt=PCnt+CurC(iii)*range**(iii-1)
                ENDDO
!                write(*,*) CurC, PCnt,RCut(PCnt)
                onerow(ii)=rcut(PCnt)
                rowpos(ii)=Pcnt
              ENDDO
!              write(*,*) 'Z vnitrni smycky'
!              write(*,*) 'Pocitam derivace'
              DO iii=0,range-1
                rcutbck(rowpos(iii))=NumDerivative(onerow,range,1.0D0,iii,1)
!                write(*,*) onerow(iii),rcutbck(rowpos(iii))
              ENDDO
!              write(*,*) 'Dopocitano'
!              write(*,*) StartC, EndC
 !             write(*,*)
              IF (.not.NextCoo(Dim,StartC,EndC,RunC,AxOrdTmp)) EXIT
            ENDDO
            rcut=rcutbck
          ENDDO
!          write(*,*) 'Mam derivaci'
!          read(*,*)
        ENDIF
!        write(*,*) 'Po numder:',rcut(0:range**Dim-1)

1000      DO Adim=Dim,1,-1 !Going from Dim to 1, at each cycle interpolating Adim-1 dimensional section of rcut
          Pcnt=1 !counter of the old rho
	  Scnt=0 !counter of the new, Adim-1 dimensional rho
	  DO i=0,range**ADim-1 ! repeat over all pixels of old rcut
            x(Pcnt)=float(Orig(AxOrd(Dim-Adim+1))+Pcnt-1)
	    y(Pcnt)=rcut(i) !fill one row
            PCnt=PCnt+1
	    IF (PCnt==range+1) THEN !if the row is full, interpolate the value at pt
	      PCnt=1
	      CALL spline(x,y,range,y2)
	      CALL splint(x,y,y2,range,pt(AxOrd(Dim-Adim+1)),rval) !rval(0) contains the density, rval(1) the first derivative and rval(2) the second derivative
!              rcut(Scnt)=rval(DerOrd(AxOrd(Dim-Adim+1)))
              rcut(Scnt)=rval(0) !assign the result to rho(Scnt).
	      						!It is perfectly safe to save to the same rho, that point will not be needed in this cycle anymore.
	      SCnt=SCnt+1
	    END IF
!            continue to the next row
          END DO
          ! the ADim-1 dimensional "first slice" of rcut now contains the interpolated section and we can continue with next step
        END DO
	rout=rcut(0)
!	write(*,*) rout
!	STOP
      END SUBROUTINE RhoAtPoint
    
      SUBROUTINE RhoAtPointFast(Dim,range,DerOrd,AxOrd,pt,orig,inrho,rout,y,y2) !returns electron density at pt and parameters y,y2 od the spline interpolation
        IMPLICIT NONE
    INTEGER :: i,Pcnt,SCnt ! counters
        INTEGER :: Dim,ADim !Dimension of the rhocut, actual dimension during the spline interpolation
        INTEGER :: range !length of the vertices of the "hypercube"
        DOUBLE PRECISION,DIMENSION(*) :: pt !point, where the density is to be determined
        INTEGER, DIMENSION(*) :: Orig !origin of the rhocut
        DOUBLE PRECISION :: inrho(0:range**Dim-1) !Input array containing the cut of the density around pt
        DOUBLE PRECISION :: rcut(0:range**Dim-1) !Working array
        DOUBLE PRECISION,DIMENSION(range) :: x,y,y2 !arrays for the routines spline and splint
        INTEGER, DIMENSION(*) :: AxOrd !order of the axes in the output rcut
        INTEGER, DIMENSION(*) :: DerOrd  ! Array defining the derivative of the function returned by RhoAtPoint
        DOUBLE PRECISION, DIMENSION(0:2) :: rval ! Density and its derivatives at the point pt, output
        DOUBLE PRECISION :: rout !output

        rcut=inrho
        DO Adim=Dim,1,-1 !Going from Dim to 1, at each cycle interpolating Adim-1 dimensional section of rcut
          Pcnt=1 !counter of the old rho
      Scnt=0 !counter of the new, Adim-1 dimensional rho
      DO i=0,range**ADim-1 ! repeat over all pixels of ald rcut
            x(Pcnt)=float(Orig(AxOrd(Dim-Adim+1))+Pcnt-1)
        y(Pcnt)=rcut(i) !fill one row
            PCnt=PCnt+1
        IF (PCnt==range+1) THEN !if the row is full, interpolate the value at pt
          PCnt=1
          CALL spline(x,y,range,y2)
          CALL splint(x,y,y2,range,pt(AxOrd(Dim-Adim+1)),rval) !rval(0) contains the density, rval(1) the first derivative and rval(2) the second derivative
              rcut(Scnt)=rval(DerOrd(AxOrd(Dim-Adim+1))) !assign the result to rho(Scnt).
                                !It is perfectly safe to save to the same rho, that point will not be needed in this cycle anymore.
          SCnt=SCnt+1
        END IF
!            continue to the next row
          END DO
          ! the ADim-1 dimensional "first slice" of rcut now contains the interpolated section and we can continue with next step
        END DO
    rout=rcut(0)
      END SUBROUTINE RhoAtPointFast


      SUBROUTINE Spline(x,y,n,y2) !copied from Numerical Recipies
      INTEGER :: i,k,n
      DOUBLE PRECISION :: x(n),y(n),y2(n)
      DOUBLE PRECISION :: p,sig,un,u(n)
      u(1)=0.
      y2(1)=0.
      do i=2,n-1
        sig=(x(i)-x(i-1))/(x(i+1)-x(i-1))
        p=sig*y2(i-1)+2.
        y2(i)=(sig-1.)/p
        u(i)=(6.*((y(i+1)-y(i))/(x(i+1)-x(i))-(y(i)-y(i-1))/(x(i)-x(i-1)))/(x(i+1)-x(i-1))-sig*u(i-1))/p
      end do
      un=0.
      y2(n)=0.
      do k=n-1,1,-1
        y2(k)=y2(k)*y2(k+1)+u(k)
      end do
      return
      END SUBROUTINE Spline

      SUBROUTINE splint(xa,ya,y2a,n,x,y) !copied from Num.Rec., strongly simplified and modified
      INTEGER :: n,khi,klo
      DOUBLE PRECISION :: x,xa(n),y2a(n),ya(n)
      DOUBLE PRECISION, DIMENSION(0:2) :: y ! value and its derivatives
      DOUBLE PRECISION :: a,b
      
!     x lies in our case always exactly in the middle of the range ->
!     no need to bisect, we can calculate klo and khi directly
      klo=int(x-xa(1)+dble(n))-n+1
      khi=klo+1
!     moreover, the spacing between the x-values is always 1 -> we can forget h; h=1
      a=xa(khi)-x
      b=x-xa(klo)
!     value
      y(0)=a*ya(klo)+b*ya(khi)+((a**3-a)*y2a(klo)+(b**3-b)*y2a(khi))/6.
!     the first derivative
      y(1)=(ya(khi)-ya(klo))+((1-3*a**2)*y2a(klo)+(3*b**2-1)*y2a(khi))/6.
!     the second derivative
      y(2)=a*y2a(klo)+b*y2a(khi)
      return
      END

   double precision function NumDerivative(arr,npoints,step,pos,derord)
      IMPLICIT NONE
      integer :: j,k,npoints,derord,pos,a, GetA,factorial
      double precision, DIMENSION(0:npoints-1) :: arr
      double precision:: step
      double precision, dimension(0:npoints-1) :: d

      !Calculate the coefficients d
      DO j=0,npoints-1
        IF (j==pos) CYCLE
        a=GetA(derord,npoints,pos,j)
        d(j)=dble(Factorial(derord)*a)/dble(factorial(j)*factorial(npoints-1-j))
        IF (mod(derord-j,2)/=0) d(j)=-d(j)
      ENDDO
      d(pos)=0
      DO j=0,npoints-1
        IF (j/=pos) d(pos)=d(pos)-d(j)
      ENDDO
      NumDerivative=0.0
      DO j=0,npoints-1
        NumDerivative=NumDerivative+d(j)*arr(j)
      ENDDO
      NumDerivative=NumDerivative/(step**derord)
    END FUNCTION NumDerivative

    integer function Factorial(n)
      integer :: n,k
      Factorial=1
      IF (n>0) THEN
        DO k=1,n
          Factorial=Factorial*k
        ENDDO
      ENDIF
    END FUNCTION Factorial

    integer function GetA(derord,npoints,i,j)
      implicit none
      integer :: derord,npoints,i,j
      integer :: k,l,PartA
      integer, dimension(0:npoints-3) :: plist
      IF (derord>2) STOP 'I am not ready for higher derivatives!'
      IF (i==j) STOP 'Internal error in numerical derivatives.'
      l=-1
      DO k=0,npoints-1
        IF ((k/=i).and.(k/=j)) THEN
          l=l+1
          plist(l)=k-i
        ENDIF
      ENDDO
      GetA=0

      IF (DerOrd==1) THEN
        GetA=1
        DO k=0,npoints-3
          GetA=GetA*plist(k)
        ENDDO
      ELSEIF (DerOrd==2) THEN
        GetA=0
        DO l=0,npoints-3
          PartA=1
          DO k=0,npoints-3
            IF (k/=l) PartA=PartA*plist(k)
          ENDDO
          GetA=GetA+PartA
        ENDDO
      ENDIF
    end function GetA
